MODULE convert
!
!  #[ description
!
! this module contains some functions to enable conversions
! between windvectors (u,v) and windvectors (speed,dir) and to convert between
! different coordinate systems on the Earths surface
!
! Part of this software was developed within the context of the EUMETSAT
! Satellite Application Facility on Numerical Weather Prediction (NWP SAF),
! under the Cooperation Agreement dated 16 December, 2003, between EUMETSAT
! and the Met Office, UK, by one or more partners within the NWP SAF.
! The partners in the NWP SAF are the Met Office, ECMWF, KNMI and
! Meteo France.
!
!     Unit Name:         convert
!
!     Created on:        15-09-2004
!
!     Last Modified on:  $Date$
!
!     Modifications Log:
!
!  $Id$
!
!
!  #]
  !  #[ USE statements
  USE numerics, only: missing_real, missing_indicator_real
  USE numerics, only: r_, r8_
  !  #]
  !  #[ parameters
  implicit none
  private ! all is private to this module, unless stated otherwise

  real(r8_), parameter :: dpi     = 3.14159265358979323846_r8_ ! = pi
  real(r_),  parameter :: pi      = 3.1415926536_r_  ! = pi
  real(r_),  parameter :: two_pi  = 2.0_r_ * pi   ! = pi*2
  real(r_),  parameter :: rad2deg = 180.0_r_/pi   ! = 180/pi
  real(r_),  parameter :: deg2rad = pi/180.0_r_   ! = pi/180

  real(r_),  parameter :: R_earth_equatorial = 6378.1_r_ ! [km]
  real(r_),  parameter :: R_earth_polar      = 6356.8_r_ ! [km]
  ! source:
  ! http://astrogeology.usgs.gov/Projects/BrowseTheSolarSystem/earth.html
  ! outdated link: http://wwwflag.wr.usgs.gov/USGSFlag/Space/wall/earth.html

  !  #]
  !  #[ user interface
  ! specify the public functions of this MODULE
  public :: speeddir_to_u, speeddir_to_v
  public :: speeddir_to_v_oce, speeddir_to_u_oce, uv_to_dir_oce
  public :: uv_to_dir, uv_to_speed
  public :: met2uv, uv2met
  public :: rotuv
  ! the following functions/routines/parameters are needed in
  ! the OrbitModule of the ADM project
  public :: latlon2xyz, xyz2latlon
  public :: orbit2xyz, xyz2orbit
  public :: WVC_Orientation
  public :: my_arctan, my_arctan_rad
  public :: dpi, pi, deg2rad, rad2deg
  public :: R_earth_equatorial, R_earth_polar
  public :: get_angle_distance, get_distance, get_angle_distance_old
  public :: get_haversine_angular_distance, get_course, get_point_from_course
  public :: relative_angle
  public :: trnsfrm
  public :: logDiff
  public :: extend_wvcs
  public :: average_2_angles
  public :: compute_orbit_elements, compute_row_cell_number
  !  #]
contains

  ! the next 2 functions are used by sdp, and are basically identical
  ! to the following 4 functions, so probably they should be merged
  ! if someone finds the time to do it ...
  ! Just leave it like this; it is more efficient so.

  subroutine met2uv(spd,dir,u,v)
    !  #[
    !
    ! Description: conversion  meteorlogical convention to (u,v)
    !
    !   Hans Bonekamp KNMI 2003
    !
    ! arguments
    real(r_), intent(in)   :: spd
    real(r_), intent(in)   :: dir   ! in degrees  clockwise  from north
    real(r_), intent(out)  :: u, v

    if (missing_real(spd)) then
      u = missing_indicator_real
      v = missing_indicator_real
    else
      u = - spd*sin(dir*deg2rad)
      v = - spd*cos(dir*deg2rad)
    endif

  end subroutine met2uv
  !  #]

  subroutine uv2met(u,v,spd,dir)
    !  #[
    !
    ! Description: conversion  meteorlogical convention to (u,v)
    !   See http://www.atd.ucar.edu/rtf/facilities/isff/wind_ref.shtml
    !
    !   Hans Bonekamp KNMI 2003
    !
    ! arguments
    real(r_), intent(in)    :: u, v
    real(r_), intent(out)   :: spd
    real(r_), intent(out)   :: dir   ! in degrees  clockwise  from north

    !------------------------------------------------------------------------------!
    ! REMARK: Note that the output of atan2 lies within (-pi,+pi]. As formulated   !
    ! below, the direction (in degrees) is in the interval (0,360] - as it should  !
    ! be in the meteorological convention. The construction                        !
    !   dir=atan2(-u,-v)*rad2deg                                                   !
    ! seems more efficient, but returns dir in (-180,+180].                        !
    !------------------------------------------------------------------------------!

    if (missing_real(u)) then
      spd = missing_indicator_real
      dir = missing_indicator_real
    else
      spd  =  sqrt(u*u+v*v)
      if (spd .le. 0.0) then
        dir = 180.0
      else
        dir = atan2(u,v)*rad2deg + 180.0  ! clockwise rotating
      endif
    endif

  end subroutine uv2met
  !  #]

  !--------------------------------
  !  #[ note for the following 4 functions:
  !--------------------------------
  !
  ! winddirection is with respect to south,
  ! (so a southward wind has direction 0)
  ! clockwise increment=positive
  ! (meteorological convention, see notebook Jos, p.48)
  ! u=eastward component
  ! v=northward component
  !--------------------------------
  !  #]
  !--------------------------------
  function speeddir_to_u(sp,dir) result(u)
    !  #[ uses meteorological convention
    real(r_) :: sp,dir ! input
    real(r_) :: u      ! output

    if (missing_real(sp) .or. missing_real(dir)) then
      u = missing_indicator_real
    else
      u = -1.0*sp*sin(dir*deg2rad)
    endif

  end function speeddir_to_u
  !  #]

  function speeddir_to_v(sp,dir) result(v)
    !  #[ uses meteorological convention
    real(r_) :: sp,dir ! input
    real(r_) :: v      ! output

    if (missing_real(sp) .or. missing_real(dir)) then
      v = missing_indicator_real
    else
      v = -1.0*sp*cos(dir*deg2rad)
    endif

  end function speeddir_to_v
  !  #]

  function uv_to_dir(u,v) result(dir)
    !  #[ uses meteorological convention
    real(r_) :: u,v ! input
    real(r_) :: dir ! output

    if (missing_real(u) .or. missing_real(v)) then
      dir = missing_indicator_real
    else
      dir = 270.0-my_arctan(u,v)
      if (dir .lt. 0.0) dir = 360.0 + dir
    endif

  end function uv_to_dir
  !  #]

  function uv_to_speed(u,v) result(speed)
    !  #[ does not depend on winddir convention
    real(r_) :: u,v   ! input
    real(r_) :: speed ! output

    if (missing_real(u) .or. missing_real(v)) then
      speed = missing_indicator_real
    else
      speed = sqrt(u*u+v*v)
    endif

  end function uv_to_speed
  !  #]

  !--------------------------------
  ! same functions but now for oceanographic convention
  !--------------------------------
  function speeddir_to_u_oce(sp,dir) result(u)
    !  #[ uses oceanographic convention
    real(r_) :: sp,dir ! input
    real(r_) :: u      ! output

    ! local variable
    real(r_) :: tmp_dir

    tmp_dir = dir
    if (.not. missing_real(tmp_dir)) then
       tmp_dir = tmp_dir + 180.0
       if (tmp_dir .gt. 360.0) tmp_dir = tmp_dir - 360.0
    endif
    u = speeddir_to_u(sp, tmp_dir)

  end function speeddir_to_u_oce
    !  #]
  function speeddir_to_v_oce(sp,dir) result(v)
    !  #[ uses oceanographic convention
    real(r_) :: sp,dir ! input
    real(r_) :: v      ! output

    ! local variable
    real(r_) :: tmp_dir

    tmp_dir = dir
    if (.not. missing_real(tmp_dir)) then
       tmp_dir = tmp_dir + 180.0
       if (tmp_dir .gt. 360.0) tmp_dir = tmp_dir - 360.0
    endif
    v = speeddir_to_v(sp, tmp_dir)

  end function speeddir_to_v_oce
    !  #]
  function uv_to_dir_oce(u,v) result(dir)
    !  #[ uses oceanographic convention
    real(r_) :: u,v ! input
    real(r_) :: dir ! output

    dir = uv_to_dir(u,v)
    if (.not. missing_real(dir)) then
       dir = dir + 180.0
       if (dir .gt. 360.0) dir = dir - 360.0
    endif

  end function uv_to_dir_oce
    !  #]
  !--------------------------------
  subroutine rotuv(alpha,u,v,ur,vr)
    !  #[
    !  rotate wind vector
    ! this routine was first part of the Ambrem2DVAR module.
    ! It rotates a vector (u,v) over an angle alpha to obtain (ur,vr).
    !
    implicit none
    ! arguments:
    real ,intent(in)   :: alpha
    real ,intent(in)   :: u
    real ,intent(in)   :: v
    real ,intent(out)  :: ur
    real ,intent(out)  :: vr
    ! local
    real :: CosAlpha
    real :: SinAlpha
    !
    CosAlpha = cos(alpha)
    SinAlpha = sin(alpha)
    ur       =     u*CosAlpha + v*SinAlpha
    vr       =  -  u*SinAlpha + v*CosAlpha

  end subroutine rotuv
  !  #]

  ! these functions were first part of the Grib module, and calculate
  ! the angle-distance and km-distance between 2 latlon pairs

  function get_angle_distance_old(lat1,lon1,lat2,lon2) result(angle_dist)
    !  #[
    ! determine the angle (in degrees) along a circle on the earths surface
    ! between 2 sets of lat-lon coordinates
    ! (assuming that the shape of the earth is a perfect globe)

    ! compare with the old PRESCAT code in NODEQC.F, lines 210/214
    ! calculation of zDlat, zDlon

    ! NOTE BY J.VOGELZANG, 25-09-2017
    ! This function produces very bad results near the South Pole: it may be
    ! off by 50% or more. Though the effect for ocesn vector winds may be small,
    ! it is safer to use the new function below.

    real(r_), intent(in) :: lat1,lon1,lat2,lon2
    real(r_) :: angle_dist

    ! loal variables
    real(r_) :: lat_diff_deg, lon_diff_deg
    real(r_) :: lat_min_deg, lat_min_rad
    real(r_) :: distance_sq

    ! get the differences
    lat_diff_deg = lat2 - lat1
    lon_diff_deg = lon2 - lon1

    ! determine lowest latitude of both positions
    lat_min_deg = min( abs(lat1),abs(lat2) )

    ! convert lowest latitude to radians
    ! remind: 180./pi = 57.295780
    lat_min_rad = lat_min_deg/57.29577951

    ! account for the fact that lon=-179.9 and lon=179.9 are neighbouring
    IF (lon_diff_deg .gt.  180.) lon_diff_deg = 360. - lon_diff_deg
    IF (lon_diff_deg .lt. -180.) lon_diff_deg = 360. + lon_diff_deg

    ! the triangle in lat-lon grid defining the distance between the
    ! two nodes, has sides of length (R=radius of the earth):
    !     R*lat_difference_rad                      (in north-south direction)
    ! and R*cos(lat_minimum_rad)*lon_difference_rad (in east-west direction)
    !
    ! so the square of the third side (and the distance) is:
    !     R^2*lat_difference_rad^2 +
    !     R^2*cos(lat_minimum_rad)^2*lon_difference_rad^2

    ! For equality of 2 nodes, we use the condition that they are
    ! within a distance equivalent to 0.25 degrees north-south
    ! distance, so within R*0.25/57.29577951
    ! This is in km a distance of ca. 40000/(360*4) = 27.8 km

    ! size of the earth: 12756.3 * !dpi = 40075.098 km (omtrek)
    ! see http://nl.wikipedia.org/wiki/Aarde

    ! for simplicity, we remove the factors R and 57.29577951
    ! from this calculation, and the following remains:

    distance_sq = lon_diff_deg**2 * cos(lat_min_rad)**2 + lat_diff_deg**2
    angle_dist = sqrt(distance_sq)

  end function get_angle_distance_old
  !  #]

  function get_angle_distance(lat1,lon1,lat2,lon2) result(angle_dist)
    !  #[
    ! determine the angle (in degrees) along a circle on the earths surface
    ! between 2 sets of lat-lon coordinates (in degrees)
    ! (assuming that the shape of the earth is a perfect globe)
    ! from Ed Williams Avaiation formulary, http://www.edwilliams.org/avform.htm
    ! the formula used here is more complicated than the standard one, but
    ! less prone to roundoff errors at small distances

    implicit none

    real(r_), intent(in)  :: lat1,lon1           ! latitude and longitude of point 1 (degrees)
    real(r_), intent(in)  :: lat2,lon2           ! latitude and longitude of point 2 (degrees)
    real(r_)              :: angle_dist          ! distance in degrees

    real(r_)              :: clat1,clat2
    real(r_)              :: sdlat,sdlon
    real(r_)              :: arg

    clat1 = cos(deg2rad*lat1)
    clat2 = cos(deg2rad*lat2)
    sdlat = sin(deg2rad*0.5*(lat1 - lat2))
    sdlon = sin(deg2rad*0.5*(lon1 - lon2))

    arg = sqrt(sdlat**2 + clat1*clat2*sdlon**2)
    angle_dist = rad2deg * 2.0*asin(arg)

    end function get_angle_distance
  !  #]

  function get_distance(lat1,lon1,lat2,lon2) result(distance)
    !  #[
    real(r_),intent(in) :: lat1,lon1,lat2,lon2 ! input
    real(r_)            :: distance            ! result

    ! local variables
    real(r_) :: angle_distance

    angle_distance = get_angle_distance(lat1,lon1,lat2,lon2)
    distance = 2.*pi*R_earth_equatorial*angle_distance/360.

  end function get_distance
    !  #]

  ! some more routines for spherical trigomometry
  !
  subroutine get_haversine_angular_distance(lat1,lon1,lat2,lon2,d)
  ! [#
  ! calculate angular distance between points with the Haversine formula
  ! that is less prone to roundoff errors for small distances.
  ! formulas from Aviation Formulary v1.46, Haversine formula.
  ! http://williams.best.vwh.net/avform.htm

    real, intent(in)  :: lat1, lon1
    real, intent(in)  :: lat2, lon2
    real, intent(out) :: d

    real              :: clat1, clat2
    real              :: A, B

    clat1 = cos(Deg2Rad*lat1)
    clat2 = cos(Deg2Rad*lat2)

    A = sin(0.5*Deg2Rad*(lat1 - lat2))
    B = sin(0.5*Deg2Rad*(lon1 - lon2))

    d = 2.0*asin(sqrt(A*A + clat1*clat2*B*B))

  end subroutine get_haversine_angular_distance
  ! #]

  subroutine get_course(lat1,lon1 , lat2,lon2 , tau)
  ! #[
  !
  ! Calculates the initial course when going from point 1 to point 2 on a great
  ! circle. The course is defined in radians measured clockwise with respect to
  ! the North (radians true North).
  ! Course pi/2 means that point 2 lies to the east of point 1, so when going
  ! from point 1 to point 2 one should set course pi/2 = 90 deg true North, so
  ! one should go eastward. Similarly, course -pi/2 (or 3*pi/2) means that
  ! point 2 lies to the west West of point 1, so to get from point 1 to point 2
  ! one should be heading -90 deg (or 270 deg) true North. Go west, young man!
  !
  ! NOTE: this routine differs a minus sign from the original formula in
  ! Aviation Formulary V 1.46 by Ed Williams,
  ! http://williams.best.vwh.net/avform.htm#GCF.
  !
    implicit none

    real, intent(in)  :: lat1, lon1            ! longitude and latitude of point 1 (degrees)
    real, intent(in)  :: lat2, lon2            ! longitude and latitude of point 2 (degrees)
    real, intent(out) :: tau                   ! initial course in radians (degrees clockwise w.r.t. North)

    real              :: slat1, slat2
    real              :: clat1, clat2
    real              :: sdlon, cdlon
    real              :: a1,a2

    clat1 = cos(deg2rad*lat1)

    if (clat1 < 1.0E-6) then           ! start point is a pole,
      if (lat1 > 0.0) then             ! ... it is the North Pole
        tau = pi
      else                             ! ... it is the South Pole
        tau = 0.0
      end if
    else                               ! start point is not a pole
      slat1 = sin(deg2rad*lat1)
      slat2 = sin(deg2rad*lat2)
      clat1 = cos(deg2rad*lat1)
      clat2 = cos(deg2rad*lat2)
      sdlon = sin(deg2rad*(lon2 - lon1))
      cdlon = cos(deg2rad*(lon2 - lon1))

      a1 = sdlon*clat2
      a2 = clat1*slat2 - slat1*clat2*cdlon
      tau  = atan2(a1,a2)
    end if

  end subroutine get_course
  ! #]

  subroutine get_point_from_course(lat1,lon1 , tau,delta , lat2,lon2)
  ! #[
  !
  ! Get the coordinates of point 2 starting at point 1 with course tau over an
  ! angular distance d. The course is defined in radians measured clockwise with
  ! respect to the North, as given by routine get_course.
  !
  ! NOTE: this routine differs a minus sign from the original formula in
  ! Aviation Formulary V 1.46 by Ed Williams,
  ! http://williams.best.vwh.net/avform.htm#GCF.
  !
    implicit none

    real, intent(in)  :: lat1, lon1              ! longitude and latitude of point 1 (degrees)
    real, intent(in)  :: tau                     ! initial course in radians (clockwise w.r.t. North)
    real, intent(in)  :: delta                   ! distance in radians
    real, intent(out) :: lat2, lon2              ! longitude and latitude of point 2 (degrees)

    real              :: slat1, clat1
    real              :: sind, cosd
    real              :: stau, ctau
    real              :: slat2
    real              :: a1, a2
    real              :: dlon

    slat1 = sin(deg2rad*lat1)
    clat1 = cos(deg2rad*lat1)
    sind  = sin(delta)
    cosd  = cos(delta)
    stau  = sin(tau)
    ctau  = cos(tau)

    slat2 = slat1*cosd + clat1*sind*ctau

    a1 = stau*sind*clat1
    a2 = cosd - slat1*slat2
    dlon = atan2(a1,a2)

    lat2 = rad2deg * asin(slat2)
    lon2 = lon1 + rad2deg*dlon

  end subroutine get_point_from_course
  ! #]


  ! these 2 routines are taken from SDP/sws/SwsSupport.F90
  subroutine xyz2latlon(x,y,z,lat,lon)
    !  #[
    !      lat lon  of  x,y,z
    ! R = radius of the circular satellite orbit
    !     = assumed to be 1 inside this routine

    ! arguments:
    real(r_), intent(in)   :: x,y,z
    real(r_), intent(out)  :: lat,lon
    !local:
    real(r_)               :: rcos
    ! --
    rcos   = sqrt(x**2+y**2)
    if ((z == 0.0d0) .and. (rcos == 0.0d0)) then
       lat = 0.0d0
    else
       lat = atan2(z,rcos)*rad2deg
    endif
    if ((y == 0.0d0) .and. (x == 0.0d0)) then
       lon = 0.0d0
    else
       lon = atan2(y,x)*rad2deg
    endif
    !-
  end subroutine xyz2latlon
  !  #]

  subroutine latlon2xyz(lat,lon,x,y,z)
    !  #[
    !
    ! Description:
    ! R = radius of the circular satellite orbit
    !     = assumed to be 1 inside this routine
    !
    !   Hans Bonekamp  KNMI, 2003
    !
    implicit NONE
    ! arguments:
    real(r_), intent(in)   :: lat,lon
    real(r_), intent(out)  :: x,y,z
    ! local
    real(r_)               :: rlon,rlat
    real(r_)               :: coslat
    !--
    rlon   = lon*deg2rad
    rlat   = lat*deg2rad
    coslat = cos(rlat)
    x      = cos(rlon)*coslat
    y      = sin(rlon)*coslat
    z      = sin(rlat)
  end subroutine latlon2xyz
  !  #]

  ! routines to convert orbit angles to x,y,z (and from there to lat-lon)
  ! and vice versa (see notebook Jos, p. 71, for formulas)

  subroutine orbit2xyz(p,a,rot_z,x,y,z)
    !  #[
    real(r_), intent(in)  :: p,a,rot_z
    real(r_), intent(out) :: x,y,z

    ! R = radius of the circular satellite orbit
    !     = assumed to be 1 inside this routine
    ! p = the angle of the orbit plane wrt the xz plane in deg
    ! a = the angle within the orbit in deg (angle=0 equals lat=0,lon=0)
    ! rot_z = rotation angle around the z-axis
    !         (corresponding to a start lon-value, and an offset determined
    !          from the time delay*360/24

    ! local variables
    real(r_) :: tmp1_x, tmp1_y, tmp1_z, rot_x
    real(r_) :: tmp2_x, tmp2_y, tmp2_z, a_rad

    a_rad = a*deg2rad

    tmp1_x = cos(a_rad)
    tmp1_y = 0.
    tmp1_z = sin(a_rad)

    rot_x = p ! still in degrees

    call rotate_around_x(tmp1_x,tmp1_y,tmp1_z, rot_x, tmp2_x,tmp2_y,tmp2_z)

    call rotate_around_z(tmp2_x,tmp2_y,tmp2_z, rot_z, x,y,z)

  end subroutine orbit2xyz
    !  #]

  subroutine xyz2orbit(x,y,z,p,a1,rot_z1,a2,rot_z2)
    !  #[
    real(r_), intent(in)   :: x,y,z,p
    real(r_), intent(out)  :: a1,a2,rot_z1,rot_z2

    real(r_), parameter :: eps = 1.0e-3

    ! R = radius of the circular satellite orbit
    !     = assumed to be 1 inside this routine
    ! p = the angle of the orbit plane wrt the xz plane in deg
    ! a = the angle within the orbit in deg (angle=0 equals lat=0,lon=0)
    ! rot_z = rotation angle around the z-axis
    !         (corresponding to a start lon-value, and an offset determined
    !          from the time delay*360/24
    ! note: since from the x,y,z position it can not be inferred if the
    ! satellite is in the ascending or descending part of the orbit,
    ! two possible results are given for a and rot_z.

    ! local variables
    real(r_) :: p_rad, x_prime, y_prime, x_tmp, y_tmp, z_tmp, tmp
    real(r_) :: rot_z1a, rot_z1b, rot_z2a, rot_z2b, r_sq

    integer :: which_sol_a1, which_sol_a2
    logical :: debug

    ! convert input to radians
    p_rad = p*deg2rad

    ! check whether radius = 1
    r_sq = x*x+y*y+z*z
    if (abs(r_sq-1) .gt. eps) then
       print *,"ERROR inside  xyz2orbit()"
       print *,"orbit radius is assumed to be 1 in this routine,"
       print *,"but the supplied x,y,z result in a radius of: ",sqrt(r_sq)
       stop
    end if

    debug = .false.
    !debug = .true.
    !if (p .eq. 90.) debug = .true.

    ! following calculation is in radians
    if (abs(abs(p)-90.) .lt. eps) then
       ! orbit is exactly above the equator now, so z is always zero
       if (abs(z) .lt. eps) then
          a1 = asin(y)
          a2 = pi - a1
       else
          print *,"ERROR in xyz2orbit():"
          print *,"for p=90. z should always be zero, but it is not...."
       end if

       rot_z1 = 0.
       rot_z2 = 0.

       if (debug) then
          print *,"rot_z1,a1 = ",rot_z1,a1
          print *,"rot_z2,a2 = ",rot_z2,a2
       end if

    else
       a1 = asin(z/cos(p_rad))
       a2 = pi - a1

       ! first case 1
       x_prime = cos(a1) ; y_prime = sin(a1)*sin(p_rad)

       if (abs(x*x+y*y-x_prime*x_prime-y_prime*y_prime) .gt. eps) then
          print *,"x,y check problems (a1) !!!!"
          print *,"length of vector [x,y] should be identical to lenght"
          print *,"of vector [x_prime,y_prime], since they only differ"
          print *,"because of a rotation around the z-axis..."
          print *,"p=",p," a1 = ",a1*rad2deg
          print *,"x=",x," y=",y," z=",z
          print *,"x_prime=",x_prime," y_prime=",y_prime
          print *,"x*x+y*y=",x*x+y*y
          print *,"x_prime*x_prime+y_prime*y_prime=",&
               x_prime*x_prime+y_prime*y_prime
          stop
       end if

!       tmp = (x_prime*y - y_prime*x)/(x_prime*x_prime + y_prime*y_prime)
       tmp = (x_prime*x + y_prime*y)/(x_prime*x_prime + y_prime*y_prime)

       ! protect the acos() function from the numerical coincidence that
       ! tmp is a little above 1 (this does happen sometimes !!)
       if (tmp .gt.  1.) tmp =  1.
       if (tmp .lt. -1.) tmp = -1.

!       rot_z1a = asin(tmp)! possible solutions are rot_z or pi-rot_z
!       rot_z1b = pi - rot_z1a

       rot_z1a = acos(tmp)    ! possible solutions are rot_z or -1*rot_z
       rot_z1b = -1.*rot_z1a

       ! then case 2
       x_prime = cos(a2) ; y_prime = sin(a2)*sin(p_rad)

       if (abs(x*x+y*y-x_prime*x_prime-y_prime*y_prime) .gt. eps) then
          print *,"x,y check problems (a2) !!!!"
          print *,"length of vector [x,y] should be identical to lenght"
          print *,"of vector [x_prime,y_prime], since they only differ"
          print *,"because of a rotation around the z-axis..."
          print *,"p=",p," a2 = ",a2*rad2deg
          print *,"x=",x," y=",y," z=",z
          print *,"x_prime=",x_prime," y_prime=",y_prime
          print *,"x*x+y*y=",x*x+y*y
          print *,"x_prime*x_prime+y_prime*y_prime=",&
               x_prime*x_prime+y_prime*y_prime
          stop
       end if

!       tmp = (x_prime*y - y_prime*x)/(x_prime*x_prime + y_prime*y_prime)
       tmp = (x_prime*x + y_prime*y)/(x_prime*x_prime + y_prime*y_prime)

       ! protect the acos() function from the numerical coincidence that
       ! tmp is a little above 1 (this does happen sometimes !!)
       if (tmp .gt.  1.) tmp =  1.
       if (tmp .lt. -1.) tmp = -1.

!       rot_z2a = asin(tmp)! possible solutions are rot_z or pi-rot_z
!       rot_z2b = pi - rot_z2a

       rot_z2a = acos(tmp)    ! possible solutions are rot_z or -1*rot_z
       rot_z2b = -1.*rot_z2a

       ! now we have 4 possible combinations of a and rot_z
       ! so do the reverse for all of them, and see what x,y,z we find

       which_sol_a1 = -1
       which_sol_a2 = -1

       ! case 1
       call orbit2xyz(p,a1*rad2deg,rot_z1a*rad2deg,x_tmp,y_tmp,z_tmp)
       if (debug) then
          print "(a,6(f10.5,1x))","p,a1,rot_z1a,x_tmp,y_tmp,z_tmp:",&
               p,a1*rad2deg,rot_z1a*rad2deg,x_tmp,y_tmp,z_tmp
       end if

       if ( (abs(x-x_tmp) .lt. eps) .and. &
            (abs(y-y_tmp) .lt. eps) .and. &
            (abs(z-z_tmp) .lt. eps)       ) then
          if (debug) print *,"case 1 is the wanted solution..."
          which_sol_a1 = 1
          rot_z1 = rot_z1a
       end if

       ! case 2
       call orbit2xyz(p,a1*rad2deg,rot_z1b*rad2deg,x_tmp,y_tmp,z_tmp)
       if (debug) then
          print "(a,6(f10.5,1x))","p,a1,rot_z1b,x_tmp,y_tmp,z_tmp:",&
               p,a1*rad2deg,rot_z1b*rad2deg,x_tmp,y_tmp,z_tmp
       end if

       if ( (abs(x-x_tmp) .lt. eps) .and. &
            (abs(y-y_tmp) .lt. eps) .and. &
            (abs(z-z_tmp) .lt. eps)       ) then
          if (debug) print *,"case 2 is the wanted solution..."
          if (which_sol_a1 .ne. -1) then
             !print *,"multiple solutions found !!!!!!! using the first one."
          else
             which_sol_a1 = 2
             rot_z1 = rot_z1b
          end if
       end if

       if (which_sol_a1 .eq. -1) then
          ! this should never happenn
          print *,"no valid solution found for a1=",a1*rad2deg
          stop
       endif

       ! case 3
       call orbit2xyz(p,a2*rad2deg,rot_z2a*rad2deg,x_tmp,y_tmp,z_tmp)
       if (debug) then
          print "(a,6(f10.5,1x))","p,a2,rot_z2a,x_tmp,y_tmp,z_tmp:",&
               p,a2*rad2deg,rot_z2a*rad2deg,x_tmp,y_tmp,z_tmp
       end if

       if ( (abs(x-x_tmp) .lt. eps) .and. &
            (abs(y-y_tmp) .lt. eps) .and. &
            (abs(z-z_tmp) .lt. eps)       ) then
          if (debug) print *,"case 3 is the wanted solution..."
          which_sol_a2 = 3
          rot_z2 = rot_z2a
       end if

       ! case 4
       call orbit2xyz(p,a2*rad2deg,rot_z2b*rad2deg,x_tmp,y_tmp,z_tmp)
       if (debug) then
          print "(a,6(f10.5,1x))","p,a2,rot_z2b,x_tmp,y_tmp,z_tmp:",&
               p,a2*rad2deg,rot_z2b*rad2deg,x_tmp,y_tmp,z_tmp
       end if

       if ( (abs(x-x_tmp) .lt. eps) .and. &
            (abs(y-y_tmp) .lt. eps) .and. &
            (abs(z-z_tmp) .lt. eps)       ) then
          if (debug) print *,"case 4 is the wanted solution..."
          if (which_sol_a2 .ne. -1) then
             !print *,"multiple solutions found !!!!!!! using the first one"
          else
             which_sol_a2 = 4
             rot_z2 = rot_z2b
          end if
       end if

       if (which_sol_a2 .eq. -1) then
          ! this should never happen
          print *,"no valid solution found for a2=",a2*rad2deg
          stop
       endif

    end if

    ! convert output to degrees
    a1 = a1*rad2deg ; rot_z1  = rot_z1 *rad2deg
    a2 = a2*rad2deg ; rot_z2  = rot_z2 *rad2deg

  end subroutine xyz2orbit
    !  #]
  subroutine rotate_around_x(x,y,z,rotx,xnew,ynew,znew)
    !  #[
    ! apply rotation around x-axis to orbit
    ! angle is given in degrees
    real(r_), intent(in)  :: x,y,z,rotx
    real(r_), intent(out) :: xnew,ynew,znew

    xnew = x ! no change when rotation around x
    ynew = z*sin(rotx*deg2rad)+y*cos(rotx*deg2rad)
    znew = z*cos(rotx*deg2rad)-y*sin(rotx*deg2rad)
  end subroutine rotate_around_x
    !  #]

  subroutine rotate_around_z(x,y,z,rotz,xnew,ynew,znew)
    !  #[
    ! apply rotation around z-axis to orbit
    ! angle is given in degrees
    real(r_), intent(in)  :: x,y,z,rotz
    real(r_), intent(out) :: xnew,ynew,znew

    xnew = x*cos(rotz*deg2rad)-y*sin(rotz*deg2rad)
    ynew = x*sin(rotz*deg2rad)+y*cos(rotz*deg2rad)
    znew = z ! no change when rotation around z
  end subroutine rotate_around_z
    !  #]

  subroutine WVC_Orientation(Lam1,Phi1,Lam2,Phi2,Alfa1,Alfa2)
    !  #[
!------------------------------------------------------------------------------!
! WVC_Orientation calculates the orientations of two WVC's with coordinates    !
! (Lam1,Phi1) and (Lam2,Phi2) in the same row, using a great spherical         !
! triangle with the North Pole as third point. If Lam1 = Lam2 the orientation  !
! is set to -90 degrees (i.e. pointing to the west).                           !
! The orientation is given as the angle in degrees of the local along-track    !
! direction measured counterclockwise from the north.                          !
! This routine is needed by 2DVAR for the transformation to diagonalize the    !
! error covariance matrices in the spatial frequency domain.                   !
!------------------------------------------------------------------------------!
  implicit none

  real, intent(in)   :: Lam1,Phi1      ! Geographical coordinates of WVC1 (deg)
  real, intent(in)   :: Lam2,Phi2      ! Geographical coordinates of WVC2 (deg)
  real, intent(out)  :: Alfa1,Alfa2    ! WVC orientations (deg)
!------------------------------------------------------------------------------!
! Local declarations.                                                          !
!------------------------------------------------------------------------------!
  real     :: SinPhi1, CosPhi1         ! Sines and cosines
  real     :: SinPhi2, CosPhi2
  real     :: dLam                     ! Difference in longitude,
  real     :: CosdLam, SinDLam         ! with its sine and cosine
  real     :: CosC, SinC_inv           ! Cosine of angle C (North Pole) and inverse of its sine
  real     :: Aux
!------------------------------------------------------------------------------!
! Initialize. Check positions.                                                 !
!------------------------------------------------------------------------------!
  Alfa1=missing_indicator_real
  Alfa2=missing_indicator_real

  if (missing_real(Lam1) .or. missing_real(Phi1)) return
  if (missing_real(Lam2) .or. missing_real(Phi2)) return
!------------------------------------------------------------------------------!
! Set angles, sines and cosines. Calculate orientations.                       !
!------------------------------------------------------------------------------!
  DLam=Lam2 - Lam1
  if (DLam == 0.0) then                ! Both points have same longitude
    Alfa1=90.0
    Alfa2=90.0
  else                                 ! Both points have different longitude
    SinPhi1=sin(Deg2Rad*Phi1)
    CosPhi1=cos(Deg2Rad*Phi1)
    SinPhi2=sin(Deg2Rad*Phi2)
    CosPhi2=cos(Deg2Rad*Phi2)

    SinDLam=sin(Deg2Rad*DLam)
    CosDLam=cos(Deg2Rad*DLam)

    CosC=SinPhi1*SinPhi2 + CosPhi1*CosPhi2*CosDLam
    SinC_inv=1.0/sqrt(1.0 - CosC*CosC)

    Aux = CosPhi2*SinDLam*SinC_inv
    if (Aux .ge.  1.0) Aux =  1.0
    if (Aux .le. -1.0) Aux = -1.0
    Alfa1=Rad2Deg*acos(Aux)

    Aux = CosPhi1*SinDLam*SinC_inv
    if (Aux .ge.  1.0) Aux =  1.0
    if (Aux .le. -1.0) Aux = -1.0
    Alfa2=Rad2Deg*acos(Aux)
  endif

  ! The acos function used above always yields angles between 0 and 180 degrees.
  ! This is valid for sun-synchronous polar orbiters but not for platforms
  ! moving from left to right over the world map, like ISS.
  ! Correct for this when necessary.
  if (Phi1 > Phi2) then       ! Lat of point 1 is larger than lat of point 2
    Alfa1=360.0-Alfa1
    Alfa2=360.0-Alfa2
  endif

  end subroutine WVC_Orientation
    !  #]

  function my_arctan(x,y) result(angle)
    !  #[
    ! a small support function, used by the above uv_to_dir conversion
    ! a function to calculate my own version of atan()
    ! making sure the result is between 0-360 (output in deg)

    real(r_) :: x,y   ! input
    real(r_) :: angle ! output

    IF (x .gt. 0.0) THEN
      IF (y .ge. 0.0) THEN
        angle = atan(y/x)*rad2deg
      ELSE
        angle = 360.0-atan(abs(y/x))*rad2deg
      END IF
    ELSE IF (x .lt. 0.0) THEN
      IF (y .ge. 0.0) THEN
        angle = 180.0-atan(abs(y/x))*rad2deg
      ELSE
        angle = 180.0+atan(abs(y/x))*rad2deg
      END IF
    ELSE  ! uitzondering, anders veroorzaakt x=0 delen door 0
      IF (y .ge. 0.0) THEN
        angle = 90.0
      ELSE
        angle = 270.0
      END IF
    END IF

  end function my_arctan
  !  #]

  function my_arctan_rad(x,y) result(angle)
    !  #[
    ! a similar function in radians (not used by me at the moment)
    ! a function to calculate my own version of atan()
    ! making sure the result is between 0-2*pi (output in radians)

    real(r_) :: x,y   ! input
    real(r_) :: angle ! output

    IF (x .gt. 0.0) THEN
      IF (y .ge. 0.0) THEN
        angle = atan(y/x)
      ELSE
        angle = two_pi-atan(abs(y/x))
      END IF
    ELSE IF (x .lt. 0.0) THEN
      IF (y .ge. 0.0) THEN
        angle = pi-atan(abs(y/x))
      ELSE
        angle = pi+atan(abs(y/x))
      END IF
    ELSE  ! uitzondering, anders veroorzaakt x=0 delen door 0
      IF (y .ge. 0.0) THEN
        angle = 0.5*pi
      ELSE
        angle = 1.5*pi
      END IF
    END IF

  end function my_arctan_rad
  !  #]

  function relative_angle(windDir, antennaDir, &
       windConvention, antennaConvention) result(relWinddir)
    !  #[ Calculate the relative wind direction
    !**********************************************************************
    ! Calculate the relative angle between two directions.
    ! E.g. The wind direction and the antenna upwind azimuth angle.
    !
    ! Input   :
    !   windDir                      - wind direction (deg)
    !   antennaDir                   - antenna azimuth angle (deg)
    !   windConvention (optional)    - convention used for windDir
    !   antennaConvention (optional) - convention used for antennaDir
    !
    ! Output  :
    !   relWindir  - Wind direction angle difference
    !                in interval [0..360).
    !
    ! Note :
    ! 1. ERS/ASCAT:
    !     windDir    : meteorological convention for BUFR files,
    !                  oceanographic convention for netCDF files
    !     antennaDir : meteorological convention
    !     GMF        : relative to upwind antenna direction. Add 180. deg
    !                  to the antenna azimuth direction!
    !    OCEANSAT/SEAWINDS:
    !     windDir    : meteorological convention for BUFR files,
    !                  oceanographic convention for netCDF files
    !     antennaDir : oceanographic convention
    !     GMF        : relative to upwind antenna direction. Add 180. deg
    !                  to the antenna azimuth direction!
    !**********************************************************************

    !use inversion, only: angle_convention_meteorological ! = 1
    !use inversion, only: angle_convention_oceanographic  ! = 2

    real, intent(in)              :: windDir, antennaDir
    integer, intent(in), optional :: windConvention, antennaConvention
    real                          :: relWinddir ! result

    relWinddir = missing_indicator_real
    if (missing_real(windDir)) return
    if (missing_real(antennaDir)) return

    relWinddir =  modulo(windDir-antennaDir, 360.)
    if (present(windConvention)) then
      if (present(antennaConvention)) then
        if (windConvention /= antennaConvention) then
          relWinddir =  modulo(windDir-antennaDir+180., 360.)
        endif
      endif
    endif

    ! The modulo function returns values between [0..360]
    ! instead of between [0..360) due to machine roundoffs of reals:
    if (relWinddir >= 360.) relWinddir = relWinddir-360.
  end function relative_angle
  !  #]

  function trnsfrm(a, trans) result(a_tr)
    !  #[
    !**********************************************************************
    ! Transformation between linear, logarithmic and z-space.
    !
    ! Input:
    !  a     -
    !  trans - 'lin_2_lin', 'lin_2_log', 'lin_2_z',
    !          'log_2_lin', 'log_2_log', 'log_2_z',
    !          'z_2_lin'  , 'z_2_log'  , 'z_2_z'
    !
    ! Output:
    !  a_tr  - a transformed
    !**********************************************************************

    real, intent(in) :: a
    character(len=*), intent(in) :: trans
    real :: a_tr ! result

    ! Locals:
    real, parameter :: minimum_db_value = -40 ! dB
    real :: a_dum

    a_tr = a

    if (missing_real(a)) then
      return
    endif

    select case(trim(trans))
    case('lin_2_lin')
      !a_tr = a
    case('lin_2_log')
      if (a /= 0.) then
        if (a > 0.) then
          a_tr = 10.*log10(a)
        elseif (a < 0.) then
          a_tr = minimum_db_value
        endif
      else ! a == 0.
        ! a_tr = 0.
      endif
    case('lin_2_z')
      if (a >= 0.) then
        a_tr = a**0.625
      else
        a_tr = -(-a)**0.625
      endif
    case('log_2_lin')
      a_tr = 10.**(a/10.)
    case('log_2_log')
      !a_tr = a
    case('log_2_z')
      a_dum = 10.**(a/10.)
      a_tr = a_dum**0.625
    case('z_2_lin')
      if (a >= 0.) then
        a_tr = a**1.6
      else
        a_tr = -(-a)**1.6
      endif
    case('z_2_log')
      if (a < 0.) then
        a_tr = minimum_db_value
      elseif (a > 0.) then
        a_dum = a**1.6
        a_tr = 10.*log10(a_dum)
      else ! (a == 0.)
        !a_tr = 0.
      endif
    case('z_2_z')
      !a_tr = a
    case default
      write(6, *) 'trnsfrm ERROR - Unknown transformation type.'
      write(6, *) ' trans: ', trans
      write(6, *)
    end select

  end function trnsfrm
    !  #]

  function logDiff(a, b) result(c)
    !  #[
    !**********************************************************************
    ! Calculate the difference of two linear values a and b,
    ! being the log of the quotient of a and b.
    !
    ! Input   : a, b - input linear values
    !
    ! Output  :
    !           c    - output log value =sign(a/b)*10*log10(abs(a/b))
    !**********************************************************************

    real, intent(in) :: a, b ! linear
    real :: c ! result in log space

    c = 0.
    if (b /= 0.) then
      c = a/b
    endif
    if (c /= 0.) then
      if (c > 0.) then
        c = 10.*log10(c)
      elseif (c < 0.) then
        c = -10.*log10(-c)
      else ! c == 0.
        ! c = 0.
      endif
    endif

  end function logDiff
    !  #]

  subroutine extend_wvcs(lat1,lon1,lat2,lon2,latnew,lonnew)
    !  #[
    !
    ! Description: given the coordinates of two WVCs (1 and 2),
    !   the coordinates of a third WVC (new) that is located along the
    !   connecting line between 1 and 2 and with the same distance to 2
    !   as the distance between 1 and 2
    !
    !   x-----x-----x
    !   1     2    new
    !
    !   This routine can be used to extend the swath.
    !   The approach here is rather symplistic, but it appears that
    !   the position error is less than 1.5 km, even near the poles.
    !
    !   Anton Verhoef, KNMI 2010
    !
    ! arguments
    real(r_), intent(in)   :: lat1
    real(r_), intent(in)   :: lon1
    real(r_), intent(in)   :: lat2
    real(r_), intent(in)   :: lon2
    real(r_), intent(out)  :: latnew
    real(r_), intent(out)  :: lonnew

    real(r_) :: lon2tmp

    latnew = 2.0 * lat2 - lat1

    lon2tmp = lon2
    if (lon2tmp - lon1 .gt.  180.0) lon2tmp = lon2tmp - 360.0
    if (lon2tmp - lon1 .lt. -180.0) lon2tmp = lon2tmp + 360.0
    lonnew = 2.0 * lon2tmp - lon1
    lonnew = mod(lonnew+3600.0,360.0)

  end subroutine extend_wvcs
    !  #]

  function average_2_angles(dir1, dir2) result(avg_dir)
    !  #[
    ! A simple function to average 2 angels (in degrees)
    ! taking into account the circular property of angles
    ! (i.e. 359 and 1 are close together)
    !
    ! Results are wrapped to the range 0-360 degrees

    ! interface
    real, intent(in)  :: dir1, dir2
    real              :: avg_dir ! result

    ! local variables
    real :: angle_diff

    ! take difference between both direction
    angle_diff = dir2-dir1

    ! map to range -180 upto +180
    angle_diff = modulo(angle_diff+180., 360.) - 180.

    ! add half the difference to dir1
    avg_dir = dir1 + 0.5*angle_diff

    ! remap to the range 0 - 360
    avg_dir = modulo(avg_dir, 360.)

  end function average_2_angles
    !  #]

  subroutine compute_orbit_elements(x_pos,y_pos,z_pos, &
                                    x_vel,y_vel,z_vel, &
                                    orb_inclination,orb_smaj_axis,orb_eccen, &
                                    long_asc_node,arg_lat,nodal_period)
    !  #[
    !
    ! Routine to compute the six orbit elements from the satellite
    ! position vector (in km) and velocity vector (in km/s)
    !
    ! Acknowledgement: this software is to a great extent based on software which
    ! was kindly provided by the NOAA Ocean Winds Science Team
    ! (NOAA/NESDIS/Center for Satellite Applications and Research).

    ! arguments
    real,intent(in)  :: x_pos,y_pos,z_pos
    real,intent(in)  :: x_vel,y_vel,z_vel
    real,intent(out) :: orb_inclination,orb_smaj_axis,orb_eccen
    real,intent(out) :: long_asc_node,arg_lat,nodal_period

    ! parameters
    real(r8_),parameter :: twopi = 6.283185307179586     ! 2*pi
    real(r8_),parameter :: rad2deg = 57.29577951308233   ! conversion from radians to degrees
    real(r8_),parameter :: mu = 398600.5                 ! Earth gravitational param G*M (km^3/s^2)
    real(r8_),parameter :: j2 = 0.0010826                ! Earth gravitational harmonic
    real(r8_),parameter :: aearth2 = 40680622.66137769   ! square of Earth equatorial radius (km^2)

    ! local variables
    real :: ux,uy,uz
    real :: wx,wy,wz
    real :: rr,rv,vv,r0
    real :: se,ce,param
    real :: sini,cosi
    real :: ree2,cosi2

    rr = x_pos * x_pos + y_pos * y_pos + z_pos * z_pos
    rv = x_pos * x_vel + y_pos * y_vel + z_pos * z_vel
    vv = x_vel * x_vel + y_vel * y_vel + z_vel * z_vel

    ! semi-major axis in km
    r0 = sqrt(rr)
    orb_smaj_axis = 1.0 / real(2.0 / r0 - vv / mu)

    ! eccentricity
    ce = 1.0 - r0 / orb_smaj_axis
    se = real(rv / sqrt(mu * orb_smaj_axis))
    orb_eccen = sqrt(se * se + ce * ce)

    ! inclination
    param = orb_smaj_axis * (1.0 - orb_eccen * orb_eccen)
    ux   = x_pos / r0
    uy   = y_pos / r0
    uz   = z_pos / r0
    wx   = real((rr * x_vel - rv * x_pos) / (r0 * sqrt(mu * param)))
    wy   = real((rr * y_vel - rv * y_pos) / (r0 * sqrt(mu * param)))
    wz   = real((rr * z_vel - rv * z_pos) / (r0 * sqrt(mu * param)))
    sini = sqrt(uz * uz + wz * wz)
    cosi = sqrt((ux + wy)**2 + (uy - wx)**2) - 1.0

    orb_inclination = atan2(sini,cosi)

    ! longitude of ascending node
    long_asc_node = atan2((uy * wz - wy * uz),(ux * wz - wx * uz))

    ! argument of latitude
    arg_lat = atan2(uz,wz)

    ! compute nodal period
    ree2 = 1.0 - orb_eccen*orb_eccen
    cosi2 = real(cos(orb_inclination)**2)
    nodal_period = real(                               &
     (twopi*sqrt(orb_smaj_axis**3/mu))*                &
      (1.+(0.750*j2*aearth2/(orb_smaj_axis**2))*       &
       ((1.0-3.0*cosi2)*sqrt(ree2)+                    &
        (1.0-5.0*cosi2))/(ree2*ree2) )                 &
    )

    ! convert radians to degrees and bring angles between 0 and 360
    orb_inclination = real(orb_inclination * rad2deg)

    long_asc_node = real(long_asc_node * rad2deg)
    if (long_asc_node .lt. 0.0) long_asc_node = long_asc_node + 360.0

    arg_lat = real(arg_lat * rad2deg)
    if (arg_lat .lt. 0.0) arg_lat = arg_lat + 360.0

  end subroutine compute_orbit_elements
    !  #]

  subroutine compute_row_cell_number(orb_inclination,orb_smaj_axis,orb_eccen, &
                                     long_asc_node,arg_lat,nodal_period, &
                                     lat,lon, &
                                     wvc_size,lgrid,wgrid, &
                                     wvc_i,wvc_j)
    !  #[
    !
    ! Routine to compute the along track row number wvc_i and across track cell
    ! number wvc_j from the orbit elements and a given lat and lon
    !
    ! Auxilliary parameters are wvc_size in meters, lgrid (number of grid points along track
    ! in an orbit) and wgrid (number of grid points across track)
    !
    ! wvc_i (row number) starts counting at the south pole
    ! wvc_j (cell number) starts counting at the left hand side of the swath as seen
    ! in flight direction
    !
    ! Acknowledgement: this software is to a great extent based on software which
    ! was kindly provided by the NOAA Ocean Winds Science Team
    ! (NOAA/NESDIS/Center for Satellite Applications and Research).

    ! arguments
    real,intent(in)     :: orb_inclination,orb_smaj_axis,orb_eccen
    real,intent(in)     :: long_asc_node,arg_lat,nodal_period
    real,intent(in)     :: lat,lon
    integer,intent(in)  :: wvc_size,lgrid,wgrid
    integer,intent(out) :: wvc_i,wvc_j

    ! parameters
    real(r8_),parameter :: pi = 3.141592653589793        ! pi
    real(r8_),parameter :: twopi = 6.283185307179586     ! 2*pi
    real(r8_),parameter :: halfpi = 1.570796326794897    ! 0.5*pi
    real(r8_),parameter :: deg2rad = 0.0174532925199433  ! conversion from degrees to radians
    real(r8_),parameter :: j2 = 0.0010826                ! Earth gravitational harmonic
    real(r8_),parameter :: ae = 6378.1363                ! Earth equatorial radius (km)
    real(r8_),parameter :: ee = 6.6943849995879496e-3    ! ellipticity of Earth ellipsoid^2
    real(r8_),parameter :: roterth = 4.37527e-3

    ! local variables
    integer   :: iteration_count
    real(r8_) :: sinlat,coslat,tanlat
    real(r8_) :: diff,lip
    real(r8_) :: sininc,cosinc
    real(r8_) :: slr
    real(r8_) :: prat,pnode,phipi
    real(r8_) :: arglmod
    real(r8_) :: lit0
    real(r8_) :: sinlit,coslit,sinlit0,coslit0
    real(r8_) :: d,ddif
    real(r8_) :: lip1,lit
    real(r8_) :: sphipi
    logical   :: ascend

    ! some common conversions of orbit element variables
    sininc = sin(orb_inclination * deg2rad)
    cosinc = cos(orb_inclination * deg2rad)

!   arglmod = arg_lat * deg2rad - halfpi  ! to start counting at the north pole
    arglmod = arg_lat * deg2rad + halfpi  ! to start counting at the south pole
    if (arglmod .gt. twopi) arglmod = arglmod - twopi

    ! compute orbit & nodal precession periods
    slr   = orb_smaj_axis / ae * (1.0 - orb_eccen * orb_eccen)
    pnode = -1.5 * twopi * j2 * cosinc / (nodal_period / 60.0 * slr * slr)
    prat  = nodal_period / 60.0 * (roterth - pnode) / twopi

    ! compute bin coordinates
    sinlat  = sin(lat*deg2rad)
    coslat  = cos(lat*deg2rad)
    tanlat  = sinlat/coslat
    lit0    = (lon - long_asc_node) * deg2rad
    sinlit0 = sin(lit0)
    coslit0 = cos(lit0)

    ascend  = .true.
    if (coslit0 .lt. 0.0) ascend = .false.

    lip = atan((cosinc * sinlit0 + (1.0 - ee) * sininc * tanlat) / coslit0)
    if (.not. ascend)  lip = lip + pi
    if (lip .lt. 0.0)  lip = lip + twopi
!   lip = lip  - halfpi  ! to start counting at the north pole
    lip = lip  + halfpi  ! to start counting at the south pole
    if (lip .gt. twopi) lip = lip - twopi
    if (lip .lt. 0.0)   lip = lip + twopi

    ! begin iteration to get along-track longitude
    diff = 1.0  ! to start the iteration loop
    iteration_count = 0

    do while (diff .ge. 1.E-5 .and. iteration_count .lt. 10)

      diff = lip - arglmod
      d    = pi - abs(pi - abs(diff))
      ddif = diff
      if (diff .lt. -pi) then
        ddif = d
      endif
      if (diff .gt. pi) then
        ddif = -d
      endif

      lit    = (lon - long_asc_node) * deg2rad + prat * ddif
      sinlit = sin(lit)
      coslit = cos(lit)

      ascend = .true.
      if (coslit .lt. 0.0) ascend = .false.

      lip1 = atan((cosinc * sinlit + (1.0 - ee) * sininc * tanlat) / coslit)
      if (.not. ascend)  lip1 = lip1 + pi
      if (lip1 .lt. 0.0) lip1 = lip1 + twopi
!     lip1 = lip1 - halfpi  ! to start counting at the north pole
      lip1 = lip1 + halfpi  ! to start counting at the south pole
      if (lip1 .gt. twopi) lip1 = lip1 - twopi
      if (lip1 .lt. 0.0)   lip1 = lip1 + twopi
      diff = abs(lip - lip1)
      lip  = lip1

      iteration_count = iteration_count + 1

    enddo

    ! along track row number: wvc_i
    wvc_i = nint(dble(lgrid) * lip / twopi + 0.5)
    if (wvc_i .gt. lgrid) wvc_i = wvc_i - lgrid

    ! across track cell number: wvc_j
    sphipi = (1.0 - ee) * cosinc * sinlat - sininc * coslat * sinlit
    sphipi = sphipi / sqrt(1.0 - ee * sinlat * sinlat)
    phipi  = asin(sphipi)

    wvc_j  = nint(phipi * ae / dble(wvc_size * 0.001) - 0.5)
    wvc_j  = (wgrid / 2) - wvc_j

    ! return wvc number -2 if the WVC is very far out of the swath, the rest of the scan will be rejected
    ! return wvc number -1 if the WVC is far out of the swath, this observation will be rejected
    if (wvc_j .lt. -2 .or. wvc_j .gt. wgrid + 3) then
      wvc_j = -2
    elseif (wvc_j .lt. 0 .or. wvc_j .gt. wgrid + 1) then
      wvc_j = -1
    else
      if (wvc_j .lt. 1)     wvc_j = 1
      if (wvc_j .gt. wgrid) wvc_j = wgrid
    endif

  end subroutine compute_row_cell_number
    !  #]

end module convert
