program bufrdecode
!
! program to read a BUFR input file given as argument and to write its contents to
! the screen
! based on ecCodes, mimics the old program based on BUFRDC
!
! Anton Verhoef, April 2020
!
  use Compiler_Features,only: getarg_genscat,iargc_genscat
  use eccodes

  implicit none

  integer            :: ifile,ibufr,iret
  integer            :: imsg,isub,ides,irep,ival
  integer            :: nargs,nmsg
  integer            :: totalLength,edition,masterTableNumber
  integer            :: bufrHeaderCentre,bufrHeaderSubCentre,updateSequenceNumber,section1Flags
  integer            :: dataCategory,internationalDataSubCategory,dataSubCategory
  integer            :: masterTablesVersionNumber,localTablesVersionNumber
  integer            :: typicalYear,typicalMonth,typicalDay,typicalHour,typicalMinute,typicalSecond
  integer            :: numberOfSubsets,section3Flags,compressedData
  integer            :: replications,start_dr,length_dr
  character(len=256) :: filename
  integer(kind=4),dimension(:),allocatable   :: descriptors,unexp_descr
  real(kind=8),dimension(:),allocatable      :: values
  character(len=64),dimension(:),allocatable :: names
  character(len=24),dimension(:),allocatable :: units
  character(len=64),dimension(:),allocatable :: abbreviations
  character(len=80),dimension(:),allocatable :: cvals_tmp
  character(len=20),dimension(:),allocatable :: cvals

  101 format(A,I10)         ! for file and header information
  102 format(I6,I8.6,2X,A)  ! for data descriptors

  ! get command line arguments
  nargs = iargc_genscat()
  if (nargs .eq. 0) then
     write(*,'(A)') "Usage: bufrdecode <BUFR input-file>"
     stop
  endif
  call getarg_genscat(1,filename)

  ! open input file
  call codes_open_file(ifile,filename,'r')

  ! get number of messages in file
  call codes_count_in_file(ifile,nmsg,iret)
  if (iret .ne. 0) then
    nmsg = 9999
  endif

  ! write file information
  write(*,*) ""
  write(*,*) "       FILE INFORMATION"
  write(*,*) ""
  if (iret .ne. 0) then
    write(*,*) "ERROR: NUMBER OF BUFR MESSAGES IN FILE COULD NOT BE OBTAINED"
  else
    write(*,101) " NUMBER OF BUFR MESSAGES IN FILE     ",nmsg
  endif

  ! loop over the messages
  do imsg = 1,nmsg

    ! get the message handle ibufr and unpack the message
    call codes_bufr_new_from_file(ifile,ibufr,iret)
    if (iret .eq. CODES_END_OF_FILE) then
      write(*,*) "ERROR: UNEXPECTED END OF FILE ENCOUNTERED"
      exit
    endif
    call codes_set(ibufr,'skipExtraKeyAttributes',1)  ! this should enhance the performance
    call codes_set(ibufr,'unpack',1)

    ! read header information
    ! not all information is available in older BUFR edition numbers, print these fields as '-9999'
    call codes_get(ibufr,'totalLength',totalLength)
    call codes_get(ibufr,'edition',edition,iret)
    if (iret .ne. 0) edition = -9999
    call codes_get(ibufr,'masterTableNumber',masterTableNumber,iret)
    if (iret .ne. 0) masterTableNumber = -9999
    call codes_get(ibufr,'bufrHeaderCentre',bufrHeaderCentre,iret)
    if (iret .ne. 0) bufrHeaderCentre = -9999
    call codes_get(ibufr,'bufrHeaderSubCentre',bufrHeaderSubCentre,iret)
    if (iret .ne. 0) bufrHeaderSubCentre = -9999
    call codes_get(ibufr,'updateSequenceNumber',updateSequenceNumber,iret)
    if (iret .ne. 0) updateSequenceNumber = -9999
    call codes_get(ibufr,'section1Flags',section1Flags,iret)
    if (iret .ne. 0) section1Flags = -9999
    call codes_get(ibufr,'dataCategory',dataCategory,iret)
    if (iret .ne. 0) dataCategory = -9999
    call codes_get(ibufr,'internationalDataSubCategory',internationalDataSubCategory,iret)
    if (iret .ne. 0) internationalDataSubCategory = -9999
    call codes_get(ibufr,'dataSubCategory',dataSubCategory,iret)
    if (iret .ne. 0) dataSubCategory = -9999
    call codes_get(ibufr,'masterTablesVersionNumber',masterTablesVersionNumber,iret)
    if (iret .ne. 0) masterTablesVersionNumber = -9999
    call codes_get(ibufr,'localTablesVersionNumber',localTablesVersionNumber,iret)
    if (iret .ne. 0) localTablesVersionNumber = -9999
    call codes_get(ibufr,'typicalYear',typicalYear,iret)
    if (iret .ne. 0) typicalYear = -9999
    call codes_get(ibufr,'typicalMonth',typicalMonth,iret)
    if (iret .ne. 0) typicalMonth = -9999
    call codes_get(ibufr,'typicalDay',typicalDay,iret)
    if (iret .ne. 0) typicalDay = -9999
    call codes_get(ibufr,'typicalHour',typicalHour,iret)
    if (iret .ne. 0) typicalHour = -9999
    call codes_get(ibufr,'typicalMinute',typicalMinute,iret)
    if (iret .ne. 0) typicalMinute = -9999
    call codes_get(ibufr,'typicalSecond',typicalSecond,iret)
    if (iret .ne. 0) typicalSecond = -9999
    call codes_get(ibufr,'numberOfSubsets',numberOfSubsets,iret)
    call codes_get(ibufr,'section3Flags',section3Flags,iret)
    call codes_get(ibufr,'compressedData',compressedData,iret)

    ! write header information
    write(*,*) ""
    write(*,*) ""
    write(*,*) "       HEADER INFORMATION"
    write(*,*) ""
    write(*,101) " TOTAL LENGTH OF BUFR MESSAGE (BYTES)",totalLength
    write(*,101) " BUFR EDITION NUMBER                 ",edition
    write(*,101) " BUFR MASTER TABLE                   ",masterTableNumber
    write(*,101) " ORIGINATING CENTRE                  ",bufrHeaderCentre
    write(*,101) " ORIGINATING SUB-CENTRE              ",bufrHeaderSubCentre
    write(*,101) " UPDATE SEQUENCE NUMBER              ",updateSequenceNumber
    write(*,101) " FLAG (PRESENCE OF SECTION 2)        ",section1Flags
    write(*,101) " DATA CATEGORY                       ",dataCategory
    write(*,101) " DATA SUB-CATEGORY                   ",internationalDataSubCategory
    write(*,101) " LOCAL DATA SUB-CATEGORY             ",dataSubCategory
    write(*,101) " VERSION NUMBER OF MASTER TABLE      ",masterTablesVersionNumber
    write(*,101) " VERSION NUMBER OF LOCAL TABLE       ",localTablesVersionNumber
    write(*,101) " YEAR                                ",typicalYear
    write(*,101) " MONTH                               ",typicalMonth
    write(*,101) " DAY                                 ",typicalDay
    write(*,101) " HOUR                                ",typicalHour
    write(*,101) " MINUTE                              ",typicalMinute
    write(*,101) " SECOND                              ",typicalSecond
    write(*,101) " NUMBER OF DATA SUBSETS              ",numberOfSubsets
    write(*,101) " FLAG (DATA TYPE/DATA COMPRESSION)   ",section3Flags
    write(*,101) " DATA COMPRESSION                    ",compressedData

    ! read unexpanded data descriptor information
    call codes_get(ibufr,'unexpandedDescriptors',unexp_descr)

    ! write unexpanded data descriptor information
    write(*,*) ""
    write(*,*) ""
    write(*,*) "       DATA DESCRIPTORS (UNEXPANDED)"
    write(*,*) ""
    do ides = 1,size(unexp_descr)
      write(*,102) ides,unexp_descr(ides)
    enddo

    ! read expanded data descriptor information
    call codes_get(ibufr,'expandedDescriptors',descriptors)
    allocate(abbreviations(size(descriptors)))
    allocate(names(size(descriptors)))
    allocate(units(size(descriptors)))
    call codes_get_string_array(ibufr,"expandedNames",names)
    call codes_get_string_array(ibufr,"expandedUnits",units)
    call codes_get_string_array(ibufr,"expandedAbbreviations",abbreviations)

    ! handle delayed data descriptors if present
    ! look for the start of the part containing delayed descriptors
    replications = 0
    start_dr     = size(descriptors)
    length_dr    = 0
    call codes_get(ibufr,"delayedDescriptorReplicationFactor",replications,iret)
    if (iret .ne. 0) then
      call codes_get(ibufr,"shortDelayedDescriptorReplicationFactor",replications,iret)
    endif
    if (iret .eq. 0) then
      do ides = 1,size(descriptors)

        ! keep the start and lenght of the delayed descriptors part
        if (trim(abbreviations(ides)) .eq. "delayedDescriptorReplicationFactor" .or. &
            trim(abbreviations(ides)) .eq. "shortDelayedDescriptorReplicationFactor") then
          start_dr  = ides - 1
          length_dr = size(descriptors) - ides
          exit
        endif
      enddo
    endif

    ! get the expanded data values
    if (allocated(values)) deallocate(values)
    call codes_get(ibufr,'numericValues',values)

    ! if there are multiple delayed replications only the first part will be shown
    if (start_dr+(replications*length_dr) .ne. size(values)/numberOfSubsets) then
      write(*,*) ""
      write(*,*) "       MESSAGE SEEMS TO CONTAIN MULTIPLE DELAYED REPLICATIONS"
      write(*,*) "       ONLY THE PART UNTIL THE FIRST REPLICATION IS SHOWN"
      length_dr = 1
      replications = 0
    endif

    ! allocate array for character data
    if (allocated(cvals)) deallocate(cvals)
    allocate(cvals(size(values)))
    cvals = ""

    ! write expanded data descriptor information
    write(*,*) ""
    write(*,*) ""
    write(*,*) "       DATA DESCRIPTORS (EXPANDED)"
    write(*,*) ""
    if (length_dr .gt. 0) then

      ! delayed descriptors present, print fixed part, info about delayed descriptors,
      ! and repeat the delayed descriptor part as appropriate
      do ides = 1,start_dr-1
        write(*,102) ides,descriptors(ides),names(ides)
      enddo
      write(*,102) start_dr,descriptors(start_dr+1),names(start_dr+1)
      do irep = 1,replications
        do ides = 1,length_dr
          write(*,102) start_dr+(irep-1)*length_dr+ides,descriptors(start_dr+ides+1),names(start_dr+ides+1)
        enddo
      enddo
    else

      ! no delayed descriptors present, simply print the list of descriptors
      do ides = 1,size(descriptors)
        write(*,102) ides,descriptors(ides),names(ides)
      enddo

      ! character fields have unit 'CCITT IA5', if this unit is found then
      ! read the character values and copy them to the cvals array
      ! note that this is only done when no delayed descriptors are present, for simplicity
      ! note that only the first 20 characters of the data are kept, to save memory
      if (allocated(cvals_tmp)) deallocate(cvals_tmp)
      allocate(cvals_tmp(numberOfSubsets))
      do ides = 1,size(descriptors)
        if (units(ides)(1:5) .eq. 'CCITT') then
          call codes_get_string_array(ibufr,abbreviations(ides),cvals_tmp)
          do isub = 1,numberOfSubsets
            cvals((isub-1)*size(descriptors)+ides) = cvals_tmp(isub)(1:20)
          enddo
        endif
      enddo

    endif

    ! loop over subsets and write data values
    ! ival is the pointer to the values array and runs from 1 to len(values)
    ival = 1
    do isub = 1,numberOfSubsets
      write(*,*) ""
      ides = 1
      if (length_dr .gt. 0) then

        ! delayed descriptors present, print the three parts separately
        do ides = 1,start_dr-1
          call write_data_value_line(ides,names(ides),units(ides),values(ival),cvals(ival))
          ival = ival+1
        enddo
        call write_data_value_line(start_dr,names(start_dr+1),units(start_dr+1),values(ival),cvals(ival))
        ival = ival+1
        do irep = 1,replications
          do ides = 1,length_dr
            call write_data_value_line(start_dr+(irep-1)*length_dr+ides,names(start_dr+ides+1), &
                                       units(start_dr+ides+1),values(ival),cvals(ival))
            ival = ival+1
          enddo
        enddo
      else

        ! no delayed descriptors present, only one part to print
        do ides = 1,size(descriptors)
          call write_data_value_line(ides,names(ides),units(ides),values(ival),cvals(ival))
          ival = ival+1
        enddo
      endif
    enddo

    deallocate(abbreviations)
    deallocate(names)
    deallocate(units)

    ! important: release handle before getting the next one
    call codes_release(ibufr)
  enddo  ! messages loop

  call codes_close_file(ifile)

end program bufrdecode

subroutine write_data_value_line(nr,exp_name,exp_unit,num_value,str_value)

  integer,intent(in)          :: nr
  character(len=*),intent(in) :: exp_name
  character(len=*),intent(in) :: exp_unit
  real(kind=8),intent(in)     :: num_value
  character(len=*),intent(in) :: str_value

  if (num_value .gt. -1.0E12 .and. num_value .lt. 1.0E12) then
    write(*,'(I6,A,A25,F21.4,A,A25,A20)') nr," ",exp_name,num_value," ",exp_unit,str_value
  else
    write(*,'(I6,A,A25,A21,A,A25,A20)') nr," ",exp_name,trim("MISSING")," ",exp_unit,str_value
  endif

end subroutine write_data_value_line
