module BufrIO
!
!  #[ Description:
!
! This module contains several routines to read, decode, encode and write
! BUFR messages.
! It is an interface between scatterometer processing programs and the BUFR
! software in the ECMWF ecCodes library
! the following subroutines from ecCodes are used:
!     codes_open_file
!     codes_close_file
!     codes_count_in_file
!     codes_bufr_new_from_file
!     codes_release
!     codes_set
!     codes_get
!     codes_get_string_array
!     codes_bufr_new_from_samples
!     codes_write
!
! Written by: Anton Verhoef
!
! This software was developed within the context of the EUMETSAT Satellite
! Application Facility on Numerical Weather Prediction (NWP SAF), under the
! Cooperation Agreement dated 29 June, 2011, between EUMETSAT and the
! Met Office, UK, by one or more partners within the NWP SAF. The partners
! in the NWP SAF are the Met Office, ECMWF, KNMI and Meteo France.
!
! Copyright 2020, EUMETSAT, All Rights Reserved
!
!     Unit Name:             bufrio_module.F90
!
!     Created on:            04-05-2020
!
!     Last Modified on:      $Date$
!
!     Modifications Log:
!
! $Id$
! 
!  #]
!  #[ Use statements
  use eccodes,only: codes_open_file,codes_close_file,codes_count_in_file, &
                    codes_bufr_new_from_file,codes_release, &
                    codes_set,codes_get,codes_get_string_array, &
                    codes_bufr_new_from_samples,codes_write, &
                    CODES_END_OF_FILE,CODES_MISSING_DOUBLE

  use numerics,only: r8_,missing_indicator_real,missing_indicator_real_r8, &
                     missing_indicator_integer,missing_real

  implicit none
!  #]
!  #[ Parameters
  ! max possible length for input path+filename
  integer,public,parameter :: max_bufrfilename_len     = 256

  integer,public,parameter :: bufrmod_no_error         =   0
  integer,public,parameter :: bufrmod_fileopen_error   =  -1
  integer,public,parameter :: bufrmod_fileclose_error  =  -2
  integer,public,parameter :: bufrmod_corrupted_msg    =  -3
  integer,public,parameter :: bufrmod_create_error     =  -4
  integer,public,parameter :: bufrmod_read_error       =  -5
  integer,public,parameter :: bufrmod_save_error       =  -6
  integer,public,parameter :: bufrmod_expand_error     =  -7
  integer,public,parameter :: bufrmod_allocation_error =  -8

  integer,public,parameter :: BufrVerbosityError  = 0  ! error   verbosity
  integer,public,parameter :: BufrVerbosityWarn   = 1  ! warning verbosity
  integer,public,parameter :: BufrVerbosityReport = 2  ! warning verbosity

  ! set default verbosity level
  integer,public,save      :: BufrVerbosity = BufrVerbosityError

  ! BUFR format identifier of last message
  integer,public           :: unexpanded_descr_1 = 0
!  #]
!  #[ Data types

  !----------------------------------
  ! BUFR data type
  !----------------------------------
  type,public :: bufr_data_type
    !----------------
    integer   :: bufr_length          ! ksec0( 2) length of BUFR message in bytes
    integer   :: bufr_edition         ! ksec0( 3)
    !-------- --------
    integer   :: bufr_centre          ! ksec1( 3)
    integer   :: update_number        ! ksec1( 4)
    integer   :: section1_flags       ! ksec1( 5)
    integer   :: data_category        ! ksec1( 6)
    integer   :: data_subcategory     ! ksec1( 7) message subtype or local data sub-category
    integer   :: local_tablesversion  ! ksec1( 8)
    integer   :: year                 ! ksec1( 9) century year
    integer   :: month                ! ksec1(10)
    integer   :: day                  ! ksec1(11)
    integer   :: hour                 ! ksec1(12)
    integer   :: minute               ! ksec1(13)
    integer   :: master_tablenumber   ! ksec1(14)
    integer   :: master_tablesversion ! ksec1(15)
    integer   :: bufr_subcentre       ! ksec1(16) only for BUFR edition >= 3
    integer   :: int_subcategory      ! ksec1(17) only for BUFR edition >= 4
    integer   :: second               ! ksec1(18) only for BUFR edition >= 4
    !-------- --------
    integer   :: section3_flags       ! ksec3( 4) compressed(0,1) observed (0,1)
    !-------- --------
    integer   :: nelements            ! number of expanded elements in BUFR template
    integer   :: nsubsets             ! ksup(6) actual number of subsets
    integer   :: nvalues              ! ksup(7) actual number of values
    !-------- --------
    integer   :: nreplications        ! number of delayed replications
    integer   :: nunexpanded          ! number of unexpanded descriptors
    integer   :: nexpanded            ! number of expanded descriptors
    !-------- --------
    integer   :: prev_unexp_1         ! housekeeping: first element of unexp descr in previous message
    integer   :: prev_repl            ! housekeeping: number of replications in previous message

    ! list of unexpanded data descriptors
    integer,dimension(:),allocatable           :: unexpanded_descr
    ! list of expanded data descriptors
    integer,dimension(:),allocatable           :: expanded_descr
    ! list of names
    character(len=64),dimension(:),allocatable :: names
    ! list of units
    character(len=24),dimension(:),allocatable :: units
    ! list of real values
    real(kind=8),dimension(:),allocatable      :: values
    ! character values
    character(len=20),dimension(:),allocatable :: cvals

  end type bufr_data_type

  !----------------------------------
  ! BUFR file attributes type
  !----------------------------------
  type bufr_file_attr_data
    integer            :: nr_of_BUFR_messages
    character(len=max_bufrfilename_len) :: bufr_filename
    integer            :: bufr_fileunit
    integer            :: file_size
    ! pointer to the message number in the file which was last read
    integer            :: curr_msg
    logical            :: file_open
    logical            :: writemode
    logical            :: arrays_allocated
    ! set this to false in case of a corrupted/truncated etc. message
    logical,dimension(:),pointer :: message_is_valid
  end type bufr_file_attr_data

!  #]
  contains

  subroutine set_BUFR_fileattributes(filename_bufr,bufr_file_attributes, &
                                     error_flag)
    !  #[
    ! init the bufr_file_attributes struct
    ! and set the filename for the file to be opened
    !
    ! arguments
    character(len=max_bufrfilename_len),intent(in) :: filename_bufr
    type(bufr_file_attr_data),intent(inout)        :: bufr_file_attributes
    integer,intent(out)                            :: error_flag

    error_flag = bufrmod_no_error

    bufr_file_attributes%bufr_filename    = filename_bufr
    bufr_file_attributes%curr_msg         = 0
    bufr_file_attributes%file_open        = .false.
    bufr_file_attributes%arrays_allocated = .false.

  end subroutine set_BUFR_fileattributes
    !  #]

  function BUFR_msg_is_valid(bufr_file_attributes,which_msg) result(n)
    !  #[
    ! inquiry functions, to see if a given BUFR message in the
    ! file is flagged as invalid, or not.
    !
    ! arguments
    type(bufr_file_attr_data),intent(in) :: bufr_file_attributes
    integer,intent(in)                   :: which_msg
    logical                              :: n

    n = bufr_file_attributes%message_is_valid(which_msg)

  end function BUFR_msg_is_valid
    !  #]

  subroutine set_BUFR_msg_to_invalid(bufr_file_attributes,which_msg)
    !  #[
    ! set the invalid flag, so that this message is no longer used
    !
    ! arguments
    type(bufr_file_attr_data),intent(inout) :: bufr_file_attributes
    integer,intent(in)                      :: which_msg

    bufr_file_attributes%message_is_valid(which_msg) = .false.

  end subroutine set_BUFR_msg_to_invalid
    !  #]

  function get_BUFR_nr_of_messages(bufr_file_attributes) result(n)
    !  #[
    ! inquiry functions, to get the total nr of BUFR messages in the file
    !
    ! arguments
    type(bufr_file_attr_data),intent(in) :: bufr_file_attributes
    integer                              :: n

    if (.not. bufr_file_attributes%file_open) then
      print *,'WARNING: get_BUFR_nr_of_messages() should only be called when a BUFR file is open'
      n = -1
      return
    endif
    n = bufr_file_attributes%nr_of_BUFR_messages

  end function get_BUFR_nr_of_messages
    !  #]

  subroutine open_BUFR_file(bufr_file_attributes, & ! input/output
                            error_flag, &           ! output
                            writemode,appendmode)   ! optional input
    !  #[
    ! open a BUFR file and collect some info about this file
    ! if opened for reading
    !
    ! arguments
    type(bufr_file_attr_data),intent(inout) :: bufr_file_attributes
    integer,intent(out)                     :: error_flag
    logical,optional,intent(in)             :: writemode,appendmode

    ! local variables
    integer   :: i,ibufr
    logical   :: use_writemode,use_appendmode
    character :: mode

    ! init and handle the optional input parameter
    error_flag = bufrmod_no_error

    use_writemode = .false.
    if (present(writemode)) then
      if (writemode) use_writemode = .true.
    endif

    use_appendmode = .false.
    if (present(appendmode)) then
      if (appendmode) then
        use_appendmode = .true.
        use_writemode = .true.
      endif
    endif

    if (bufr_file_attributes%file_open) then

      if (BufrVerbosity .ge. BufrVerbosityWarn) then
        print *,"File: ",trim(bufr_file_attributes%bufr_filename)," was already open"
        if (bufr_file_attributes%writemode) then
          if (use_appendmode) then
            print *,"(for appending)"
          else
            print *,"(for writing)"
          endif
        else
          print *,"(for reading)"
        endif
      endif

      error_flag = bufrmod_fileopen_error
      return
    else

      ! open the file (returns a file handle)
      if (BufrVerbosity .ge. BufrVerbosityReport) then
        print *,"opening BUFR file..."
      endif

      bufr_file_attributes%writemode = .false.
      mode = 'r'
      if (use_writemode) then
        bufr_file_attributes%writemode = .true.
        if (use_appendmode) then
          mode = 'a'
        else
          mode = 'w'
        endif
      endif

      call codes_open_file(bufr_file_attributes%bufr_fileunit, &
                           bufr_file_attributes%bufr_filename,mode,error_flag)
      if (error_flag .ne. 0) then
        print *,"open_BUFR_file: cannot open ",trim(bufr_file_attributes%bufr_filename)
        error_flag = bufrmod_fileopen_error
        return
      else
        if (BufrVerbosity .ge. BufrVerbosityReport) then
          print *,"open_BUFR_file has succesfully opened file: ",trim(bufr_file_attributes%bufr_filename)
          if (bufr_file_attributes%writemode) then
            if (use_appendmode) then
              print *,"(for writing/appending)"
            else
              print *,"(for writing)"
            endif
          else
            print *,"(for reading)"
          endif
        endif
      endif
    endif

    if (bufr_file_attributes%writemode) then
      bufr_file_attributes%nr_of_BUFR_messages = 0
    else

      ! get the number of messages in the file
      call codes_count_in_file(bufr_file_attributes%bufr_fileunit, &
                               bufr_file_attributes%nr_of_BUFR_messages,error_flag)

      ! if the number cannot be retrieved, read the messages one by one
      if (error_flag .ne. 0) then
        if (BufrVerbosity .ge. BufrVerbosityWarn) then
          print *,"WARNING in open_BUFR_file: problem counting number of messages in file"
          print *,"WARNING number of messages will be counted by retrieving them"
        endif

        bufr_file_attributes%nr_of_BUFR_messages = 0
        do while (bufr_file_attributes%nr_of_BUFR_messages .lt. 99999)
          call codes_bufr_new_from_file(bufr_file_attributes%bufr_fileunit,ibufr,error_flag)
          call codes_release(ibufr)
          if (error_flag .eq. CODES_END_OF_FILE) exit
          bufr_file_attributes%nr_of_BUFR_messages = bufr_file_attributes%nr_of_BUFR_messages + 1
        enddo

        ! close and open file to reset message pointer
        call codes_close_file(bufr_file_attributes%bufr_fileunit,error_flag)
        call codes_open_file(bufr_file_attributes%bufr_fileunit, &
                             bufr_file_attributes%bufr_filename,'r',error_flag)
        error_flag = bufrmod_no_error
      endif
      if (BufrVerbosity .ge. BufrVerbosityReport) then
        print *,"number of BUFR messages found in file:",bufr_file_attributes%nr_of_BUFR_messages
      endif

      ! allocate array indicating validity of messages
      if (bufr_file_attributes%arrays_allocated) then
        deallocate(bufr_file_attributes%message_is_valid)
        nullify(bufr_file_attributes%message_is_valid)
        bufr_file_attributes%arrays_allocated = .false.
      endif

      allocate(bufr_file_attributes%message_is_valid(bufr_file_attributes%nr_of_BUFR_messages), &
        STAT=error_flag)
      if (error_flag .ne. 0) then
        print *,"ERROR in open_BUFR_file: problem allocating message_is_valid array"
        error_flag = bufrmod_allocation_error
        return
      endif
      bufr_file_attributes%arrays_allocated = .true.
    endif

    ! messages are valid, untill proven otherwise
    do i = 1,bufr_file_attributes%nr_of_BUFR_messages
      bufr_file_attributes%message_is_valid(i) = .true.
    enddo

    ! ok, the file was succesfully opened, set message pointer to zero
    bufr_file_attributes%file_open = .true.
    bufr_file_attributes%curr_msg  = 0

  end subroutine open_BUFR_file
    !  #]

  subroutine close_BUFR_file(bufr_file_attributes,error_flag)
    !  #[
    ! close a BUFR file
    !
    ! arguments
    type(bufr_file_attr_data),intent(inout) :: bufr_file_attributes
    integer,intent(out)                     :: error_flag

    error_flag = bufrmod_no_error
    if (bufr_file_attributes%file_open) then
      call codes_close_file(bufr_file_attributes%bufr_fileunit,error_flag)
      if (error_flag .ne. 0) then
        print *,"close_BUFR_file: file could not be closed: ",trim(bufr_file_attributes%bufr_filename)
        error_flag = bufrmod_fileclose_error
        return
      endif

      if (BufrVerbosity .ge. BufrVerbosityReport) then
        print *,"close_BUFR_file has succesfully closed file: ",trim(bufr_file_attributes%bufr_filename)
      endif

      bufr_file_attributes%nr_of_BUFR_messages = 0
      bufr_file_attributes%file_open = .false.
    else
      error_flag = bufrmod_fileclose_error
    endif

  end subroutine close_BUFR_file
    !  #]

  subroutine get_BUFR_message(bufr_file_attributes,which_msg, &
                              bufr_data,error_flag)
    !  #[
    ! routine to read a chosen BUFR message from the opened file
    ! and put the contents in the bufr_data_type structure
    !
    ! arguments
    type(bufr_file_attr_data),intent(inout) :: bufr_file_attributes
    integer,intent(in)                      :: which_msg
    type(bufr_data_type),intent(inout)      :: bufr_data
    integer,intent(out)                     :: error_flag

    ! local variables
    integer      :: ibufr,ides,irep,isub,iret,i
    integer      :: replications,start_dr,length_dr
    logical      :: repl_present
    integer,dimension(:),allocatable,save           :: expanded_descr_tmp
    character(len=64),dimension(:),allocatable,save :: names_tmp
    character(len=24),dimension(:),allocatable,save :: units_tmp
    character(len=64),dimension(:),allocatable,save :: abbreviations
    character(len=80),dimension(:),allocatable,save :: cvals_tmp

    error_flag = bufrmod_no_error

    if (bufr_file_attributes%writemode) then
      error_flag = bufrmod_read_error
      print *,"ERROR in get_BUFR_message: reading from a file is not possible when it is open for writing"
      return
    endif

    if (.not. bufr_file_attributes%file_open) then
      error_flag = bufrmod_read_error
      print *,"ERROR in get_BUFR_message: reading from this file is not possible, it should be opened first"
      return
    endif

    if (.not. bufr_file_attributes%message_is_valid(which_msg)) then
      error_flag = bufrmod_corrupted_msg
      print *,"ERROR in get_BUFR_message: message,",which_msg,"is invalid"
      return
    endif

    ! if the requested message was already read, we need to close and open the file
    ! since ecCodes cannot read backwards
    if (which_msg .le. bufr_file_attributes%curr_msg) then
      if (BufrVerbosity .ge. BufrVerbosityReport) then
        print *,"get_BUFR_message will close and re-open the file to read from the start"
      endif
      call codes_close_file(bufr_file_attributes%bufr_fileunit,error_flag)
      call codes_open_file(bufr_file_attributes%bufr_fileunit, &
                           bufr_file_attributes%bufr_filename,'r',error_flag)
      bufr_file_attributes%curr_msg = 0
    endif

    ! move forward if the requested message is not following directly the previously read one
    do while (bufr_file_attributes%curr_msg .lt. which_msg - 1)
      call codes_bufr_new_from_file(bufr_file_attributes%bufr_fileunit,ibufr,error_flag)
      bufr_file_attributes%curr_msg = bufr_file_attributes%curr_msg + 1
      call codes_release(ibufr)
    enddo

    ! get the message handle
    call codes_bufr_new_from_file(bufr_file_attributes%bufr_fileunit,ibufr,error_flag)
    bufr_file_attributes%curr_msg = bufr_file_attributes%curr_msg + 1
    if (error_flag .ne. 0) then
      bufr_file_attributes%message_is_valid(which_msg) = .false.
      error_flag = bufrmod_corrupted_msg
      print *,"ERROR in get_BUFR_message: message,",which_msg,"can not be read"
      return
    endif

    ! unpack the message
    call codes_set(ibufr,'skipExtraKeyAttributes',1)  ! this should enhance the performance
    call codes_set(ibufr,'unpack',1)

    ! fill heading of bufr_data structure
    call codes_get(ibufr,'totalLength',bufr_data%bufr_length)
    call codes_get(ibufr,'edition',bufr_data%bufr_edition)
    call codes_get(ibufr,'masterTableNumber',bufr_data%master_tablenumber)
    call codes_get(ibufr,'bufrHeaderCentre',bufr_data%bufr_centre)
    call codes_get(ibufr,'updateSequenceNumber',bufr_data%update_number)
    call codes_get(ibufr,'section1Flags',bufr_data%section1_flags)
    call codes_get(ibufr,'dataCategory',bufr_data%data_category)
    call codes_get(ibufr,'dataSubCategory',bufr_data%data_subcategory)
    call codes_get(ibufr,'masterTablesVersionNumber',bufr_data%master_tablesversion)
    call codes_get(ibufr,'localTablesVersionNumber',bufr_data%local_tablesversion)

    ! the fields below are only present for BUFR edition >= 2
    if (bufr_data%bufr_edition .ge. 2) then
      call codes_get(ibufr,'typicalMonth',bufr_data%month)
      call codes_get(ibufr,'typicalDay',bufr_data%day)
      call codes_get(ibufr,'typicalHour',bufr_data%hour)
      call codes_get(ibufr,'typicalMinute',bufr_data%minute)
    endif

    ! the fields below are only present for BUFR edition >= 3
    if (bufr_data%bufr_edition .ge. 3) then
      call codes_get(ibufr,'bufrHeaderSubCentre',bufr_data%bufr_subcentre)
      call codes_get(ibufr,'typicalYear',bufr_data%year)
    endif

    ! the fields below are only present for BUFR edition >= 4
    if (bufr_data%bufr_edition .ge. 4) then
      call codes_get(ibufr,'internationalDataSubCategory',bufr_data%int_subcategory)
      call codes_get(ibufr,'typicalSecond',bufr_data%second)
    endif

    call codes_get(ibufr,'section3Flags',bufr_data%section3_flags)
    call codes_get(ibufr,'numberOfSubsets',bufr_data%nsubsets)
    if (allocated(bufr_data%unexpanded_descr)) deallocate(bufr_data%unexpanded_descr)
    call codes_get(ibufr,'unexpandedDescriptors',bufr_data%unexpanded_descr)
    bufr_data%nunexpanded = size(bufr_data%unexpanded_descr)

    ! get the real data values
    if (allocated(bufr_data%values)) deallocate(bufr_data%values)
    call codes_get(ibufr,'numericValues',bufr_data%values,error_flag)
    if (error_flag .ne. 0) then
      error_flag = bufrmod_expand_error
      print *,"ERROR in get_BUFR_message: error reading real values from BUFR message"
      return
    endif
    bufr_data%nvalues   = size(bufr_data%values)
    bufr_data%nelements = bufr_data%nvalues / bufr_data%nsubsets

    ! set values to missing indicator when they are reported missing
    ! missing values from ecCodes are -1.E100
    do i = 1,size(bufr_data%values)
      if (abs(bufr_data%values(i)) .gt. missing_indicator_real_r8) then
        bufr_data%values(i) = missing_indicator_real_r8
      endif
    enddo

    ! check if delayed replications are present in the template and get the number
    repl_present = .true.
    call codes_get(ibufr,"delayedDescriptorReplicationFactor",replications,iret)
    if (iret .ne. 0) then
      call codes_get(ibufr,"shortDelayedDescriptorReplicationFactor",replications,iret)
    endif
    if (iret .ne. 0) then
      repl_present = .false.
      replications = 0
    endif
    bufr_data%nreplications = replications
    bufr_data%nexpanded     = bufr_data%nelements

    ! fill the local lists with expanded descriptors, abbreviations, names and units
    ! do this only if the unexpanded descriptor has changed w.r.t the previous call
    if (bufr_data%unexpanded_descr(1) .ne. bufr_data%prev_unexp_1) then

      if (allocated(expanded_descr_tmp)) deallocate(expanded_descr_tmp)
      call codes_get(ibufr,'expandedDescriptors',expanded_descr_tmp)

      if (allocated(abbreviations)) deallocate(abbreviations)
      allocate(abbreviations(size(expanded_descr_tmp)))
      call codes_get_string_array(ibufr,"expandedAbbreviations",abbreviations)

      if (allocated(names_tmp)) deallocate(names_tmp)
      allocate(names_tmp(size(expanded_descr_tmp)))
      call codes_get_string_array(ibufr,"expandedNames",names_tmp)

      if (allocated(units_tmp)) deallocate(units_tmp)
      allocate(units_tmp(size(expanded_descr_tmp)))
      call codes_get_string_array(ibufr,"expandedUnits",units_tmp)
    endif

    ! ecCodes returns the lists of descriptors, names, and units without repeating
    ! the blocks when there are delayed replications
    ! therefore we will do this here
    ! only when the unexpanded descriptor or number of replications have changed
    if (bufr_data%unexpanded_descr(1) .ne. bufr_data%prev_unexp_1 .or. &
        replications .ne. bufr_data%prev_repl) then

      ! look for the start of the part containing delayed descriptors
      start_dr  = size(expanded_descr_tmp)
      length_dr = 0
      if (repl_present) then
        do ides = 1,size(expanded_descr_tmp)

          ! keep the start and lenght of the delayed descriptors part
          if (trim(abbreviations(ides)) .eq. "delayedDescriptorReplicationFactor" .or. &
              trim(abbreviations(ides)) .eq. "shortDelayedDescriptorReplicationFactor") then
            start_dr  = ides - 1
            length_dr = size(expanded_descr_tmp) - ides
            exit
          endif
        enddo
      endif

      ! if there are multiple delayed replications only the first part of some info will be shown
      if (start_dr+(replications*length_dr) .ne. bufr_data%nexpanded) then
        if (BufrVerbosity .ge. BufrVerbosityWarn) then
          print *,"BUFR message contains multiple delayed replications"
          print *,"only the names/units until the first replication are shown"
        endif
        length_dr = 1
        replications = 0
      endif

      ! allocate arrays in bufr_data structure with enough length to contain
      ! the repeating replication blocks
      if (allocated(bufr_data%expanded_descr)) deallocate(bufr_data%expanded_descr)
      allocate(bufr_data%expanded_descr(bufr_data%nexpanded))
      if (allocated(bufr_data%names)) deallocate(bufr_data%names)
      allocate(bufr_data%names(bufr_data%nexpanded))
      if (allocated(bufr_data%units)) deallocate(bufr_data%units)
      allocate(bufr_data%units(bufr_data%nexpanded))

      ! copy expanded descriptors, names, and units into the bufr_data
      if (length_dr .gt. 0) then

        ! delayed descriptors present, copy fixed part, info about delayed descriptors,
        ! and repeat the delayed descriptor part as appropriate
        do ides = 1,start_dr-1
          bufr_data%expanded_descr(ides) = expanded_descr_tmp(ides)
          bufr_data%names(ides)          = names_tmp(ides)
          bufr_data%units(ides)          = units_tmp(ides)
        enddo
        bufr_data%expanded_descr(start_dr) = expanded_descr_tmp(start_dr+1)
        bufr_data%names(start_dr)          = names_tmp(start_dr+1)
        bufr_data%units(start_dr)          = units_tmp(start_dr+1)
        do irep = 1,replications
          do ides = 1,length_dr
            bufr_data%expanded_descr(start_dr+(irep-1)*length_dr+ides) = expanded_descr_tmp(start_dr+ides+1)
            bufr_data%names(start_dr+(irep-1)*length_dr+ides)          = names_tmp(start_dr+ides+1)
            bufr_data%units(start_dr+(irep-1)*length_dr+ides)          = units_tmp(start_dr+ides+1)
          enddo
        enddo
      else

        ! no delayed descriptors present, simply copy the list of descriptors
        do ides = 1,size(expanded_descr_tmp)
          bufr_data%expanded_descr(ides) = expanded_descr_tmp(ides)
          bufr_data%names(ides)          = names_tmp(ides)
          bufr_data%units(ides)          = units_tmp(ides)
        enddo
      endif

    endif

    ! if there are no delayed replications, read the character fields into the cvals array
    ! this is not implemented for messages containing delayed replications
    if (.not. repl_present) then

      ! allocate character data array, only when more space is needed than previously
      if (size(bufr_data%cvals) .lt. bufr_data%nvalues) then
        if (allocated(bufr_data%cvals)) deallocate(bufr_data%cvals)
        allocate(bufr_data%cvals(bufr_data%nvalues),STAT=error_flag)
        if (error_flag .ne. 0) then
          print *,"ERROR in get_BUFR_message: problem allocating cvals array"
          error_flag = bufrmod_allocation_error
          return
        endif
        bufr_data%cvals = ""
      endif

      ! character fields have unit 'CCITT IA5', if this unit is found then
      ! read the character values and copy them to the cvals array
      ! note that this is only done when no delayed descriptors are present, for simplicity
      ! note that only the first 20 characters of the data are kept, to save memory
      if (allocated(cvals_tmp)) deallocate(cvals_tmp)
      allocate(cvals_tmp(bufr_data%nsubsets))
      do ides = 1,bufr_data%nelements
        if (bufr_data%units(ides)(1:5) .eq. 'CCITT') then
          call codes_get_string_array(ibufr,abbreviations(ides),cvals_tmp,error_flag)
          if (error_flag .ne. 0) then
            error_flag = bufrmod_expand_error
            print *,"ERROR in get_BUFR_message: error reading character values from BUFR message"
            return
          endif
          do isub = 1,bufr_data%nsubsets
            bufr_data%cvals((isub-1)*bufr_data%nelements+ides) = cvals_tmp(isub)(1:20)
          enddo
        endif
      enddo
    endif

    ! important: release handle before getting the next one
    call codes_release(ibufr)

    ! keep track of which message was last read, what type it was, number of replications
    bufr_data%prev_unexp_1 = bufr_data%unexpanded_descr(1)
    bufr_data%prev_repl    = replications

    ! keep first unexpanded descriptor as reference for last type of message read
    unexpanded_descr_1 = bufr_data%unexpanded_descr(1)

  end subroutine get_BUFR_message
    !  #]

  subroutine init_BUFR_data(bufr_data,nsubsets,centre,subcentre, &
       lversion,mversion,updatenr,data_category,data_subcategory, &
       year,month,day,hour,minute,master_tablenumber, &
       ndescriptors,int_subcategory, &
       error_flag)
    !  #[
    ! initialise the BUFR data type struct and
    ! set the number of subsets for this BUFR message
    !
    ! arguments
    type(bufr_data_type),intent(inout) :: bufr_data
    integer,intent(in)                 :: nsubsets
    integer,intent(in)                 :: centre,subcentre,lversion,mversion
    integer,intent(in)                 :: updatenr,data_category,data_subcategory
    integer,intent(in)                 :: year,month,day
    integer,intent(in)                 :: hour,minute,master_tablenumber
    integer,intent(in)                 :: ndescriptors,int_subcategory
    integer,intent(out)                :: error_flag

    ! local variables
    integer :: i

    error_flag = bufrmod_no_error

    bufr_data%nelements   = ndescriptors
    bufr_data%nsubsets    = nsubsets
    bufr_data%nvalues     = ndescriptors * nsubsets
    bufr_data%nunexpanded = size(bufr_data%unexpanded_descr)

    ! allocate real values array, only when more space is needed than previously
    if (size(bufr_data%values) .lt. bufr_data%nvalues) then
      if (allocated(bufr_data%values)) deallocate(bufr_data%values)
      allocate(bufr_data%values(bufr_data%nvalues),STAT=error_flag)
      if (error_flag .ne. 0) then
        print *,"ERROR in init_BUFR_data: problem allocating values array"
        error_flag = bufrmod_allocation_error
        return
      endif
    endif

    ! initialise real values array
    do i = 1,bufr_data%nvalues
      bufr_data%values(i) = missing_indicator_real_r8
    enddo

    ! set BUFR heading information
    bufr_data%bufr_length          = 0  ! unknown at this point
    bufr_data%bufr_edition         = 4  ! always use edition 4 for new BUFR data
    bufr_data%master_tablenumber   = master_tablenumber
    bufr_data%bufr_centre          = centre
    bufr_data%bufr_subcentre       = subcentre
    bufr_data%update_number        = updatenr
    bufr_data%section1_flags       = 0  ! no section 2 is present
    bufr_data%data_category        = data_category
    bufr_data%int_subcategory      = int_subcategory
    bufr_data%data_subcategory     = data_subcategory
    bufr_data%master_tablesversion = mversion
    bufr_data%local_tablesversion  = lversion
    bufr_data%year                 = year
    bufr_data%month                = month
    bufr_data%day                  = day
    bufr_data%hour                 = hour
    bufr_data%minute               = minute
    bufr_data%second               = 0
    bufr_data%section3_flags       = 192  ! use data compression (192 = on, 128 = off)
                                          ! 128=observed data, 64=compressed data

  end subroutine init_BUFR_data
    !  #]

  subroutine save_BUFR_message(save_file_attributes,bufr_data,error_flag)
    !  #[
    ! encode and write a BUFR message to an opened file
    !
    ! arguments
    type(bufr_file_attr_data),intent(in) :: save_file_attributes
    type(bufr_data_type),intent(inout)   :: bufr_data
    integer,intent(out)                  :: error_flag

    ! local variables
    integer :: ibufr

    error_flag = bufrmod_no_error

    if (.not. save_file_attributes%writemode) then
      error_flag = bufrmod_save_error
      print *,"ERROR in save_BUFR_message: writing to a file is not possible when it is open for reading"
      return
    endif

    if (.not. save_file_attributes%file_open) then
      error_flag = bufrmod_save_error
      print *,"ERROR in save_BUFR_message: writing to this file is not possible, it should be opened first"
      return
    endif

    ! data is always written in BUFR edition 4
    call codes_bufr_new_from_samples(ibufr,'BUFR4',error_flag)
    if (error_flag .ne. 0) then
      error_flag = bufrmod_create_error
      print *,"ERROR in save_BUFR_message: could not create message from sample"
      return
    endif
    bufr_data%bufr_edition = 4

    ! set heading information
    call codes_set(ibufr,'masterTableNumber',bufr_data%master_tablenumber)
    call codes_set(ibufr,'bufrHeaderCentre',bufr_data%bufr_centre)
    call codes_set(ibufr,'bufrHeaderSubCentre',bufr_data%bufr_subcentre)
    call codes_set(ibufr,'updateSequenceNumber',bufr_data%update_number)
    call codes_set(ibufr,'section1Flags',bufr_data%section1_flags)
    call codes_set(ibufr,'dataCategory',bufr_data%data_category)
    call codes_set(ibufr,'internationalDataSubCategory',bufr_data%int_subcategory)
    call codes_set(ibufr,'dataSubCategory',bufr_data%data_subcategory)
    call codes_set(ibufr,'masterTablesVersionNumber',bufr_data%master_tablesversion)
    call codes_set(ibufr,'localTablesVersionNumber',bufr_data%local_tablesversion)
    call codes_set(ibufr,'typicalYear',bufr_data%year)
    call codes_set(ibufr,'typicalMonth',bufr_data%month)
    call codes_set(ibufr,'typicalDay',bufr_data%day)
    call codes_set(ibufr,'typicalHour',bufr_data%hour)
    call codes_set(ibufr,'typicalMinute',bufr_data%minute)
    call codes_set(ibufr,'typicalSecond',bufr_data%second)
    call codes_set(ibufr,'section3Flags',bufr_data%section3_flags)

    ! set number of delayed replications
    if (bufr_data%nreplications .gt. 0) then
      call codes_set(ibufr,'inputDelayedDescriptorReplicationFactor',bufr_data%nreplications)
    endif

    call codes_set(ibufr,'skipExtraKeyAttributes',1)  ! this should enhance the performance
    call codes_set(ibufr,'numberOfSubsets',bufr_data%nsubsets)
    call codes_set(ibufr,'unexpandedDescriptors',bufr_data%unexpanded_descr)

    ! this does not work in ecCodes unfortunately
!   call codes_set(ibufr,'numericValues',bufr_data%values)

    call fill_BUFR_values(ibufr,bufr_data)

    call codes_set(ibufr,'pack',1,error_flag)
    if (error_flag .ne. 0) then
      print *,"ERROR in save_BUFR_message: could not pack message"
      error_flag = bufrmod_save_error
      return
    endif

    call codes_write(ibufr,save_file_attributes%bufr_fileunit,error_flag)
    if (error_flag .ne. 0) then
      print *,"ERROR in save_BUFR_message: could not write message to file"
      error_flag = bufrmod_save_error
      return
    endif

    call codes_release(ibufr)

    if (BufrVerbosity .ge. BufrVerbosityReport) then
      print *,"succesfully written BUFR message to this file"
    endif

  end subroutine save_BUFR_message
    !  #]

  subroutine fill_BUFR_values(ibufr,bufr_data)
    !  #[
    ! fill real values of BUFR message from the BUFR data type struct
    !
    ! arguments
    integer,intent(in)              :: ibufr
    type(bufr_data_type),intent(in) :: bufr_data

    ! local variables
    integer(kind=4),dimension(:),allocatable,save   :: descriptors
    character(len=64),dimension(:),allocatable,save :: abbreviations,keys,abbrev_tmp
    real(kind=8),dimension(:),allocatable           :: values_tmp
    integer,save :: replications,start_dr,length_dr
    integer,save :: previous_descr = -1
    integer      :: i,j,ides,isub,nequal,iret

    ! create a list of keys, equal to what you would get from
    ! bufr_dump -p -S1 <file>
    ! look for descriptors with the same abbreviation and add a counter with
    ! the format '#002#' in front of the name to make them unique
    ! example: #001#windSpeedAt10M, #002#windSpeedAt10M
    ! the new unique abbreviations are in the array 'keys'
    !
    ! since this process is quite time consuming we only do it when this routine is
    ! called first or when the BUFR template changes while writing the file
    ! (which will be quite rare)
    ! the arrays are of type save so they will be kept for the next call
    if (bufr_data%unexpanded_descr(1) .ne. previous_descr) then
      previous_descr = bufr_data%unexpanded_descr(1)

      ! read expanded data descriptor information
      if (allocated(descriptors)) deallocate(descriptors)
      call codes_get(ibufr,'expandedDescriptors',descriptors)

      ! read temporary list of data descriptor abbreviations
      if (allocated(abbrev_tmp)) deallocate(abbrev_tmp)
      allocate(abbrev_tmp(size(descriptors)))
      call codes_get_string_array(ibufr,"expandedAbbreviations",abbrev_tmp)

      ! handle delayed data descriptors if present
      ! look for the start of the part containing delayed descriptors
      replications = 0
      start_dr     = size(descriptors)
      length_dr    = 0
      call codes_get(ibufr,"delayedDescriptorReplicationFactor",replications,iret)

      if (iret .eq. 0) then  ! message contains delayed replications
        do i = 1,size(descriptors)

          ! determine the start and lenght of the delayed descriptors part of the list
          if (trim(abbrev_tmp(i)) .eq. "delayedDescriptorReplicationFactor") then
            start_dr  = i - 1
            length_dr = size(descriptors) - i
            exit
          endif
        enddo

        ! build an extended list of abbreviations repeating the replications
        if (allocated(abbreviations)) deallocate(abbreviations)
        allocate(abbreviations(start_dr+length_dr*replications))

        ! the first part of the list is a plain copy
        do i = 1,start_dr-1
          abbreviations(i) = abbrev_tmp(i)
        enddo

        ! the list contains both the key '104000' and the 'delayedDescriptorReplicationFactor'
        ! so the list should be offset by 1
        abbreviations(start_dr) = abbrev_tmp(start_dr+1)

        ! for each replication, extend the list with the descriptor sequence
        do i = 1,replications
          do j = 1,length_dr
            abbreviations(start_dr+(i-1)*length_dr+j) = abbrev_tmp(start_dr+j+1)
          enddo
        enddo

      else  ! message does not contain delayed replications

        ! copy list of data descriptor abbreviations from temporary list
        if (allocated(abbreviations)) deallocate(abbreviations)
        allocate(abbreviations(size(descriptors)))
        do i = 1,size(descriptors)
          abbreviations(i) = abbrev_tmp(i)
        enddo
      endif

      ! build the list of keys with the #counter# prefix if needed
      if (allocated(keys)) deallocate(keys)
      allocate(keys(start_dr+length_dr*replications))
      keys = abbreviations

      do i=size(keys),1,-1
        if (keys(i)(1:1) .eq. '#') cycle
        nequal = 0
        do j = 1,i-1
          if (abbreviations(j) .eq. abbreviations(i)) then
            nequal = nequal + 1
            write(keys(j),'(A,I3.3,2A)') '#',nequal,'#',trim(abbreviations(j))
          endif
        enddo
        nequal = nequal + 1
        if (nequal .gt. 1) then
          write(keys(i),'(A,I3.3,2A)') '#',nequal,'#',trim(abbreviations(i))
        endif
      enddo

      if (BufrVerbosity .ge. BufrVerbosityReport) then
        print *,"list of keys in BUFR message for assignment of values"
        do i=1,size(keys)
          print *,i,trim(keys(i))
        enddo
      endif
    endif

    ! loop over the list of descriptors
    ! copy the real values in all subsets belonging to this key to a temporary array
    ! and set the corresponding values in the BUFR message
    allocate(values_tmp(bufr_data%nsubsets))
    do ides=1,size(keys)

      ! skip key for delayed replication, it is fixed and cannot be changed
      if (ides .eq. start_dr .and. replications .gt. 0) cycle

      do isub = 1,bufr_data%nsubsets
        values_tmp(isub) = bufr_data%values(size(keys)*(isub-1)+ides)
        if (missing_real(values_tmp(isub))) values_tmp(isub) = CODES_MISSING_DOUBLE
      enddo
      call codes_set(ibufr,keys(ides),values_tmp)
    enddo

    ! deallocate array
    if (allocated(values_tmp)) deallocate(values_tmp)

  end subroutine fill_BUFR_values
    !  #]

  subroutine print_BUFR_data(bufr_data)
    !  #[
    ! print heading of BUFR message
    !
    ! arguments
    type(bufr_data_type),intent(in) :: bufr_data

    print *,"bufr_data%nelements                 = ",bufr_data%nelements
    print *,"bufr_data%nsubsets                  = ",bufr_data%nsubsets
    print *,"bufr_data%nvalues                   = ",bufr_data%nvalues

    print *,"bufr_data%bufr_length               = ",bufr_data%bufr_length
    print *,"bufr_data%bufr_edition              = ",bufr_data%bufr_edition

    print *,"bufr_data%bufr_centre               = ",bufr_data%bufr_centre
    print *,"bufr_data%update_number             = ",bufr_data%update_number
    print *,"bufr_data%section1_flags            = ",bufr_data%section1_flags
    print *,"bufr_data%data_category             = ",bufr_data%data_category
    print *,"bufr_data%data_subcategory          = ",bufr_data%data_subcategory
    print *,"bufr_data%local_tablesversion       = ",bufr_data%local_tablesversion
    if (bufr_data%bufr_edition .ge. 3) then
      print *,"bufr_data%year                      = ",bufr_data%year
    endif
    if (bufr_data%bufr_edition .ge. 2) then
      print *,"bufr_data%month                     = ",bufr_data%month
      print *,"bufr_data%day                       = ",bufr_data%day
      print *,"bufr_data%hour                      = ",bufr_data%hour
      print *,"bufr_data%minute                    = ",bufr_data%minute
    endif
    print *,"bufr_data%master_tablenumber        = ",bufr_data%master_tablenumber
    print *,"bufr_data%master_tablesversion      = ",bufr_data%master_tablesversion
    if (bufr_data%bufr_edition .ge. 3) then
      print *,"bufr_data%bufr_subcentre            = ",bufr_data%bufr_subcentre
    endif
    if (bufr_data%bufr_edition .ge. 4) then
      print *,"bufr_data%int_subcategory           = ",bufr_data%int_subcategory
      print *,"bufr_data%second                    = ",bufr_data%second
    endif

    print *,"bufr_data%section3_flags            = ",bufr_data%section3_flags

  end subroutine print_BUFR_data
    !  #]

  function BUFR_real2int(value) result(res)
    !  #[
    ! convert BUFR real value to integer value and set it as a
    ! missing value if appropriate
    !
    ! arguments
    real,intent(in) :: value
    integer         :: res

    if (missing_real(value)) then
      res = missing_indicator_integer
    else
      res = nint(value)
    endif

  end function BUFR_real2int
    !  #]

  function BUFR_int2real(value) result(res)
    !  #[
    ! convert integer value to BUFR real value and set it as a
    ! missing value if appropriate
    !
    ! arguments
    integer,intent(in) :: value
    real               :: res

    if (value .ge. missing_indicator_integer) then
      res = missing_indicator_real
    else
      res = real(value)
    endif

  end function BUFR_int2real
    !  #]

end module BufrIO
