  program grib2uv
!------------------------------------------------------------------------------!
! Program for generating (u,v) ASCII files from GRIB files for use as input    !
! for spectral calculations by WSC.                                            !
!------------------------------------------------------------------------------!
  use Compiler_Features, only: getarg_genscat, iargc_genscat
  use grib_api

  implicit none

  character*256                     :: inputfile, gribfile, path, stmp, uvfile
  character*4                       :: inputtype
  integer                           :: nerr, nargs, nmsgs, imsg, msgid, filenr, io
  integer(kind=4)                   :: fileunit, iret, itmp, nx, ny
  real                              :: rtmp, MinLon, MaxLon, MinLat, MaxLat
!!real                              :: dLat, dLon
  integer                           :: i, j, k, jstart, jend, jstride
  real, dimension(:), allocatable   :: values
  real, dimension(:,:), allocatable :: u, v, lsm, ice
  integer                           :: verbosity=0

  integer, parameter                :: maxx = 10240, maxy = 5120  ! size of local arrays to hold data
  integer, parameter                :: inputunit=1, outputunit=2

  inputtype='file'

  ! read command line
  nargs = iargc_genscat()
  if (nargs.ne.3) then
    write(*,*) 'Usage: grib2uv <inputtype> <inputfile> <outputfile>'
    write(*,*) '  with inputtype is file (default) or list'
    stop
  endif
  call getarg_genscat(1,inputtype)
  call getarg_genscat(2,inputfile)
  call getarg_genscat(3,uvfile)

  ! check setting of environment variable
  call getenv("GRIB_DEFINITION_PATH",path)
  if (len_trim(path) .eq. 0) then
    write(*,*) ' Environment variable $GRIB_DEFINITION_PATH is not set'
    write(*,*) ' Set it to (...)/genscat/support/eccodes/definitions'
    stop
  endif

  ! open input and output files
  filenr=0
  if (inputtype == 'list') open (inputunit, file=inputfile, status='old')
  open (outputunit , file=uvfile , status='unknown')

  ! start loop over input files; set grib file name
  fileloop: do
    filenr=filenr + 1
    if (inputtype == 'list') then
      read (inputunit,*,iostat=io) gribfile
      if (io < 0) exit fileloop
    else
      if (filenr > 1) exit fileloop
      gribfile=inputfile
    end if

    ! open grib file
    call grib_open_file(fileunit, trim(gribfile), 'r', status = iret)
    if (iret .ne. 0) then
      write(*,*) ' Return code from grib_open_file =', iret
      stop
    endif

    ! count messages in file
    call grib_count_in_file(fileunit, nmsgs, status = iret)
    if (iret .ne. 0) then
      write(*,*) ' Return code from grib_count_in_file =', iret
      stop
    endif
    if (nmsgs < 1) then
      write (*,*) 'ERROR: number of messages equals',nmsgs
      write (*,*) 'GRIB file : ',trim(gribfile)
      stop
    end if

    allocate(u(maxx,maxy))
    allocate(v(maxx,maxy))
    allocate(lsm(maxx,maxy))
    allocate(ice(maxx,maxy))

    ! loop through GRIB products in file
    nerr = 0
    msgloop: do imsg = 1, nmsgs
      write(*,*) ' '
      write(*,*) 'Message',imsg,' of',nmsgs

      ! read the message
      call grib_new_from_file(fileunit, msgid, status = iret)
      if (iret .ne. 0) then
        write(*,*) ' Return code from grib_new_from_file =', iret
        nerr = nerr + 1
      endif

      write(*,'(A)') ' '
      write(*,'(A)') 'Header information'
      write(*,'(A)') '---------------------------------------'
      call grib_get(msgid, 'totalLength', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Length of GRIB message (octets)             ', itmp

      call grib_get(msgid, 'editionNumber', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'GRIB Edition Number                         ', itmp
      write(*,*) ' '
      call grib_get(msgid, 'centre', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Originating centre identifier               ', itmp

      call grib_get(msgid, 'subCentre', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Sub-centre identifier                       ', itmp

      call grib_get(msgid, 'generatingProcessIdentifier', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Generating Process Identifier               ', itmp

      call grib_get(msgid, 'gridDefinition', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Grid definition                             ', itmp

      call grib_get(msgid, 'paramId', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Parameter identifier (Code Table 2)         ', itmp

      call grib_get(msgid, 'shortName', stmp, status = iret)
      if (iret .eq. 0) write(*,'(A,A10)') 'Parameter short name                        ', trim(stmp)

      call grib_get(msgid, 'levtype', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Type of level (Code Table 3) or satellite id', itmp

      call grib_get(msgid, 'level', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Value of level (Code Table 3)               ', itmp

      call grib_get(msgid, 'dataDate', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Reference date                              ', itmp

      call grib_get(msgid, 'dataTime', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Reference time                              ', itmp

      call grib_get(msgid, 'stepRange', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Time range                                  ', itmp

      call grib_get(msgid, 'decimalScaleFactor', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Units decimal scaling factor                ', itmp

      call grib_get(msgid, 'decimalPrecision', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Decimal precision                           ', itmp

      write(*,'(A)') ' '

      call grib_get(msgid, 'typeOfGrid', stmp, status = iret)
      if (iret .eq. 0) write(*,'(A,A10)') 'Type of grid                                ', trim(stmp)

      call grib_get(msgid, 'dataRepresentationType', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Data represent type (Code Table 6)          ', itmp

      call grib_get(msgid, 'Nx', nx, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of points along a parallel or X-axis ', nx
      if (nx < 1) then
         write (*,*) 'ERROR in nx =',nx
         write (*,*) 'GRIB file : ',trim(gribfile)
         stop
      end if
      if (nx > maxx) then
         write (*,*) 'ERROR nx is too large, enlarge maxx to',nx
         stop
      end if

      call grib_get(msgid, 'Ny', ny, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of points along a meridian or Y-axis ', ny
      if (ny < 1) then 
         write (*,*) 'ERROR in ny =',ny
         write (*,*) 'GRIB file : ',trim(gribfile)
         stop
      end if
      if (ny > maxy) then
         write (*,*) 'ERROR ny is too large, enlarge maxy to',ny
         stop
      end if

      ! grid description parameters for lat/lon and gaussian grids
      call grib_get(msgid, 'latitudeOfFirstGridPointInDegrees', MinLat, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of the first grid point            ', MinLat

      call grib_get(msgid, 'latitudeOfLastGridPointInDegrees', MaxLat, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of the last grid point             ', MaxLat

      call grib_get(msgid, 'numberOfParallelsBetweenAPoleAndTheEquator', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of parallels between pole and equator', itmp

      call grib_get(msgid, 'jScansPositively', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'j direction (North-South) scans positively  ', itmp

      call grib_get(msgid, 'longitudeOfFirstGridPointInDegrees', MinLon, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of the first grid point           ', MinLon

      call grib_get(msgid, 'longitudeOfLastGridPointInDegrees', MaxLon, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of the last grid point            ', MaxLon

      call grib_get(msgid, 'iDirectionIncrementInDegrees', rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'i direction (East-West) increment           ', rtmp

      call grib_get(msgid, 'iScansNegatively', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'i direction (East-West) scans negatively    ', itmp

      ! grid description parameters for space view
      call grib_get(msgid, 'latitudeOfSubSatellitePointInDegrees', rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of sub-satellite point             ', rtmp

      call grib_get(msgid, 'longitudeOfSubSatellitePointInDegrees', rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of sub-satellite point            ', rtmp

      call grib_get(msgid, 'apparentDiameterOfEarthInGridLengthsInXDirection', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Diameter of the earth in x direction        ', itmp

      call grib_get(msgid, 'apparentDiameterOfEarthInGridLengthsInYDirection', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Diameter of the earth in y direction        ', itmp

      call grib_get(msgid, 'xCoordinateOfSubSatellitePoint', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'X coordinate of sub-satellite point         ', itmp

      call grib_get(msgid, 'yCoordinateOfSubSatellitePoint', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Y coordinate of sub-satellite point         ', itmp

      call grib_get(msgid, 'orientationOfTheGridInDegrees', rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Orientation of the grid                     ', rtmp

      call grib_get(msgid, 'altitudeOfTheCameraFromTheEarthSCenterMeasuredInUnitsOfTheEarth', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Altitude of the camera                      ', itmp

      call grib_get(msgid, 'xCoordinateOfOriginOfSectorImage', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'X coordinate of origin of sector image      ', itmp

      call grib_get(msgid, 'yCoordinateOfOriginOfSectorImage', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Y coordinate of origin of sector image      ', itmp

      ! general grid description parameters
      call grib_get(msgid, 'earthIsOblate', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Earth flag                                  ', itmp

      call grib_get(msgid, 'uvRelativeToGrid', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Flag for u and v components w.r.t. grid     ', itmp

      call grib_get(msgid, 'NV', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of vertical coordinate parameters    ', itmp

      write(*,'(A)') ' '
      call grib_get(msgid, 'numberOfValues', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of data values coded/decoded         ', itmp
 
      call grib_get(msgid, 'bitsPerValue', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Number of bits per data value               ', itmp

      call grib_get(msgid, 'gridType', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Type of data       (0=grid pt, 128=spectral)', itmp

      call grib_get(msgid, 'packingType', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Type of packing    (0=simple, 64=complex)   ', itmp

      write(*,'(A)') ' '
      call grib_get(msgid, 'missingValue', itmp, status = iret)
      if (iret .eq. 0) write(*,'(A,I10)') 'Missing data value for integer data         ', itmp

      call grib_get(msgid, 'missingValue', rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F10.3)') 'Missing data value for real data            ', rtmp

      call grib_get_size(msgid, 'values', itmp)
      allocate(values(itmp), stat=iret)

      ! get data values
      call grib_get(msgid, 'values', values)

      select case(imsg)
      case (1)
        do j=1,ny
          do i=1,nx
            k=(j-1)*nx + i
            u(i,j) = values(k)
          enddo
        enddo

        if (verbosity > 2) then
          do i=1,nx,128
            write (*,*) ' '
            write (*,*) 'u-values for Meridian',i
            write (*,'(10f12.6)') (u(i,j),j=1,ny)
          enddo
        end if
      case (2)
        do j=1,ny
          do i=1,nx
            k=(j-1)*nx + i
            v(i,j) = values(k)
          enddo
        enddo

        if (verbosity > 2) then
          do i=1,nx,128
            write (*,*) ' '
            write (*,*) 'v-values for Meridian',i
            write (*,'(10f12.6)') (v(i,j),j=1,ny)
          enddo
        end if
      case (3)
        do j=1,ny
          do i=1,nx
            k=(j-1)*nx + i
            lsm(i,j) = values(k)
          enddo
        enddo

        if (verbosity > 2) then
          do i=1,nx,128
            write (*,*) ' '
            write (*,*) 'lsm-values for Meridian',i,'  longitude',MinLon + real(i - 1)*(MaxLon - MinLon)/real(nx)
            write (*,'(10f12.6)') (lsm(i,j),j=1,ny)
          enddo
        end if
      case (4)
        do j=1,ny
          do i=1,nx
            k=(j-1)*nx + i
            ice(i,j) = values(k)
          enddo
        enddo

        if (verbosity > 2) then
          do i=1,nx,128
            write (*,*) ' '
            write (*,*) 'lsm-values for Meridian',i,'  longitude',MinLon + real(i - 1)*(MaxLon - MinLon)/real(nx)
            write (*,'(10f12.6)') (ice(i,j),j=1,ny)
          enddo
        end if
      case default
        write (*,*) 'ERROR in grib2uv: grib file contains more than three messages'
        stop
      end select

      ! get max, min and average of data
      write(*,'(A)') ' '
      call grib_get(msgid,'max',rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F16.7)') 'Maximum value in field is ', rtmp
      call grib_get(msgid,'min',rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F16.7)') 'Minimum value in field is ', rtmp
      call grib_get(msgid,'average',rtmp, status = iret)
      if (iret .eq. 0) write(*,'(A,F16.7)') 'Average value in field is ', rtmp
 
      ! release message from memory
      call grib_release(msgid)
      if (allocated(values)) deallocate(values)
    enddo msgloop

    ! close grib file
    call grib_close_file(fileunit, status = iret)
    if (iret .ne. 0) then
      write(*, *) ' Return code from grib_close_file =', iret
      stop
    endif

    write(*,*)
    write(*,'(A)') 'End of file reached'
    write(*,'(2(A,I5))') 'Number of products =', (nmsgs), ', number of decoding errors =', nerr
    write (*,*)
    write(*,'(A)') 'Writing output'

    ! write output
    write (*,*) 'nx,ny=',nx,ny

!!  dLat=(MaxLat - MinLat)/real(ny - 1)
!!  dLon=(MaxLon - MinLon)/real(nx)

    write (outputunit,*) nx,ny,MinLat,MaxLat,MinLon,MaxLon

    do i=1,nx
      if (mod(i,2) == 0) then
        jstart=ny
        jend=1
        jstride=-1
      else
        jstart=1
        jend=ny
        jstride=1
      end if

      do j=jstart,jend,jstride
        write (outputunit,*) u(i,j),v(i,j),int(lsm(i,j)),ice(i,j)
      enddo
    enddo

    if (allocated(u)) deallocate(u)
    if (allocated(v)) deallocate(v)
    if (allocated(lsm)) deallocate(lsm)
    if (allocated(ice)) deallocate(ice)

  enddo fileloop
  
  write(*,*) '(uv) data written on file ',trim(uvfile)
  close (outputunit , status='keep')
  if (inputtype == 'list') close (inputunit , status='keep')

  end program grib2uv
