program gribinfo

  ! a small program to print some properties of a GRIB file
  ! using the standard routines in the ECMWF GRIB library

  use Compiler_Features, only: getarg_genscat, iargc_genscat

  use grib_api, only: grib_open_file, grib_count_in_file, grib_new_from_file, &
       grib_get, grib_close_file, grib_get_size, grib_release

  implicit none

  character*256   :: gribfile, path, stmp
  integer         :: nerr, nargs, nmsgs, imsg, msgid
  integer(kind=4) :: fileunit, iret, itmp
  real            :: rtmp
  real, dimension(:), allocatable :: values

  ! read command line
  nargs = iargc_genscat()
  if (nargs.ne.1) then
    write(*,*) 'Usage: gribinfo <inputfile>'
    stop
  endif
  call getarg_genscat(1,gribfile)

  ! check setting of environment variable
  call getenv("GRIB_DEFINITION_PATH",path)
  if (len_trim(path) .eq. 0) then
    write(*,*) ' Environment variable $GRIB_DEFINITION_PATH is not set'
    write(*,*) ' Set it to (...)/genscat/support/grib/definitions'
    stop
  endif

  ! open file
  call grib_open_file(fileunit, trim(gribfile), 'r', status = iret)
  if (iret .ne. 0) then
    write(*,*) ' Return code from grib_open_file =', iret
    stop
  endif

  ! count messages in file
  call grib_count_in_file(fileunit, nmsgs, status = iret)
  if (iret .ne. 0) then
    write(*,*) ' Return code from grib_count_in_file =', iret
    stop
  endif

  ! loop through GRIB products in file
  nerr = 0
  msgloop: do imsg = 1, nmsgs

    ! read the message
    call grib_new_from_file(fileunit, msgid, status = iret)
    if (iret .ne. 0) then
      write(*,*) ' Return code from grib_new_from_file =', iret
      nerr = nerr + 1
    endif

    write(*,'(A)') ' '
    write(*,'(A)') 'Header information'
    write(*,'(A)') '---------------------------------------'
    call grib_get(msgid, 'totalLength', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Length of GRIB message (octets)             ', itmp

    call grib_get(msgid, 'editionNumber', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'GRIB Edition Number                         ', itmp
    write(*,*) ' '
    call grib_get(msgid, 'centre', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Originating centre identifier               ', itmp

    call grib_get(msgid, 'subCentre', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Sub-centre identifier                       ', itmp

    call grib_get(msgid, 'generatingProcessIdentifier', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Generating Process Identifier               ', itmp

    call grib_get(msgid, 'gridDefinition', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Grid definition                             ', itmp

    call grib_get(msgid, 'paramId', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Parameter identifier (Code Table 2)         ', itmp

    call grib_get(msgid, 'shortName', stmp, status = iret)
    if (iret .eq. 0) write(*,'(A,A10)') 'Parameter short name                        ', trim(stmp)

    call grib_get(msgid, 'levtype', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Type of level (Code Table 3) or satellite id', itmp

    call grib_get(msgid, 'level', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Value of level (Code Table 3)               ', itmp

    call grib_get(msgid, 'dataDate', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Reference date                              ', itmp

    call grib_get(msgid, 'dataTime', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Reference time                              ', itmp

    call grib_get(msgid, 'stepRange', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Time range                                  ', itmp

    call grib_get(msgid, 'decimalScaleFactor', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Units decimal scaling factor                ', itmp

    call grib_get(msgid, 'decimalPrecision', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Decimal precision                           ', itmp

    write(*,'(A)') ' '

    call grib_get(msgid, 'typeOfGrid', stmp, status = iret)
    if (iret .eq. 0) write(*,'(A,A10)') 'Type of grid                                ', trim(stmp)

    call grib_get(msgid, 'dataRepresentationType', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Data represent type (Code Table 6)          ', itmp

    call grib_get(msgid, 'Nx', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of points along a parallel or X-axis ', itmp

    call grib_get(msgid, 'Ny', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of points along a meridian or Y-axis ', itmp

    ! grid description parameters for lat/lon and gaussian grids
    call grib_get(msgid, 'latitudeOfFirstGridPointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of the first grid point            ', rtmp

    call grib_get(msgid, 'latitudeOfLastGridPointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of the last grid point             ', rtmp

    call grib_get(msgid, 'numberOfParallelsBetweenAPoleAndTheEquator', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of parallels between pole and equator', itmp

    call grib_get(msgid, 'jScansPositively', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'j direction (North-South) scans positively  ', itmp

    call grib_get(msgid, 'longitudeOfFirstGridPointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of the first grid point           ', rtmp

    call grib_get(msgid, 'longitudeOfLastGridPointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of the last grid point            ', rtmp

    call grib_get(msgid, 'iDirectionIncrementInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'i direction (East-West) increment           ', rtmp

    call grib_get(msgid, 'iScansNegatively', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'i direction (East-West) scans negatively    ', itmp

    ! grid description parameters for space view
    call grib_get(msgid, 'latitudeOfSubSatellitePointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Latitude of sub-satellite point             ', rtmp

    call grib_get(msgid, 'longitudeOfSubSatellitePointInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Longitude of sub-satellite point            ', rtmp

    call grib_get(msgid, 'apparentDiameterOfEarthInGridLengthsInXDirection', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Diameter of the earth in x direction        ', itmp

    call grib_get(msgid, 'apparentDiameterOfEarthInGridLengthsInYDirection', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Diameter of the earth in y direction        ', itmp

    call grib_get(msgid, 'xCoordinateOfSubSatellitePoint', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'X coordinate of sub-satellite point         ', itmp

    call grib_get(msgid, 'yCoordinateOfSubSatellitePoint', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Y coordinate of sub-satellite point         ', itmp

    call grib_get(msgid, 'orientationOfTheGridInDegrees', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Orientation of the grid                     ', rtmp

    call grib_get(msgid, 'altitudeOfTheCameraFromTheEarthSCenterMeasuredInUnitsOfTheEarth', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Altitude of the camera                      ', itmp

    call grib_get(msgid, 'xCoordinateOfOriginOfSectorImage', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'X coordinate of origin of sector image      ', itmp

    call grib_get(msgid, 'yCoordinateOfOriginOfSectorImage', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Y coordinate of origin of sector image      ', itmp

    ! general grid description parameters
    call grib_get(msgid, 'earthIsOblate', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Earth flag                                  ', itmp

    call grib_get(msgid, 'uvRelativeToGrid', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Flag for u and v components w.r.t. grid     ', itmp

    call grib_get(msgid, 'NV', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of vertical coordinate parameters    ', itmp

    write(*,'(A)') ' '
    call grib_get(msgid, 'numberOfValues', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of data values coded/decoded         ', itmp

    call grib_get(msgid,'numberOfPoints', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of data values                       ', itmp

    call grib_get_size(msgid,'values', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Length of values array                      ', itmp

    call grib_get(msgid, 'bitsPerValue', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Number of bits per data value               ', itmp

    call grib_get(msgid, 'gridType', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Type of data       (0=grid pt, 128=spectral)', itmp

    call grib_get(msgid, 'packingType', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Type of packing    (0=simple, 64=complex)   ', itmp

    write(*,'(A)') ' '
    call grib_get(msgid, 'missingValue', itmp, status = iret)
    if (iret .eq. 0) write(*,'(A,I10)') 'Missing data value for integer data         ', itmp

    call grib_get(msgid, 'missingValue', rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F10.3)') 'Missing data value for real data            ', rtmp

    call grib_get_size(msgid, 'values', itmp)
    allocate(values(itmp), stat=iret)

    ! get data values
    write(*,'(A)') ' '
    write(*,'(A)') 'REALACC(6)'
    call grib_get(msgid, 'values', values)
    write(*,'(A)') 'First 20 data values'
    write(*,'(F16.7)') (values(itmp),itmp=1,20)

    ! get max, min and average of data
    write(*,'(A)') ' '
    call grib_get(msgid,'max',rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F16.7)') 'Maximum value in field is ', rtmp
    call grib_get(msgid,'min',rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F16.7)') 'Minimum value in field is ', rtmp
    call grib_get(msgid,'average',rtmp, status = iret)
    if (iret .eq. 0) write(*,'(A,F16.7)') 'Average value in field is ', rtmp

    write(*,'(A)') 'ENDREALACC'
    
    ! release message from memory
    call grib_release(msgid)
    if (allocated(values)) deallocate(values)

  enddo msgloop

  write(*,*)
  write(*,'(A)') 'End of file reached'
  write(*,'(2(A,I5))') 'Number of products =', (nmsgs), ', number of decoding errors =', nerr

  ! close file
  call grib_close_file(fileunit, status = iret)
  if (iret .ne. 0) then
    write(*, *) ' Return code from grib_close_file =', iret
    stop
  endif

  end program gribinfo
