module GribIO
!
!  #[ Description
!
!
! this module defines functions for reading GRIB messages
! and provides a data structure for storing the data
! It is an interface between scatterometer processing programs and the GRIB
! software in the ECMWF ECCODES library
! the following subroutines from ECCODES are used:
!     grib_open_file
!     grib_close_file
!     grib_count_in_file
!     grib_new_from_file
!     grib_get_size
!     grib_get
!     grib_set
!     grib_release
!
! Written by: Jos de Kloe and Anton Verhoef
!
! Part of this software was developed within the context of the EUMETSAT
! Satellite Application Facility on Numerical Weather Prediction (NWP SAF),
! under the Cooperation Agreement dated 16 December, 2003, between EUMETSAT
! and the Met Office, UK, by one or more partners within the NWP SAF.
! The partners in the NWP SAF are the Met Office, ECMWF, KNMI
! and Meteo France.
!
!     Unit Name:             gribio_module.F90
!
!     Created on:            15-07-2003
!
!     Last Modified on:      $Date$
!
!     Modifications Log:
!
! $Id$
!
!
!  #]
!  #[ Use statements
  use DateTimeMod, only: conv_date_to_daycount, conv_daycount_to_date
  use DateTimeMod, only: convert_to_derived_datetime

  use convert, only: get_angle_distance

  use grib_api, only: grib_open_file, grib_new_from_file, grib_close_file, &
       grib_release, grib_multi_support_on, grib_get, grib_set, &
       grib_get_size, GRIB_END_OF_FILE

  use numerics, only: r4_, r8_, missing_indicator_real, missing_real

  implicit none ! no implicit variable typing

  ! these are defined in numerics. It is useless to define them again
  ! here, since numerics is needed by datetimemod and convert anyway,
  ! so must allways be linked to the program.
  !integer, parameter :: r4_ = Selected_Real_Kind( 6, 37)  ! = real*4
  !integer, parameter :: r8_ = Selected_Real_Kind(15,307)  ! = real*8
  !integer, parameter :: s_  = r4_
  !integer, parameter :: l_  = r4_ ! should be r8_ to keep naming consistent

  !real(l_), parameter :: missing_indicator_real = 1.7E38
  real(r4_), parameter :: grib_missing_value     = 1.0E09

  !  #]
  !  #[ Parameters

  !---------------------------------------------------------
  ! define a data structure to use for GRIB data
  !---------------------------------------------------------

  public ! the following data structure is public

  ! some type definitions

  ! note: this datetime type is simpler then the one in DateTimeMod
  ! since it inly takes the hours into account as an integer
  ! and stores the date as YYYYMMDD in an integer
  type AnDateTime
     integer :: date
     integer :: hour
  end type AnDateTime

  ! maximum number of GRIB files to be open simultaneously
  integer, parameter :: max_nr_open_grib_files = 55

  ! maximum number of GRIB messages that are expected to be in a file
  integer, parameter :: maxnr_of_grib_messages = 400

  ! maximum number of GRIB messages for each GRIB file to be present in memory
  integer, parameter :: max_nr_open_grib_msgs  = 100

  ! define the grib parameters that we need
  integer, parameter :: par_id_u     = 131    ! wind u component
  integer, parameter :: par_id_v     = 132    ! wind v component
  integer, parameter :: par_id_q     = 133    ! Q specific humidity (kg/kg)

  ! soil temperature level 1 is above sea equivalent to sst?
  integer, parameter :: par_id_stl1  = 139    ! soil temperature level1 (K)
  integer, parameter :: par_id_chnk  = 148    ! chnk
  integer, parameter :: par_id_mslp  = 151    ! mean sea level pressure
  integer, parameter :: par_id_lnsp  = 152    ! lnsp
  integer, parameter :: par_id_2t    = 167    ! 2m temperature (K)
  integer, parameter :: par_id_2d    = 168    ! 2m dewpoint temperature (K)
  integer, parameter :: par_id_skt   = 235    ! skin temperature (K)

  integer, parameter :: par_id_istl1 = 35     ! Ice surface temperature layer 1
  integer, parameter :: par_id_ci    = 31     ! sea-ice cover

  integer, parameter :: par_id_10u   = 165    ! 10m wind u component
  integer, parameter :: par_id_10v   = 166    ! 10m wind v component
  integer, parameter :: par_id_lsm   = 172    ! land-sea mask

  integer, parameter :: par_id_u10n  = 228131 ! 10m neutral wind u component
  integer, parameter :: par_id_v10n  = 228132 ! 10m neutral wind v component

  ! parameters used in non-ECMWF models like NCEP
  integer, parameter :: par_id_sst_ncep = 130 ! SST, use also level_type 1 and level 0
  integer, parameter :: par_id_t_ncep   = 999 ! T, dummy value since 130 is already used for SST

  ! generic parameters, set equal to those in ECMWF model
  integer, parameter :: par_id_sst   = 34     ! sea surface temperature
  integer, parameter :: par_id_t     = 130    ! T (K)

  ! these special commands are no longer used
  ! except for the test programs test_read_GRIB1.F90 and
  ! test_read_GRIB3.F90
  ! they will be removed soon... (JK)

  ! some special commands for getting time/date from a grib file
  ! (the lat-parameter is used to pass this command to the subroutine
  !  get_from_GRIB_file, so make sure the code is not a valid latitude)
  real(r4_), parameter :: xgrib_get_time = -999. ! command to get the grib time
  real(r4_), parameter :: xgrib_get_date = -998. ! command to get the grib date
  ! command to get the derived grib time
  real(r4_), parameter :: xgrib_get_derived_time = -997.
  ! command to get the derived grib date
  real(r4_), parameter :: xgrib_get_derived_date = -996.

  ! possible results of the inquire_GRIB_filelist subroutine
  integer, parameter :: inq_is_present            = 1
  integer, parameter :: inq_not_present           = 2
  integer, parameter :: inq_multiple_msgs_present = 3

  ! the no_error error code is public
  integer, public, parameter :: gribio_no_error                = 0

  ! private error codes for the gribio_module (range -100..-199)
!  integer, private, parameter :: gribio_fileopen_error         = -100
!  integer, private, parameter :: gribio_close_GRIB_file        = -101
!  integer, private, parameter :: gribio_file_not_open_error    = -102
  integer, private, parameter :: gribio_GRIB_number_too_large  = -104
!  integer, private, parameter :: gribio_GRIB_read_problem      = -105
!  integer, private, parameter :: gribio_InvalidGRIBstring      = -106
  integer, private, parameter :: gribio_GRIB_decode_problem    = -107
!  integer, private, parameter :: gribio_wrong_time_unit_ind    = -108

  ! verbosity control
  integer, public, parameter :: GribVerbosityError  = 0  ! error     verbosity
  integer, public, parameter :: GribVerbosityWarn   = 1  ! warning   verbosity
  integer, public, parameter :: GribVerbosityReport = 2  ! reporting verbosity
  integer, public, parameter :: GribVerbosityInfo   = 3  ! info      verbosity
  integer, public, parameter :: GribVerbosityDebug  = 4  ! debug     verbosity
  ! set default verbosity level
  integer, public, save      :: GribVerbosity = GribVerbosityError

  !  #]
  !  #[ Variables, Types etc

  ! a structure containing info on the current GRIB file
  type grib_file_attr_data
    private
    ! the contents of this structure are private to this module,
    ! so any acces from the main program should be done by
    ! functions/subroutines
    !
    integer :: nr_of_GRIB_messages
    character(len=256) :: grib_filename
    integer :: grib_fileunit
    integer :: file_size
    logical :: file_open
    ! note: the index in these next 9 lists is what is called
    !       message_pos_in_file or msg_pos
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_message_ids ! assigned by ECCODES
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_level
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_level_type
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_date
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_hour
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_analyse
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_derived_date
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_derived_hour
    integer, dimension(1:maxnr_of_grib_messages)::list_of_GRIB_par_id
  end type grib_file_attr_data

  ! a structure containing info on the current GRIB message
  type grib_message_data
!!  private
    ! meta data
    integer  :: message_pos_in_file
    integer  :: message_id   ! assigned by ECCODES
    !real(r4_) :: date ! no longer used
    !real(r4_) :: time! no longer used
    !real(r4_) :: derived_date ! = date + time/24 ! no longer used
    !real(r4_) :: derived_time ! = mod(time,24)   ! no longer used
    !integer  :: total_message_size ! no longer used
    integer  :: vals_size
    logical  :: is_decoded   ! are the data present in the real_values_xy array
    ! grid details
    integer  :: nr_lon_points
    integer  :: nr_lat_points
    real(r4_) :: lat_of_first_gridpoint ! lat. range: 90 .. -90
    real(r4_) :: lat_of_last_gridpoint
    real(r4_) :: lon_of_first_gridpoint ! lon. range:  0 .. 360
    real(r4_) :: lon_of_last_gridpoint
    real(r4_) :: lat_step
    character(len=30) :: grid_type
    integer, dimension(:), pointer :: nr_lons_per_lat  ! for reduced grids
    real,    dimension(:), pointer :: lon_step_per_lat
    ! the decoded data
    real(r4_), dimension(:,:), pointer :: real_vals_xy  ! one dimensional array
    ! this one is converted to a local array in get_GRIB_data_values()
    !real(r4_), dimension(:), pointer   :: real_values   ! one dimensional array
  end type grib_message_data

  type list_of_grib_files_type
    type(grib_file_attr_data)             :: grib_file_attributes
    type(grib_message_data), &
         dimension(max_nr_open_grib_msgs) :: list_of_GRIB_msgs
  end type list_of_grib_files_type

  ! this is the global structure containing the GRIB data
  ! it is allocatable, but of fixed size max_nr_open_grib_files
  type(list_of_grib_files_type), allocatable, dimension(:) :: list_of_GRIB_files

  integer, save :: nr_of_GRIB_files_in_list
  integer, save :: t_interval = 99
  integer, save :: prev_hour  = -1
  !  #]
  interface operator(.eq.)
     module procedure isequal
  end interface
  interface operator(.gt.)
     module procedure islarger
  end interface
  interface operator(.lt.)
     module procedure issmaller
  end interface
  contains
    !---------------------------------------
    subroutine init_AnDateTime(dt)
      !  #[ init to missing
      type(AnDateTime), intent(out) :: dt

      dt%date = -1
      dt%hour = -1

    end subroutine init_AnDateTime
      !  #]
    subroutine set_AnDateTime(dt,date,hour)
      !  #[ define
      type(AnDateTime), intent(out) :: dt ! output
      integer, intent(in) :: date,hour ! input

      dt%date = date
      dt%hour = hour

    end subroutine set_AnDateTime
      !  #]
    function missing_AnDateTime(dt) result(is_missing)
      !  #[ check for missing
      type(AnDateTime), intent(in) :: dt
      logical :: is_missing ! result

      is_missing = .false.
      if (dt%date .eq. -1) is_missing = .true.
      if (dt%hour .eq. -1) is_missing = .true.

    end function missing_AnDateTime
    !  #]
    function islarger(dt1,dt2) result(is_larger)
      !  #[ check if dt1 is the larger one
      type(AnDateTime), intent(in) :: dt1, dt2
      logical :: is_larger ! result

      is_larger = .false.
      if (dt1%date .gt. dt2%date) then
         is_larger = .true.
         return
      end if
      if (dt1%date .eq. dt2%date) then
         if (dt1%hour .gt. dt2%hour) then
            is_larger = .true.
            return
         end if
      end if

    end function islarger
      !  #]
    function isequal(dt1,dt2) result(is_equal)
      !  #[ check if dt1 is equal to dt2
      type(AnDateTime), intent(in) :: dt1, dt2
      logical :: is_equal ! result

      is_equal = .true.
      if (dt1%date .ne. dt2%date) then
         is_equal = .false.
         return
      end if
      if (dt1%hour .ne. dt2%hour) then
         is_equal = .false.
         return
      end if

    end function isequal
      !  #]
    function issmaller(dt1,dt2) result(is_smaller)
      !  #[ check if dt1 is the smaller one
      type(AnDateTime), intent(in) :: dt1, dt2
      logical :: is_smaller ! result

      is_smaller = .false.
      if (dt1%date .lt. dt2%date) then
         is_smaller = .true.
         return
      end if
      if (dt1%date .eq. dt2%date) then
         if (dt1%hour .lt. dt2%hour) then
            is_smaller = .true.
            return
         end if
      end if

    end function issmaller
      !  #]
    !---------------------------------------

  subroutine init_GRIB_module(Verbosity, ByteSwapping)
    !  #[
    ! input parameters
    integer, optional, intent(in) :: Verbosity
    logical, optional, intent(in) :: ByteSwapping ! not used any more, kept for compatibility

    ! local variables
    integer :: i, j, allocstatus
    character(len=256) :: path

    call getenv("GRIB_DEFINITION_PATH",path)
    if (len_trim(path) .eq. 0) then
      print *,"ERROR: environment variable $GRIB_DEFINITION_PATH is not set"
      print *,"ERROR: set it to (...)/genscat/support/eccodes/definitions"
      stop 1
    endif

    if (.not. allocated(list_of_GRIB_files)) then
      allocate(list_of_GRIB_files(1:max_nr_open_grib_files), STAT=allocstatus)
      if (allocstatus .ne. 0) then
        print *,"ERROR: allocation failure for list_of_GRIB_files in gribio_module"
        stop 1
      endif
    endif

    ! set all GRIB file entries to closed and initialise message elements
    do i = 1,max_nr_open_grib_files
      list_of_GRIB_files(i)%grib_file_attributes%file_open = .false.
      do j = 1,max_nr_open_grib_msgs
        list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%message_pos_in_file = -1

        list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%is_decoded = .false.
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%nr_lons_per_lat)
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%lon_step_per_lat)
        !nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%real_values)
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%real_vals_xy)
      enddo
    enddo

    if (present(Verbosity)) then
      GribVerbosity = Verbosity
    endif

    ! this code gives segmentation faults for pgf90 for high verbosity
    ! in case the optional ByteSwapping is not present.
    !if (GribVerbosity .gt. GribVerbosityReport) then
    !  print *,"ByteSwapping is set to: ",ByteSwapping
    !endif

  end subroutine init_GRIB_module
  !---------------------------------------------------
  !  #]
! can be made private
  function get_GRIB_msgnr(grib_file_attributes, par_id, &
                          alt_grib_par_id, &
                          derived_date, derived_hour, &
                          date, hour, analyse, level, level_type, &
                          allow_multiple_matches, &
                          start_at_msg, silent) result(n)
    !  #[ search for a grib msg in a given grib file
    ! input parameters
    type(grib_file_attr_data), intent(in) :: grib_file_attributes
    integer, intent(in) :: par_id

    ! optional input parameters
    integer, intent(in),  optional :: alt_grib_par_id
    integer, intent(in),  optional :: derived_date, derived_hour
    integer, intent(in),  optional :: date, hour, analyse, level, level_type
    logical, intent(in),  optional :: allow_multiple_matches
    integer, intent(in),  optional :: start_at_msg
    logical, intent(in),  optional :: silent

    ! local variables
    integer :: n, i, start_msg
    logical :: be_silent

    n = -1

    be_silent = .false.
    if (present(silent)) then
      if (silent) be_silent = .true.
    endif

    if (.not. grib_file_attributes%file_open) then
      if (GribVerbosity .ge. GribVerbosityWarn) then
         print *,"WARNING: get_GRIB_msgnr should only be called "//&
              "when a GRIB file is open"
      endif
      return
    endif

    start_msg = 1
    if (present(start_at_msg)) start_msg = start_at_msg

    if (GribVerbosity .ge. GribVerbosityDebug) &
         print *,'DEBUG: start_msg = ', start_msg

    ! loop over the messages in the GRIB file and see
    ! if they meet the selection criteria
    do i = start_msg, grib_file_attributes%nr_of_GRIB_messages

       if (GribVerbosity .ge. GribVerbosityDebug) &
            print *,'DEBUG: looking at msg nr. ', i

       if (grib_file_attributes%list_of_GRIB_par_id(i) .ne. par_id) then

          ! first try the alternative par_id
          if (present(alt_grib_par_id)) then
             if (grib_file_attributes%list_of_GRIB_par_id(i) .ne. &
                  alt_grib_par_id) then
                if (GribVerbosity .ge. GribVerbosityDebug) &
                     print *,'DEBUG: par_id and alt_par_id do not match'
                cycle
             end if
          else
             if (GribVerbosity .ge. GribVerbosityDebug) &
                  print *,'DEBUG: par_id_does not match'
             cycle
          endif
       endif

       if (present(derived_date)) then
          if (grib_file_attributes%list_of_GRIB_derived_date(i) .ne. &
               derived_date) then
             if (GribVerbosity .ge. GribVerbosityDebug) then
                print *,'DEBUG: derived_date does not match'
                print *,'DEBUG: current grib msg has: ' , &
                     grib_file_attributes%list_of_GRIB_derived_date(i), &
                     ' but requested value is: ' , derived_date
             end if
             cycle
          end if
       endif

       if (present(derived_hour)) then
          if (grib_file_attributes%list_of_GRIB_derived_hour(i) .ne. &
               derived_hour)then
             if (GribVerbosity .ge. GribVerbosityDebug) then
                print *,'DEBUG: derived_hour does not match'
                print *,'DEBUG: current grib msg has: ' , &
                     grib_file_attributes%list_of_GRIB_derived_hour(i), &
                     ' but requested value is: ' , derived_hour
             end if
             cycle
          end if
       endif

       if (present(date)) then
          if (grib_file_attributes%list_of_GRIB_date(i) .ne. date) then
             if (GribVerbosity .ge. GribVerbosityDebug) then
                print *,'DEBUG: date does not match'
                print *,'DEBUG: current grib msg has: ' , &
                     grib_file_attributes%list_of_GRIB_date(i), &
                     ' but requested value is: ' , date
             end if
             cycle
          end if
       endif

       if (present(hour)) then
          if (grib_file_attributes%list_of_GRIB_hour(i) .ne. hour) then
             if (GribVerbosity .ge. GribVerbosityDebug) then
                print *,'DEBUG: hour does not match'
                print *,'DEBUG: current grib msg has: ' , &
                     grib_file_attributes%list_of_GRIB_hour(i), &
                     ' but requested value is: ' , hour
             end if
             cycle
          end if
       endif

       if (present(analyse)) then
          if (grib_file_attributes%list_of_GRIB_analyse(i) .ne. analyse) then
             if (GribVerbosity .ge. GribVerbosityDebug) &
                  print *,'DEBUG: analyse does not match'
             cycle
          end if
       endif

       if (present(level)) then
          if (grib_file_attributes%list_of_GRIB_level(i) .ne. level) then
             if (GribVerbosity .ge. GribVerbosityDebug) &
                  print *,'DEBUG: level does not match'
             cycle
          end if
       endif

       if (present(level_type)) then
          if (grib_file_attributes%list_of_GRIB_level_type(i) .ne. &
               level_type) then
             if (GribVerbosity .ge. GribVerbosityDebug) &
                  print *,'DEBUG: level_type does not match'
             cycle
          end if
       endif

       ! it seems that we have found the message we were looking for
       if (present(allow_multiple_matches)) then
          if (allow_multiple_matches) then
             n = i
             return
          end if
       end if

       ! otherwise check if it is unique in the file; if not then return
       if (n .eq. -1) then
          n = i
       else
          if (.not. be_silent .and. GribVerbosity .ge. GribVerbosityWarn) then
             call display_req_GRIB_msg_properties(par_id, &
                  alt_grib_par_id, derived_date, derived_hour, &
                  date,hour, analyse, level, level_type)
             call display_GRIB_message_properties(grib_file_attributes)
             print *,"WARNING: multiple GRIB messages in this file have"
             print *,"WARNING: this par_id"
             if (present(date))         print *,"WARNING: and date"
             if (present(hour))         print *,"WARNING: and hour"
             if (present(derived_date)) print *,"WARNING: and derived_date"
             if (present(derived_hour)) print *,"WARNING: and derived_hour"
             if (present(analyse))      print *,"WARNING: and analyse"
             if (present(level))        print *,"WARNING: and level"
             if (present(level_type))   print *,"WARNING: and level_type"
             print *,"WARNING: impossible to choose which one to use"
          endif
          n = -1
          return
       endif
    enddo

    if ( n .eq. -1 .and. .not. be_silent .and. &
         GribVerbosity .ge. GribVerbosityWarn) then
      call display_req_GRIB_msg_properties(par_id, &
             alt_grib_par_id, derived_date, derived_hour, &
             date,hour, analyse,level, level_type)
      call display_GRIB_message_properties(grib_file_attributes)
      print *,"WARNING: the requested GRIB message was not found in this file"
    endif

  end function get_GRIB_msgnr
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine display_req_GRIB_msg_properties(par_id, alt_grib_par_id, &
       derived_date, derived_hour, date, hour, analyse, level, level_type)
    !  #[ print the given optional variables in a uniform way

    ! input parameters
    integer, intent(in) :: par_id

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: derived_date, derived_hour
    integer, intent(in), optional :: date, hour, analyse, level, level_type

    ! local variables
    character(len=6) :: string_par_id
    character(len=6) :: string_alt_par_id
    character(len=8) :: string_date
    character(len=3) :: string_hour
    character(len=3) :: string_analyse
    character(len=8) :: string_derived_date
    character(len=3) :: string_derived_hour
    character(len=5) :: string_level
    character(len=3) :: string_level_type

    write(string_par_id,"(I6)") par_id

    if (present(alt_grib_par_id)) then
      write(string_alt_par_id,"(I6)") alt_grib_par_id
    else
      string_alt_par_id = "     -"
    endif

    if (present(date)) then
      write(string_date,"(I8)") date
    else
      string_date = "     any"
    endif

    if (present(hour)) then
      write(string_hour,"(I3)") hour
    else
      string_hour = "any"
    endif

    if (present(analyse)) then
      write(string_analyse,"(I3)") analyse
    else
      string_analyse = "any"
    endif

    if (present(derived_date)) then
      write(string_derived_date,"(I8)") derived_date
    else
      string_derived_date = "     any"
    endif

    if (present(derived_hour)) then
      write(string_derived_hour,"(I3)") derived_hour
    else
      string_derived_hour = "any"
    endif

    if (present(level)) then
      write(string_level,"(I5)") level
    else
      string_level = "  any"
    endif

    if (present(level_type)) then
      write(string_level_type,"(I3)") level_type
    else
      string_level_type = "any"
    endif

    print *,"======================================================="
    print *,"Requested properties of GRIB message"
    print *,"par_id      date hour analyse der_date der_hour level  level_type  alt_par_id"
    print "(1X,A6,2X,A8,2X,A3,5X,A3,1X,A8,6X,A3,1X,A5,9X,A3,6X,A6)",&
         string_par_id, string_date, string_hour, string_analyse, &
         string_derived_date, string_derived_hour, &
         string_level, string_level_type,string_alt_par_id
    print *,"======================================================="

  end subroutine display_req_GRIB_msg_properties
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine display_GRIB_message_properties(grib_file_attributes, par_id_only)
    !  #[ print properties for all grib msgs in a grib file

    ! input parameters
    type(grib_file_attr_data), intent(in) :: grib_file_attributes
    integer, optional, intent(in) :: par_id_only

    ! local variables
    integer :: i

    print *,"======================================================="
    print *,"Nr. of GRIB messages in file ", &
         trim(grib_file_attributes%grib_filename), ":", &
         grib_file_attributes%nr_of_GRIB_messages
    print *,"Properties of each message"
    print *,"par_id      date hour analyse der_date der_hour level  level_type"
    do i = 1,grib_file_attributes%nr_of_GRIB_messages
       if (present(par_id_only)) then
          if (grib_file_attributes%list_of_GRIB_par_id(i) .ne. par_id_only) cycle
       end if
       print "(1X,I6,2X,I8,2X,I3,5X,I3,1X,I8,6X,I3,1X,I5,9X,I3)",&
            grib_file_attributes%list_of_GRIB_par_id(i),&
            grib_file_attributes%list_of_GRIB_date(i),&
            grib_file_attributes%list_of_GRIB_hour(i),&
            grib_file_attributes%list_of_GRIB_analyse(i),&
            grib_file_attributes%list_of_GRIB_derived_date(i),&
            grib_file_attributes%list_of_GRIB_derived_hour(i),&
            grib_file_attributes%list_of_GRIB_level(i),&
            grib_file_attributes%list_of_GRIB_level_type(i)
    enddo
    print *,"======================================================="

  end subroutine display_GRIB_message_properties
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine open_GRIB_file(filename_grib, grib_file_attributes, error_flag)
  !  #[ open a grib file and read header info for all its msgs

    ! open a GRIB file and collect some info about this file
    ! loop over all GRIB messages, and store some crucial info for
    ! each message in the grib_file_attributes structure

    ! input and output parameters
    character(len=*), intent(in)           :: filename_grib
    type(grib_file_attr_data), intent(out) :: grib_file_attributes
    integer, intent(out)                   :: error_flag

    ! local variables
    integer :: msg_nr,msg_id

    error_flag = gribio_no_error

    grib_file_attributes%grib_filename = filename_grib
    grib_file_attributes%file_open = .false.

    ! open the file
    call grib_open_file(grib_file_attributes%grib_fileunit, &
                        grib_file_attributes%grib_filename,'r', &
                        status = error_flag)
    if (error_flag .ne. 0) then
      print *,"ERROR: cannot open ", trim(grib_file_attributes%grib_filename)
      return
    else
      if (GribVerbosity .ge. GribVerbosityReport) then
        print *,"succesfully opened file: ", trim(grib_file_attributes%grib_filename)
      endif
    endif

    ! ok, the file was succesfully opened
    grib_file_attributes%file_open = .true.

    ! set multi mode on, otherwise e.g. u and v fields may be reported as one field
    ! rather than two
    call grib_multi_support_on()

    if (GribVerbosity .ge. GribVerbosityReport) then
      print *,"================================="
      print *,"=== start analysing GRIB file ==="
    endif

    msg_nr = 0
    gribmsgloop: do while (.true.)

      ! read the message
      msg_id = 0
      call grib_new_from_file(grib_file_attributes%grib_fileunit,msg_id, &
                              status = error_flag)

      if (error_flag .eq. GRIB_END_OF_FILE) then
        error_flag = 0
        exit gribmsgloop
      endif
      if (error_flag .ne. 0) then
        print *,'ERROR: problem reading GRIB message'
        return
      endif

      ! check if the message still fits
      msg_nr = msg_nr + 1
      if (msg_nr .gt. maxnr_of_grib_messages) then
        print *,"ERROR: too many messages in GRIB file, more than",maxnr_of_grib_messages
        print *,"ERROR: change the parameter maxnr_of_grib_messages and compile again"
        error_flag = gribio_GRIB_number_too_large
        return
      endif

      grib_file_attributes%list_of_GRIB_message_ids(msg_nr) = msg_id

      ! extract the requested information from it
      call read_GRIB_header_info(grib_file_attributes, &
                                 msg_nr,grib_file_attributes%list_of_GRIB_message_ids(msg_nr), &
                                 error_flag)
      if (error_flag .ne. gribio_no_error) return

    enddo gribmsgloop

    grib_file_attributes%nr_of_GRIB_messages = msg_nr

    if (GribVerbosity .ge. GribVerbosityReport) then
      print *,"number of GRIB messages in this file = ", &
              grib_file_attributes%nr_of_GRIB_messages
    endif

    if (GribVerbosity .ge. GribVerbosityReport) then
      print *,"===  end analysing GRIB file   ==="
      print *,"=================================="
    endif

  end subroutine open_GRIB_file
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine read_GRIB_header_info(grib_file_attributes, which_grib_message, message_id, &
                                   error_flag)
    !  #[ read header info for a given grib msg
    ! this subroutine extracts some data for a selected GRIB message

    ! input and output parameters
    type(grib_file_attr_data), intent(inout) :: grib_file_attributes
    integer, intent(in)                      :: which_grib_message, message_id
    integer, intent(out)                     :: error_flag

    ! local variables
    integer :: date, hour, an_hour, st_hour, date_offset, daycount
    integer :: nr_lon_points, nr_lat_points
    integer :: centre_id, par_id
    integer :: ierr
    character(len=30) :: levtstring, typestring
    character(len=10) :: datestring, timestring, stepstring

    error_flag = gribio_no_error

    if (GribVerbosity .ge. GribVerbosityReport) then
      print *,'read header information of GRIB message', which_grib_message
      print *,'message id = ', message_id
    endif

    call grib_get(message_id,'paramId', par_id)

    ! see if this message is from ECMWF (originating centre 98)
    ! if not, convert the parameter ids which are different for other models, to
    ! the values used by ECMWF
    call grib_get(message_id,'centre',centre_id)
    if (centre_id .ne. 98) then
      if (par_id .eq. par_id_sst_ncep) par_id = par_id_sst
      if (par_id .eq. par_id_t_ncep)   par_id = par_id_t
    endif

    grib_file_attributes%list_of_GRIB_par_id(which_grib_message) = par_id

    call grib_get(message_id,'level', &
                  grib_file_attributes%list_of_GRIB_level(which_grib_message))

    ! read the type of level as a string value and convert it back to numerical
    ! GRIB 1 key value
    call grib_get(message_id,'typeOfLevel',levtstring)
    if (trim(levtstring) .eq. 'surface') then
      grib_file_attributes%list_of_GRIB_level_type(which_grib_message) = 1
    elseif (trim(levtstring) .eq. 'isobaricInhPa') then
      grib_file_attributes%list_of_GRIB_level_type(which_grib_message) = 100
    elseif (trim(levtstring) .eq. 'heightAboveSea') then
      grib_file_attributes%list_of_GRIB_level_type(which_grib_message) = 103
    elseif (trim(levtstring) .eq. 'heightAboveGround') then
      grib_file_attributes%list_of_GRIB_level_type(which_grib_message) = 105
    else
      grib_file_attributes%list_of_GRIB_level_type(which_grib_message) = 0
    endif

    ! get date and time attributes, set default in case string conversion fails
    call grib_get(message_id,'dataDate',datestring)
    date = 19900101
    read(datestring,'(I8)',iostat=ierr) date

    call grib_get(message_id,'dataTime',timestring)
    an_hour = 0
    read(timestring,'(I2)',iostat=ierr) an_hour

    call grib_get(message_id,'stepRange',stepstring)
    st_hour = 0
    read(stepstring,'(I2)',iostat=ierr) st_hour

    ! analyse time + step
    hour = an_hour + st_hour

    grib_file_attributes%list_of_GRIB_date(which_grib_message) = date
    grib_file_attributes%list_of_GRIB_hour(which_grib_message) = hour
    grib_file_attributes%list_of_GRIB_analyse(which_grib_message) = an_hour

    ! try to determine the time interval between subsequent forecasts
    if (st_hour .gt. 0) then
      if (prev_hour .ne. -1) then
        if ((st_hour - prev_hour) .gt. 0 .and. (st_hour - prev_hour) .lt. t_interval) then
          t_interval = st_hour - prev_hour
        endif
      endif
      prev_hour = st_hour
    endif

    ! determine a new date/time combination which can be easier compared
    ! to BUFR date/time fields (by limiting the hour to the range 0-23)
    date_offset = hour/24
    daycount = conv_date_to_daycount(date)
    daycount = daycount + date_offset
    date     = conv_daycount_to_date(daycount)
    hour     = hour - date_offset*24

    grib_file_attributes%list_of_GRIB_derived_date(which_grib_message) = date
    grib_file_attributes%list_of_GRIB_derived_hour(which_grib_message) = hour

    if (GribVerbosity .ge. GribVerbosityReport) then

      print *,'date/time valid = ',date,hour,' H ', &
           'time offset = ',an_hour,' H'

      select case (grib_file_attributes%list_of_GRIB_par_id(which_grib_message))
      case(par_id_t)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : t [temperature: K]'
      case(par_id_u)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : u [u velocity: m/s]'
      case(par_id_v)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : v [v velocity: m/s]'
      case(par_id_q)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : q [specific humidity: kg/kg]'
      case(par_id_lnsp)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : lnsp [log of surface pressure: -]'
      case(par_id_skt)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : skt [skin temperature: K]'
      case(par_id_sst)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : sst [sea surface temperature: K]'
      case(par_id_stl1)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : stl1 [soil temperature level 1: K]'
        print *,'(is this SST above sea ?)'
      case(par_id_chnk)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : chnk [Charnock parameter: -]'
        print *,'?? unused for ERS40 ?? = Surface stress (SS) before 19980519'
      case(par_id_mslp)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : msl [mean sea level pressure: Pa]'
      case(par_id_10u)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : 10u [10 metre U wind component: m/s]'
      case(par_id_10v)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : 10v [10 metre V wind component: m/s]'
      case(par_id_lsm)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : lsm [land-sea mask: -]'
      case(par_id_u10n)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : u10n [10 metre neutral U wind component: m/s]'
      case(par_id_v10n)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : v10n [10 metre neutral V wind component: m/s]'
      case(par_id_2t)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : 2t [2 metre temperature: K]'
      case(par_id_2d)
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : 2d [2 metre dewpoint temperature: K]'
      case default
        print *,'parameter = ',grib_file_attributes%list_of_GRIB_par_id(which_grib_message), &
             ' : not yet implemented parameter'
      end select
    endif

    if (GribVerbosity .ge. GribVerbosityReport) then
      call grib_get(message_id,'Nx',nr_lon_points)
      call grib_get(message_id,'Ny',nr_lat_points)
      call grib_get(message_id, 'gridType', typestring)
      print *,'number of lon points = ', nr_lon_points
      print *,'number of lat points = ', nr_lat_points
      print *,'grid type            = ', trim(typestring)
    endif

    error_flag = gribio_no_error

  end subroutine read_GRIB_header_info
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine get_GRIB_data_values(grib_file_attributes, &   ! input
                                  which_grib_message, &     ! input
                                  grib_message, error_flag) ! output
    !  #[ extract the data for a given grib msg
    ! decode a GRIB message and put the data values into the grib_message

    ! input and output parameters
    type(grib_file_attr_data), intent(in)  :: grib_file_attributes
    integer, intent(in)                    :: which_grib_message
    type(grib_message_data), intent(inout) :: grib_message
    integer, intent(out)                   :: error_flag

    ! local variables
    integer :: message_id, grid_info
    integer :: allocstatus, i, j, k
    character(len=30) :: typestring
    real(r4_), dimension(:), pointer :: real_values ! one dimensional array

    error_flag = gribio_no_error

    if (GribVerbosity .ge. GribVerbosityReport) then
      print *,'get data values of GRIB message', which_grib_message
    endif

    ! copy some meta information from the file attributes
    grib_message%message_id = &
         grib_file_attributes%list_of_GRIB_message_ids(which_grib_message)

    grib_message%message_pos_in_file = which_grib_message

    message_id = grib_message%message_id

    if (.not. grib_message%is_decoded) then

      if (associated(grib_message%real_vals_xy)) then
        print *,"ERROR: in get_GRIB_data_values, this grib_message%real_vals_xy pointer"
        print *,"ERROR: is already associated with an allocated block of memory!"
        print *,"ERROR: cannot read this GRIB message"
        error_flag = gribio_GRIB_decode_problem
        return
      endif

      ! check grid type
      call grib_get(message_id, 'gridType', typestring)
      if (trim(typestring) .ne. 'regular_ll' .and. &
          trim(typestring) .ne. 'regular_gg' .and. &
          trim(typestring) .ne. 'reduced_gg') then
        print *,"ERROR: regular or (reduced) gaussian lat-lon grid expected"
        print *,"ERROR: grid type ", trim(typestring), " not implemented"
        error_flag = gribio_GRIB_decode_problem
        return
      endif
      grib_message%grid_type = typestring

      ! the scan direction in j or y (positively or negatively) does not matter
      ! for the i or x direction we force a positive scan direction, otherwise
      ! we need to complicate the code for the -180 to +180 boundary
      call grib_get(message_id, 'iScansNegatively', grid_info)
      if (grid_info .ne. 0) then
        print *,"ERROR: expected scanning West to East, found the opposite"
        error_flag = gribio_GRIB_decode_problem
        return
      endif

      ! consecutive points in the data values need to be in i or x direction,
      ! i.e., horizontal scanning
      call grib_get(message_id, 'jPointsAreConsecutive', grid_info)
      if (grid_info .ne. 0) then
        print *,"ERROR: expected adjacent grid points to be consecutive,"
        print *,"ERROR: but this is not the case"
        error_flag = gribio_GRIB_decode_problem
        return
      endif

      ! retrieve some grid details
      call grib_get(message_id, 'Ny', grib_message%nr_lat_points)
      call grib_get(message_id, 'Nx', grib_message%nr_lon_points)

      ! read array containing number of lon points for each lat
      ! if this information is not present, take a constant number of lons per
      ! lat (regular grid)
      allocate(grib_message%nr_lons_per_lat(grib_message%nr_lat_points))
      call grib_get(message_id, 'PLPresent', grid_info)
      if (grid_info .eq. 1) then
        call grib_get_size(message_id, 'pl', grid_info)
        if (grid_info .ne. grib_message%nr_lat_points) then
          print *,"ERROR: size of pl is not equal to Ny"
          error_flag = gribio_GRIB_decode_problem
          return
        endif
        call grib_get(message_id, 'pl', grib_message%nr_lons_per_lat)
      else
        if (grib_message%nr_lon_points .lt. 1) then
          print *,"ERROR: expected either a positive Nx or PLPresent,"
          print *,"ERROR: but this is not the case"
          error_flag = gribio_GRIB_decode_problem
          return
        endif
        do i = 1,grib_message%nr_lat_points
          grib_message%nr_lons_per_lat(i) = grib_message%nr_lon_points
        enddo
      endif

      call grib_get(message_id, 'numberOfPoints', grib_message%vals_size)
      call grib_get(message_id, 'latitudeOfFirstGridPointInDegrees', &
           grib_message%lat_of_first_gridpoint)
      call grib_get(message_id, 'longitudeOfFirstGridPointInDegrees', &
           grib_message%lon_of_first_gridpoint)
      call grib_get(message_id, 'latitudeOfLastGridPointInDegrees', &
           grib_message%lat_of_last_gridpoint)
      call grib_get(message_id, 'longitudeOfLastGridPointInDegrees', &
           grib_message%lon_of_last_gridpoint)

      ! ensure that lon_of_last_gridpoint is greater than lon_of_first_gridpoint
      ! sometimes we get GRIB messages going from e.g. +180.3 to 180.0
      ! when this happens, change the lon_of_first_gridpoint to -179.7
      if (grib_message%lon_of_last_gridpoint .lt. &
          grib_message%lon_of_first_gridpoint) then
        grib_message%lon_of_last_gridpoint = &
             grib_message%lon_of_last_gridpoint + 360.0
        if (grib_message%lon_of_first_gridpoint .gt. 180.0) then
          grib_message%lon_of_first_gridpoint = &
               grib_message%lon_of_first_gridpoint - 360.0
          grib_message%lon_of_last_gridpoint = &
               grib_message%lon_of_last_gridpoint - 360.0
        endif
      endif

      grib_message%lat_step = &
           (grib_message%lat_of_last_gridpoint - &
            grib_message%lat_of_first_gridpoint)/&
            real(grib_message%nr_lat_points-1,r4_)

      allocate(grib_message%lon_step_per_lat(grib_message%nr_lat_points))
      do i = 1,grib_message%nr_lat_points
         if (trim(typestring) .eq. 'reduced_gg') then
            ! For reduced Gaussian grid, just take 360/n to calc lon step
            grib_message%lon_step_per_lat(i) = &
                 360./real(grib_message%nr_lons_per_lat(i),r4_)
         else
            ! this is wrong for reduced Gaussian!
            ! lon_of_last_gridpoint only gives maximum value for the
            ! equator, and does not take into account that is should
            ! change for each latitude !
            ! This next calculation is only to be used for regular grids.
            grib_message%lon_step_per_lat(i) = &
                 (grib_message%lon_of_last_gridpoint - &
                 grib_message%lon_of_first_gridpoint)/&
                 real(grib_message%nr_lons_per_lat(i)-1,r4_)
         end if
      enddo

      ! read the data values
      allocate(real_values(1:grib_message%vals_size), &
           STAT=allocstatus)
      if (allocstatus .ne. 0) then
        print *,"ERROR: allocation failure for real_values in gribio_module"
        error_flag = gribio_GRIB_decode_problem
        return
      endif

      if (GribVerbosity .ge. GribVerbosityReport) then
        print *,"succesfully allocated ",grib_message%vals_size,&
             " elements for real_values"
        print *,"start decoding data values of the GRIB message"
      endif

      call grib_set(message_id, 'missingValue', grib_missing_value)
      call grib_get(message_id, 'values', real_values, error_flag)
      if (error_flag .gt. gribio_no_error) then
        print *,"ERROR: grib data values not read properly"
        print *,'ERROR: grib_get error code: ',error_flag
        return
      endif

      ! set the missing values to missing_indicator_real
      do i = 1,grib_message%vals_size
        if (real_values(i) .gt. grib_missing_value*0.99) then
           real_values(i) = missing_indicator_real
        endif
      enddo

      ! transform the one dimensional values array into a two dimensional
      ! array
      allocate(grib_message%real_vals_xy(maxval(grib_message%nr_lons_per_lat), &
           grib_message%nr_lat_points), STAT=allocstatus)
      if (allocstatus .ne. 0) then
        print *,"ERROR: allocation failure for real_vals_xy in gribio_module"
        error_flag = gribio_GRIB_decode_problem
        return
      endif
      k = 1
      do j = 1,grib_message%nr_lat_points
        do i = 1,grib_message%nr_lons_per_lat(j)
          grib_message%real_vals_xy(i,j) = real_values(k)
          k = k + 1
        enddo
      enddo

      ! deallocate one dimensional values array since we don't need it any more
      if (associated(real_values)) then
        deallocate(real_values)
        !nullify(real_values)
      endif

      grib_message%is_decoded = .true.
    endif

  end subroutine get_GRIB_data_values
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine dealloc_GRIB_message(grib_message, error_flag)
    !  #[ discard data for a given grib msg

    ! input and output parameters
    type(grib_message_data), intent(inout) :: grib_message
    integer, intent(out)                   :: error_flag

    ! local variables
    integer :: allocstatus

    error_flag = gribio_no_error

    if (grib_message%is_decoded) then
      if (associated(grib_message%real_vals_xy)) then
        deallocate(grib_message%real_vals_xy, STAT=allocstatus)
        nullify(grib_message%real_vals_xy)
        grib_message%is_decoded = .false.
        if (allocstatus .ne. 0) then
          print *,'deallocation problem in dealloc_GRIB_message, status = ',allocstatus
        endif
      else
        print *,'pointer seemed not associated: grib_message%real_vals_xy'
      endif

      if (associated(grib_message%nr_lons_per_lat)) then
        deallocate(grib_message%nr_lons_per_lat)
        nullify(grib_message%nr_lons_per_lat)
      endif
      if (associated(grib_message%lon_step_per_lat)) then
        deallocate(grib_message%lon_step_per_lat)
        nullify(grib_message%lon_step_per_lat)
      endif
    endif

  end subroutine dealloc_GRIB_message
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine dealloc_all_GRIB_messages()
    !  #[ deallocate all grib data and close all open grib files

    ! local variables
    integer :: i,j,error_flag

    if (.not. allocated(list_of_GRIB_files)) then
      if (GribVerbosity .ge. GribVerbosityReport) then
        print *, &
          'list of GRIB files is not in memory, it cannot be deallocated by dealloc_all_GRIB_messages'
      endif
      return
    endif

    do i = 1,max_nr_open_grib_files

      ! deallocate messages
      do j = 1,max_nr_open_grib_msgs
        if (list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%is_decoded) then
          if (GribVerbosity .ge. GribVerbosityReport) then
            print *,"deallocating grib message ", &
              list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%message_pos_in_file, &
              " from file number", i
          endif
          call dealloc_GRIB_message( &
            list_of_GRIB_files(i)%list_of_GRIB_msgs(j),error_flag)
        endif
      enddo

      ! close file
      if (list_of_GRIB_files(i)%grib_file_attributes%file_open) then

        ! release messages from memory
        do j = 1, list_of_GRIB_files(i)%grib_file_attributes%nr_of_GRIB_messages
          call grib_release(list_of_GRIB_files(i)%grib_file_attributes%list_of_GRIB_message_ids(j))
          list_of_GRIB_files(i)%grib_file_attributes%list_of_GRIB_message_ids(j) = 0
        enddo

        call grib_close_file(list_of_GRIB_files(i)%grib_file_attributes%grib_fileunit, &
        error_flag)
        if (error_flag .ne. 0) then
          print *,"ERROR: closing of GRIB file", &
            trim(list_of_GRIB_files(i)%grib_file_attributes%grib_filename), &
            " failed in dealloc_all_GRIB_messages"
          stop 1
        endif

        list_of_GRIB_files(i)%grib_file_attributes%file_open = .false.
      endif
    enddo

    deallocate(list_of_GRIB_files)

  end subroutine dealloc_all_GRIB_messages
  !---------------------------------------------------
  !  #]
! can be made private
  subroutine extract_data_from_GRIB_message(grib_message, lat, lon, &    ! input
                                            logn_to_linear_conversion, & ! optional
                                            val, error_flag)             ! output
    !  #[ extract interpolated data for a given lat-lon from a given grib msg
    ! extract data at the 4 grid-points around the requested lat-lon
    ! position, and interpolate the result.
    !
    ! some parameters are stored in GRIB in logarithmic scale. Prior to
    ! the spatial and the temporal interpolations, a conversion to linear
    ! scale is therefore required. The latter is enabled by setting the
    ! option logn_to_linear_conversion.

    ! input and output parameters
    type(grib_message_data), intent(in) :: grib_message
    real(r4_), intent(in)                :: lat, lon
    logical, intent(in), optional       :: logn_to_linear_conversion
    real(r8_), intent(out)              :: val
    integer, intent(out)                :: error_flag

    ! local variables
    real(r4_)  :: val11, val12, val21, val22, val1, val2
    real(r4_)  :: lon_tmp, diff_lat, diff_lon1, diff_lon2
    integer   :: lat_index1, lat_index2
    integer   :: lon_index11, lon_index12, lon_index21, lon_index22
    integer   :: lat_indexc, lon_indexc
    logical   :: do_logn_to_linear_conversion

! WARNING: SST uses the value of 0 for missing or no data !!!!!!!
!          this happens above land.
!          Therefore it is better to use soil-temperrature-level1 (stl1)
!          which is equal to sst above sea, and causes no problems
!          when used for interpolations near land.

    do_logn_to_linear_conversion = .false.
    if (present(logn_to_linear_conversion)) then
      if (logn_to_linear_conversion) do_logn_to_linear_conversion = .true.
    endif

    error_flag = gribio_no_error
    val = missing_indicator_real

    if (.not. grib_message%is_decoded) then
      print *,"ERROR: can not extract data from GRIB message since it is not yet decoded"
      error_flag = gribio_GRIB_decode_problem
      return
    endif

    lat_index1 = floor((lat - grib_message%lat_of_first_gridpoint) / grib_message%lat_step) + 1
    lat_index2 = lat_index1 + 1

    ! return with missing value if location is outside latitude boundaries
    if (lat_index1 .lt. 1 .or. lat_index2 .gt. grib_message%nr_lat_points) then
      return
    endif

    ! store lon in local variable which can be adjusted
    lon_tmp = lon
    if (lon_tmp .lt. grib_message%lon_of_first_gridpoint) then
      lon_tmp = lon_tmp + 360.0
    endif
    if (lon_tmp .gt. grib_message%lon_of_last_gridpoint) then
      lon_tmp = lon_tmp - 360.0
    endif

    lon_index11 = &
        floor((lon_tmp - grib_message%lon_of_first_gridpoint) / grib_message%lon_step_per_lat(lat_index1)) + 1
    lon_index21 = lon_index11 + 1

    lon_index12 = &
        floor((lon_tmp - grib_message%lon_of_first_gridpoint) / grib_message%lon_step_per_lat(lat_index2)) + 1
    lon_index22 = lon_index12 + 1

    ! see if this GRIB field is global (i.e., covers 360 degrees in longitude)
    if (grib_message%lon_of_last_gridpoint - grib_message%lon_of_first_gridpoint .gt. &
        359.9 - grib_message%lon_step_per_lat(lat_index1)) then

      ! jump from +180 to -180 if needed for periodic boundary conditions
      if (lon_index11 .gt. grib_message%nr_lons_per_lat(lat_index1)) &
        lon_index11 = lon_index11 - grib_message%nr_lons_per_lat(lat_index1)
      if (lon_index11 .lt. 1) &
        lon_index11 = lon_index11 + grib_message%nr_lons_per_lat(lat_index1)
      if (lon_index21 .gt. grib_message%nr_lons_per_lat(lat_index1)) &
        lon_index21 = lon_index21 - grib_message%nr_lons_per_lat(lat_index1)
      if (lon_index21 .lt. 1) &
        lon_index21 = lon_index21 + grib_message%nr_lons_per_lat(lat_index1)

      if (lon_index12 .gt. grib_message%nr_lons_per_lat(lat_index2)) &
        lon_index12 = lon_index12 - grib_message%nr_lons_per_lat(lat_index2)
      if (lon_index12 .lt. 1) &
        lon_index12 = lon_index12 + grib_message%nr_lons_per_lat(lat_index2)
      if (lon_index22 .gt. grib_message%nr_lons_per_lat(lat_index2)) &
        lon_index22 = lon_index22 - grib_message%nr_lons_per_lat(lat_index2)
      if (lon_index22 .lt. 1) &
        lon_index22 = lon_index22 + grib_message%nr_lons_per_lat(lat_index2)
    else

      ! return with missing value if location is outside longitude boundaries
      ! of regional field
      if (lon_index11 .lt. 1 .or. lon_index21 .gt. grib_message%nr_lons_per_lat(lat_index1) .or. &
          lon_index12 .lt. 1 .or. lon_index22 .gt. grib_message%nr_lons_per_lat(lat_index2)) then
        return
      endif
    endif

    ! grid values surrounding the lat/lon position O of interest
    !
    !    X val12  X val2  X val22
    !
    !             O val
    !
    !   X val11   X val1   X val21
    !
    val11 = grib_message%real_vals_xy(lon_index11,lat_index1)
    val21 = grib_message%real_vals_xy(lon_index21,lat_index1)
    val12 = grib_message%real_vals_xy(lon_index12,lat_index2)
    val22 = grib_message%real_vals_xy(lon_index22,lat_index2)

    ! the interpolation can be done only when all 4 values are
    ! not missing
    if (val11 .lt. missing_indicator_real*0.99 .and. &
        val12 .lt. missing_indicator_real*0.99 .and. &
        val21 .lt. missing_indicator_real*0.99 .and. &
        val22 .lt. missing_indicator_real*0.99) then

      ! convert GRIB data in logn to linear scale
      if (do_logn_to_linear_conversion) then
        val11 = exp(val11)
        val12 = exp(val12)
        val21 = exp(val21)
        val22 = exp(val22)
      endif

      ! step 1: interpolate in x direction between val11 and val21, to obtain val1
      diff_lon1 = lon_tmp - grib_message%lon_of_first_gridpoint - &
                  (lon_index11-1) * grib_message%lon_step_per_lat(lat_index1)
      if (diff_lon1 .gt.  180.) diff_lon1 = diff_lon1 - 360.
      if (diff_lon1 .lt. -180.) diff_lon1 = diff_lon1 + 360.
      val1 = val11 + (val21 - val11) * diff_lon1 / grib_message%lon_step_per_lat(lat_index1)

      ! step 2: interpolate in x direction between val12 and val22, to obtain val2
      diff_lon2 = lon_tmp - grib_message%lon_of_first_gridpoint - &
                  (lon_index12-1) * grib_message%lon_step_per_lat(lat_index2)
      if (diff_lon2 .gt.  180.) diff_lon2 = diff_lon2 - 360.
      if (diff_lon2 .lt. -180.) diff_lon2 = diff_lon2 + 360.
      val2 = val12 + (val22 - val12) * diff_lon2 / grib_message%lon_step_per_lat(lat_index2)

      ! step 3: interpolate in y direction between val1 and val2 to obtain val
      diff_lat = lat - grib_message%lat_of_first_gridpoint - &
                 (lat_index1-1) * grib_message%lat_step

      val = val1 + (val2 - val1) * diff_lat / grib_message%lat_step

    else
      ! if one or more values are missing, report the closest value
      ! (which may be missing as well)
      lat_indexc = nint((lat - grib_message%lat_of_first_gridpoint)/&
                        grib_message%lat_step) + 1

      ! return with missing value if location is outside latitude boundaries
      if (lat_indexc .lt. 1 .or. lat_indexc .gt. grib_message%nr_lat_points) then
        return
      endif

      lon_indexc = nint((lon_tmp - grib_message%lon_of_first_gridpoint)/&
                        grib_message%lon_step_per_lat(lat_indexc)) + 1

      ! see if this GRIB field is global (i.e., covers 360 degrees in longitude)
      if (grib_message%lon_of_last_gridpoint - grib_message%lon_of_first_gridpoint .gt. &
          359.9 - grib_message%lon_step_per_lat(lat_indexc)) then

        ! jump from +180 to -180 if needed for periodic boundary conditions
        if (lon_indexc .gt. grib_message%nr_lons_per_lat(lat_indexc)) &
          lon_indexc = lon_indexc - grib_message%nr_lons_per_lat(lat_indexc)
        if (lon_indexc .lt. 1) &
          lon_indexc = lon_indexc + grib_message%nr_lons_per_lat(lat_indexc)
      else

        ! return with missing value if location is outside longitude boundaries
        ! of regional field
        if (lon_indexc .lt. 1 .or. lon_indexc .gt. grib_message%nr_lons_per_lat(lat_indexc)) then
          return
        endif
      endif

      val = grib_message%real_vals_xy(lon_indexc,lat_indexc)

      ! convert GRIB data in logn to linear scale
      if (do_logn_to_linear_conversion .and. val .lt. missing_indicator_real*0.99) then
        val = exp(val)
      endif
    endif

  end subroutine extract_data_from_GRIB_message
  !---------------------------------------------------
  !  #]
! note: as far as I can see this routine is nowhere used
! in our software. For now disable it by changing its name
! before actually removing it
! only the test programs test_read_GRIB2.F90 and
! test_read_GRIB3.F90 call it
  subroutine xget_field_from_GRIB_file(filename_grib, grib_par_id, &
             max_lat_size, max_lon_size, &
             lat_of_first_gridpoint, lat_step, nr_lat_points, &
             lon_of_first_gridpoint, lon_step, nr_lon_points, &
             field, &
             derived_date, derived_hour, &
             date, hour, analyse, level, level_type)
    !  #[ extract a 2D field from a given grib msg

! to do: add code to select proper field for a given
!        year, month, day, or time in case multiple fields
!        with the same code are present in a GRIB file

    ! input parameters
    character(len=256), intent(in) :: filename_grib
    integer, intent(in) :: grib_par_id
    integer, intent(in) :: max_lat_size, max_lon_size

    ! output parameters
    real(r4_), intent(out) :: lat_of_first_gridpoint, lat_step
    integer, intent(out)  :: nr_lat_points
    real(r4_), intent(out) :: lon_of_first_gridpoint, lon_step
    integer, intent(out)  :: nr_lon_points

    ! output parameter (should be allocated by calling program)
    real(r4_), dimension(1:max_lat_size,1:max_lon_size), intent(out) :: field

    ! optional input parameters
    integer, intent(in), optional :: derived_date, derived_hour
    integer, intent(in), optional :: date, hour, analyse, level, level_type

    ! local variables
    type(grib_file_attr_data) :: grib_file_attributes
    type(grib_message_data)   :: grib_message       ! the message itself
    integer :: i, j, error_flag
    integer :: which_grib_message

    ! to enable signalling the calling program that an error occurred
    nr_lat_points = -1
    nr_lon_points = -1

    ! initialise grib message
    grib_message%message_pos_in_file = -1
    grib_message%is_decoded = .false.
    nullify(grib_message%nr_lons_per_lat)
    nullify(grib_message%lon_step_per_lat)
    !nullify(grib_message%real_values)
    nullify(grib_message%real_vals_xy)

    ! open the file, count messages and get some information on each message
    call open_GRIB_file(filename_grib, grib_file_attributes, error_flag)
    if (error_flag .ne. gribio_no_error) then
      stop 1
    endif

    ! see which message has the requested par_id
    which_grib_message = &
         get_GRIB_msgnr(grib_file_attributes, grib_par_id, &
                        derived_date, derived_hour, &
                        date, hour, analyse, level, level_type)
    if (which_grib_message .eq. -1) then
      ! a warning was already printed inside get_GRIB_msgnr
      return
    else
      if (GribVerbosity .ge. GribVerbosityReport) then
        print *,"GRIB message ", which_grib_message, &
                " contains par_id: ", grib_par_id
      endif
    endif

    ! retrieve the data values
    call get_GRIB_data_values(grib_file_attributes, &
                              which_grib_message, grib_message, error_flag)
    if (error_flag .ne. gribio_no_error) then
      stop 1
    endif

    ! this can't be done with reduced grids
    if (grib_message%grid_type .ne. 'regular_ll' .and. &
        grib_message%grid_type .ne. 'regular_gg') then
      print *,"ERROR: in get_field_from_GRIB_file, a regular_ll or"
      print *,"ERROR: regular_gg GRIB field must be used"
      stop 1
    endif

    lat_of_first_gridpoint = grib_message%lat_of_first_gridpoint
    lat_step               = grib_message%lat_step
    nr_lat_points          = grib_message%nr_lat_points

    lon_of_first_gridpoint = grib_message%lon_of_first_gridpoint
    lon_step               = grib_message%lon_step_per_lat(1)
    nr_lon_points          = grib_message%nr_lon_points

    if (nr_lat_points .gt. max_lat_size) then
      print *,"WARNING: latitude dimension of field is too small to contain"
      print *,"WARNING: the read data; truncating the array !!!!!"
      print *,"original:  nr_lat_points = ",nr_lat_points
      nr_lat_points = max_lat_size
      print *,"truncated: nr_lat_points = ",nr_lat_points
    endif

    if (nr_lon_points .gt. max_lon_size) then
      print *,"WARNING: longitude dimension of field is too small to contain"
      print *,"WARNING: the read data; truncating the array !!!!!"
      print *,"original:  nr_lon_points = ",nr_lon_points
      nr_lon_points = max_lon_size
      print *,"truncated: nr_lon_points = ",nr_lon_points
    endif

    ! copy the data for this field to the output array
    do i = 1,nr_lon_points
      do j = 1,nr_lat_points
        field(j,i) = grib_message%real_vals_xy(i,j)
      enddo
    enddo

    ! deallocate the GRIB message and close the GRIB file
    call dealloc_GRIB_message(grib_message, error_flag)
    if (error_flag .ne. gribio_no_error) then
      stop 1
    endif

    if (grib_file_attributes%file_open) then

      ! release messages from memory
      do j = 1, grib_file_attributes%nr_of_GRIB_messages
        call grib_release(grib_file_attributes%list_of_GRIB_message_ids(j))
        grib_file_attributes%list_of_GRIB_message_ids(j) = 0
      enddo

      call grib_close_file(grib_file_attributes%grib_fileunit, error_flag)
      if (error_flag .ne. 0) then
        print *,"ERROR: closing of GRIB file", trim(grib_file_attributes%grib_filename), &
          " failed in get_field_from_GRIB_file"
        stop 1
      endif

      grib_file_attributes%file_open = .false.
    else
      if (GribVerbosity .ge. GribVerbosityWarn) then
        print *,"WARNING: file ", trim(grib_file_attributes%grib_filename), &
          " already seems to be closed"
      endif
    endif

  end subroutine xget_field_from_GRIB_file
  !---------------------------------------------------------------
  !  #]

! note: as far as I can see this routine is nowhere used
! in our software. For now disable it by changing its name
! before actually removing it
! only the test programs test_read_GRIB1.F90 and
! test_read_GRIB3.F90 call it
  function xget_from_GRIB_file(filename_grib, lat, lon, grib_par_id, &
                              alt_grib_par_id, &
                              derived_date, derived_hour, &
                              date, hour, analyse, &
                              level, level_type, &
                              logn_to_linear_conversion) result(val)
    !  #[ extract interpolated data from a grib file

    ! some special commands for getting time/date from a grib file
    ! (the lat-parameter is used to pass this command to the subroutine
    !  get_from_GRIB_file, so make sure the code is not a valid latitude)
    !         grib_get_time = -999 ! command to get the grib time
    !         grib_get_date = -998 ! command to get the grib date
    !         grib_get_derived_time = -997.  ! cmd to get the derived grib time
    !         grib_get_derived_date = -996.  ! cmd to get the derived grib date
    !
    ! (not very beautifull but I cannot use optional input parameters
    !  to do this, because this
    !  routine is the interface that is called by the c-wrapper when
    !  using it for IDL, and I don't know how to handle optional
    !  parameters in c ....)
    !
    ! some parameters are stored in GRIB in logarithmic scale. Prior to
    ! the spatial and the temporal interpolations, a conversion to linear
    ! scale is therefore required. The latter is enabled by setting the
    ! option logn_to_linear_conversion.

    ! input parameters
    character(len=*), intent(in)  :: filename_grib
    real(r4_), intent(in)          :: lat, lon
    integer, intent(in)           :: grib_par_id

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: derived_date, derived_hour
    integer, intent(in), optional :: date, hour, analyse, level, level_type
    logical, intent(in), optional :: logn_to_linear_conversion

    ! output, type r8_ is needed to be able to pass the date via this field !
    real(r8_)                     :: val

    ! local parameters
    integer :: i, j, file_pos, msg_pos, error_flag
    integer :: found_msg_nr

    if (.not. allocated(list_of_GRIB_files)) then
      print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
      print *,"ERROR: init_GRIB_module should be called first"
      stop 1
    endif

    val = missing_indicator_real

    ! find out what the position of this file is in the list
    file_pos = -1
    fileloop: do i = 1,max_nr_open_grib_files
      if (list_of_GRIB_files(i)%grib_file_attributes%file_open) then
        if (trim(filename_grib) .eq. &
            trim(list_of_GRIB_files(i)%grib_file_attributes%grib_filename)) then
          file_pos = i
          exit fileloop
        endif
      endif
    enddo fileloop

    ! file is not yet open, so open it if possible
    if (file_pos .eq. -1) then

      ! find an empty spot in the file list struct
      fileloop2: do i = 1,max_nr_open_grib_files
        if (.not. list_of_GRIB_files(i)%grib_file_attributes%file_open) then
          file_pos = i
          exit fileloop2 ! break the do looop
        endif
      enddo fileloop2

      if (file_pos .eq. -1) then
        print *,"ERROR: only ",max_nr_open_grib_files," GRIB files can be opened"
        print *,"ERROR: simultaneously inside the routine get_from_GRIB_file"
        print *,"ERROR: unable to open this GRIB file!"
        print *,"ERROR: change the parameter max_nr_open_grib_files and compile again"
        ! TO DO:
        ! close a file to make room for the new one
        stop 1
      endif

      ! open the file using the found file_pos, count messages and get
      ! some information on each message
      call open_GRIB_file(filename_grib, &
               list_of_GRIB_files(file_pos)%grib_file_attributes,error_flag)
      if (error_flag .ne. gribio_no_error) then
        stop 1
      endif
    endif ! end of code for opening the GRIB file

    ! see which message has the requested par_id
    found_msg_nr = &
         get_GRIB_msgnr(list_of_GRIB_files(file_pos)%grib_file_attributes, &
                        grib_par_id, &
                        alt_grib_par_id, &
                        derived_date, derived_hour, &
                        date, hour, analyse, level, level_type)
    if (found_msg_nr .eq. -1) then
      return
    endif

    ! some information is already set by read_GRIB_header_info
    if (lat .eq. xgrib_get_time) then
      val = list_of_GRIB_files(file_pos)%grib_file_attributes%list_of_GRIB_hour(found_msg_nr)
      return
    endif

    ! get_analyse_time not yet implemented

    if (lat .eq. xgrib_get_date) then
      val = list_of_GRIB_files(file_pos)%grib_file_attributes%list_of_GRIB_date(found_msg_nr)
      return
    endif

    if (lat .eq. xgrib_get_derived_time) then
      val = list_of_GRIB_files(file_pos)%grib_file_attributes%list_of_GRIB_derived_hour(found_msg_nr)
      return
    endif

    if (lat .eq. xgrib_get_derived_date) then
      val = list_of_GRIB_files(file_pos)%grib_file_attributes%list_of_GRIB_derived_date(found_msg_nr)
      return
    endif

    ! see if this message is already decoded
    msg_pos = -1
    msgloop: do j = 1,max_nr_open_grib_msgs
      if (list_of_GRIB_files(file_pos)%list_of_GRIB_msgs(j)%message_pos_in_file &
          .eq. found_msg_nr) then
        msg_pos = j
        exit msgloop
      endif
    enddo msgloop

    ! message is not yet decoded
    if (msg_pos .eq. -1) then

      ! find an empty spot in the msg list struct
      msgloop2: do j = 1,max_nr_open_grib_msgs
        if (list_of_GRIB_files(file_pos)%list_of_GRIB_msgs(j)%message_pos_in_file &
             .eq. -1) then
          msg_pos = j
          exit msgloop2
        endif
      enddo msgloop2

      if (msg_pos .eq. -1) then
        print *,"ERROR: only ",max_nr_open_grib_msgs," GRIB messages can be opened"
        print *,"ERROR: simultaneously inside the routine get_from_GRIB_file"
        print *,"ERROR: unable to open this GRIB message!"
        print *,"ERROR: change the parameter max_nr_open_grib_msgs and compile again"
        ! TO DO:
        ! close a message to make room for the new one
        ! ...this is probably not a very good thing to do (bad for efficiency !)
        stop 1
!       ! a new GRIB message from the same file is needed, so delete the
!       ! old one first
!       print *,"INFO: deleting GRIB message, and reading new one"
!       call dealloc_GRIB_message(list_of_GRIB_files(file_pos)%grib_message,&
!                                 error_flag)
      endif
    endif ! end of (re)loading GRIB message

    if (.not. list_of_GRIB_files(file_pos)%list_of_GRIB_msgs(msg_pos)%is_decoded) then

      ! retrieve the data values
      call get_GRIB_data_values( &
           list_of_GRIB_files(file_pos)%grib_file_attributes, &
           found_msg_nr, &
           list_of_GRIB_files(file_pos)%list_of_GRIB_msgs(msg_pos), &
           error_flag)
      if (error_flag .ne. gribio_no_error) then
        stop 1
      endif
    endif

    ! get interpolated data value
    call extract_data_from_GRIB_message( &
         list_of_GRIB_files(file_pos)%list_of_GRIB_msgs(msg_pos), &
         lat, lon, logn_to_linear_conversion, val, error_flag)
    if (error_flag .ne. gribio_no_error) then
      stop 1
    endif

  end function xget_from_GRIB_file
  !---------------------------------------------------------------

  !  #]
  subroutine set_GRIB_filelist(GRIB_filenames, nr_of_GRIB_files)
    !  #[ set the list of grib files to be used

    ! input parameters
    integer, intent(in)                            :: nr_of_GRIB_files
    character(len=256), dimension(1:*), intent(in) :: GRIB_filenames

    ! local variables
    integer :: i,j, error_flag, filecount

    if (.not. allocated(list_of_GRIB_files)) then
      print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
      print *,"ERROR: init_GRIB_module should be called first"
      stop 1
    endif

    if (nr_of_GRIB_files .gt. max_nr_open_grib_files) then
      print *,"ERROR: only ",max_nr_open_grib_files," GRIB files can be used"
      print *,"ERROR: simultaneously inside the routine set_GRIB_filelist"
      print *,"ERROR: change the parameter max_nr_open_grib_files and compile again"
      stop 1
    endif

    ! set all GRIB file entries to closed and initialise message elements
    do i = 1,max_nr_open_grib_files
      list_of_GRIB_files(i)%grib_file_attributes%file_open = .false.
      do j = 1,max_nr_open_grib_msgs
        list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%message_pos_in_file = -1

        list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%is_decoded = .false.
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%nr_lons_per_lat)
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%lon_step_per_lat)
        !nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%real_values)
        nullify(list_of_GRIB_files(i)%list_of_GRIB_msgs(j)%real_vals_xy)
      enddo
    enddo

    filecount = 0
    fileloop: do i = 1,nr_of_GRIB_files
      !print *,'DEBUG: trying to load: ',trim(GRIB_filenames(i))

      ! check to prevent using files more than once
      do j = 1,i-1
        if (trim(GRIB_filenames(j)) .eq. trim(GRIB_filenames(i))) then
          if (GribVerbosity .ge. GribVerbosityWarn) then
            print *,"WARNING: file ",trim(GRIB_filenames(i))
            print *,"was found earlier in the list, and will be skipped now"
          endif
          cycle fileloop
        endif
      enddo

      filecount = filecount + 1

      ! open the file, count messages and get some information on each message
      call open_GRIB_file(GRIB_filenames(i), &
               list_of_GRIB_files(filecount)%grib_file_attributes, error_flag)

      if (error_flag .ne. gribio_no_error) then
        stop 1
      endif

    enddo fileloop

    nr_of_GRIB_files_in_list = filecount

  end subroutine set_GRIB_filelist
  !---------------------------------------------------------------
  !  #]
  subroutine add_to_GRIB_filelist(GRIB_filename)
    !  #[ add a file to the list of grib files
    character(len=256) :: GRIB_filename
    integer            :: filecount, error_flag,i,j

    if (.not. allocated(list_of_GRIB_files)) then
      print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
      print *,"ERROR: init_GRIB_module should be called first"
      stop 1
    endif

    if (nr_of_GRIB_files_in_list .ge. max_nr_open_grib_files) then
      print *,"ERROR: only ",max_nr_open_grib_files," GRIB files can be used"
      print *,"ERROR: simultaneously inside the routine add_to_GRIB_filelist"
      print *,"ERROR: change the parameter max_nr_open_grib_files and compile again"
      stop 1
    endif

    ! see if the file is already open
    do i = 1,nr_of_GRIB_files_in_list
      if (trim(list_of_GRIB_files(i)%grib_file_attributes%grib_filename) .eq. &
          trim(GRIB_filename)) then
        if (GribVerbosity .ge. GribVerbosityWarn) then
          print *,"WARNING: file ", trim(GRIB_filename), " was already present in"
          print *,"list of GRIB files, and will be skipped now"
        endif
        return
      endif
    enddo

    filecount = nr_of_GRIB_files_in_list + 1

    ! open the file, count messages and get some information on each message
    call open_GRIB_file(GRIB_filename, &
             list_of_GRIB_files(filecount)%grib_file_attributes, &
             error_flag)
    if (error_flag .ne. gribio_no_error) then
      stop 1
    endif

    ! initialise message elements
    do j = 1,max_nr_open_grib_msgs
      list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%message_pos_in_file = -1

      list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%is_decoded = .false.
      nullify(list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%nr_lons_per_lat)
      nullify(list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%lon_step_per_lat)
      !nullify(list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%real_values)
      nullify(list_of_GRIB_files(filecount)%list_of_GRIB_msgs(j)%real_vals_xy)
    enddo

    nr_of_GRIB_files_in_list = filecount

  end subroutine add_to_GRIB_filelist
  !---------------------------------------------------------------
  !  #]
  function get_from_GRIB_filelist(lat, lon, grib_par_id,&
                                  alt_grib_par_id, &
                                  derived_date, derived_hour, &
                                  date, hour, analyse, level, level_type,&
                                  getmeanval, radius,&
                                  logn_to_linear_conversion) result(val)
    !  #[ extract interpolated data from the list of grib files
    ! as extra option (added 11-05-2004) the possibility to check all
    ! values on the gridpoints within a given radius around the given
    ! lat, lon values, and return the maximum found value (usefull for LSM)

    ! -some parameters are stored in GRIB in logarithmic scale. Prior to
    !  the spatial and the temporal interpolations, a conversion to linear
    !  scale is therefore required. The latter is enabled by setting the
    !  option logn_to_linear_conversion.


    ! input parameters
    real(r4_), intent(in)          :: lat, lon
    integer, intent(in)            :: grib_par_id

    ! optional input parameters
    integer,  intent(in), optional :: alt_grib_par_id
    integer,  intent(in), optional :: derived_date, derived_hour
    integer,  intent(in), optional :: date, hour, analyse, level, level_type
    logical,  intent(in), optional :: getmeanval
    real(r4_), intent(in), optional :: radius
    logical,  intent(in), optional :: logn_to_linear_conversion

    ! output, type r8_ is needed to be able to pass the date via this field !
    real(r8_)                      :: val

    ! local variables
    integer   :: i,j,m, error_flag
    integer   :: found_file_nr, found_msg_nr, msg_pos
    logical   :: do_getmeanval
    logical   :: multiple_found_msg_written = .false.
    logical   :: not_found_msg_written      = .false.
    logical   :: coarse_res_msg_written     = .false.
    integer   :: nr_lat_points, nr_lon_points
    integer   :: lat_index1, lat_index2, lat_index, tmp_index
    integer   :: lon_index1, lon_index2, lon_index, tmp_lon_index
    real(r4_)  :: lat1, lat2, lon1, lon2, cos_lat, latlon_dist
    real(r4_)  :: lat_step, lon_step, this_lat, this_lon
    real(r4_)  :: lat_of_first_gridpoint, lon_of_first_gridpoint
    real(r4_)  :: angle_dist
    real(r4_)  :: average_val, sum_vals, sum_weights, weight
    real(r8_) :: first_val
    integer   :: nr_of_points
    logical   :: do_logn_to_linear_conversion, all_values_identical

    if (.not. allocated(list_of_GRIB_files)) then
      print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
      print *,"ERROR: init_GRIB_module should be called first"
      stop 1
    endif

    val = missing_indicator_real

    ! check whether getmaxval and radius are present
    do_getmeanval = .false.
    if (present(getmeanval)) then
      if (getmeanval) then
        if (present(radius)) then
          do_getmeanval = .true.
        else
          print *,"ERROR: when the option getmeanval is used, a radius"
          print *,"ERROR: must also be supplied"
          print *,"ERROR: wrong usage of get_from_GRIB_filelist, quitting ..."
          stop 1
        endif
      endif
    endif

    do_logn_to_linear_conversion = .false.
    if (present(logn_to_linear_conversion)) then
      if (logn_to_linear_conversion) do_logn_to_linear_conversion = .true.
    endif

    ! find out which gribfile matches with the requested
    ! input parameters:
    found_file_nr = -1
    found_msg_nr  = -1

    fileloop: do i = 1,nr_of_GRIB_files_in_list

       if (.not. list_of_GRIB_files(i)%grib_file_attributes%file_open) cycle

       j = get_GRIB_msgnr(list_of_GRIB_files(i)%grib_file_attributes, &
                          grib_par_id, alt_grib_par_id, &
                          derived_date, derived_hour, &
                          date, hour, analyse, level, level_type, &
                          silent = .true.)

       if (j .ne. -1) then
          if ((found_file_nr .eq. -1) .and. (found_msg_nr .eq. -1)) then
             found_file_nr = i
             found_msg_nr  = j

             if ( multiple_found_msg_written .or. &
                  GribVerbosity .lt. GribVerbosityWarn) exit fileloop
          else

             ! write warning message only once, not for each node again
             if (.not. multiple_found_msg_written .and. GribVerbosity .ge. GribVerbosityWarn) then
                call display_req_GRIB_msg_properties(grib_par_id, &
                     alt_grib_par_id, derived_date, derived_hour, &
                     date, hour, analyse, level, level_type)
                call display_GRIB_message_properties( &
                     list_of_GRIB_files(found_file_nr)%grib_file_attributes)
                call display_GRIB_message_properties( &
                     list_of_GRIB_files(i)%grib_file_attributes)
                print *,"WARNING: multiple GRIB messages in the filelist "//&
                     "match the requested parameters"
                print *,"WARNING: the processor chooses the first GRIB "//&
                     "message from the filelist"
                multiple_found_msg_written = .true.
             endif
             exit fileloop
          endif
       endif
    enddo fileloop ! loop over all GRIB files in the list

    if (found_file_nr .ne. -1 .and. found_msg_nr .ne. -1)  then

      ! see if this message is loaded into memory
      msg_pos = -1
      msgloop: do j = 1,max_nr_open_grib_msgs
         if (list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(j)%&
              message_pos_in_file .eq. found_msg_nr) then
            msg_pos = j
            exit msgloop
         endif
      enddo msgloop

      if (msg_pos .eq. -1) then

        ! find an empty spot in the msg list struct and (re)load message
        msgloop2: do j = 1,max_nr_open_grib_msgs
          if (list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(j)%message_pos_in_file .eq. &
              -1) then
            msg_pos = j
            exit msgloop2
          endif
        enddo msgloop2

        if (msg_pos .eq. -1) then
          print *,"ERROR: only ",max_nr_open_grib_msgs," GRIB messages can be opened"
          print *,"ERROR: simultaneously inside the routine get_from_GRIB_filelist"
          print *,"ERROR: unable to open this GRIB message!"
          print *,"ERROR: change the parameter max_nr_open_grib_msgs and compile again"
          ! TO DO:
          ! close a message to make room for the new one
          stop 1
        endif

        if (list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%message_pos_in_file .eq. &
            -1) then
          if (GribVerbosity .ge. GribVerbosityReport) then
            print *,"(re)reading msg ",found_msg_nr," from file number ",found_file_nr
          endif

          ! retrieve the data values
          call get_GRIB_data_values( &
               list_of_GRIB_files(found_file_nr)%grib_file_attributes, &
               found_msg_nr, &
               list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos), &
               error_flag)
          if (error_flag .ne. gribio_no_error) then
            stop 1
          endif
        endif
      endif ! end of (re)loading GRIB message

      if (do_getmeanval) then

        ! the do_getmeanval only works for regular grids
        if (list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%grid_type .ne. 'regular_ll' &
            .and. &
            list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%grid_type .ne. 'regular_gg' &
           ) then
          print *,"ERROR: when the option getmeanval is used, a regular_ll or"
          print *,"ERROR: regular_gg GRIB field must be used"
          print *,"ERROR: wrong usage of get_from_GRIB_filelist, quitting ..."
          stop 1
        endif

        ! see if this GRIB field is global (i.e., covers 360 degrees in longitude)
        if (list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%lon_of_last_gridpoint - &
            list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%lon_of_first_gridpoint .lt. &
            359.9 - list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos)%lon_step_per_lat(1)) then
          print *,"ERROR: when the option getmeanval is used, a global GRIB field"
          print *,"ERROR: covering 360 degrees in longitude must be used"
          print *,"ERROR: wrong usage of get_from_GRIB_filelist, quitting ..."
          stop 1
        endif

        lat_of_first_gridpoint = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%lat_of_first_gridpoint
        lon_of_first_gridpoint = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%lon_of_first_gridpoint
        lat_step = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%lat_step
        lon_step = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%lon_step_per_lat(1)
        nr_lat_points = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%nr_lat_points
        nr_lon_points = list_of_GRIB_files(found_file_nr)%&
             list_of_GRIB_msgs(msg_pos)%nr_lon_points

        latlon_dist = radius * 360./40000. ! convert km to degrees
        lat1 = lat - latlon_dist
        lat2 = lat + latlon_dist
        lat_index1 = floor((lat1 - lat_of_first_gridpoint)/lat_step)+2
        lat_index2 = floor((lat2 - lat_of_first_gridpoint)/lat_step)+1

        ! number of lon steps increases at higher latitudes
        cos_lat = cos(lat/57.29577951) ! = lat * 2pi/360
        if (cos_lat .lt. 0.01) cos_lat = 0.01
        lon1 = lon - latlon_dist/cos_lat
        lon2 = lon + latlon_dist/cos_lat
        lon_index1 = floor((lon1 - lon_of_first_gridpoint)/lon_step)+2
        lon_index2 = floor((lon2 - lon_of_first_gridpoint)/lon_step)+1

        ! see if the indexes need to be swapped to enable the do loops below
        ! this occurs when lat_step or lon_step are negative
        ! in this case, we can take two steps less since the rounding by
        ! floor is in the wrong direction
        if (lat_index1 .gt. lat_index2) then
          tmp_index  = lat_index1
          lat_index1 = lat_index2
          lat_index2 = tmp_index
          lat_index1 = lat_index1+1
          lat_index2 = lat_index2-1
        endif

        if (lon_index1 .gt. lon_index2) then
          tmp_index  = lon_index1
          lon_index1 = lon_index2
          lon_index2 = tmp_index
          lon_index1 = lon_index1+1
          lon_index2 = lon_index2-1
        endif

        if (lat_index1 .lt. 1            ) lat_index1 = 1
        if (lat_index2 .lt. 1            ) lat_index2 = 1
        if (lat_index1 .gt. nr_lat_points) lat_index1 = nr_lat_points
        if (lat_index2 .gt. nr_lat_points) lat_index2 = nr_lat_points

        ! do a check whether all GRIB values surrounding the lat and lon have
        ! the same value, e.g. in case of a land-sea mask far away from the
        ! coast
        ! if this is the case, we can skip the time consuming loop below to
        ! compute the average value
        first_val = missing_indicator_real
        all_values_identical = .true.
        lat_testloop: do lat_index = lat_index1, lat_index2
          lon_testloop: do tmp_lon_index = lon_index1, lon_index2
            lon_index = tmp_lon_index
            if (lon_index .lt. 1) lon_index = lon_index + nr_lon_points
            if (lon_index .gt. nr_lon_points) &
                                    lon_index = lon_index - nr_lon_points
            val = list_of_GRIB_files(found_file_nr)%&
                    list_of_GRIB_msgs(msg_pos)%&
                    real_vals_xy(lon_index,lat_index)

            if (val .gt. missing_indicator_real*0.99) cycle

            if (do_logn_to_linear_conversion) val = exp(val)

            if (first_val .eq. missing_indicator_real) first_val = val
            if (val .ne. first_val) then
              all_values_identical = .false.
              exit lat_testloop
            endif
          enddo lon_testloop
        enddo lat_testloop

        if (all_values_identical) then
          sum_vals     = real(val,r4_)
          sum_weights  = 1.0
          nr_of_points = 4
        else

          ! do the loop over the surroundings of the given lat-lon pair,
          ! and calculate the average of the field on the grid points,
          ! using 1/distance**2 as weight
          sum_vals     = 0.0
          sum_weights  = 0.0
          nr_of_points = 0
          latloop: do lat_index = lat_index1, lat_index2

            this_lat = lat_of_first_gridpoint + (lat_index-1)*lat_step
            if (abs(this_lat - lat) .gt. latlon_dist) cycle ! performance!

            lonloop: do tmp_lon_index = lon_index1, lon_index2
              lon_index = tmp_lon_index
              if (lon_index .lt. 1) lon_index = lon_index + nr_lon_points
              if (lon_index .gt. nr_lon_points) &
                                      lon_index = lon_index - nr_lon_points
              this_lon = lon_of_first_gridpoint + (lon_index-1)*lon_step

              angle_dist = get_angle_distance(lat,lon,this_lat,this_lon)
              if (angle_dist .le. latlon_dist) then
                val = list_of_GRIB_files(found_file_nr)%&
                        list_of_GRIB_msgs(msg_pos)%&
                        real_vals_xy(lon_index,lat_index)

                if (val .gt. missing_indicator_real*0.99) cycle

                if (do_logn_to_linear_conversion) val = exp(val)

                nr_of_points = nr_of_points + 1

                ! 0.001 to prevent division by zero
                weight = 1.0/(angle_dist*angle_dist+0.001)
                sum_vals = sum_vals + real(val,r4_)*weight
                sum_weights = sum_weights + weight
              endif
            enddo lonloop
          enddo latloop
        endif

        if (nr_of_points .lt. 4) then
          if (.not. coarse_res_msg_written .and. GribVerbosity .ge. GribVerbosityWarn) then
            print *,"WARNING: less than 4 neighbouring points found in function get_from_GRIB_filelist with"
            print *,"WARNING: option getmeanval"
            print *,"WARNING: maybe the radius is set too small for the GRIB field resolution?"
            print *,"WARNING: GRIB parameter: ", grib_par_id, "lat and lon: ", lat, lon
            print *,"WARNING: returning with interpolated value"
            coarse_res_msg_written = .true.
          endif

          call extract_data_from_GRIB_message( &
                  list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos), &
                  lat, lon, logn_to_linear_conversion, val, error_flag)
          if (error_flag .ne. gribio_no_error) then
            stop 1
          endif
        else
          average_val = sum_vals/sum_weights
          val = average_val
        endif
      else
        call extract_data_from_GRIB_message( &
                list_of_GRIB_files(found_file_nr)%list_of_GRIB_msgs(msg_pos), &
                lat, lon, logn_to_linear_conversion, val, error_flag)
        if (error_flag .ne. gribio_no_error) then
          stop 1
        endif
      endif
    else

      ! write warning message only once, not for each node again
      if (.not. not_found_msg_written .and. GribVerbosity .ge. GribVerbosityWarn) then
        call display_req_GRIB_msg_properties(grib_par_id, &
               alt_grib_par_id, derived_date, derived_hour, &
               date, hour, analyse, level, level_type)
        print *,"List of available GRIB files:"
        do m = 1, nr_of_GRIB_files_in_list
          call display_GRIB_message_properties(list_of_GRIB_files(m)%grib_file_attributes, &
                                               par_id_only=grib_par_id)
        enddo
        print *,"WARNING: no GRIB message in the filelist matches the requested parameters"
        not_found_msg_written = .true.
      endif
      return
    endif

  end function get_from_GRIB_filelist
  !---------------------------------------------------------------
  !  #]
  subroutine inquire_GRIB_filelist(inquire_result, &
                                   grib_par_id, alt_grib_par_id, &
                                   derived_date, derived_hour, &
                                   date, hour, analyse, level, level_type)
    !  #[ check if needed data is present in the file list or not

    ! output
    integer, intent(out)          :: inquire_result

    ! input parameters
    integer, intent(in)           :: grib_par_id

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: derived_date, derived_hour
    integer, intent(in), optional :: date, hour, analyse, level, level_type

    ! local variables
    integer :: i,j, next_msg_to_inspect
    integer :: found_file_nr, found_msg_nr

    ! possible results (defined above in the heading of this module)
    !integer, parameter :: inq_is_present            = 1
    !integer, parameter :: inq_not_present           = 2
    !integer, parameter :: inq_multiple_msgs_present = 3

    if (.not. allocated(list_of_GRIB_files)) then
      print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
      print *,"ERROR: init_GRIB_module should be called first"
      stop 1
    endif

    ! init results
    inquire_result = inq_not_present

    ! init internal results
    found_file_nr = -1
    found_msg_nr  = -1

    ! find out which grib file and message match
    ! with the requested input parameters
    do i = 1,nr_of_GRIB_files_in_list

      if (.not. list_of_GRIB_files(i)%grib_file_attributes%file_open) cycle

      next_msg_to_inspect = 1
      keep_trying: do
         ! NOTE: get_GRIB_msgnr()
         ! will return -1 if multiple messages within the same file
         ! match the search criteria!
         ! this means inq_multiple_msgs_present will never be reached
         ! unless the matching messages come from different grib files!
         !
         ! NOTE: if allow_multiple_matches is set to true,
         ! get_GRIB_msgnr will return the first match and not check for
         ! doubles in the list. If you use this, then you should
         ! use the keep_trying logic as is done in the current loop
         ! to ensure the inquire_result is defined properly

         j = get_GRIB_msgnr(list_of_GRIB_files(i)%grib_file_attributes, &
                            grib_par_id, alt_grib_par_id, &
                            derived_date, derived_hour, &
                            date, hour, analyse, level, level_type, &
                            allow_multiple_matches = .true., &
                            start_at_msg = next_msg_to_inspect, &
                            silent = .true.)

         ! no matches in the remainder of the grib msg list for this file
         if (j .eq. -1) exit keep_trying

         if (j .ne. -1) then
            next_msg_to_inspect = j+1
            if ((found_file_nr .eq. -1) .and. (found_msg_nr .eq. -1)) then
               found_file_nr = i
               found_msg_nr  = j
               inquire_result = inq_is_present
            else
               ! is a second match is found, just return,
               ! don't check for even more matches.
               inquire_result = inq_multiple_msgs_present
               return
            endif
         endif

      end do keep_trying
    enddo ! loop over all GRIB files in the list

  end subroutine inquire_GRIB_filelist
  !---------------------------------------------------------------
  !  #]
! can be made private
  subroutine get_analyse_dates_and_times(analyse_t1, analyse_t2, &
                                         analyse_d1, analyse_d2, &
                                         inquire_result, &
                                         grib_par_id, field_date, field_time, &
                                         name, &
                                         level, level_type)
    !  #[ find analyse dates/times before or around a given date/time

    ! Assumption:
    ! this routine only returns the first or second match
    ! for the analysis date/time and assumes the grib file will
    ! never have more than 2 fields with identical derived validity date/time
    ! If we ever need to handle more than 2, this routine needs
    ! to be adapted.

    ! output
    integer, intent(out)          :: analyse_t1
    integer, intent(out)          :: analyse_t2
    integer, intent(out)          :: analyse_d1
    integer, intent(out)          :: analyse_d2
    integer, intent(out)          :: inquire_result

    ! input parameters
    integer, intent(in)           :: grib_par_id
    integer, intent(in)           :: field_date
    integer, intent(in)           :: field_time
    character(len=*), intent(in)  :: name

    ! optional input parameters
    integer, intent(in), optional :: level, level_type

    ! local variables
    integer :: i,j, next_msg_to_inspect
    integer :: found_file_nr, found_msg_nr

    if (.not. allocated(list_of_GRIB_files)) then
       print *,"ERROR: list_of_GRIB_files is not allocated in gribio_module"
       print *,"ERROR: init_GRIB_module should be called first"
       stop 1
    endif

    ! return -1 when no fields are present, so set defaults to -1
    analyse_t1 = -1
    analyse_t2 = -1
    analyse_d1 = -1
    analyse_d2 = -1
    inquire_result = inq_not_present

    ! init internal results
    found_file_nr = -1
    found_msg_nr  = -1

    ! find out which grib file and message match
    ! with the requested input parameters
    fileloop: do i = 1,nr_of_GRIB_files_in_list

       if (.not. list_of_GRIB_files(i)%grib_file_attributes%file_open) cycle

       next_msg_to_inspect = 1
       keep_trying: do
          j = get_GRIB_msgnr(list_of_GRIB_files(i)%grib_file_attributes, &
                             grib_par_id, &
                             derived_date=field_date, &
                             derived_hour=field_time, &
                             level=level, level_type=level_type, &
                             allow_multiple_matches = .true., &
                             start_at_msg = next_msg_to_inspect, &
                             silent = .true.)

          ! no matches in the remainder of the grib msg list for this file
          if (j .eq. -1) exit keep_trying

          if (j .ne. -1) then
             next_msg_to_inspect = j+1
             if ((found_file_nr .eq. -1) .and. (found_msg_nr .eq. -1)) then
                found_file_nr = i
                found_msg_nr  = j

                ! at least 1 field is present, so get analyse time
                ! of first matching field
                analyse_t1 = list_of_GRIB_files(i)%grib_file_attributes%&
                                                   list_of_GRIB_analyse(j)
                analyse_d1 = list_of_GRIB_files(i)%grib_file_attributes%&
                                                   list_of_GRIB_date(j)
                inquire_result = inq_is_present
             else

                ! this point is only reached if multiple fields match,
                ! for this case, get also the analyse time
                ! of the second matching field
                analyse_t2 = list_of_GRIB_files(i)%grib_file_attributes%&
                                                   list_of_GRIB_analyse(j)
                analyse_d2 = list_of_GRIB_files(i)%grib_file_attributes%&
                                                   list_of_GRIB_date(j)
                inquire_result = inq_multiple_msgs_present
                ! return
                exit fileloop
             endif
          endif
       end do keep_trying
    enddo fileloop ! loop  over all GRIB files in the list

    if (GribVerbosity .gt. GribVerbosityReport) then
       print "(A10,A1,1X,A3,I8,A3,I3,1X,A5,I8,1X,I2,1X,A5,I8,1X,I2,1X,A5,I1)",&
            name,":",&
            "d=",field_date," t=",field_time,&
            'an1=',analyse_d1,analyse_t1,&
            'an2=',analyse_d2,analyse_t2,&
            'inq=',inquire_result
    endif


  end subroutine get_analyse_dates_and_times
  !---------------------------------------------------------------
  !  #]
! can be made private
  function  check_proximity_to_analyse(time,date,an_time,an_date,name,&
                           min_analyse_distance) result(dont_use_this_field)
    !  #[ check the time between measurement and analyse
    ! take date difference into account, since we have derived
    ! dates and times at this point

    real(r4_) :: time                  ! input
    integer  :: date                  ! input
    integer  :: an_time, an_date      ! input
    character(len=*) :: name          ! input
    integer  :: min_analyse_distance  ! input
    logical  :: dont_use_this_field   ! result

    ! local variables
    integer :: datediff

    dont_use_this_field = .false.

    if ((an_date .eq. -1) .or. (an_time .eq. -1)) then
       ! this field seems to be missing, so should not be used
       dont_use_this_field = .true.
       return
    endif

    datediff = conv_date_to_daycount(date) - conv_date_to_daycount(an_date)

    if ((time/3600. - an_time + 24*datediff) .lt. min_analyse_distance) then
       if (GribVerbosity .ge. GribVerbosityReport) then
          PRINT *,"REALACC(6)"
          print *,"WARNING: measurement time: ",time/3600.
          PRINT *,"ENDREALACC"
          print *,"is too close to analyse time ",trim(name),":",an_time
          print *,"for date = ",date," and an_date = ",an_date
       endif
       dont_use_this_field = .true.
    endif

  end function check_proximity_to_analyse
  !---------------------------------------------------------------
  !  #]

! can be made private
  subroutine choose_analyse_to_use_3pt( time, date, t_step, t_offset, &
       grib_par_id, alt_grib_par_id, &
       level, level_type,&
       do_proximity_check, min_analyse_distance, &
       two_analyses_times_enabled, &
       analyse_time_to_use, analyse_date_to_use,&
       t1, t2, t3, date1, date2, date3)
    !  #[ collected logic to choose which analyses should be used
    ! input parameters
    real(r4_), intent(in) :: time
    integer, intent(in) :: date, t_step, t_offset
    integer, intent(in) :: grib_par_id
    logical, intent(in) :: do_proximity_check, two_analyses_times_enabled
    integer, intent(in) :: min_analyse_distance
    integer, intent(out) :: analyse_time_to_use ! result
    integer, intent(out) :: analyse_date_to_use ! result
    integer, intent(out) :: t1, t2, t3, date1, date2, date3 ! result

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: level, level_type

    ! local variables
    logical :: dont_use_prev1, dont_use_prev2
    logical :: dont_use_next1, dont_use_next2
    integer :: an_t1_prev_prev, an_t2_prev_prev
    integer :: an_t1_prev,      an_t2_prev
    integer :: an_t1_next,      an_t2_next
    integer :: an_t1_next_next, an_t2_next_next
    integer :: an_d1_prev_prev, an_d2_prev_prev
    integer :: an_d1_prev,      an_d2_prev
    integer :: an_d1_next,      an_d2_next
    integer :: an_d1_next_next, an_d2_next_next
    integer :: t_prev_field, t_prev_prev_field
    integer :: t_next_field, t_next_next_field
    integer :: date_prev_field, date_prev_prev_field
    integer :: date_next_field, date_next_next_field
    integer :: hour
    integer  :: inquire_result
    integer  :: inquire_prev_prev, inquire_prev
    integer  :: inquire_next_next, inquire_next


    ! init outputs
    t1 = -1
    t2 = -1
    t3 = -1

    date1 = -1
    date2 = -1
    date3 = -1

    analyse_time_to_use = -1
    analyse_date_to_use = -1

    ! set the date to current date to start with
    date_prev_prev_field = date
    date_prev_field      = date
    date_next_field      = date
    date_next_next_field = date

    hour = floor(time/3600.)

    ! first get the times for the previous and following GRIB field
    t_prev_field = t_offset + t_step*floor(real(hour/t_step))
    t_next_field = t_prev_field + t_step

    ! then get the times for the prevprev and nextnext fields
    t_prev_prev_field = t_prev_field - t_step
    t_next_next_field = t_next_field + t_step

    ! then convert these date/time pairs to derived date/time
    ! note: this subroutine changes the input variables, so the
    !       result is available in the same variables after the call
    call convert_to_derived_datetime(date_prev_prev_field,t_prev_prev_field)
    call convert_to_derived_datetime(date_prev_field,     t_prev_field     )
    call convert_to_derived_datetime(date_next_field,     t_next_field     )
    call convert_to_derived_datetime(date_next_next_field,t_next_next_field)

    ! then see if a field for each date-time pair is present
    ! and if so, what analyses time it has
    call get_analyse_dates_and_times( &
         an_t1_prev_prev, an_t2_prev_prev, & ! output
         an_d1_prev_prev, an_d2_prev_prev, & ! output
         inquire_prev_prev, &                ! output
         grib_par_id, date_prev_prev_field, t_prev_prev_field, & ! input
         'prev_prev', & ! input
         level = level, level_type = level_type) ! optional input
    if (GribVerbosity .ge. GribVerbosityDebug) &
         print *,'DEBUG: inquire_prev_prev = ', inquire_prev_prev
    call get_analyse_dates_and_times( &
         an_t1_prev, an_t2_prev, & ! output
         an_d1_prev, an_d2_prev, & ! output
         inquire_prev, &           ! output
         grib_par_id, date_prev_field, t_prev_field, & ! input
         'prev', & ! input
         level = level, level_type = level_type) ! optional input
    if (GribVerbosity .ge. GribVerbosityDebug) &
         print *,'DEBUG: inquire_prev = ', inquire_prev
    call get_analyse_dates_and_times( &
         an_t1_next, an_t2_next, & ! output
         an_d1_next, an_d2_next, & ! output
         inquire_next, &           ! output
         grib_par_id, date_next_field, t_next_field, & ! input
         'next', & ! input
         level = level, level_type = level_type) ! optional input
    if (GribVerbosity .ge. GribVerbosityDebug) &
         print *,'DEBUG: inquire_next = ', inquire_next
    call get_analyse_dates_and_times( &
         an_t1_next_next, an_t2_next_next, & ! output
         an_d1_next_next, an_d2_next_next, & ! output
         inquire_next_next, &                ! output
         grib_par_id, date_next_next_field, t_next_next_field, & ! input
         'next_next', & ! input
         level = level, level_type = level_type) ! optional input
    if (GribVerbosity .ge. GribVerbosityDebug) &
         print *,'DEBUG: inquire_next_next = ', inquire_next_next

    ! check if any date-time pair has an analyses time too close to the
    ! measurement and mark it if this is the case

    ! check the time between measurement and analyse
    ! unless requested otherwise
    dont_use_prev1 = .false.
    dont_use_next1 = .false.
    dont_use_prev2 = .false.
    dont_use_next2 = .false.

    if (do_proximity_check) then
       dont_use_prev1 = check_proximity_to_analyse(time,date,an_t1_prev,&
                                 an_d1_prev,"prev1",min_analyse_distance)
       dont_use_next1 = check_proximity_to_analyse(time,date,an_t1_next,&
                                 an_d1_next,"next1",min_analyse_distance)

       dont_use_prev2 = check_proximity_to_analyse(time,date,an_t2_prev,&
                                 an_d2_prev,"prev2",min_analyse_distance)
       dont_use_next2 = check_proximity_to_analyse(time,date,an_t2_next,&
                                 an_d2_next,"next2",min_analyse_distance)
    end if

    ! then decide which analyse to use
    analyse_time_to_use = -1
    analyse_date_to_use = -1

    ! when doing 2 point interpolation, and first matching grib fields
    ! have identical analysis date/time, then use it
    ! NOTE: no logic is yet available to use the most recent analysis
    ! in case both could be used for the dame interpolation.
    if ((an_t1_prev .eq. an_t1_next) .and. (an_d1_prev .eq. an_d1_next)) then
       if ((.not. dont_use_prev1) .and. (.not. dont_use_next1)) then
          analyse_time_to_use = an_t1_prev
          analyse_date_to_use = an_d1_prev
       endif
    endif

    ! case when on prev time 2 fields are available
    if (inquire_prev .eq. inq_multiple_msgs_present) then

       ! if this newer analyse is also available for the next time then use it
       ! note: this assumes the messages in the input grib file are sorted
       ! by analysis. This is no requirement for valid grib files, so may
       ! fail one day ...
       if ((an_t2_prev .eq. an_t1_next) .and. (an_d2_prev .eq. an_d1_next)) then
          if ((.not. dont_use_prev2) .and. (.not. dont_use_next1)) then
             analyse_time_to_use = an_t2_prev
             analyse_date_to_use = an_d2_prev
          endif
       endif
    endif

    ! case when only next time 2 fields are available
    if (inquire_next .eq. inq_multiple_msgs_present) then

       ! if this newer analyse is also available for the prev time then use it
       if ((an_t1_prev .eq. an_t2_next) .and. (an_d1_prev .eq. an_d2_next)) then
          if ((.not. dont_use_prev1) .and. (.not. dont_use_next2)) then
             analyse_time_to_use = an_t1_prev
             analyse_date_to_use = an_d1_prev
          endif
       endif
    endif

    ! case when on both times 2 fields are available
    ! this second field is newer than the first field, so this code
    ! overrides the analyse_*_to_use settings, which is ok
    ! note: this again assumes the messages in the input grib file are sorted
    ! by analysis. This is no requirement for valid grib files, so may
    ! fail one day ...
    if ( (inquire_prev .eq. inq_multiple_msgs_present) .and. &
         (inquire_next .eq. inq_multiple_msgs_present)       ) then
       if ( (an_t2_prev .eq. an_t2_next) .and. &
            (an_d2_prev .eq. an_d2_next)       ) then
          if ( (.not. dont_use_prev2) .and. &
               (.not. dont_use_next2)       ) then
             analyse_time_to_use = an_t2_prev
             analyse_date_to_use = an_d2_prev
          endif
       endif
    endif

    if (analyse_time_to_use .eq. -1 .or. analyse_date_to_use .eq. -1) then
       if (.not. two_analyses_times_enabled) then
          if (GribVerbosity .ge. GribVerbosityWarn) then
             print *,"WARNING: no suitable analyse time was found for which"
             print *,"a prev and a next field is available"
          endif
          return
       endif
    endif

    if (GribVerbosity .gt. GribVerbosityReport) then
       if (two_analyses_times_enabled) then
          print *,"analyse_to_use = allow mixing forecasts from "//&
               "different analyses during interpolation"
       else
          print *,"analyse_time_to_use = ",analyse_time_to_use," h"
          print *,"analyse_date_to_use = ",analyse_date_to_use
       endif
    endif

    ! set times and dates for 3 point interpolation

    ! take 2 GRIB files before the measurement and 1 after
    ! (since earlier files probably have a better forecast, this is
    !  better than taking 1 before and 2 after)
    ! unless not possible, due to start of the day,
    ! or a jump in analyse time
    t1 = t_prev_prev_field
    t2 = t_prev_field
    t3 = t_next_field

    date1 = date_prev_prev_field
    date2 = date_prev_field
    date3 = date_next_field

    ! now check if a GRIB file for t1 exists
    if (two_analyses_times_enabled) then
       inquire_result = inquire_prev_prev
    else

       ! check also if one of the GRIB files for t1 has the correct analyse time
       inquire_result = inq_not_present
       if (analyse_time_to_use .eq. an_t1_prev_prev .and. &
            analyse_date_to_use .eq. an_d1_prev_prev) then
          inquire_result = inq_is_present
       endif
       if (analyse_time_to_use .eq. an_t2_prev_prev .and. &
            analyse_date_to_use .eq. an_d2_prev_prev) then
          inquire_result = inq_is_present
       endif
    endif

    if (inquire_result .ne. inq_is_present) then

       ! now try to take 1 GRIB file before, and 2 after the BUFR time
       t1 = t_prev_field
       t2 = t_next_field
       t3 = t_next_next_field

       date1 = date_prev_field
       date2 = date_next_field
       date3 = date_next_next_field
    endif

    !print *,'debug: t1,t2,t3',t1,t2,t3
    !print *,'debug: d1,d2,d3',date1,date2,date3

  end subroutine choose_analyse_to_use_3pt
    !  #]
! can be made private
  subroutine choose_analyse_to_use_2pt( time, date, t_step, t_offset, &
       grib_par_id, alt_grib_par_id, &
       level, level_type,&
       do_proximity_check, min_analyse_distance, &
       two_analyses_times_enabled, &
       analyse_time_to_use, analyse_date_to_use,&
       t1, t2, date1, date2)
    !  #[ collected logic to choose which analyses should be used
    ! input parameters
    real(r4_), intent(in) :: time
    integer, intent(in) :: date, t_step, t_offset
    integer, intent(in) :: grib_par_id
    logical, intent(in) :: do_proximity_check, two_analyses_times_enabled
    integer, intent(in) :: min_analyse_distance
    integer, intent(out) :: analyse_time_to_use ! result
    integer, intent(out) :: analyse_date_to_use ! result
    integer, intent(out) :: t1, t2, date1, date2 ! result

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: level, level_type

    ! local variables
    logical :: dont_use_prev1, dont_use_prev2
    logical :: dont_use_next1, dont_use_next2
    integer :: an_t1_prev, an_t2_prev, an_t1_next, an_t2_next
    integer :: an_d1_prev, an_d2_prev, an_d1_next, an_d2_next
    integer :: t_prev_field, t_next_field
    integer :: date_prev_field, date_next_field
    integer :: inquire_prev, inquire_next
    integer :: hour

    type(AnDateTime) :: an_dt1_prev, an_dt1_next
    type(AnDateTime) :: an_dt2_prev, an_dt2_next
    type(AnDateTime) :: dt1, dt2, an_dt_to_use
    type(AnDateTime) :: oldest_an_dt, newest_an_dt

    ! init outputs
    t1 = -1
    t2 = -1
    date1 = -1
    date2 = -1
    call init_AnDateTime(dt1)
    call init_AnDateTime(dt2)

    analyse_time_to_use = -1
    analyse_date_to_use = -1
    call init_AnDateTime(an_dt_to_use)

    ! set the date to current date to start with
    date_prev_field = date
    date_next_field = date

    hour = floor(time/3600.)

    ! first get the times for the previous and following GRIB field
    t_prev_field = t_offset + t_step*floor(real(hour/t_step))
    t_next_field = t_prev_field + t_step

    ! then convert these date/time pairs to derived date/time
    ! note: this subroutine changes the input variables, so the
    !       result is available in the same variables after the call
    call convert_to_derived_datetime(date_prev_field, t_prev_field)
    call convert_to_derived_datetime(date_next_field, t_next_field)

    ! set times and dates for linear interpolation
    t1 = t_prev_field
    t2 = t_next_field
    date1 = date_prev_field
    date2 = date_next_field

!    call set_AnDateTime(dt_prev,date_prev_field,t_prev_field)
!    call set_AnDateTime(dt_next,date_next_field,t_next_field)

    ! then see if a field for each date-time pair is present
    ! and if so, what analyses time it has
    call get_analyse_dates_and_times( &
         an_t1_prev, an_t2_prev, & ! output of first match
         an_d1_prev, an_d2_prev, & ! output of second match
         inquire_prev, &           ! output
         grib_par_id, date_prev_field, t_prev_field, & ! input
         'prev', & ! input
         level = level, level_type = level_type) ! optional input
    call get_analyse_dates_and_times( &
         an_t1_next, an_t2_next, & ! output of first match
         an_d1_next, an_d2_next, & ! output of second match
         inquire_next, &           ! output
         grib_par_id, date_next_field, t_next_field, & ! input
         'next', & ! input
         level = level, level_type = level_type) ! optional input

    ! combine found date and times in the combined AnDateTime type
    call set_AnDateTime(an_dt1_prev, an_d1_prev, an_t1_prev)
    call set_AnDateTime(an_dt2_prev, an_d2_prev, an_t2_prev)
    call set_AnDateTime(an_dt1_next, an_d1_next, an_t1_next)
    call set_AnDateTime(an_dt2_next, an_d2_next, an_t2_next)

    ! check if any date-time pair has an analyses time too close to the
    ! measurement and mark it if this is the case

    ! check the time between measurement and analyse
    ! unless requested otherwise
    dont_use_prev1 = .false.
    dont_use_next1 = .false.
    dont_use_prev2 = .false.
    dont_use_next2 = .false.

    if (do_proximity_check) then
       ! dont use will be set to true if a datetime pair is missing
       ! or if it is too close to the analyses
       dont_use_prev1 = check_proximity_to_analyse(time,date,an_t1_prev,&
                                 an_d1_prev,"prev1",min_analyse_distance)
       dont_use_next1 = check_proximity_to_analyse(time,date,an_t1_next,&
                                 an_d1_next,"next1",min_analyse_distance)

       dont_use_prev2 = check_proximity_to_analyse(time,date,an_t2_prev,&
                                 an_d2_prev,"prev2",min_analyse_distance)
       dont_use_next2 = check_proximity_to_analyse(time,date,an_t2_next,&
                                 an_d2_next,"next2",min_analyse_distance)
    end if

    call init_AnDateTime(oldest_an_dt)
    call init_AnDateTime(newest_an_dt)

    ! handle the required first match for prev
    if (.not. dont_use_prev1) newest_an_dt = an_dt1_prev

    ! handle the required first match for next
    if (.not. dont_use_next1) then
       if (an_dt1_next .gt. newest_an_dt) then
          oldest_an_dt = newest_an_dt ! shift
          newest_an_dt = an_dt1_next  ! and fill newest
       end if
    end if

    ! handle the optional an_dt2_prev
    ! note: it is assumed that no more than two model runs are combined
    ! in a single grib message, so this way we will always get both
    ! the datetimes if present
    if (.not. missing_AnDateTime(an_dt2_prev)) then
       if (an_dt2_prev .gt. newest_an_dt) then
          oldest_an_dt = newest_an_dt ! shift
          newest_an_dt = an_dt2_prev  ! and fill newest
       end if
    end if

    ! handle the optional an_dt2_prev
    if (.not. missing_AnDateTime(an_dt2_next)) then
       if (an_dt2_next .gt. newest_an_dt) then
          oldest_an_dt = newest_an_dt ! shift
          newest_an_dt = an_dt2_next  ! and fill newest
       end if
    end if

    ! then decide which analyse to use

    ! init is done above
    !analyse_time_to_use = -1
    !analyse_date_to_use = -1
    !call init_AnDateTime(an_dt_to_use)

    !an_dt_to_use = newest_an_dt

    ! when doing 2 point interpolation, and first matching grib fields
    ! have identical analysis date/time, then use it
    !if ( (an_dt1_prev .eq. newest_an_dt) .or. &
    !     (an_dt2_prev .eq. newest_an_dt)      ) then
    !   if ( (an_dt1_next .eq. newest_an_dt) .or. &
    !        (an_dt2_next .eq. newest_an_dt)      ) then
    !      ! newest analyse is available before and after, so use it
    !      an_dt_to_use = newest_an_dt
    !   end if
    !end if

    if (an_dt1_prev .eq. an_dt1_next) then
       if ((.not. dont_use_prev1) .and. (.not. dont_use_next1)) then
          an_dt_to_use = an_dt1_prev
       endif
    endif

    if (an_dt1_prev .eq. an_dt2_next) then
       if ((.not. dont_use_prev1) .and. (.not. dont_use_next2)) then
          if (an_dt1_prev .gt. an_dt_to_use) an_dt_to_use = an_dt1_prev
       endif
    endif

    if (an_dt2_prev .eq. an_dt1_next) then
       if ((.not. dont_use_prev2) .and. (.not. dont_use_next1)) then
          if (an_dt2_prev .gt. an_dt_to_use) an_dt_to_use = an_dt2_prev
       endif
    endif

    if (an_dt2_prev .eq. an_dt2_next) then
       if ((.not. dont_use_prev2) .and. (.not. dont_use_next2)) then
          if (an_dt2_prev .gt. an_dt_to_use) an_dt_to_use = an_dt2_prev
       endif
    endif

    if (missing_AnDateTime(an_dt_to_use)) then
       if (two_analyses_times_enabled) then
          if (GribVerbosity .gt. GribVerbosityReport) then
             print *,"analyse_to_use = allow mixing forecasts from "//&
                  "different analyses during interpolation"
          end if
       else
          if (GribVerbosity .ge. GribVerbosityWarn) then
             print *,"WARNING: no suitable analyse time was found for which"
             print *,"a prev and a next field is available"
          endif
       endif
       return
    endif

    if (GribVerbosity .gt. GribVerbosityReport) then
       print *,"analyse_time_to_use = ",an_dt_to_use%hour," h"
       print *,"analyse_date_to_use = ",an_dt_to_use%date
    endif

    ! sat analyse to use
    analyse_time_to_use = an_dt_to_use%hour
    analyse_date_to_use = an_dt_to_use%date

    !print *,'debug: t1,t2',t1,t2
    !print *,'debug: d1,d2',date1,date2

  end subroutine choose_analyse_to_use_2pt
    !  #]
  function get_colloc_from_GRIB_filelist(lat, lon, &
                                  time, date, &
                                  grib_par_id, alt_grib_par_id, &
                                  derived_date, derived_hour, &
                                  level, level_type,&
                                  use_3point_interpolation,&
                                  set_time_step,&
                                  set_time_offset, &
                                  enable_two_analyses_times,&
                                  set_min_analyse_distance,&
                                  set_proximity_check, &
                                  logn_to_linear_conversion) &
                                  result(val)
    !  #[ extract a time/place interpolated value from the file list

    ! This function interpolates a given parameter both in
    ! position (latlon) in between the grid points defined by the GRIB file,
    ! and in time, inbetween 2 GRIB files (typically two following forecasts
    ! 3 hours apart.)
    ! In case only forecasts for every 6 hours are available also
    ! a 3 point parabolic interpolation can be done to improve
    ! the accuracy.
    ! -time must be given in seconds since midnight.
    ! -date must have the format: YYYYMMDD
    !  (this does not fit into a 4 byte real, therefore an integer is used)
    ! -generally it is not a good idea to interpolate windfields resulting
    !  from two different analyses, so a special option must be given
    !  to enable this (enable_two_analyses_times).
    !
    ! some parameters are stored in GRIB in logarithmic scale. Prior to
    ! the spatial and the temporal interpolations, a conversion to linear
    ! scale is therefore required. The latter is enabled by setting the
    ! option logn_to_linear_conversion.

    ! input parameters
    real(r4_), intent(in)         :: lat, lon, time
    integer, intent(in)           :: grib_par_id, date

    ! optional input parameters
    integer, intent(in), optional :: alt_grib_par_id
    integer, intent(in), optional :: derived_date, derived_hour
    integer, intent(in), optional :: level, level_type
    integer, intent(in), optional :: set_time_step,set_time_offset
    logical, intent(in), optional :: use_3point_interpolation
    logical, intent(in), optional :: enable_two_analyses_times
    integer, intent(in), optional :: set_min_analyse_distance  ! in hours
    logical, intent(in), optional :: set_proximity_check
    logical, intent(in), optional :: logn_to_linear_conversion

    ! output, type r8_ is needed to be able to pass the date via this field !
    real(r8_)                     :: val

    ! local variables
    logical  :: do_3point_interpolation
    logical  :: two_analyses_times_enabled
    integer  :: t1,t2,t3, hour, date1, date2, date3, t_step, t_offset
    integer  :: t1_sec, t2_sec, t3_sec
    integer  :: analyse_time_to_use
    integer  :: analyse_date_to_use
    real(r4_) :: val1, val2, val3, a, b, c, t_prime

    integer :: min_analyse_distance, datediff
    integer :: time_min
    logical :: do_proximity_check, request_analyses

    ! init outputs
    val = missing_indicator_real

    ! init internal variables
    t1 = -1
    t2 = -1
    t3 = -1

    date1 = -1
    date2 = -1
    date3 = -1

    ! take 3 hour forecast steps as default, which is what I use for era40
    t_step = 3

    ! if the forecast interval was succesfully computed when reading the
    ! files, use it here to change the step interval
    if (t_interval .lt. 99) t_step = t_interval

    ! use this optional parameter to change it if needed
    if (present(set_time_step)) then
      t_step = set_time_step
    endif

    t_offset=0
    if (present(set_time_offset)) then
      t_offset = set_time_offset
    end if

    ! don't allow the data to be closer to an analyse
    ! used for interpolation then t_step hours, to make sure that the
    ! same cell-data was not used to produce this analyse in the first
    ! place (which would introduce unwanted dependencies)
    min_analyse_distance = t_step

    ! allow exceptions only if explicitly requested
    if (present(set_min_analyse_distance)) then
      min_analyse_distance = set_min_analyse_distance
    endif

    do_3point_interpolation = .false.
    if (present(use_3point_interpolation)) then
      if (use_3point_interpolation) do_3point_interpolation = .true.
    endif

    two_analyses_times_enabled = .false.
    if (present(enable_two_analyses_times)) then
      if (enable_two_analyses_times) then
        two_analyses_times_enabled = .true.
      endif
    endif

    if (present(alt_grib_par_id)) then
      if (GribVerbosity .ge. GribVerbosityWarn) then
         print *,"WARNING: alt_grib_par_id is neglected in "//&
              "get_colloc_from_GRIB_filelist"
      endif
    endif

    if (present(derived_date)) then
      if (GribVerbosity .ge. GribVerbosityWarn) then
         print *,"WARNING: derived_date is neglected in "//&
              "get_colloc_from_GRIB_filelist"
      endif
    endif

    if (present(derived_hour)) then
      if (GribVerbosity .ge. GribVerbosityWarn) then
         print *,"WARNING: derived_hour is neglected in "//&
              "get_colloc_from_GRIB_filelist"
      endif
    endif

    do_proximity_check=.true.
    if (present(set_proximity_check)) do_proximity_check=set_proximity_check

    hour = floor(time/3600.)

    if (GribVerbosity .gt. GribVerbosityReport) then
       time_min   = int((time-3600*hour)/60.)
       print *,"==========================="
       print *,"get collocation data for GRIB parameter",grib_par_id
       print "(A,I8,A,I2,A,I2,A)"," obs. d=",date,&
            " t=",hour,' h ',time_min,' m'
       !print "(A,I2,A,I2,A,I2)","hour=",hour," &
       !  minutes = ",floor((time-real(3600*hour))/60.),&
       !  " seconds = ",nint(time)-3600*hour-60*floor((time-real(3600*hour))/60.)
    end if

    ! decide which analyse to use
    if (do_3point_interpolation) then
       call choose_analyse_to_use_3pt(time, date, t_step, t_offset, &
            grib_par_id, alt_grib_par_id, &
            level, level_type,&
            do_proximity_check, min_analyse_distance, &
            two_analyses_times_enabled, &
            analyse_time_to_use, analyse_date_to_use, &
            t1, t2, t3, date1, date2, date3)
    else
       call choose_analyse_to_use_2pt(time, date, t_step, t_offset, &
            grib_par_id, alt_grib_par_id, &
            level, level_type,&
            do_proximity_check, min_analyse_distance, &
            two_analyses_times_enabled, &
            analyse_time_to_use, analyse_date_to_use, &
            t1, t2, date1, date2)
    end if

    ! get common data for 2-point and 3-point interpolation
    request_analyses = .true.
    if ( (analyse_time_to_use .eq. -1) .or. &
         (analyse_date_to_use .eq. -1)      ) then
       if (two_analyses_times_enabled) then
          ! no suitable analyses datetime was found, so allow
          ! any matching data in this cvase
          request_analyses = .false.
       else
          if (GribVerbosity .ge. GribVerbosityDebug) then
             print *,"ERROR: no data found"
             print *,"ERROR: please either probide this program with proper"
             print *,"ERROR: GRIB files, so it can interpolate to the requested"
             print *,"ERROR: date and time,"
             print *,"ERROR: or set the enable_two_analyses_times switch in"
             print *,"ERROR: the call to get_colloc_from_GRIB_filelist() to"
             print *,"ERROR: allow interpolations between different model runs."
          end if
          return
       end if
    end if

    if (request_analyses) then
       val1 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                            analyse=analyse_time_to_use, &
                            date=analyse_date_to_use, &
                            derived_hour=t1, derived_date=date1, &
                            level=level, level_type=level_type, &
                            logn_to_linear_conversion=&
                            logn_to_linear_conversion),r4_)
       val2 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                            analyse=analyse_time_to_use, &
                            date=analyse_date_to_use, &
                            derived_hour=t2, derived_date=date2, &
                            level=level, level_type=level_type, &
                            logn_to_linear_conversion=&
                            logn_to_linear_conversion),r4_)
    else
       val1 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                            derived_hour=t1, derived_date=date1, &
                            level=level, level_type=level_type, &
                            logn_to_linear_conversion=&
                            logn_to_linear_conversion),r4_)
       val2 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                            derived_hour=t2, derived_date=date2, &
                            level=level, level_type=level_type, &
                            logn_to_linear_conversion=&
                            logn_to_linear_conversion),r4_)
    end if

    if (GribVerbosity .ge. GribVerbosityDebug) then
       print *,"DEBUG: grib_par_id = ",grib_par_id ! alt_grib_par_id
       print *,'DEBUG: val1=',val1
       print *,'DEBUG: val2=',val2
    end if

    ! return if missing value
    if (val1 .gt. missing_indicator_real*0.99 .or. &
        val2 .gt. missing_indicator_real*0.99) then
      return
    endif

    ! convert the times to seconds for later use during the actual interpolation
    ! REMIND: we have derived times at this point so
    ! t1 might be 18h and t2 0h for the next day, which would break
    ! the interpolation. Therefore the day difference must be
    ! accounted for in the second calculation
    datediff = conv_date_to_daycount(date1) - conv_date_to_daycount(date)
    t1_sec = (t1 + datediff*24)*60*60
    datediff = conv_date_to_daycount(date2) - conv_date_to_daycount(date)
    t2_sec = (t2 + datediff*24)*60*60

    if (do_3point_interpolation) then
      !  #[ do the 3-point interpolation

      ! get data used for 3-point interpolation only
      if (two_analyses_times_enabled) then
        val3 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                              derived_hour=t3, derived_date=date3, &
                              level=level, level_type=level_type, &
                              logn_to_linear_conversion=logn_to_linear_conversion),r4_)
      else
        val3 = real(get_from_GRIB_filelist(lat, lon, grib_par_id, &
                              analyse=analyse_time_to_use, &
                              date=analyse_date_to_use, &
                              derived_hour=t3, derived_date=date3, &
                              level=level, level_type=level_type, &
                              logn_to_linear_conversion=logn_to_linear_conversion),r4_)
      endif

      ! return if missing value
      if (val3 .gt. missing_indicator_real*0.99) then
        return
      endif

      datediff = conv_date_to_daycount(date3) - conv_date_to_daycount(date)
      t3_sec = (t3 + datediff*24)*60*60

      ! extra safety check: difference between t1, t2 and t3
      ! should be t_step, also after converting to seconds
      if (t2_sec-t1_sec .ne. t_step*60*60 .or. &
          t3_sec-t2_sec .ne. t_step*60*60) then
        print *,"Programming ERROR inside get_colloc_from_GRIB_filelist:"
        print *,"difference between t1_sec, t2_sec and t3_sec is not equal"
        print *,"to t_step*3600 = ",t_step*3600
        print *,"=="
        print *,"time = ",time," date = ",date
        print *,"t1 = ",t1," date1 = ",date1
        print *,"t2 = ",t2," date2 = ",date2
        print *,"t3 = ",t3," date3 = ",date3
        print *,"analyse_time_to_use = ",analyse_time_to_use
        print *,"=="
        stop 1
      endif

      ! do parabolic interpolation (see my notebook, p.69),
      ! (method is identical to the one in PRESCAT, file guesswnd.f)

      ! transform the time to simplify the formulas
      t_prime = (time - t2_sec) / (t3_sec - t2_sec)

      ! interpolate
      a = (val3 + val1 - 2*val2)/2
      b = (val3 - val1)/2
      c = val2
      val = a*t_prime*t_prime + b*t_prime + c

      !  #]
    else
      !  #[ do the 2 point interpolation

      ! do a linear interpolation ! time is in seconds, so this should work OK
      ! (zie mijn waarn.boek p.6 voor details)
      val = val1 + (val2 - val1) * (time-t1_sec) / (t2_sec-t1_sec)
      !  #]
    endif

  end function get_colloc_from_GRIB_filelist
  !---------------------------------------------------------------
  !  #]

end module GribIO
