/*
 *  This module contains interface routines for calling HDF5
 *  routines in C from Fortran.
 */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <hdf5.h>

/*
 * Function to open HDF5 file for reading.
 * After that, switch off error reporting from HDF5 library.
 */
int h5f_open_c_(char *f_name, int *verbosity)
{
  int f_id;

  f_id = (int)H5Fopen(f_name, H5F_ACC_RDONLY, H5P_DEFAULT);

  if (*verbosity <= 1) {
    H5Eset_auto(NULL, NULL, H5P_DEFAULT);
  }

  return f_id;
}

/*
 * Function to open HDF5 file for writing - existing file will be overwritten.
 * After that, switch off error reporting from HDF5 library.
 */
int h5f_create_c_(char *f_name, int *verbosity)
{
  int f_id;

  f_id = (int)H5Fcreate(f_name, H5F_ACC_TRUNC, H5P_DEFAULT, H5P_DEFAULT);

  if (*verbosity <= 1) {
    H5Eset_auto(NULL, NULL, H5P_DEFAULT);
  }

  return f_id;
}

/*
 * Function to close HDF5 file.
 */
int h5f_close_c_(int *f_id)
{
  int status;

  status = (int)H5Fclose(*f_id);
  return status;
}

/*
 * Function to open HDF5 group in file.
 */
int h5g_open_c_(int *f_id, char *g_name)
{
  int g_id;

  g_id = (int)H5Gopen(*f_id, g_name, H5P_DEFAULT);
  return g_id;
}

/*
 * Function to create HDF5 group in file.
 */
int h5g_create_c_(int *f_id, char *g_name)
{
  int g_id;

  g_id = (int)H5Gcreate(*f_id, g_name,
			H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
  return g_id;
}

/*
 * Function to close HDF5 group.
 */
int h5g_close_c_(int *g_id)
{
  int status;

  status = (int)H5Gclose(*g_id);
  return status;
}

/*
 * Function to open HDF5 attribute in location (f_id or g_id).
 */
int h5a_open_name_c_(int *l_id, char *a_name)
{
  int a_id;

  a_id = (int)H5Aopen_name(*l_id, a_name);
  return a_id;
}

/*
 * Function to create HDF5 attribute in location (f_id or g_id).
 */
int h5a_create_c_(int *l_id, char *a_name, int *t_id, int *s_id)
{
  int a_id;

  a_id = (int)H5Acreate(*l_id, a_name, *t_id, *s_id,
			H5P_DEFAULT, H5P_DEFAULT);
  return a_id;
}

/*
 * Function to get type of HDF5 attribute.
 */
int h5a_get_type_c_(int *a_id)
{
  int t_id;

  t_id = (int)H5Aget_type(*a_id);
  return t_id;
}

/*
 * Function to read string HDF5 attribute.
 */
int h5a_read_char_c_(int *a_id, int *t_id, char *sdata)
{
  int status;

  status = (int)H5Aread(*a_id, *t_id, sdata);
  return status;
}


/*
 * Function to read int HDF5 attribute (LZ).
 */
int h5a_read_int_c_(int *a_id, int *t_id, int *sdata)
{
  int status;

  status = (int)H5Aread(*a_id, *t_id, sdata);
  return status;
}

/*
 * Function to read float HDF5 attribute (LZ).
 */
int h5a_read_flt_c_(int *a_id, int *t_id, float *sdata)
{
  int status;

  status = (int)H5Aread(*a_id, *t_id, sdata);
  return status;
}

/*
 * Function to write string HDF5 attribute.
 */
int h5a_write_char_c_(int *a_id, int *t_id, char *sdata)
{
  int status;

  status = (int)H5Awrite(*a_id, *t_id, sdata);
  return status;
}

/*
 * Function to close HDF5 attribute.
 */
int h5a_close_c_(int *a_id)
{
  int status;

  status = (int)H5Aclose(*a_id);
  return status;
}

/*
 * Function to open HDF5 dataset in location (f_id or g_id).
 */
int h5d_open_c_(int *l_id, char *d_name)
{
  int d_id;

  d_id = (int)H5Dopen(*l_id, d_name, H5P_DEFAULT);
  return d_id;
}

/*
 * Function to create integer HDF5 dataset in location (f_id or g_id).
 */
int h5d_create_int_c_(int *l_id, char *d_name, int *s_id)
{
  int d_id;

  d_id = (int)H5Dcreate(*l_id, d_name, H5T_NATIVE_INT, *s_id,
			H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
  return d_id;
}

/*
 * Function to create float HDF5 dataset in location (f_id or g_id).
 */
int h5d_create_float_c_(int *l_id, char *d_name, int *s_id)
{
  int d_id;

  d_id = (int)H5Dcreate(*l_id, d_name, H5T_NATIVE_FLOAT, *s_id,
			H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
  return d_id;
}

/*
 * Function to create string HDF5 dataset in location (f_id or g_id).
 */
int h5d_create_char_c_(int *l_id, char *d_name, int *t_id, int *s_id)
{
  int d_id;

  d_id = (int)H5Dcreate(*l_id, d_name, *t_id, *s_id,
			H5P_DEFAULT, H5P_DEFAULT, H5P_DEFAULT);
  return d_id;
}

/*
 * Function to close HDF5 dataset.
 */
int h5d_close_c_(int *d_id)
{
  int status;

  status = (int)H5Dclose(*d_id);
  return status;
}

/*
 * Function to get type of HDF5 dataset.
 */
int h5d_get_type_c_(int *d_id)
{
  int t_id;

  t_id = (int)H5Dget_type(*d_id);
  return t_id;
}

/*
 * Function to read integer HDF5 dataset.
 */
int h5d_read_int_c_(int *d_id, int *idata)
{
  int status;

  status = (int)H5Dread(*d_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, idata);
  return status;
}

/*
 * Function to read float HDF5 dataset.
 */
int h5d_read_float_c_(int *d_id, float *rdata, int *idata)
{
  int status;

  status = (int)H5Dread(*d_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, rdata);

  /* when no success, try to read the data as integers */
  if (status < 0) {
    status = (int)H5Dread(*d_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, idata);
    if (status == 0) status = 33;
  }
  return status;
}

/*
 * Function to read string HDF5 dataset.
 */
int h5d_read_char_c_(int *d_id, int *t_id, char *sdata)
{
  int status;

  status = (int)H5Dread(*d_id, *t_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, sdata);
  return status;
}

/*
 * Function to read part of integer HDF5 dataset.
 */
int h5d_read_part_int_c_(int *d_id, int *ms_id, int *fs_id, int *idata)
{
  int status;

  status = (int)H5Dread(*d_id, H5T_NATIVE_INT, *ms_id, *fs_id, H5P_DEFAULT, idata);
  return status;
}

/*
 * Function to read part of float HDF5 dataset.
 */
int h5d_read_part_float_c_(int *d_id, int *ms_id, int *fs_id, float *rdata)
{
  int status;

  status = (int)H5Dread(*d_id, H5T_NATIVE_FLOAT, *ms_id, *fs_id, H5P_DEFAULT, rdata);
  return status;
}

/*
 * Function to write integer HDF5 dataset.
 */
int h5d_write_int_c_(int *d_id, int *idata)
{
  int status;

  status = (int)H5Dwrite(*d_id, H5T_NATIVE_INT, H5S_ALL, H5S_ALL, H5P_DEFAULT, idata);
  return status;
}

/*
 * Function to write float HDF5 dataset.
 */
int h5d_write_float_c_(int *d_id, float *rdata)
{
  int status;

  status = (int)H5Dwrite(*d_id, H5T_NATIVE_FLOAT, H5S_ALL, H5S_ALL, H5P_DEFAULT, rdata);
  return status;
}

/*
 * Function to write string HDF5 dataset.
 */
int h5d_write_char_c_(int *d_id, int *t_id, char *sdata)
{
  int status;

  status = (int)H5Dwrite(*d_id, *t_id, H5S_ALL, H5S_ALL, H5P_DEFAULT, sdata);
  return status;
}

/*
 * Function to get space of HDF5 dataset.
 */
int h5d_get_space_c_(int *d_id)
{
  int s_id;

  s_id = (int)H5Dget_space(*d_id);
  return s_id;
}

/*
 * Function to get number of points of HDF5 dataset space.
 */
int h5s_get_select_npoints_c_(int *s_id)
{
  int npoints;

  npoints = (int)H5Sget_select_npoints(*s_id);
  return npoints;
}

/*
 * Function to select a HDF5 hyperslab region.
 */
int h5s_select_hyperslab_c_(int *s_id, int *start, int *stride, int *count)
{
  int status,rank,i,ii;
  hsize_t *c_start, *c_stride, *c_count;

  /* Transpose array since storage orders in C and Fortran are different */
  rank     = H5Sget_simple_extent_ndims(*s_id);
  c_start  = malloc(sizeof(hsize_t) * rank);
  c_stride = malloc(sizeof(hsize_t) * rank);
  c_count  = malloc(sizeof(hsize_t) * rank);
  for (i = 0; i < rank ; i++) {
    ii = (rank - i) - 1;
    c_start[i]  = (hsize_t)start[ii];
    c_stride[i] = (hsize_t)stride[ii];
    c_count[i]  = (hsize_t)count[ii];
  }

  status = (int)H5Sselect_hyperslab(*s_id, H5S_SELECT_SET, c_start, c_stride, c_count, NULL);
  return status;
}

/*
 * Function to create a new simple dataspace and to open it for access.
 */
int h5s_create_simple_c_(int *rank, int *dims)
{
  int s_id,i;
  hsize_t *c_dims;

  /* Transpose array since storage orders in C and Fortran are different */
  c_dims = malloc(sizeof(hsize_t) * (*rank));
  for (i = 0; i < *rank ; i++) {
    c_dims[i] =  dims[*rank - i - 1];
  }

  s_id = (int)H5Screate_simple(*rank, c_dims, NULL);

  free(c_dims);
  return s_id;
}

/*
 * Function to close HDF5 dataset space.
 */
int h5s_close_c_(int *s_id)
{
  int status;

  status = (int)H5Sclose(*s_id);
  return status;
}

/*
 * Function to get size of HDF5 type.
 */
int h5t_get_size_c_(int *t_id)
{
  int size;

  size = (int)H5Tget_size(*t_id);
  return size;
}

/*
 * Function to set size of HDF5 string type.
 */
int h5t_set_string_size_c_(int *len)
{
  int t_id, status;

  t_id = (int)H5Tcopy(H5T_C_S1);
  status = (int)H5Tset_size(t_id, *len);
  if (status < 0) {
    return status;
  } else {
    return t_id;
  }
}

/*
 * Function to close HDF5 type.
 */
int h5t_close_c_(int *t_id)
{
  int status;

  status = (int)H5Tclose(*t_id);
  return status;
}
