module NetCDFMod
!
!  #[ Description:
!
!   This module contains interface routines for calling UCAR/UNIDATA NetCDF
!   routines in C from Fortran.
!
!
! Current Code Owner:  Anton Verhoef
!
! Code Description:
!     Language:           Fortran 90
!     Software Standards: "European Standards for Writing and
!     Documenting Exchangeable Fortran 90 Code"
!
! This software was developed within the context of the EUMETSAT Satellite
! Application Facility on Numerical Weather Prediction (NWP SAF), under the
! Cooperation Agreement dated 16 December, 2003, between EUMETSAT and the
! Met Office, UK, by one or more partners within the NWP SAF. The partners
! in the NWP SAF are the Met Office, ECMWF, KNMI and Meteo France.
!
! Copyright 2008, EUMETSAT, All Rights Reserved.
!
!     Unit Name:             NetCDFMod
!
!     Created on:            Sep. 2008
!
!     Last Modified on:      $Date$
!
!     Modifications Log:
!
! $Id$
! 
!
!  #]
!  #[ modules used:
!
  use numerics, only: r8_, r4_, i4_, i2_, missing_real
!
!  #]
!
  implicit none
!
!  #[ public declarations

!
  real(r8_),parameter    :: undef_double = -999999.9_r8_
  real(r4_),parameter    :: undef_float  = -999999.9_r4_
  integer(i4_),parameter :: undef_int    = -2147483647_i4_
  integer(i2_),parameter :: undef_short  = -32767_i2_

  ! defined in line 137 of support/netcdf/unidata/netcdf.h
  integer,parameter      :: nc_global  = -1

  ! attribute types, copied from enum type nc_type,
  ! defined in lines 48-56 of support/netcdf/unidata/netcdf.h
  !
  integer,parameter :: nc_type_nat =    0 ! NAT = 'Not A Type' (c.f. NaN)
  integer,parameter :: nc_type_byte =   1 ! signed 1 byte integer
  integer,parameter :: nc_type_char =   2 ! ISO/ASCII character
  integer,parameter :: nc_type_short =  3 ! signed 2 byte integer
  integer,parameter :: nc_type_int =    4 ! signed 4 byte integer
  integer,parameter :: nc_type_float =  5 ! single precision fl. point number
  integer,parameter :: nc_type_double = 6 ! double precision fl. point number

! WARNING: if we ever start using the nc_type_int64, please be very carefull with the use
! of the c-long type in the c-to-fortran interface. See the file netcdfio.c for more details.

!  #]
!  #[ private declarations
!
!
!  #]
!  #[ interfaces
!
  interface
     function get_bytesize_short() result(size_short)
       use numerics, only: i1_
       implicit none
       integer(i1_) :: size_short
     end function get_bytesize_short
  end interface

  interface
     function get_bytesize_int() result(size_int)
       use numerics, only: i1_
       implicit none
       integer(i1_) :: size_int
     end function get_bytesize_int
  end interface

  interface
     function get_bytesize_long() result(size_long)
       use numerics, only: i1_
       implicit none
       integer(i1_) :: size_long
     end function get_bytesize_long
  end interface

  interface
     function get_bytesize_float() result(size_float)
       use numerics, only: i1_
       implicit none
       integer(i1_) :: size_float
     end function get_bytesize_float
  end interface

  interface
     function get_bytesize_double() result(size_double)
       use numerics, only: i1_
       implicit none
       integer(i1_) :: size_double
     end function get_bytesize_double
  end interface

  interface
     function nc_open(filename, mode, ncid_ptr) result(status)
       implicit none
       character(len=*),intent(in) :: filename
       character(len=1),intent(in) :: mode ! 'r' or 'w'
       integer,intent(out)         :: ncid_ptr
       integer                     :: status
     end function nc_open
  end interface

  interface
     function nc_inq(ncid, ndims, nvars, natts, unlimdimid) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(out) :: ndims
       integer, intent(out) :: nvars
       integer, intent(out) :: natts
       integer, intent(out) :: unlimdimid
       integer              :: status
     end function nc_inq
  end interface

  interface
     function nc_inq_attname(ncid, varid, count, attr_name) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: count
       character(len=*),intent(out) :: attr_name
       integer              :: status
     end function nc_inq_attname
  end interface

  interface
     function nc_inq_attid(ncid, varid, attrname, attrid) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(in) :: attrname
       integer, intent(out) :: attrid
       integer              :: status
     end function nc_inq_attid
  end interface

  interface
     function nc_inq_atttype(ncid, varid, attrname, attrtype) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(in) :: attrname
       integer, intent(out) :: attrtype
       integer              :: status
     end function nc_inq_atttype
  end interface

  interface
     function nc_inq_attlen(ncid, varid, attrname, attrlen) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(in) :: attrname
       integer, intent(out) :: attrlen
       integer              :: status
     end function nc_inq_attlen
  end interface

  interface
     function nc_get_att_text(ncid, varid, attrname, attrvalue) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(in) :: attrname
       character(len=*),intent(out) :: attrvalue
       integer              :: status
     end function nc_get_att_text
  end interface

  interface
     function nc_get_att_short(ncid, varid, attrname, attrvalue) result(status)
       use numerics, only: i2_
       implicit none
       integer,          intent(in)  :: ncid
       integer,          intent(in)  :: varid
       character(len=*), intent(in)  :: attrname
       integer(i2_),     intent(out) :: attrvalue
       integer                       :: status
     end function nc_get_att_short
  end interface

  interface
     function nc_get_att_int(ncid, varid, attrname, attrvalue) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(in) :: attrname
       integer, intent(out) :: attrvalue
       integer              :: status
     end function nc_get_att_int
  end interface

  interface
     function nc_get_att_float(ncid, varid, attrname, attrvalue) result(status)
       use numerics, only: r4_
       implicit none
       integer,         intent(in)  :: ncid
       integer,         intent(in)  :: varid
       character(len=*),intent(in)  :: attrname
       real(r4_),       intent(out) :: attrvalue
       integer                      :: status
     end function nc_get_att_float
  end interface

  interface
     function nc_get_att_double(ncid, varid, attrname, attrvalue) result(status)
       use numerics, only: r8_
       implicit none
       integer,         intent(in)  :: ncid
       integer,         intent(in)  :: varid
       character(len=*),intent(in)  :: attrname
       real(r8_),       intent(out) :: attrvalue
       integer                      :: status
     end function nc_get_att_double
  end interface

  interface
     function nc_get_att_int_1d(ncid, varid, size0, attrname, int_data) &
          result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       character(len=*), intent(in) :: attrname
       integer, dimension(0:size0-1), intent(out) :: int_data
       integer              :: status
     end function nc_get_att_int_1d
  end interface

  interface
     function nc_inq_varname(ncid, varid, varname) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       character(len=*),intent(out) :: varname
       integer              :: status
     end function nc_inq_varname
  end interface

  interface
     function nc_inq_vartype(ncid, varid, vartype) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(out) :: vartype
       integer              :: status
     end function nc_inq_vartype
  end interface

  interface
     function nc_inq_varndims(ncid, varid, ndims) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(out) :: ndims
       integer              :: status
     end function nc_inq_varndims
  end interface

  interface
     function nc_inq_vardimid(ncid, varid, dimid0, dimid1, dimid2) &
          result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(out) :: dimid0
       integer, intent(out) :: dimid1
       integer, intent(out) :: dimid2
       integer              :: status
     end function nc_inq_vardimid
  end interface

  interface
     function nc_inq_varnatts(ncid, varid, natts) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(out) :: natts
       integer              :: status
     end function nc_inq_varnatts
  end interface

  interface
     function nc_inq_dim(ncid, dimid, name, length) result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: dimid
       character(len=*),intent(out) :: name
       integer, intent(out) :: length
       integer              :: status
     end function nc_inq_dim
  end interface

  interface
     function nc_get_var_byte_1d(ncid, varid, size0, byte_data) &
          result(status)
       use numerics, only: i1_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer(i1_), dimension(0:size0-1), intent(out) :: byte_data
       integer              :: status
     end function nc_get_var_byte_1d
  end interface

  interface
     function nc_get_var_byte_2d(ncid, varid, size0, size1, byte_data) &
          result(status)
       use numerics, only: i1_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       integer(i1_), dimension(0:size0-1,0:size1-1), intent(out) :: byte_data
       integer              :: status
     end function nc_get_var_byte_2d
  end interface

  interface
     function nc_get_var_text_2d(ncid, varid, size0, size1, char_data) &
          result(status)

       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       character(len=1), dimension(0:size0-1,0:size1-1), intent(out) :: &
                                char_data
       integer              :: status
     end function nc_get_var_text_2d
  end interface

  interface
     function nc_get_var_short_1d(ncid, varid, size0, short_data) &
          result(status)
       use numerics, only: i2_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer(i2_), dimension(0:size0-1), intent(out) :: short_data
       integer              :: status
     end function nc_get_var_short_1d
  end interface

  interface
     function nc_get_var_short_2d(ncid, varid, size0, size1, short_data) &
          result(status)
       use numerics, only: i2_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       integer(i2_), dimension(0:size0-1,0:size1-1), intent(out) :: short_data
       integer              :: status
     end function nc_get_var_short_2d
  end interface

  interface
     function nc_get_var_short_3d(ncid, varid, size0, size1, size2, &
                                  short_data) result(status)
       use numerics, only: i2_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0, size1, size2
       integer(i2_), dimension(0:size0-1,0:size1-1,0:size2-1), &
            intent(out) :: short_data
       integer              :: status
     end function nc_get_var_short_3d
  end interface

  interface
     function nc_get_var_int_1d(ncid, varid, size0, int_data) &
          result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, dimension(0:size0-1), intent(out) :: int_data
       integer              :: status
     end function nc_get_var_int_1d
  end interface

  interface
     function nc_get_var_int_2d(ncid, varid, size0, size1, int_data) &
          result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       integer, dimension(0:size0-1,0:size1-1), intent(out) :: int_data
       integer              :: status
     end function nc_get_var_int_2d
  end interface

  interface
     function nc_get_var_int_3d(ncid, varid, size0, size1, size2, int_data) &
          result(status)
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0, size1, size2
       integer, dimension(0:size0-1,0:size1-1,0:size2-1), &
            intent(out) :: int_data
       integer              :: status
     end function nc_get_var_int_3d
  end interface

  interface
     function nc_get_var_float_1d(ncid, varid, size0, float_data) &
          result(status)
       use numerics, only: r4_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       real(r4_), dimension(0:size0-1), intent(out) :: float_data
       integer              :: status
     end function nc_get_var_float_1d
  end interface

  interface
     function nc_get_var_float_2d(ncid, varid, size0, size1, float_data) &
          result(status)
       use numerics, only: r4_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       real(r4_), dimension(0:size0-1,0:size1-1), intent(out) :: float_data
       integer              :: status
     end function nc_get_var_float_2d
  end interface

  interface
     function nc_get_var_float_3d(ncid, varid, size0, size1, size2, &
                                  float_data) result(status)
       use numerics, only: r4_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0, size1, size2
       real(r4_), dimension(0:size0-1,0:size1-1,0:size2-1), &
            intent(out) :: float_data
       integer              :: status
     end function nc_get_var_float_3d
  end interface

  interface
     function nc_get_var_double_1d(ncid, varid, size0, double_data) &
          result(status)
       use numerics, only: r8_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       real(r8_), dimension(0:size0-1), intent(out) :: double_data
       integer              :: status
     end function nc_get_var_double_1d
  end interface

  interface
     function nc_get_var_double_2d(ncid, varid, size0, size1, double_data) &
          result(status)
       use numerics, only: r8_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0
       integer, intent(in)  :: size1
       real(r8_), dimension(0:size0-1,0:size1-1), intent(out) :: double_data
       integer              :: status
     end function nc_get_var_double_2d
  end interface

  interface
     function nc_get_var_double_3d(ncid, varid, size0, size1, size2, &
                                   double_data) result(status)
       use numerics, only: r8_
       implicit none
       integer, intent(in)  :: ncid
       integer, intent(in)  :: varid
       integer, intent(in)  :: size0, size1, size2
       real(r8_), dimension(0:size0-1,0:size1-1,0:size2-1), &
            intent(out) :: double_data
       integer              :: status
     end function nc_get_var_double_3d
  end interface

  interface
     function nc_create(path, ncid_ptr) result(status)
       implicit none
       character(len=*),intent(in) :: path
       integer,intent(out)         :: ncid_ptr
       integer                     :: status
     end function nc_create
  end interface

  interface
     function nc_enddef(ncid) result(status)
       implicit none
       integer,intent(in) :: ncid
       integer            :: status
     end function nc_enddef
  end interface

  interface
     function nc_close(ncid) result(status)
       implicit none
       integer,intent(in) :: ncid
       integer            :: status
     end function nc_close
  end interface

  interface
     function nc_def_dim(ncid, name, size, dimid) result(status)
       implicit none
       integer,intent(in)          :: ncid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: size
       integer,intent(out)         :: dimid
       integer                     :: status
     end function nc_def_dim
  end interface

  interface
     function nc_def_var_double(ncid, name, ndims, dimids, varidp) result(status)
       implicit none
       integer,intent(in)          :: ncid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: ndims
       integer,dimension(0:ndims-1),intent(in) :: dimids
       integer,intent(out)         :: varidp
       integer                     :: status
     end function nc_def_var_double
  end interface

  interface
     function nc_def_var_float(ncid, name, ndims, dimids, varidp) result(status)
       implicit none
       integer,intent(in)          :: ncid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: ndims
       integer,dimension(0:ndims-1),intent(in) :: dimids
       integer,intent(out)         :: varidp
       integer                     :: status
     end function nc_def_var_float
  end interface

  interface
     function nc_def_var_int(ncid, name, ndims, dimids, varidp) result(status)
       implicit none
       integer,intent(in)          :: ncid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: ndims
       integer,dimension(0:ndims-1),intent(in) :: dimids
       integer,intent(out)         :: varidp
       integer                     :: status
     end function nc_def_var_int
  end interface

  interface
     function nc_def_var_short(ncid, name, ndims, dimids, varidp) result(status)
       implicit none
       integer,intent(in)          :: ncid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: ndims
       integer,dimension(0:ndims-1),intent(in) :: dimids
       integer,intent(out)         :: varidp
       integer                     :: status
     end function nc_def_var_short
  end interface

  interface
     function nc_put_att_double(ncid, varid, name, value_in) result(status)
       use numerics, only: r8_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       real(r8_),intent(in)        :: value_in
       integer                     :: status
     end function nc_put_att_double
  end interface

  interface
     function nc_put_att_double_arr(ncid, varid, name, nelems, &
                                    value_in) result(status)
       use numerics, only: r8_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: nelems
       real(r8_),dimension(0:nelems-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_att_double_arr
  end interface

  interface
     function nc_put_att_float(ncid, varid, name, value_in) result(status)
       use numerics, only: r4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       real(r4_),intent(in)        :: value_in
       integer                     :: status
     end function nc_put_att_float
  end interface

  interface
     function nc_put_att_float_arr(ncid, varid, name, nelems, &
                                   value_in) result(status)
       use numerics, only: r4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: nelems
       real(r4_),dimension(0:nelems-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_att_float_arr
  end interface

  interface
     function nc_put_att_int(ncid, varid, name, value_in) result(status)
       use numerics, only: i4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer(i4_),intent(in)     :: value_in
       integer                     :: status
     end function nc_put_att_int
  end interface

  interface
     function nc_put_att_int_arr(ncid, varid, name, nelems, &
                                 value_in) result(status)
       use numerics, only: i4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: nelems
       integer(i4_),dimension(0:nelems-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_att_int_arr
  end interface

  interface
     function nc_put_att_short(ncid, varid, name, value_in) result(status)
       use numerics, only: i2_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer(i2_),intent(in)     :: value_in
       integer                     :: status
     end function nc_put_att_short
  end interface

  interface
     function nc_put_att_short_arr(ncid, varid, name, nelems, &
                                   value_in) result(status)
       use numerics, only: i2_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       integer,intent(in)          :: nelems
       integer(i2_),dimension(0:nelems-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_att_short_arr
  end interface

  interface
     function nc_put_att_text(ncid, varid, name, value_in) result(status)
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       character(len=*),intent(in) :: name
       character(len=*),intent(in) :: value_in
       integer                     :: status
     end function nc_put_att_text
  end interface

  interface
     function nc_put_vara_double_1d(ncid, varid, start, edges, &
                                    value_in) result(status)
       use numerics, only: r8_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start
       integer,intent(in)          :: edges
       real(r8_),dimension(start:start+edges-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_double_1d
  end interface

  interface
     function nc_put_vara_double_2d(ncid, varid, start0, start1, &
                                    edges0, edges1, value_in) result(status)
       use numerics, only: r8_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       real(r8_),dimension(start0:start0+edges0-1,&
                           start1:start1+edges1-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_double_2d
  end interface

  interface
     function nc_put_vara_double_3d(ncid, varid, start0, start1, start2, &
                             edges0, edges1, edges2, value_in) result(status)
       use numerics, only: r8_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0, start1, start2
       integer,intent(in)          :: edges0, edges1, edges2
       real(r8_),dimension(start0:start0+edges0-1,&
                           start1:start1+edges1-1,&
                           start2:start2+edges2-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_double_3d
  end interface

  interface
     function nc_put_vara_float_1d(ncid, varid, start, edges, &
                                   value_in) result(status)
       use numerics, only: r4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start
       integer,intent(in)          :: edges
       real(r4_),dimension(start:start+edges-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_float_1d
  end interface

  interface
     function nc_put_vara_float_2d(ncid, varid, start0, start1, &
                                   edges0, edges1, value_in) result(status)
       use numerics, only: r4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       real(r4_),dimension(start0:start0+edges0-1,&
                           start1:start1+edges1-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_float_2d
  end interface

  interface
     function nc_put_vara_float_3d(ncid, varid, start0, start1, start2, &
                                   edges0, edges1, edges2, &
                                   value_in) result(status)
       use numerics, only: r4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0, start1, start2
       integer,intent(in)          :: edges0, edges1, edges2
       real(r4_),dimension(start0:start0+edges0-1,&
                           start1:start1+edges1-1,&
                           start2:start2+edges2-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_float_3d
  end interface

  interface
     function nc_put_vara_int_1d(ncid, varid, start, edges, &
                                 value_in) result(status)
       use numerics, only: i4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start
       integer,intent(in)          :: edges
       integer(i4_),dimension(start:start+edges-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_int_1d
  end interface

  interface
     function nc_put_vara_int_2d(ncid, varid, start0, start1, &
                                 edges0, edges1, value_in) result(status)
       use numerics, only: i4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       integer(i4_),dimension(start0:start0+edges0-1,&
                              start1:start1+edges1-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_int_2d
  end interface

  interface
     function nc_put_vara_int_3d(ncid, varid, start0, start1, start2, &
                             edges0, edges1, edges2, value_in) result(status)
       use numerics, only: i4_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: start2
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       integer,intent(in)          :: edges2
       integer(i4_),dimension(start0:start0+edges0-1,&
                              start1:start1+edges1-1,&
                              start2:start2+edges2-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_int_3d
  end interface

  interface
     function nc_put_vara_short_1d(ncid, varid, start, edges, &
                                   value_in) result(status)
       use numerics, only: i2_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start
       integer,intent(in)          :: edges
       integer(i2_),dimension(start:start+edges-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_short_1d
  end interface

  interface
     function nc_put_vara_short_2d(ncid, varid, start0, start1, &
                                   edges0, edges1, value_in) result(status)
       use numerics, only: i2_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       integer(i2_),dimension(start0:start0+edges0-1,&
                              start1:start1+edges1-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_short_2d
  end interface

  interface
     function nc_put_vara_short_3d(ncid, varid, start0, start1, start2, &
                            edges0, edges1,edges2, value_in) result(status)
       use numerics, only: i2_
       implicit none
       integer,intent(in)          :: ncid
       integer,intent(in)          :: varid
       integer,intent(in)          :: start0
       integer,intent(in)          :: start1
       integer,intent(in)          :: start2
       integer,intent(in)          :: edges0
       integer,intent(in)          :: edges1
       integer,intent(in)          :: edges2
       integer(i2_),dimension(start0:start0+edges0-1,&
                              start1:start1+edges1-1,&
                              start2:start2+edges2-1),intent(in) :: value_in
       integer                     :: status
     end function nc_put_vara_short_3d
  end interface
!
!  #]
!
contains
  function dtype_to_text(dtype) result(text)
    !  #[ convert dtype to a text string
    integer, intent(in) :: dtype
    character(len=100)   :: text

    ! local variable
    character(len=50) :: txtdtype

    write(txtdtype,*,err=999) dtype
    select case(dtype)
       case(nc_type_nat)
          text = "NAT = 'Not A Type' (c.f. NaN)"
       case(nc_type_byte)
          text = "signed 1 byte integer, siged char"
       case(nc_type_char)
          text = "ISO/ASCII character"
       case(nc_type_short)
          text = "signed 2 byte integer"
       case(nc_type_int)
          text = "signed 4 byte integer"
       case(nc_type_float)
          text = "single precision fl. point number"
       case(nc_type_double)
          text = "double precision fl. point number"
       case default
          text = "unimplemented type number"
       end select

       text = trim(text)//" [typenumber: "//trim(adjustl(txtdtype))//"]"

       return

999    text = "could not convert dtype integer to a text string"
       return

  end function dtype_to_text
    !  #]

  function nc_real_to_double(val_real,scale_fact) result(val_double)
    !  #[
    !
    ! Description: convert real value to NetCDF double value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real,intent(in)               :: val_real
    real(r8_),intent(in),optional :: scale_fact
    real(r8_)                     :: val_double

    if (missing_real(val_real)) then
       val_double = undef_double
       return
    end if

    if (abs(val_real) .lt. abs(real(undef_double))) then
      if (present(scale_fact)) then
        if (abs(scale_fact) .lt. 1.E-38_r8_) then
          val_double = undef_double   ! scale factor too small
        else
          if (abs(val_real/scale_fact) .lt. abs(real(undef_double))) then
            val_double = real(val_real/scale_fact,r8_)
          else
            val_double = undef_double ! scaled value out of range
          endif
        endif
      else
        val_double = real(val_real,r8_)
      endif
    else
      val_double = undef_double       ! unscaled value out of range
    endif

  end function nc_real_to_double
    !  #]

  function nc_real_to_float(val_real,scale_fact) result(val_float)
    !  #[
    !
    ! Description: convert real value to NetCDF float value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real,intent(in)               :: val_real
    real(r8_),intent(in),optional :: scale_fact
    real(r4_)                     :: val_float

    if (missing_real(val_real)) then
       val_float = undef_float
       return
    end if

    if (abs(val_real) .lt. abs(real(undef_float))) then
      if (present(scale_fact)) then
        if (abs(scale_fact) .lt. 1.E-38_r8_) then
          val_float = undef_float   ! scale factor too small
        else
          if (abs(val_real/scale_fact) .lt. abs(real(undef_float))) then
            val_float = real(val_real/scale_fact,r4_)
          else
            val_float = undef_float ! scaled value out of range
          endif
        endif
      else
        val_float = real(val_real,r4_)
      endif
    else
      val_float = undef_float       ! unscaled value out of range
    endif

  end function nc_real_to_float
    !  #]

  function nc_real8_to_float(val_real,scale_fact) result(val_float)
    !  #[
    !
    ! Description: convert real value to NetCDF float value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real(r8_),intent(in)          :: val_real
    real(r8_),intent(in),optional :: scale_fact
    real(r4_)                     :: val_float

    val_float = nc_real_to_float(real(val_real),scale_fact)

  end function nc_real8_to_float
    !  #]

  function nc_real_to_int(val_real,scale_fact) result(val_int)
    !  #[
    !
    ! Description: convert real value to NetCDF int value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real,intent(in)               :: val_real
    real(r8_),intent(in),optional :: scale_fact
    integer(i4_)                  :: val_int

    if (missing_real(val_real)) then
       val_int = undef_int
       return
    end if

    if (abs(val_real) .lt. abs(real(undef_int))) then
      if (present(scale_fact)) then
        if (abs(scale_fact) .lt. 1.E-38_r8_) then
          val_int = undef_int   ! scale factor too small
        else
          if (abs(val_real/scale_fact) .lt. abs(real(undef_int))) then
            val_int = nint(val_real/scale_fact,i4_)
          else
            val_int = undef_int ! scaled value out of range
          endif
        endif
      else
        val_int = nint(val_real,i4_)
      endif
    else
      val_int = undef_int       ! unscaled value out of range
    endif

  end function nc_real_to_int
    !  #]

  function nc_real8_to_int(val_real,scale_fact) result(val_int)
    !  #[
    !
    ! Description: convert real8 value to NetCDF int value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real(r8_),intent(in)          :: val_real
    real(r8_),intent(in),optional :: scale_fact
    integer(i4_)                  :: val_int

    val_int = nc_real_to_int(real(val_real),scale_fact)

  end function nc_real8_to_int
    !  #]

  function nc_real_to_short(val_real,scale_fact) result(val_short)
    !  #[
    !
    ! Description: convert real value to NetCDF short value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real,intent(in)               :: val_real
    real(r8_),intent(in),optional :: scale_fact
    integer(i2_)                  :: val_short

    if (missing_real(val_real)) then
       val_short = undef_short
       return
    end if

    if (abs(val_real) .lt. abs(real(undef_short))) then
      if (present(scale_fact)) then
        if (abs(scale_fact) .lt. 1.E-38_r8_) then
          val_short = undef_short   ! scale factor too small
        else
          if (abs(val_real/scale_fact) .lt. abs(real(undef_short))) then
            val_short = nint(val_real/scale_fact,i2_)
          else
            val_short = undef_short ! scaled value out of range
          endif
        endif
      else
        val_short = nint(val_real,i2_)
      endif
    else
      val_short = undef_short       ! unscaled value out of range
    endif

  end function nc_real_to_short
    !  #]

  function nc_real8_to_short(val_real,scale_fact) result(val_short)
    !  #[
    !
    ! Description: convert real value to NetCDF short value and apply scale factor
    !              replace by undefined if input is outside the valid range
    !
    real(r8_),intent(in)          :: val_real
    real(r8_),intent(in),optional :: scale_fact
    integer(i2_)                  :: val_short

    val_short = nc_real_to_short(real(val_real),scale_fact)

  end function nc_real8_to_short
    !  #]

  subroutine nc_create_var_double(unit_nc,var_name,ndims,dimids,standard_name,long_name,units, &
                                  scale_factor,add_offset,valid_min,valid_max, &
                                  varid,ierror)
    !  #[

    !
    ! Description: create double NetCDF variable with some standard settings
    !
    integer,intent(in)          :: unit_nc,ndims
    character(len=*),intent(in) :: var_name,standard_name,long_name,units
    integer,dimension(0:ndims-1),intent(in) :: dimids
    real(r8_),intent(in)        :: scale_factor,add_offset
    real(r8_),intent(in)        :: valid_min,valid_max
    integer,intent(out)         :: varid,ierror

    ierror = nc_def_var_double(unit_nc, trim(var_name), ndims, dimids, varid)
    if (ierror .ne. 0) return

    ierror = nc_put_att_double(unit_nc, varid, '_FillValue', undef_double)
    if (ierror .ne. 0) return

    ierror = nc_put_att_double(unit_nc, varid, 'missing_value', undef_double)
    if (ierror .ne. 0) return

    ! apply scale factor and offset to min and max values
    ! if both attributes are present, the data are scaled before the offset is added
    ! when scaled data are written, the application should first subtract the offset
    ! and then divide by the scale factor
    ierror = nc_put_att_double(unit_nc, varid, 'valid_min', &
                               real((valid_min-add_offset) / scale_factor,r8_))
    if (ierror .ne. 0) return

    ierror = nc_put_att_double(unit_nc, varid, 'valid_max', &
                               real((valid_max-add_offset) / scale_factor,r8_))
    if (ierror .ne. 0) return

    if (len_trim(standard_name) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'standard_name', standard_name)
      if (ierror .ne. 0) return
    endif

    ierror = nc_put_att_text(unit_nc, varid, 'long_name', long_name)
    if (ierror .ne. 0) return

    if (len_trim(units) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'units', units)
      if (ierror .ne. 0) return
    endif


    ! create scale factor and offset attributes if one of them differs from 1 resp. 0
    if ((abs(scale_factor - 1.0) .gt. 1.E-7) .or. (abs(add_offset) .gt. 1.E-7)) then

      ierror = nc_put_att_double(unit_nc, varid, 'scale_factor', scale_factor)
      if (ierror .ne. 0) return

      ierror = nc_put_att_double(unit_nc, varid, 'add_offset', add_offset)
      if (ierror .ne. 0) return
    endif

  end subroutine nc_create_var_double
    !  #]

  subroutine nc_create_var_float(unit_nc,var_name,ndims,dimids,standard_name,long_name,units, &
                                 scale_factor,add_offset,valid_min,valid_max, &
                                 varid,ierror)
    !  #[
    !
    ! Description: create float NetCDF variable with some standard settings
    !
    integer,intent(in)          :: unit_nc,ndims
    character(len=*),intent(in) :: var_name,standard_name,long_name,units
    integer,dimension(0:ndims-1),intent(in) :: dimids
    real(r8_),intent(in)        :: scale_factor,add_offset
    real(r8_),intent(in)        :: valid_min,valid_max
    integer,intent(out)         :: varid,ierror

    ierror = nc_def_var_float(unit_nc, trim(var_name), ndims, dimids, varid)
    if (ierror .ne. 0) return

    ierror = nc_put_att_float(unit_nc, varid, '_FillValue', undef_float)
    if (ierror .ne. 0) return

    ierror = nc_put_att_float(unit_nc, varid, 'missing_value', undef_float)
    if (ierror .ne. 0) return

    ! apply scale factor and offset to min and max values
    ! if both attributes are present, the data are scaled before the offset is added
    ! when scaled data are written, the application should first subtract the offset
    ! and then divide by the scale factor
    ierror = nc_put_att_float(unit_nc, varid, 'valid_min', &
                              real((valid_min-add_offset) / scale_factor,r4_))
    if (ierror .ne. 0) return

    ierror = nc_put_att_float(unit_nc, varid, 'valid_max', &
                              real((valid_max-add_offset) / scale_factor,r4_))
    if (ierror .ne. 0) return

    if (len_trim(standard_name) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'standard_name', standard_name)
      if (ierror .ne. 0) return
    endif

    ierror = nc_put_att_text(unit_nc, varid, 'long_name', long_name)
    if (ierror .ne. 0) return

    if (len_trim(units) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'units', units)
      if (ierror .ne. 0) return
    endif

    ! create scale factor and offset attributes if one of them differs from 1 resp. 0
    if ((abs(scale_factor - 1.0) .gt. 1.E-7) .or. (abs(add_offset) .gt. 1.E-7)) then
      ierror = nc_put_att_double(unit_nc, varid, 'scale_factor', scale_factor)
      if (ierror .ne. 0) return

      ierror = nc_put_att_double(unit_nc, varid, 'add_offset', add_offset)
      if (ierror .ne. 0) return
    endif

  end subroutine nc_create_var_float
    !  #]

  subroutine nc_create_var_int(unit_nc,var_name,ndims,dimids,standard_name,long_name,units, &
                               scale_factor,add_offset,valid_min,valid_max, &
                               varid,ierror)
    !  #[

    !
    ! Description: create int NetCDF variable with some standard settings
    !
    integer,intent(in)          :: unit_nc,ndims
    character(len=*),intent(in) :: var_name,standard_name,long_name,units
    integer,dimension(0:ndims-1),intent(in) :: dimids
    real(r8_),intent(in)        :: scale_factor,add_offset
    real(r8_),intent(in)        :: valid_min,valid_max
    integer,intent(out)         :: varid,ierror

    ierror = nc_def_var_int(unit_nc, trim(var_name), ndims, dimids, varid)
    if (ierror .ne. 0) return

    ierror = nc_put_att_int(unit_nc, varid, '_FillValue', undef_int)
    if (ierror .ne. 0) return

    ierror = nc_put_att_int(unit_nc, varid, 'missing_value', undef_int)
    if (ierror .ne. 0) return

    ! apply scale factor and offset to min and max values
    ! if both attributes are present, the data are scaled before the offset is added
    ! when scaled data are written, the application should first subtract the offset
    ! and then divide by the scale factor
    ierror = nc_put_att_int(unit_nc, varid, 'valid_min', &
                            nint((valid_min-add_offset) / scale_factor,i4_))
    if (ierror .ne. 0) return

    ierror = nc_put_att_int(unit_nc, varid, 'valid_max', &
                            nint((valid_max-add_offset) / scale_factor,i4_))
    if (ierror .ne. 0) return

    if (len_trim(standard_name) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'standard_name', standard_name)
      if (ierror .ne. 0) return
    endif

    ierror = nc_put_att_text(unit_nc, varid, 'long_name', long_name)
    if (ierror .ne. 0) return

    if (len_trim(units) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'units', units)
      if (ierror .ne. 0) return
    endif

    ! create scale factor and offset attributes if one of them differs from 1 resp. 0
    if ((abs(scale_factor - 1.0) .gt. 1.E-7) .or. (abs(add_offset) .gt. 1.E-7)) then

      ierror = nc_put_att_double(unit_nc, varid, 'scale_factor', scale_factor)
      if (ierror .ne. 0) return

      ierror = nc_put_att_double(unit_nc, varid, 'add_offset', add_offset)
      if (ierror .ne. 0) return
    endif

  end subroutine nc_create_var_int
    !  #]

  subroutine nc_create_var_short(unit_nc,var_name,ndims,dimids,standard_name,long_name,units, &
                                 scale_factor,add_offset,valid_min,valid_max, &
                                 varid,ierror)
    !  #[
    !
    ! Description: create short NetCDF variable with some standard settings
    !
    integer,intent(in)          :: unit_nc,ndims
    character(len=*),intent(in) :: var_name,standard_name,long_name,units
    integer,dimension(0:ndims-1),intent(in) :: dimids
    real(r8_),intent(in)        :: scale_factor,add_offset
    real(r8_),intent(in)        :: valid_min,valid_max
    integer,intent(out)         :: varid,ierror

    ierror = nc_def_var_short(unit_nc, trim(var_name), ndims, dimids, varid)
    if (ierror .ne. 0) return

    ierror = nc_put_att_short(unit_nc, varid, '_FillValue', undef_short)
    if (ierror .ne. 0) return

    ierror = nc_put_att_short(unit_nc, varid, 'missing_value', undef_short)
    if (ierror .ne. 0) return

    ! apply scale factor and offset to min and max values
    ! if both attributes are present, the data are scaled before the offset is added
    ! when scaled data are written, the application should first subtract the offset
    ! and then divide by the scale factor
    ierror = nc_put_att_short(unit_nc, varid, 'valid_min', &
                              nint((valid_min-add_offset) / scale_factor,i2_))
    if (ierror .ne. 0) return

    ierror = nc_put_att_short(unit_nc, varid, 'valid_max', &
                              nint((valid_max-add_offset) / scale_factor,i2_))
    if (ierror .ne. 0) return

    if (len_trim(standard_name) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'standard_name', standard_name)
      if (ierror .ne. 0) return
    endif

    ierror = nc_put_att_text(unit_nc, varid, 'long_name', long_name)
    if (ierror .ne. 0) return

    if (len_trim(units) .gt. 0) then
      ierror = nc_put_att_text(unit_nc, varid, 'units', units)
      if (ierror .ne. 0) return
    endif

    ! create scale factor and offset attributes if one of them differs from 1 resp. 0
    if ((abs(scale_factor - 1.0) .gt. 1.E-7) .or. (abs(add_offset) .gt. 1.E-7)) then

      ierror = nc_put_att_double(unit_nc, varid, 'scale_factor', scale_factor)
      if (ierror .ne. 0) return

      ierror = nc_put_att_double(unit_nc, varid, 'add_offset', add_offset)
      if (ierror .ne. 0) return
    endif

  end subroutine nc_create_var_short
    !  #]

  subroutine nc_put_array_double_1d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put double NetCDF variable in 1-dimensional array
    !
    integer,intent(in)  :: ncid
    integer,intent(in)  :: varid
    integer,intent(in)  :: start
    integer,intent(in)  :: edges
    real(r8_),dimension(start:start+edges-1),intent(in) :: value
    integer,intent(out) :: ierror

    ierror = nc_put_vara_double_1d(ncid, varid, start, edges, value)

  end subroutine nc_put_array_double_1d
    !  #]

  subroutine nc_put_array_double_2d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put double NetCDF variable in 2-dimensional array
    ! to compensate for difference in c and fortran array ordering
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:1),intent(in) :: start
    integer,dimension(0:1),intent(in) :: edges
    real(r8_),dimension(start(0):,start(1):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k
    real(r8_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        tmp_value(k) = value(i,j)
        k = k + 1
      enddo
    enddo

    ierror = nc_put_vara_double_2d(ncid, varid, start(0), start(1), &
                                   edges(0), edges(1), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_double_2d
    !  #]

  subroutine nc_get_array_double_2d(ncid, varid, size0, size1, &
                                    double_data, ierror)
    !  #[
    !
    ! Description: get double 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    real(r8_), dimension(0:size0-1,0:size1-1), intent(out) :: double_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    real(r8_),dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_double_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         double_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_double_2d
    !  #]

  subroutine nc_get_array_byte_2d(ncid, varid, size0, size1, &
                                    byte_data, ierror)
    !  #[
    !
    ! Description: get byte 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    ! added by LZ of this byte_2d type
    use numerics, only: i1_

    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer(i1_), dimension(0:size0-1,0:size1-1), intent(out) :: byte_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    integer(i1_),dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_byte_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         byte_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_byte_2d
    !  #]

  subroutine nc_get_array_text_2d(ncid, varid, size0, size1, &
                                    char_data, ierror)
    !  #[
    !
    ! Description: get char 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    ! added by LZ of this text_2d type

    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    character(len=1), dimension(0:size0-1,0:size1-1), intent(out) :: char_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    character(len=1),dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_text_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         char_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_text_2d
    !  #]

  subroutine nc_put_array_double_3d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put double NetCDF variable in 3-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:2),intent(in) :: start
    integer,dimension(0:2),intent(in) :: edges
    real(r8_),dimension(start(0):,start(1):,start(2):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k,l
    real(r8_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) * (edges(2)-start(2)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
         do l =  start(2),start(2)+edges(2)-1
            tmp_value(k) = value(i,j,l)
            k = k + 1
         enddo
      enddo
    enddo

    ierror = nc_put_vara_double_3d(ncid, varid, start(0), start(1), start(2), &
                                   edges(0), edges(1), edges(2), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_double_3d
    !  #]

  subroutine nc_get_array_double_3d(ncid, varid, size0, size1, size2, &
                                   double_data, ierror)
    !  #[
    !
    ! Description: get double 3-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer, intent(in)  :: size2
    real(r8_), dimension(0:size0-1,0:size1-1,0:size2-1), &
         intent(out) :: double_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k,l
    real(r8_),dimension(:),allocatable :: tmp_value

    isize = size0*size1*size2
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_double_3d(ncid, varid, size0, size1, size2, tmp_value)

    ! convert 3d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         do l = 0,size2-1
            double_data(i,j,l) = tmp_value(k)
            k = k + 1
         end do
      enddo
    enddo

  end subroutine nc_get_array_double_3d
    !  #]

  subroutine nc_put_array_float_1d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put float NetCDF variable in 1-dimensional array
    !
    integer,intent(in)  :: ncid
    integer,intent(in)  :: varid
    integer,intent(in)  :: start
    integer,intent(in)  :: edges
    real(r4_),dimension(start:start+edges-1),intent(in) :: value
    integer,intent(out) :: ierror

    ierror = nc_put_vara_float_1d(ncid, varid, start, edges, value)

  end subroutine nc_put_array_float_1d
    !  #]

  subroutine nc_put_array_float_2d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put float NetCDF variable in 2-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:1),intent(in) :: start
    integer,dimension(0:1),intent(in) :: edges
    real(r4_),dimension(start(0):,start(1):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k
    real(r4_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        tmp_value(k) = value(i,j)
        k = k + 1
      enddo
    enddo

    ierror = nc_put_vara_float_2d(ncid, varid, start(0), start(1), &
                                  edges(0), edges(1), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_float_2d
    !  #]

  subroutine nc_get_array_float_2d(ncid, varid, size0, size1, &
                                   float_data, ierror)
    !  #[
    !
    ! Description: get float 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    real(r4_), dimension(0:size0-1,0:size1-1), intent(out) :: float_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    real(r4_),dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_float_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         float_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_float_2d
    !  #]

  subroutine nc_put_array_float_3d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put integer NetCDF variable in 3-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:2),intent(in) :: start
    integer,dimension(0:2),intent(in) :: edges
    real(r4_),dimension(start(0):,start(1):,start(2):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k,l
    real(r4_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) * (edges(2)-start(2)) - 1
    allocate(tmp_value(0:isize))

    ! convert 3d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        do l =  start(2),start(2)+edges(2)-1
        tmp_value(k) = value(i,j,l)
        k = k + 1
        enddo
      enddo
    enddo

    ierror = nc_put_vara_float_3d(ncid, varid, start(0), start(1), start(2), &
                                  edges(0), edges(1), edges(2), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_float_3d
    !  #]

  subroutine nc_get_array_float_3d(ncid, varid, size0, size1, size2, &
                                   float_data, ierror)
    !  #[
    !
    ! Description: get float 3-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer, intent(in)  :: size2
    real(r4_), dimension(0:size0-1,0:size1-1,0:size2-1), &
         intent(out) :: float_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k,l
    real(r4_),dimension(:),allocatable :: tmp_value

    isize = size0*size1*size2
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_float_3d(ncid, varid, size0, size1, size2, tmp_value)

    ! convert 3d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         do l = 0,size2-1
            float_data(i,j,l) = tmp_value(k)
            k = k + 1
         end do
      enddo
    enddo

  end subroutine nc_get_array_float_3d
    !  #]

  subroutine nc_put_array_int_1d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put integer NetCDF variable in 1-dimensional array
    !
    integer,intent(in)  :: ncid
    integer,intent(in)  :: varid
    integer,intent(in)  :: start
    integer,intent(in)  :: edges
    integer(i4_),dimension(start:start+edges-1),intent(in) :: value
    integer,intent(out) :: ierror

    ierror = nc_put_vara_int_1d(ncid, varid, start, edges, value)

  end subroutine nc_put_array_int_1d
    !  #]

  subroutine nc_put_array_int_2d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put integer NetCDF variable in 2-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:1),intent(in) :: start
    integer,dimension(0:1),intent(in) :: edges
    integer(i4_),dimension(start(0):,start(1):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k
    integer(i4_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        tmp_value(k) = value(i,j)
        k = k + 1
      enddo
    enddo

    ierror = nc_put_vara_int_2d(ncid, varid, start(0), start(1), &
                                edges(0), edges(1), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_int_2d
    !  #]

  subroutine nc_get_array_int_2d(ncid, varid, size0, size1, &
                                 int_data, ierror)
    !  #[
    !
    ! Description: get int 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer(i4_), dimension(0:size0-1,0:size1-1), intent(out) :: int_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    integer(i4_), dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_int_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         int_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_int_2d
    !  #]

  subroutine nc_put_array_int_3d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put integer NetCDF variable in 3-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:2),intent(in) :: start
    integer,dimension(0:2),intent(in) :: edges
    integer(i4_),dimension(start(0):,start(1):,start(2):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k,l
    integer(i4_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) * (edges(2)-start(2)) - 1
    allocate(tmp_value(0:isize))

    ! convert 3d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        do l =  start(2),start(2)+edges(2)-1
        tmp_value(k) = value(i,j,l)
        k = k + 1
        enddo
      enddo
    enddo

    ierror = nc_put_vara_int_3d(ncid, varid, start(0), start(1), start(2), &
                                edges(0), edges(1), edges(2), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_int_3d
    !  #]

  subroutine nc_get_array_int_3d(ncid, varid, size0, size1, size2, &
                                 int_data, ierror)
    !  #[
    !
    ! Description: get int 3-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer, intent(in)  :: size2
    integer(i4_), dimension(0:size0-1,0:size1-1,0:size2-1), &
         intent(out) :: int_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k,l
    integer(i4_), dimension(:),allocatable :: tmp_value

    isize = size0*size1*size2
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_int_3d(ncid, varid, size0, size1, size2, tmp_value)

    ! convert 3d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         do l = 0,size2-1
            int_data(i,j,l) = tmp_value(k)
            k = k + 1
         end do
      enddo
    enddo

  end subroutine nc_get_array_int_3d
    !  #]

  subroutine nc_put_array_short_1d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put short NetCDF variable in 1-dimensional array
    !
    integer,intent(in)  :: ncid
    integer,intent(in)  :: varid
    integer,intent(in)  :: start
    integer,intent(in)  :: edges
    integer(i2_),dimension(start:start+edges-1),intent(in) :: value
    integer,intent(out) :: ierror

    ierror = nc_put_vara_short_1d(ncid, varid, start, edges, value)

  end subroutine nc_put_array_short_1d
    !  #]

  subroutine nc_put_array_short_2d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put short NetCDF variable in 2-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:1),intent(in) :: start
    integer,dimension(0:1),intent(in) :: edges
    integer(i2_),dimension(start(0):,start(1):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,k
    integer(i2_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        tmp_value(k) = value(i,j)
        k = k + 1
      enddo
    enddo

    ierror = nc_put_vara_short_2d(ncid, varid, start(0), start(1), &
                                  edges(0), edges(1), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_short_2d
    !  #]

  subroutine nc_get_array_short_2d(ncid, varid, size0, size1, &
                                   short_data, ierror)
    !  #[
    !
    ! Description: get short 2-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer(i2_), dimension(0:size0-1,0:size1-1), intent(out) :: short_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k
    integer(i2_), dimension(:),allocatable :: tmp_value

    isize = size0*size1
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_short_2d(ncid, varid, size0, size1, tmp_value)

    ! convert 2d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         short_data(i,j) = tmp_value(k)
        k = k + 1
      enddo
    enddo

  end subroutine nc_get_array_short_2d
    !  #]

  subroutine nc_put_array_short_3d(ncid,varid,start,edges,value,ierror)
    !  #[
    !
    ! Description: put short NetCDF variable in 3-dimensional array
    !
    integer,intent(in)                :: ncid
    integer,intent(in)                :: varid
    integer,dimension(0:2),intent(in) :: start
    integer,dimension(0:2),intent(in) :: edges
    integer(i2_),dimension(start(0):,start(1):,start(2):),intent(in) :: value
    integer,intent(out)               :: ierror

    integer :: isize,i,j,l,k
    integer(i2_),dimension(:),allocatable :: tmp_value

    isize = (edges(0)-start(0)) * (edges(1)-start(1)) * (edges(2)-start(2)) - 1
    allocate(tmp_value(0:isize))

    ! convert 2d-array to temporary linear array
    k = 0
    do i = start(0),start(0)+edges(0)-1
      do j = start(1),start(1)+edges(1)-1
        do l =  start(2),start(2)+edges(2)-1
        tmp_value(k) = value(i,j,l)
        k = k + 1
        enddo
      enddo
    enddo

    ierror = nc_put_vara_short_3d(ncid, varid, start(0), start(1), start(2), &
                                  edges(0), edges(1), edges(2), tmp_value)

    deallocate(tmp_value)

  end subroutine nc_put_array_short_3d
    !  #]

  subroutine nc_get_array_short_3d(ncid, varid, size0, size1, size2, &
                                   short_data, ierror)
    !  #[
    !
    ! Description: get int 3-dimensional array NetCDF variable
    ! and compensate for difference in c and fortran array ordering
    !
    integer, intent(in)  :: ncid
    integer, intent(in)  :: varid
    integer, intent(in)  :: size0
    integer, intent(in)  :: size1
    integer, intent(in)  :: size2
    integer(i2_), dimension(0:size0-1,0:size1-1,0:size2-1), &
         intent(out) :: short_data
    integer, intent(out) :: ierror

    ! local variables
    integer :: isize,i,j,k,l
    integer(i2_), dimension(:),allocatable :: tmp_value

    isize = size0*size1*size2
    allocate(tmp_value(0:isize-1))

    ierror = nc_get_var_short_3d(ncid, varid, size0, size1, size2, tmp_value)

    ! convert 3d-array to temporary linear array
    k = 0
    do i = 0,size0-1
      do j = 0,size1-1
         do l = 0,size2-1
            short_data(i,j,l) = tmp_value(k)
            k = k + 1
         end do
      enddo
    enddo

  end subroutine nc_get_array_short_3d
    !  #]

end module NetCDFMod
