  program TestNetCDF
  !
  ! Example program to create a small NetCDF test file using the interface routines
  ! and the UCAR/UNIDATA software package.
  !
  use NetCDFMod, only: nc_global, &
                       nc_create, nc_enddef, nc_close, nc_def_dim, &
                       nc_put_att_text, &
                       nc_put_array_double_1d, nc_put_array_float_1d, nc_put_array_int_1d, nc_put_array_short_1d, &
                       nc_create_var_short, nc_create_var_int, nc_create_var_float, nc_create_var_double, &
                       nc_real_to_short, nc_real_to_int, nc_real_to_float, nc_real_to_double

  use numerics, only: i2_, i4_, r4_, r8_

  implicit none
  
  integer, parameter :: num_obs = 4
  integer :: unit_nc, error_flag
  integer :: varid_dum_short, varid_dum_int, varid_dum_float, varid_dum_double
  integer :: i

  character(len=256) :: tmp_str

  integer,dimension(0:0) :: dimid_obs
  integer(i2_),dimension(0:num_obs-1) :: dum_short
  integer(i4_),dimension(0:num_obs-1) :: dum_int
  real(r4_),dimension(0:num_obs-1)    :: dum_float
  real(r8_),dimension(0:num_obs-1)    :: dum_double

  ! fill arrays with dummy data
  do i = 0, num_obs-2
    dum_short(i) = int(i, i2_)
    dum_int(i) = int(i*2, i4_)
    dum_float(i) = real(i*3, r4_) * 1.1_r4_
    dum_double(i) = real(i*4, r8_) * 1.2_r8_
  enddo
  dum_short(num_obs-1) = nc_real_to_short(1.7E38)
  dum_int(num_obs-1) = nc_real_to_int(1.7E38)
  dum_float(num_obs-1) = nc_real_to_float(1.7E38)
  dum_double(num_obs-1) = nc_real_to_double(1.7E38)

  ! create NetCDF file
  error_flag = nc_create('testfile.nc', unit_nc)
  if (error_flag .ne. 0) stop

  ! define dimensions
  error_flag = nc_def_dim(unit_nc, 'observation' // achar(0), num_obs, dimid_obs(0))
  if (error_flag .ne. 0) stop

  ! define some dummy fields
  call nc_create_var_short(unit_nc, 'dum_short', 1, dimid_obs, 'dum_short_var', &
                           'dummy short variable', '1', 1.0E-2_r8_, 0.1_r8_, 0.0_r8_, 99.0_r8_, &
                           varid_dum_short, error_flag)
  if (error_flag .ne. 0) stop

  call nc_create_var_int(unit_nc, 'dum_int', 1, dimid_obs, 'dum_int_var', &
                         'dummy integer variable', '1', 1.0_r8_, 0.0_r8_, 0.0_r8_, 999.0_r8_, &
                         varid_dum_int, error_flag)
  if (error_flag .ne. 0) stop

  call nc_create_var_float(unit_nc, 'dum_float', 1, dimid_obs, ' ', &
                           'dummy float variable', '1', 1.0_r8_, 0.0_r8_, 0.0_r8_, 999.0_r8_, &
                           varid_dum_float, error_flag)
  if (error_flag .ne. 0) stop

  call nc_create_var_double(unit_nc, 'dum_double', 1, dimid_obs, ' ', &
                            'dummy double variable', '1', 1.0_r8_, 0.0_r8_, 0.0_r8_, 999.0_r8_, &
                            varid_dum_double, error_flag)
  if (error_flag .ne. 0) stop

  ! define global attributes
  write(tmp_str,'(A)') 'Dummy NetCDF test file'
  error_flag = nc_put_att_text(unit_nc, nc_global, 'title', tmp_str)
  if (error_flag .ne. 0) stop

  ! leave define mode
  error_flag = nc_enddef(unit_nc)
  if (error_flag .ne. 0) stop

  ! put data
  call nc_put_array_short_1d(unit_nc, varid_dum_short, 0, num_obs, dum_short, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_int_1d(unit_nc, varid_dum_int, 0, num_obs, dum_int, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_float_1d(unit_nc, varid_dum_float, 0, num_obs, dum_float, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_double_1d(unit_nc, varid_dum_double, 0, num_obs, dum_double, error_flag)
  if (error_flag .ne. 0) stop

  ! close NetCDF file
  error_flag = nc_close(unit_nc)
  if (error_flag .ne. 0) stop

  write(*,'(A)') ' '
  write(*,'(A)') 'A NetCDF file called testfile.nc was created'
  write(*,'(A)') 'You can check its contents e.g. using the ncdump utility'
  write(*,'(A)') ' '

end program TestNetCDF
