/*
 *  This module contains interface routines for calling UCAR/UNIDATA NetCDFC
 *  routines in C from Fortran.
 *
 *
 *  WARNING: currently the "long int" and "unsigned long int" types
 *  are not used in our c code. Also the corresponding interfacing
 *  to netcdf types NC_UINT64 and NC_INT64 are not used.
 *  Please remind that special care is needed if we wish to introduce these.
 *  The c type can be 32 or 64 bits wide, depending on machine architecture!
 *  Therefore special runtime code is needed as workaround when interfacing
 *  fortran and c code with these types.
 *  If we decide to implement these we should:
 *  ==> activate the long size check in subroutine check_c_f90_interface
 *      in netcdfreader.F90 again.
 *  ==> add special interfaces in the fortran side of the interface
 *      in the file NetCDFMod.F90
 *      (see for examples support/c_support.F90)
 *  ==> add double function definitions (_c4_ and _c8_ versions) to
 *      the c side of the interface in the current file
 *      (see for examples support/F90_c_support.c)
 */

#include <stdio.h>
#include <stdlib.h>
#include <netcdf.h>
#include <string.h>
#include <stdbool.h>

/*
 * helper function to convert fortran string to c string
 * BEWARE: this function manually allocates memory, so remember to store
 * the returned pointer in a variable and free after use or you
 * will get memory leaks.
 */
char *string_f90_to_c(char *ftext, int len)
{
  int n;
  char *ctext = (char *) malloc(len+1);

  if(!ctext) {
      perror("malloc problem in string_f90_to_c");
      return((char*)0);
  }
  /* copy all chars, and ensure the last char is initialised to \0  */
  strncpy(ctext,ftext,len);
  ctext[len] = '\0';

  n = len-1; /* skip the final \0 in the next loop */
  while( (n > 0) && ( (ctext[n] == ' ') || (ctext[n] == '\0') ) )
    { n--; }
  /* when the first non-space character is found, add a \0 character
     which terminates a c string */
  ctext[++n]='\0';

  return(ctext);
}

/*
 * helper function to convert c string to fortran string (inplace)
 */
void string_c_to_f90(char *text, int len)
{
  int n;
  bool line_end_found;

  /* find the first \0 in the string and fill from that location
     untill the end with spaces */
  n = 0;
  line_end_found = false;
  for (n=0; n < len; n++) {
    if (text[n] == '\0')
      {
        /* printf("==>line end found at n=%i\n", n); */
        line_end_found = true; }
    if (line_end_found)
      {
        /*printf("==>inserting space at n=%i\n", n);*/
        text[n] = ' '; }
  }
}

/*
 * some helper functions to inspect variable sizes during runtime
 */
char get_bytesize_short_()
{
  short testshort;
  return sizeof(testshort);
}
char get_bytesize_int_()
{
  int testinteger;
  return sizeof(testinteger);
}
char get_bytesize_long_()
{
  long testlong;
  return sizeof(testlong);
}
char get_bytesize_float_()
{
  float testfloat;
  return sizeof(testfloat);
}
char get_bytesize_double_()
{
  double testdouble;
  return sizeof(testdouble);
}



/*
 * Function to check for NetCDF error. If error then write message
 * and close NetCDF file.
 */
void nc_check_status(int *ncid, int status)
{
  if (status != NC_NOERR) {
    printf("ERROR: %s\n", nc_strerror(status));
    nc_close(*ncid);
  }
}

/*
 * Function to open a NetCDF file.
 */
int nc_open_(char *path, char mode, int *ncid_ptr,
             int path_len) /* hidden input: path length */
{
  int status, nc_mode;
  char *cpath;

  /* printf("path_len = %i ",path_len); */

  nc_mode = NC_NOWRITE;
  if (mode=='r'){ nc_mode = NC_NOWRITE; }
  if (mode=='w'){ nc_mode = NC_WRITE; }

  cpath = string_f90_to_c(path, path_len);
  status = nc_open(cpath, nc_mode, ncid_ptr);
  nc_check_status(ncid_ptr, status);

  free(cpath);

  return status;
}

/*
 * Fuction to inquire the NetCDF for its objects
 */
int nc_inq_(int *ncid, int *ndims, int *nvars, int *natts, int *unlimdimid)
{
  int status;

  status = nc_inq(*ncid, ndims, nvars, natts, unlimdimid);
  nc_check_status(ncid, status);

  return status;
}


/*
 * Fuction to request the name of a NetCDF attribute
 */
int nc_inq_attname_(int *ncid, int *varid, int *count, char *name,
                    int name_len) /* hidden input: name length */
{
  int status;

  status = nc_inq_attname(*ncid, *varid, *count, name);
  nc_check_status(ncid, status);
  string_c_to_f90(name, name_len);

  return status;
}

/*
 * Function to request the id for a named attribute
 */
int nc_inq_attid_(int *ncid, int *varid, char *attrname, int *attrid,
                  int attrname_len) /* hidden input: attrname length */
{
  int status;
  char *cattrname;

  cattrname = string_f90_to_c(attrname,attrname_len);
  status = nc_inq_attid(*ncid, *varid, cattrname, attrid);
  nc_check_status(ncid, status);

  free(cattrname);
  return status;
}

/*
 * Fuction to request type of a named NetCDF attribute
 */
int nc_inq_atttype_(int *ncid, int *varid, char *name, char *xtype,
                    int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_inq_atttype(*ncid, *varid, cname, (nc_type *)xtype);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to request length of a named NetCDF text attribute
 */
int nc_inq_attlen_(int *ncid, int *varid, char *name, int *len,
                   int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_inq_attlen(*ncid, *varid, cname, (size_t *)len);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to get text value of a named NetCDF text attribute
 */
int nc_get_att_text_(int *ncid, int *varid, char *name, char *value,
                     int name_len, int value_len)
                     /* hidden inputs: name length, value length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_get_att_text(*ncid, *varid, cname, value);
  nc_check_status(ncid, status);

  /* note: this does not always work as expected. It seems these NetCDF text
     results do not always have a closing zero character in their strings ...
     We'll have to inspect the reported length to truncate them then.
  */
  string_c_to_f90(value, value_len);
  free(cname);

  return status;
}
/*
 * Fuction to get short value of a named NetCDF int attribute
 */
int nc_get_att_short_(int *ncid, int *varid, char *name, short *value,
                      int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_get_att_short(*ncid, *varid, cname, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to get int value of a named NetCDF int attribute
 */
int nc_get_att_int_(int *ncid, int *varid, char *name, int *value,
                    int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_get_att_int(*ncid, *varid, cname, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to get float value of a named NetCDF int attribute
 */
int nc_get_att_float_(int *ncid, int *varid, char *name, float *value,
                      int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_get_att_float(*ncid, *varid, cname, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to get double value of a named NetCDF int attribute
 */
int nc_get_att_double_(int *ncid, int *varid, char *name, double *value,
                       int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  status = nc_get_att_double(*ncid, *varid, cname, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Fuction to get 1d array of int values for a named NetCDF int attribute
 * (basically identical to nc_get_att_int_, but the fortran interfacing
 *  differs)
 */
int nc_get_att_int_1d_(int *ncid, int *varid, int *size0, char *name,
		       int *int_data_1d,
		       int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name,name_len);
  /*
  printf("getting attribute with name: %s\n", name);
  printf("getting attribute with cname: %s\n", cname);
  */
  status = nc_get_att_int(*ncid, *varid, cname, int_data_1d);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to request the id for a named attribute
 */
int nc_inq_varname_(int *ncid, int *varid, char *name,
                    int name_len) /* hidden input: name length */
{
  int status;

  status = nc_inq_varname(*ncid, *varid, name);
  nc_check_status(ncid, status);
  string_c_to_f90(name, name_len);

  return status;
}

/*
 * Fuction to request type of a NetCDF variable
 */
int nc_inq_vartype_(int *ncid, int *varid, char *xtype)
{
  int status;

  status = nc_inq_vartype(*ncid, *varid, (nc_type *)xtype);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to request ndims of a NetCDF variable
 */
int nc_inq_varndims_(int *ncid, int *varid, int *ndims)
{
  int status;

  status = nc_inq_varndims(*ncid, *varid, ndims);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to request the list of dimids of a NetCDF variable
 */
int nc_inq_vardimid_(int *ncid, int *varid,
		     int *dimid0, int *dimid1, int *dimid2)
{
  int status;
  int tmp_dimids[3];
  /* NOTE: there seems no direct way to pass a c array like dimids[3]
   to the calling fortran code, therefore it is copied here as workaround */

  tmp_dimids[0] = -1;
  tmp_dimids[1] = -1;
  tmp_dimids[2] = -1;

  status = nc_inq_vardimid(*ncid, *varid, tmp_dimids);
  nc_check_status(ncid, status);

  *dimid0 = tmp_dimids[0];
  *dimid1 = tmp_dimids[1];
  *dimid2 = tmp_dimids[2];

  return status;
}

/*
 * Fuction to inquire a NetCDF variable for its number of attached attributes
 */
int nc_inq_varnatts_(int *ncid, int *varid, int *natts)
{
  int status;

  status = nc_inq_varnatts(*ncid, *varid, natts);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to inquire a NetCDF dimension for its name and length
 */
int nc_inq_dim_(int *ncid, int *dimid, char* name, int *length,
                int name_len) /* hidden input: name length */
{
  int status;

  status = nc_inq_dim(*ncid, *dimid, name, (size_t *)length);
  nc_check_status(ncid, status);
  string_c_to_f90(name, name_len);

  return status;
}

/*
 * Fuction to get all values of a 1D NetCDF byte/char variable
 */
int nc_get_var_byte_1d_(int *ncid, int *varid,
                        int *size0, signed char *char_data_1d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
     nc_get_var_byte_1d_ added by LZ -----
   */

  status = nc_get_var_schar(*ncid, *varid, char_data_1d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF byte/char variable
 */
int nc_get_var_byte_2d_(int *ncid, int *varid,
                        int *size0, int *size1, signed char *char_data_2d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
     nc_get_var_byte_2d_ added by LZ -----
   */

  status = nc_get_var_schar(*ncid, *varid, char_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF text variable
 */
int nc_get_var_text_2d_(int *ncid, int *varid,
                       int *size0, int *size1, char *char_data_2d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
     nc_get_var_text_2d_ added by LZ -----
   */

  status = nc_get_var_text(*ncid, *varid, char_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 1D NetCDF short variable
 */
int nc_get_var_short_1d_(int *ncid, int *varid,
                         int *size0, short *short_data_1d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_short(*ncid, *varid, short_data_1d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF short variable
 */
int nc_get_var_short_2d_(int *ncid, int *varid,
                         int *size0, int *size1, short *short_data_2d)
{
  int status;
  /* note: size0 and size1 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_short(*ncid, *varid, short_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 3D NetCDF short variable
 */
int nc_get_var_short_3d_(int *ncid, int *varid,
                         int *size0, int *size1, int *size2,
			 short *short_data_3d)
{
  int status;
  /* note: size0, size1 and size2 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_short(*ncid, *varid, short_data_3d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 1D NetCDF int variable
 */
int nc_get_var_int_1d_(int *ncid, int *varid,
                       int *size0, int *int_data_1d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_int(*ncid, *varid, int_data_1d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF int variable
 */
int nc_get_var_int_2d_(int *ncid, int *varid,
                       int *size0, int *size1, int *int_data_2d)
{
  int status;
  /* note: size0 and size1 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_int(*ncid, *varid, int_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 3D NetCDF int variable
 */
int nc_get_var_int_3d_(int *ncid, int *varid,
                       int *size0, int *size1, int *size2,
		       int *int_data_3d)
{
  int status;
  /* note: size0, size1 and size2 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_int(*ncid, *varid, int_data_3d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 1D NetCDF float variable
 */
int nc_get_var_float_1d_(int *ncid, int *varid,
                         int *size0, float *float_data_1d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_float(*ncid, *varid, float_data_1d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF float variable
 */
int nc_get_var_float_2d_(int *ncid, int *varid,
                         int *size0, int *size1, float *float_data_2d)
{
  int status;
  /* note: size0 and size1 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_float(*ncid, *varid, float_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 3D NetCDF float variable
 */
int nc_get_var_float_3d_(int *ncid, int *varid,
                         int *size0, int *size1, int *size2,
			 float *float_data_3d)
{
  int status;
  /* note: size0, size1 and size2 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_float(*ncid, *varid, float_data_3d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 1D NetCDF double variable
 */
int nc_get_var_double_1d_(int *ncid, int *varid,
                          int *size0, double *double_data_1d)
{
  int status;
  /* note: size0 is not used here, but is essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_double(*ncid, *varid, double_data_1d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 2D NetCDF double variable
 */
int nc_get_var_double_2d_(int *ncid, int *varid,
                          int *size0, int *size1, double *double_data_2d)
{
  int status;
  /* note: size0 and size1 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_double(*ncid, *varid, double_data_2d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Fuction to get all values of a 3D NetCDF double variable
 */
int nc_get_var_double_3d_(int *ncid, int *varid,
                          int *size0, int *size1, int *size2,
			  double *double_data_3d)
{
  int status;
  /* note: size0, size1 and size2 are not used here, but they are essential
     for the fortran interface definition in NetCDFMod.F90.
     Without it you get lot's of ugly segmentation faults.
   */

  status = nc_get_var_double(*ncid, *varid, double_data_3d);
  nc_check_status(ncid, status);

  return status;
}

/*
 * Function to create NetCDF file. Overwrite existing file.
 */
int nc_create_(char *path, int *ncid_ptr,
               int path_len) /* hidden input: path length */
{
  int status;
  char *cpath;

  cpath = string_f90_to_c(path, path_len);
  status = nc_create(cpath, NC_CLOBBER, ncid_ptr);
  nc_check_status(ncid_ptr, status);

  free(cpath);

  return status;
}

/*
 * Fuction to end NetCDF define mode.
 */
int nc_enddef_(int *ncid)
{
  int status;

  status = nc_enddef(*ncid);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Fuction to close NetCDF file.
 */
int nc_close_(int *ncid)
{
  int status;

  status = nc_close(*ncid);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to define dimension of NetCDF array.
 */
int nc_def_dim_(int *ncid, char *name, int *size, int *dimidp,
		int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_def_dim(*ncid, cname, (size_t)*size, dimidp);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to define NetCDF variable of type double.
 */
int nc_def_var_double_(int *ncid, char *name, int *ndims, int *dimids,
		       int *varidp, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_def_var(*ncid, cname, NC_DOUBLE, *ndims, dimids, varidp);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to define NetCDF variable of type float.
 */
int nc_def_var_float_(int *ncid, char *name, int *ndims, int *dimids,
		      int *varidp, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_def_var(*ncid, cname, NC_FLOAT, *ndims, dimids, varidp);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to define NetCDF variable of type int.
 */
int nc_def_var_int_(int *ncid, char *name, int *ndims, int *dimids,
		    int *varidp, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_def_var(*ncid, cname, NC_INT, *ndims, dimids, varidp);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to define NetCDF variable of type short.
 */
int nc_def_var_short_(int *ncid, char *name, int *ndims, int *dimids,
		      int *varidp, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_def_var(*ncid, cname, NC_SHORT, *ndims, dimids, varidp);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF scalar attribute of type double.
 */
int nc_put_att_double_(int *ncid, int *varid, char *name, double *value,
		       int name_len)  /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_double(*ncid, *varid, cname, NC_DOUBLE, 1, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF array attribute of type double.
 */
int nc_put_att_double_arr_(int *ncid, int *varid, char *name, int *nelems,
			   double *value, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_double(*ncid, *varid, cname, NC_DOUBLE,
			     (size_t)*nelems, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF scalar attribute of type float.
 */
int nc_put_att_float_(int *ncid, int *varid, char *name, float *value,
		      int name_len) /* hidden input: name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_float(*ncid, *varid, cname, NC_FLOAT, 1, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF array attribute of type float.
 */
int nc_put_att_float_arr_(int *ncid, int *varid, char *name, int *nelems,
			  float *value, int name_len)
{ /* hidden input: name length */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_float(*ncid, *varid, cname, NC_FLOAT, (size_t)*nelems,
			    value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF scalar attribute of type int.
 */
int nc_put_att_int_(int *ncid, int *varid, char *name, int *value,
                    int name_len) /* hidden input name length */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_int(*ncid, *varid, cname, NC_INT, 1, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF array attribute of type int,
 * i.e. to put 1d array of int values for a named NetCDF int attribute
 * (basically identical to nc_put_att_int_, but the fortran interfacing
 *  differs)
 */
int nc_put_att_int_arr_(int *ncid, int *varid, char *name, int *nelems,
			int *value, int name_len) /* hidden input name lenght */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_int(*ncid, *varid, cname, NC_INT, (size_t)*nelems, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF scalar attribute of type short.
 */
int nc_put_att_short_(int *ncid, int *varid, char *name, short *value,
		      int name_len) /* hidden input name lenght */
{
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_short(*ncid, *varid, cname, NC_SHORT, 1, value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF array attribute of type short.
 */
int nc_put_att_short_arr_(int *ncid, int *varid, char *name, int *nelems,
			  short *value, int name_len)
{  /* hidden input name lenght */
  int status;
  char *cname;

  cname = string_f90_to_c(name, name_len);
  status = nc_put_att_short(*ncid, *varid, cname, NC_SHORT, (size_t)*nelems,
			    value);
  nc_check_status(ncid, status);

  free(cname);

  return status;
}

/*
 * Function to put NetCDF attribute of type text.
 */
int nc_put_att_text_(int *ncid, int *varid, char *name, char *textvalue,
                     int name_len, int textvalue_len) /* hidden inputs */
{
  int status, ctextvalue_len;
  char *cname, *ctextvalue;

  cname = string_f90_to_c(name, name_len);
  ctextvalue = string_f90_to_c(textvalue, textvalue_len);
  ctextvalue_len = strlen(ctextvalue);

  status = nc_put_att_text(*ncid, *varid, cname,
              (size_t) ctextvalue_len, ctextvalue);
  nc_check_status(ncid, status);

  free(cname);
  free(ctextvalue);

  return status;
}

/*
 * Function to put NetCDF 1d array of type double.
 */
int nc_put_vara_double_1d_(int *ncid, int *varid, int *start, int *edges, double *value)
{
  size_t tmp_start, tmp_edges;
  int status;

  tmp_start = *start;
  tmp_edges = *edges;
  status = nc_put_vara_double(*ncid, *varid, &tmp_start, &tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 2d array of type double.
 */
int nc_put_vara_double_2d_(int *ncid, int *varid, int *start0, int *start1, int *edges0, int *edges1, double *value)
{
  size_t tmp_start[2], tmp_edges[2];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  status = nc_put_vara_double(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 3d array of type double.
 */
int nc_put_vara_double_3d_(int *ncid, int *varid,
			   int *start0, int *start1, int *start2,
			   int *edges0, int *edges1, int *edges2,
			   double *value)
{
  size_t tmp_start[2], tmp_edges[2];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_start[2] = *start2;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  tmp_edges[2] = *edges2;
  status = nc_put_vara_double(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 1d array of type float.
 */
int nc_put_vara_float_1d_(int *ncid, int *varid, int *start, int *edges, float *value)
{
  size_t tmp_start, tmp_edges;
  int status;

  tmp_start = *start;
  tmp_edges = *edges;
  status = nc_put_vara_float(*ncid, *varid, &tmp_start, &tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 2d array of type float.
 */
int nc_put_vara_float_2d_(int *ncid, int *varid, int *start0, int *start1, int *edges0, int *edges1, float *value)
{
  size_t tmp_start[2], tmp_edges[2];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  status = nc_put_vara_float(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 3d array of type float.
 */
int nc_put_vara_float_3d_(int *ncid, int *varid,
			  int *start0, int *start1, int *start2,
			  int *edges0, int *edges1, int *edges2,
			  float *value)
{
  size_t tmp_start[3], tmp_edges[3];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_start[2] = *start2;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  tmp_edges[2] = *edges2;
  status = nc_put_vara_float(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 1d array of type int.
 */
int nc_put_vara_int_1d_(int *ncid, int *varid, int *start, int *edges, int *value)
{
  size_t tmp_start, tmp_edges;
  int status;

  tmp_start = *start;
  tmp_edges = *edges;
  status = nc_put_vara_int(*ncid, *varid, &tmp_start, &tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 2d array of type int.
 */
int nc_put_vara_int_2d_(int *ncid, int *varid, int *start0, int *start1, int *edges0, int *edges1, int *value)
{
  size_t tmp_start[2], tmp_edges[2];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  status = nc_put_vara_int(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 3d array of type int.
 */
int nc_put_vara_int_3d_(int *ncid, int *varid, int *start0, int *start1, int *start2, int *edges0, int *edges1, int *edges2, int *value)
{
  size_t tmp_start[3], tmp_edges[3];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_start[2] = *start2;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  tmp_edges[2] = *edges2;
  status = nc_put_vara_int(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}
/*
 * Function to put NetCDF 1d array of type short.
 */
int nc_put_vara_short_1d_(int *ncid, int *varid, int *start, int *edges, short *value)
{
  size_t tmp_start, tmp_edges;
  int status;

  tmp_start = *start;
  tmp_edges = *edges;
  status = nc_put_vara_short(*ncid, *varid, &tmp_start, &tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 2d array of type short.
 */
int nc_put_vara_short_2d_(int *ncid, int *varid, int *start0, int *start1, int *edges0, int *edges1, short *value)
{
  size_t tmp_start[2], tmp_edges[2];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  status = nc_put_vara_short(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}

/*
 * Function to put NetCDF 3d array of type short.
 */
int nc_put_vara_short_3d_(int *ncid, int *varid, int *start0, int *start1, int *start2, int *edges0, int *edges1, int *edges2, short *value)
{
  size_t tmp_start[3], tmp_edges[3];
  int status;

  tmp_start[0] = *start0;
  tmp_start[1] = *start1;
  tmp_start[2] = *start2;
  tmp_edges[0] = *edges0;
  tmp_edges[1] = *edges1;
  tmp_edges[2] = *edges2;
  status = nc_put_vara_short(*ncid, *varid, tmp_start, tmp_edges, value);
  nc_check_status(ncid, status);
  return status;
}
