program Bufr2Asc
!
! Program to retrieve latitude, longitude, date/time, wind speed
! and wind direction from a BUFR file containing scatterometer information
!
! Last Modified on: $Date$
!
!
  use BufrIO, only: set_BUFR_fileattributes, get_BUFR_nr_of_messages, &
                    open_BUFR_file, get_BUFR_message, close_BUFR_file, &
                    bufr_file_attr_data, bufr_data_type, BufrVerbosity

  use numerics, only: missing_real

  use bufr_data_extraction_module, only: set_bufr_template, &
      goto_subset, get_lat, get_lon, get_spd, get_dir, get_mspd, get_mdir, &
      get_qcflag, get_year, get_month, get_day, get_hour, get_hour, get_minute

  use Compiler_Features, only: getarg_genscat, iargc_genscat

  implicit none

  type(bufr_file_attr_data) :: bufr_file_attributes
  type(bufr_data_type)      :: tbd
  integer            :: kmsg, ksub, nr_of_messages
  integer            :: narg, error_flag
  real               :: lat, lon, spd, dir
  logical            :: qcflag
  logical            :: print_rejected = .false.
  logical            :: print_model    = .false.
  character(len=256) :: filename_bufr, filename_asc, tmpstr
  character(len=20)  :: dtg

  integer,parameter  :: unit_asc = 29

  ! set verbosity of BUFR module
  BufrVerbosity = 0

  ! read command line
  narg = iargc_genscat()
  if (narg .lt. 2 .or. narg .gt. 4) then
    print '(A)',' '
    print '(A)','Usage: Bufr2Asc <BUFR input file> <ASCII output file> [-rejected] [-model]'
    print '(A)',' '
    stop
  endif

  call getarg_genscat(1,filename_bufr)
  call getarg_genscat(2,filename_asc)
  if (narg .ge. 3) then
    call getarg_genscat(3,tmpstr)
    if (trim(tmpstr) .eq. '-rejected') then
      print_rejected = .true.
    elseif (trim(tmpstr) .eq. '-model') then
      print_model    = .true.
    endif
  endif
  if (narg .eq. 4) then
    call getarg_genscat(4,tmpstr)
    if (trim(tmpstr) .eq. '-rejected') then
      print_rejected = .true.
    elseif (trim(tmpstr) .eq. '-model') then
      print_model    = .true.
    endif
  endif

  ! open BUFR input file and obtain number of messages present
  call set_BUFR_fileattributes(filename_bufr, bufr_file_attributes, error_flag)
  call open_BUFR_file(bufr_file_attributes, error_flag)
  nr_of_messages = get_BUFR_nr_of_messages(bufr_file_attributes)

  ! read first BUFR message and retrieve some information from it
  call get_BUFR_message(bufr_file_attributes, 1, tbd, error_flag)
  
  ! choose which BUFR template to use when interpreting the data
  ! (i.e. determine type of message depending on BUFR descriptor)
  ! this assumes all bufr messages in the file use the same template.
  call set_BUFR_template(tbd)

  ! open ASCII output file and write heading
  open(unit_asc, file=filename_asc)
  write(unit_asc,'(A)') '#    Lat     Lon      Date  Time   Speed     Dir'

  ! loop over messages in BUFR file
  msgloop:do kmsg= 1, nr_of_messages
     call get_BUFR_message(bufr_file_attributes, kmsg, tbd, error_flag)
     
     ! choose which BUFR template to use when interpreting the data
     ! (i.e. determine type of message depending on BUFR descriptor)
     ! this assumes all bufr messages in the file use the same template.
     call set_BUFR_template(tbd)

     ! loop over Wind Vector Cells in BUFR message
     do ksub = 1, tbd%nsubsets
        ! pointer to the position just before the start of this cell
        call goto_subset(tbd, ksub)

        lat = get_lat(tbd)
        lon = get_lon(tbd)
        if (print_model) then
          spd = get_mspd(tbd)
          dir = get_mdir(tbd)
        else
          spd = get_spd(tbd)
          dir = get_dir(tbd)
        endif
        qcflag = get_qcflag(tbd)
        
        if ( .not. ( missing_real(lat) .or. missing_real(lon) .or. &
                     missing_real(spd) .or. missing_real(dir)      )  )then
           if ((print_rejected .and. qcflag) .or. &              ! print rejected
               (.not. print_rejected .and. .not. qcflag)) then   ! print accepted
              write(dtg,101) &   ! write date/time information to variable
                   get_year(tbd), get_month(tbd), get_day(tbd), &
                   get_hour(tbd), get_minute(tbd)
              write(unit_asc,102) lat, lon, trim(dtg), spd, dir
           end if
        endif
     enddo
  enddo msgloop

  ! close input and output files
  call close_BUFR_file(bufr_file_attributes, error_flag)
  close(unit_asc)

  ! format statements
  101 format (I4.4,2I2.2,2X,2I2.2)
  102 format (2F8.2,2X,A,2F8.2)

end program Bufr2Asc
