program Bufr2Nc
  !
  ! Program to retrieve wind information from a BUFR file containing
  ! scatterometer information and to write this information to a NetCDF file.
  !
  ! Last Modified on: $Date$
  !
  !
  use BufrIO, only: set_BUFR_fileattributes, get_BUFR_nr_of_messages, &
                    open_BUFR_file, get_BUFR_message, close_BUFR_file, &
                    bufr_file_attr_data, bufr_data_type

  use bufr_data_extraction_module, only: set_bufr_template, goto_subset, &
       get_sat, get_ins, get_sw_id_l1, get_sw_id_wind, get_lat, get_lon, &
       get_res, get_trk, get_julian_daynumber, get_time_seconds_in_day, &
       get_year, get_month, get_day, get_hour, get_minute, get_second, &
       get_orb, get_cln, get_mdir, get_mspd, get_iceprb, get_iceage, get_qcflags, &
       get_dir, get_spd, get_bsdist, report_angle_conv_oceanographic

  use NetCDFMod, only: nc_global, undef_int, &
                       nc_create, nc_enddef, nc_close, nc_def_dim, &
                       nc_def_var_int, nc_def_var_short, &
                       nc_put_att_int, nc_put_att_short, nc_put_att_text, &
                       nc_put_att_int_arr, &
                       nc_put_array_int_2d, nc_put_array_short_2d, &
                       nc_real_to_int, nc_real_to_short, &
                       nc_create_var_int, nc_create_var_short

  use Compiler_Features, only: getarg_genscat, iargc_genscat

  use DateTimeMod, only: date_type, time_type, increment_datetime, ymd2julian

  use numerics, only: r8_, i4_, i2_, &
                missing_int, missing_real, &
                missing_indicator_real_r4=>missing_indicator_real

  implicit none

  ! in this program, we convert all data from the BUFR module to default
  ! size reals (4 bytes, generally)
  ! to prevent rounding errors when checking for missing values, the
  ! value of missing_indicator_real is also rounded
  real, parameter :: missing_indicator_real = real(missing_indicator_real_r4)

  type(bufr_file_attr_data) :: bufr_file_attributes
  type(bufr_data_type)      :: tbd
  type(date_type)           :: date_tmp
  type(time_type)           :: time_tmp
  integer            :: unit_nc
  integer            :: kmsg, ksub, nr_of_messages
  integer            :: narg, error_flag
  integer            :: i, str_pos, length
  integer            :: julian1990, julian, orbit_nr, swath_width
  integer            :: krow, kcell, prev_kcell, kcc, krc
  integer            :: sw_id_l1, ins_cal_vers, sw_id_wind
  integer            :: year, month, day, hour, minute, second
  integer            :: lut_index, time_i
  real               :: latitude, longitude
  real               :: latl, latr, lonl, lonr
  real               :: lat_sub, lon_sub, lon_tmp
  real               :: diff, delta_lon, delta_time, lsvl_360
  real               :: lat_min, lat_max, lon_min, lon_max
  character(len=256) :: filename_bufr, filename_nc
  character(len=256) :: basename_bufr, basename_nc
  character(len=256) :: references
  character(len=512) :: tmp_str, tmp_str2

  ! attributes to be written to NetCDF file
  character(len=256) :: title           = ' '
  character(len=256) :: title_short     = ' '
  character(len=25)  :: data_source     = ' '
  character(len=25)  :: pixel_size      = ' '
  character(len=25)  :: service_type    = ' '
  character(len=25)  :: start_date      = ' '
  character(len=25)  :: start_time      = ' '
  character(len=25)  :: stop_date       = ' '
  character(len=25)  :: stop_time       = ' '
  character(len=25)  :: creation_date   = ' '
  character(len=25)  :: creation_time   = ' '
  character(len=25)  :: eq_cross_lon    = ' '
  character(len=25)  :: eq_cross_date   = ' '
  character(len=25)  :: eq_cross_time   = ' '
  character(len=25)  :: rev_orb_period  = ' '
  character(len=25)  :: orb_inclination = ' '

  ! parameters to be written to XML file
  character(len=10)  :: ASTI = ' '          ! Satellite ID
  character(len=10)  :: APNM = ' '          ! Product type
  integer            :: AVBA = -1           ! Base algorithm version
  integer            :: AVPA = -1           ! Product algorithm version
  character(len=10)  :: AIID = ' '          ! Instrument ID
  integer            :: LONS = -1           ! Start orbit number
  character(len=25)  :: LSVT = ' '          ! Ascending node crossing date/time
  real               :: LSVL = -999.9999    ! Ascending node longitude
  character(len=25)  :: SSBT = ' '          ! Sensing start date/time
  character(len=25)  :: SSST = ' '          ! Sensing end date/time

  ! lookup tables for Metop orbits: provided that the satellite has its ascending pass
  ! over the Equator at 0 degrees longitude and at time 0 seconds, what is its lon position
  ! and time when it is at latitude 0, 10, 20, 30, 40, 50, 60, 70, 80, 90, 80, ..., -10, 0
  ! note that since the satellite never reaches -90 or 90 degrees, these table values
  ! are dummies
  ! MetOp does one orbit in ~101 minutes (6081.7 seconds)
  real,parameter,dimension(0:36) :: rel_lon_vs_lat_metop = &
    (/ 0.00000,   -2.28119,   -4.62780,   -7.20285,  -10.20838,  -14.02735,  -19.54704,  -29.54643,  -64.46212,    0.00000, &
                -128.26430, -163.17370, -173.17347, -178.69314, -182.51204, -185.51759, -188.09261, -190.43919, -192.68553, &
                -194.93213, -197.27948, -199.85572, -202.86269, -206.68282, -212.20171, -222.18963, -257.02311,    0.00000, &
                -321.03384, -355.86867, -365.85656, -371.37553, -375.19578, -378.20285, -380.77924, -383.12673, -385.37350 /)
  real,parameter,dimension(0:36) :: rel_time_vs_lat_metop = &
    (/ 0.0,  169.0,  339.0,  510.0,  682.0,  856.0, 1031.0, 1215.0, 1437.0,    0.0, &
            1597.0, 1820.0, 2003.0, 2179.0, 2352.0, 2524.0, 2695.0, 2865.0, 3035.0, &
            3205.0, 3376.0, 3547.0, 3720.0, 3893.0, 4070.0, 4254.0, 4478.0,    0.0, &
            4637.0, 4862.0, 5045.0, 5222.0, 5396.0, 5568.0, 5739.0, 5910.0, 6080.0 /)

  ! lookup tables for QuikSCAT orbits
  ! QuikSCAT does one orbit in ~101 minutes (6059.1 seconds)
  real,parameter,dimension(0:36) :: rel_lon_vs_lat_quik = &
    (/ 0.00,   -2.21,   -4.56,   -7.10,  -10.10,  -13.88,  -19.37,  -29.48,  -64.65,    0.00, &
             -127.90, -163.05, -173.14, -178.68, -182.49, -185.49, -188.03, -190.37, -192.60, &
             -194.81, -197.15, -199.72, -202.69, -206.49, -212.01, -222.11, -257.35,    0.00, &
             -320.56, -355.68, -365.81, -371.34, -375.15, -378.13, -380.69, -383.02, -385.23 /)
  real,parameter,dimension(0:36) :: rel_time_vs_lat_quik = &
    (/ 0.0,  169.0,  339.0,  509.0,  680.0,  852.0, 1027.0, 1208.0, 1427.0,    0.0, &
            1603.0, 1822.0, 2003.0, 2178.0, 2350.0, 2521.0, 2692.0, 2860.0, 3030.0, &
            3199.0, 3369.0, 3538.0, 3709.0, 3881.0, 4056.0, 4237.0, 4457.0,    0.0, &
            4632.0, 4852.0, 5033.0, 5208.0, 5380.0, 5550.0, 5721.0, 5890.0, 6059.0 /)

  ! lookup tables for Oceansat-2 orbits
  ! Oceansat-2 does one orbit in ~99 minutes (5958.6 seconds)
  real,parameter,dimension(0:36) :: rel_lon_vs_lat_oce2 = &
    (/ 0.00,   -2.15,   -4.41,   -6.86,   -9.73,  -13.44,  -18.73,  -28.38,  -60.73,    0.00, &
             -131.36, -163.98, -173.68, -178.96, -182.66, -185.53, -187.98, -190.21, -192.37, &
             -194.52, -196.78, -199.28, -202.18, -205.78, -211.18, -220.78, -253.08,    0.00, &
             -323.58, -356.38, -366.08, -371.43, -375.03, -377.93, -380.43, -382.70, -384.85 /)
  real,parameter,dimension(0:36) :: rel_time_vs_lat_oce2 = &
    (/ 0.0,  166.0,  333.0,  500.0,  667.0,  863.0, 1007.0, 1184.0, 1392.0,    0.0, &
            1573.0, 1782.0, 1960.0, 2132.0, 2308.0, 2476.0, 2642.0, 2807.0, 2973.0, &
            3141.0, 3307.0, 3474.0, 3642.0, 3811.0, 3984.0, 4162.0, 4371.0,    0.0, &
            4562.0, 4770.0, 4950.0, 5122.0, 5292.0, 5458.0, 5626.0, 5793.0, 5958.6 /)

  ! lookup tables for ISS orbits
  ! ISS does one orbit in ~93 minutes (5567.1 seconds)
  real,parameter,dimension(0:36) :: rel_lon_vs_lat_iss = &
    (/ 0.00,    7.22,   14.99,   24.53,   38.00,   65.71,    0.00,    0.00,    0.00,    0.00, &
                0.00,    0.00,    0.00,  102.28,  129.72,  143.19,  152.78,  160.56,  167.72, &
              174.84,  182.56,  192.04,  205.21,  231.23,    0.00,    0.00,    0.00,    0.00, &
                0.00,    0.00,    0.00,  272.93,  298.95,  312.11,  321.48,  329.30,  336.42 /)
  real,parameter,dimension(0:36) :: rel_time_vs_lat_iss = &
    (/ 0.0,  199.0,  398.0,  610.0,  850.0, 1203.5,    0.0,    0.0,    0.0,    0.0, &
               0.0,    0.0,    0.0, 1581.0, 1930.0, 2170.0, 2381.5, 2583.5, 2781.0, &
            2976.0, 3175.0, 3387.5, 3624.0, 3960.0,    0.0,    0.0,    0.0,    0.0, &
               0.0,    0.0,    0.0, 4388.5, 4724.5, 4961.0, 5169.5, 5372.5, 5567.1 /)

  integer,dimension(0:1) :: dimids,startids,lengths
  integer,parameter      :: max_nr_rows  = 8000
!!integer,parameter      :: max_nr_cells = 82
  integer,parameter      :: max_nr_cells = 198

  ! pointers to NetCDF variables
  integer :: varid_time, varid_wvc_quality, &
             varid_wvc_index, &
             varid_lat, varid_lon, &
             varid_model_speed, varid_model_dir, varid_ice_prob, varid_ice_age, &
             varid_wind_speed, varid_wind_dir, varid_bs_distance

  ! scale factors for various variables
  real(r8_),parameter :: sf_lat_lon = 1.E-5_r8_, sf_wind_speed = 0.01_r8_, sf_wind_dir = 0.1_r8_, &
                         sf_ice_prob = 0.001_r8_, sf_ice_age = 0.01_r8_, &
                         sf_bs_distance = 0.1_r8_

  ! arrays holding scatterometer data
  integer(i4_),dimension(0:max_nr_rows-1,0:max_nr_cells-1) :: time, lat, lon, wvc_quality
  integer(i2_),dimension(0:max_nr_rows-1,0:max_nr_cells-1) :: wvc_index, model_speed, model_dir, &
                                                              wind_speed, wind_dir, &
                                                              ice_prob, ice_age, bs_distance

  integer(i4_),dimension(0:16) :: flag_masks = (/ 64, 128, 256, 512, 1024, 2048, 4096, 8192, &
                                                  16384, 32768, 65536, 131072, 262144, 524288, &
                                                  1048576, 2097152, 4194304 /)

  ! read command line
  narg = iargc_genscat()
  if (narg .ne. 2) then
     print '(A)',' '
     print '(A)','Usage: Bufr2Nc <BUFR input file> <NetCDF output file>'
     print '(A)',' '
     stop
  endif

  call getarg_genscat(1,filename_bufr)
  call getarg_genscat(2,filename_nc)

  ! set region of interest (lon ranges from 0 to 360)
  lat_min = -90.0
  lat_max =  90.0
  lon_min =   0.0
  lon_max = 360.0

  ! determine file names without path
  str_pos = -1
  length    = len_trim(filename_bufr)
  do i = 1, length
     if (filename_bufr(i:i) .eq. '/') str_pos = i
  enddo
  if (str_pos .eq. -1) then
     basename_bufr = filename_bufr
  else
     basename_bufr = filename_bufr(str_pos+1:length)
  endif

  ! open BUFR input file and obtain number of messages present
  call set_BUFR_fileattributes(filename_bufr, bufr_file_attributes, error_flag)
  call open_BUFR_file(bufr_file_attributes, error_flag)
  nr_of_messages = get_BUFR_nr_of_messages(bufr_file_attributes)

  ! read first BUFR message and retrieve some information from it
  call get_BUFR_message(bufr_file_attributes, 1, tbd, error_flag)

  ! choose which BUFR template to use when interpreting the data
  ! (i.e. determine type of message depending on BUFR descriptor)
  call set_BUFR_template(tbd)

  ! ensure wind direction angles are reported using
  ! oceanographic angle convention
  call report_angle_conv_oceanographic()

  ! take 1 Jan 1990 as time reference
  call ymd2julian(1990,1,1,julian1990)

  ! initialise
  krow          = 0
  swath_width   = 0
  prev_kcell    = max_nr_cells
  references    = ' '
  orbit_nr      = undef_int
  sw_id_l1      = undef_int
  ins_cal_vers  = 0         ! instrument calibration version, not yet in BUFR
  sw_id_wind    = undef_int

  do kcc = 0, max_nr_cells-1
     do krc = 0, max_nr_rows-1
        time(krc,kcc)        = nc_real_to_int(missing_indicator_real)
        lat(krc,kcc)         = nc_real_to_int(missing_indicator_real)
        lon(krc,kcc)         = nc_real_to_int(missing_indicator_real)
        wvc_quality(krc,kcc) = nc_real_to_int(missing_indicator_real)
        wvc_index(krc,kcc)   = nc_real_to_short(missing_indicator_real)
        model_speed(krc,kcc) = nc_real_to_short(missing_indicator_real)
        model_dir(krc,kcc)   = nc_real_to_short(missing_indicator_real)
        wind_speed(krc,kcc)  = nc_real_to_short(missing_indicator_real)
        wind_dir(krc,kcc)    = nc_real_to_short(missing_indicator_real)
        ice_prob(krc,kcc)    = nc_real_to_short(missing_indicator_real)
        ice_age(krc,kcc)     = nc_real_to_short(missing_indicator_real)
        bs_distance(krc,kcc) = nc_real_to_short(missing_indicator_real)
     enddo
  enddo

  ! loop over messages in BUFR file
  msgloop:do kmsg= 1, nr_of_messages
     call get_BUFR_message(bufr_file_attributes, kmsg, tbd, error_flag)

     ! loop over Wind Vector Cells in BUFR message
     do ksub = 1, tbd%nsubsets

        ! pointer to the position just before the start of this cell
        call goto_subset(tbd, ksub)

        ! skip cells with incomplete information
        if (missing_int(get_sw_id_l1(tbd)))   cycle
        if (missing_int(get_sat(tbd)))        cycle
        if (missing_int(get_ins(tbd)))        cycle
        if (missing_real(get_trk(tbd)))       cycle
        if (missing_int(get_julian_daynumber(tbd)))    cycle
        if (missing_int(get_time_seconds_in_day(tbd))) cycle
        if (missing_real(get_lat(tbd)))       cycle
        if (missing_real(get_lon(tbd)))       cycle
        if (missing_int(get_res(tbd)))        cycle
        if (missing_int(get_orb(tbd)))        cycle
        if (missing_int(get_sw_id_wind(tbd))) cycle

        ! find swath width
        kcell = get_cln(tbd)
        if (kcell .gt. swath_width) then
           swath_width = kcell
           if (swath_width .gt. max_nr_cells) then
              print *,'Too many cells in one row in BUFR file, enlarge max_nr_cells'
              print *,'swath_width  = ', swath_width
              print *,'max_nr_cells = ', max_nr_cells
              stop
           endif
        endif

        ! check if location is within the region of interest
        latitude  = get_lat(tbd)
        longitude = get_lon(tbd)
        if (longitude .lt.   0.0) longitude = longitude + 360.0
        if (longitude .ge. 360.0) longitude = longitude - 360.0
        if (latitude  .lt. lat_min) cycle
        if (latitude  .gt. lat_max) cycle
        if (lon_max .gt. lon_min) then
           if (longitude .lt. lon_min) cycle
           if (longitude .gt. lon_max) cycle
        else
           lon_tmp = longitude
           if (lon_tmp .gt. 180.0) lon_tmp = lon_tmp - 360.0
           if (lon_tmp .lt. (lon_min - 360.0)) cycle
           if (lon_tmp .gt. lon_max) cycle
        endif

        ! find cell and row numbers
        if (kcell .lt. prev_kcell) then
           krow = krow + 1
           if (krow .gt. max_nr_rows) then
              print *,'Too many data rows in BUFR file, enlarge max_nr_rows'
              print *,'krow        = ', krow
              print *,'max_nr_rows = ', max_nr_rows
              stop
           endif
        endif
        prev_kcell = kcell
        kcc = kcell - 1
        krc = krow - 1

        ! get time information
        year   = get_year(tbd)
        month  = get_month(tbd)
        day    = get_day(tbd)
        hour   = get_hour(tbd)
        minute = get_minute(tbd)
        second = get_second(tbd)

        ! fill date/time and position information
        julian = get_julian_daynumber(tbd)

        ! compute time_i as number of seconds into the day,
        ! 0 = 0:00:00 and 86399 = 23:59:59
        time_i = get_time_seconds_in_day(tbd)

        time(krc,kcc)      = julian - julian1990          ! days since 1 Jan 1990
        time(krc,kcc)      = time(krc,kcc) * 24 * 3600    ! convert days to seconds
        time(krc,kcc)      = time(krc,kcc) + time_i       ! add time of day in seconds

        lat(krc,kcc)       = nc_real_to_int(latitude,sf_lat_lon)
        lon(krc,kcc)       = nc_real_to_int(longitude,sf_lat_lon)
        wvc_index(krc,kcc) = nc_real_to_short(real(get_cln(tbd)))

        ! sensing start date/time
        if (SSBT .eq. ' ') then
           write(SSBT,101) year, month, day, hour, minute, second
           write(start_date,102) year, month, day
           write(start_time,103) hour, minute, second
        endif

        ! sensing end date/time
        write(SSST,101) year, month, day, hour, minute, second
        write(stop_date,102) year, month, day
        write(stop_time,103) hour, minute, second

        ! satellite ID
        if (ASTI .eq. ' ') then
           select case(get_sat(tbd))
           case(1)
              ASTI = 'ERS'
              data_source = 'ERS-1 SCAT'
           case(2)
              ASTI = 'ERS'
              data_source = 'ERS-2 SCAT'
           case(3)
              ASTI = 'M01'
              data_source = 'MetOp-B ASCAT'
           case(4)
              ASTI = 'M02'
              data_source = 'MetOp-A ASCAT'
           case(5)
              ASTI = 'M03'
              data_source = 'MetOp-C ASCAT'
           case(281)
              ASTI = 'QUIKSCAT'
              data_source = 'QuikSCAT SeaWinds'
           case(421)
              ASTI = 'OCEANSAT2'
              data_source = 'Oceansat-2 OSCAT'
           case(422)
              ASTI = 'SCATSAT1'
              data_source = 'ScatSat-1 OSCAT'
           case(502, 990)
              ASTI = 'HY-2A'
              data_source = 'HY-2A HSCAT'
           case(503)
              ASTI = 'HY-2B'
              data_source = 'HY-2B HSCAT'
           case(504)
              ASTI = 'HY-2C'
              data_source = 'HY-2C HSCAT'
           case(555, 802)
              ASTI = 'CFOSAT'
              data_source = 'CFOSAT SCAT'
           case(801, 991)
              ASTI = 'ISS'
              data_source = 'ISS RapidScat'
           case default
              print '(A,I7)','Unknown satellite identifier: ', get_sat(tbd)
              stop
           end select
        endif

        ! instrument ID and product type
        if (AIID .eq. ' ' .or. APNM .eq. ' ') then
           select case (get_ins(tbd))
           case (8)
              AIID = 'SWINDS'
              if (get_res(tbd) .eq. 100000) then
                 APNM = 'OSWW100'
                 title_short = 'SeaWinds-L2-100km'
              elseif (get_res(tbd) .eq. 50000) then
                 APNM = 'OSWW050'
                 title_short = 'SeaWinds-L2-50km'
              elseif (get_res(tbd) .eq. 25000) then
                 APNM = 'OSWW025'
                 title_short = 'SeaWinds-L2-25km'
              elseif (get_res(tbd) .eq. 12500) then
                 APNM = 'OSWW012'
                 title_short = 'SeaWinds-L2-12.5km'
              else
                 print '(A,I7)','Invalid SeaWinds WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = 'SeaWinds Product Manual, ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case (142)
              AIID = 'ERS'
              APNM = 'OERW025'
              title_short = 'SCATT-L2-25km'
              references = 'ERS Scatterometer Product User Manual, ' // &
                   'http://www.knmi.nl/scatterometer/'
           case (190)
              AIID = 'ASCA'
              if (get_res(tbd) .eq. 25000) then
                 APNM = 'OASW025'
                 if (get_sat(tbd) .eq. 3) then
                    title_short = 'ASCATB-L2-25km'
                 elseif (get_sat(tbd) .eq. 4) then
                    title_short = 'ASCATA-L2-25km'
                 else
                    title_short = 'ASCATC-L2-25km'
                 endif
              elseif (get_res(tbd) .eq. 12500) then
                 if (index(filename_bufr,"COA") .gt. 0 .or. index(filename_bufr,"coa") .gt. 0 .or. &
                     index(filename_nc,"COA")   .gt. 0 .or. index(filename_nc,"coa")   .gt. 0) then
                    if (str_pos .gt. 0) then
                       APNM = 'OASWC12'
                    else
                       APNM = 'OASC12'
                    endif
                    if (get_sat(tbd) .eq. 3) then
                       title_short = 'ASCATB-L2-Coastal'
                    elseif (get_sat(tbd) .eq. 4) then
                       title_short = 'ASCATA-L2-Coastal'
                    else
                       title_short = 'ASCATC-L2-Coastal'
                    endif
                 else
                    if (str_pos .gt. 0) then
                       APNM = 'OASW012'
                    else
                       APNM = 'OAS012'
                    endif
                    if (get_sat(tbd) .eq. 3) then
                       title_short = 'ASCATB-L2-12.5km'
                    elseif (get_sat(tbd) .eq. 4) then
                       title_short = 'ASCATA-L2-12.5km'
                    else
                       title_short = 'ASCATC-L2-12.5km'
                    endif
                 endif
              elseif (get_res(tbd) .eq. 6250) then
                 APNM = 'OASWC06'
                 if (get_sat(tbd) .eq. 3) then
                    title_short = 'ASCATB-L2-6.25km'
                 elseif (get_sat(tbd) .eq. 4) then
                    title_short = 'ASCATA-L2-6.25km'
                 else
                    title_short = 'ASCATC-L2-6.25km'
                 endif
              elseif (get_res(tbd) .eq. 5650) then
                 APNM = 'OASWC05'
                 if (get_sat(tbd) .eq. 3) then
                    title_short = 'ASCATB-L2-5.7km'
                 elseif (get_sat(tbd) .eq. 4) then
                    title_short = 'ASCATA-L2-5.7km'
                 else
                    title_short = 'ASCATC-L2-5.7km'
                 endif
              else
                 print '(A,I7)','Invalid ASCAT WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = 'ASCAT Wind Product User Manual, ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case (288)
              AIID = 'OSCAT'
              if (get_res(tbd) .eq. 50000) then
                 APNM = 'OOSW050'
                 title_short = 'OSCAT-L2-50km'
              elseif (get_res(tbd) .eq. 25000) then
                 APNM = 'OOSW025'
                 title_short = 'OSCAT-L2-25km'
              else
                 print '(A,I7)','Invalid OSCAT WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = 'Oceansat-2 Wind Product User Manual, ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case(686)
              AIID = 'HSCAT'
              if (get_res(tbd) .eq. 50000) then
                 APNM = 'OHSW050'
                 title_short = 'HSCAT-L2-50km'
              elseif (get_res(tbd) .eq. 25000) then
                 APNM = 'OHSW025'
                 title_short = 'HSCAT-L2-25km'
              else
                 print '(A,I7)','Invalid HY-2A WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = ' ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case(943)
              AIID = 'SCAT'
              if (get_res(tbd) .eq. 50000) then
                 APNM = 'OCSW050'
                 title_short = 'CFOSAT-L2-50km'
              elseif (get_res(tbd) .eq. 25000) then
                 APNM = 'OCSW025'
                 title_short = 'CFOSAT-L2-25km'
              else
                 print '(A,I7)','Invalid CFOSAT WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = ' ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case(314, 991)
              AIID = 'RSCAT'
              if (get_res(tbd) .eq. 50000) then
                 APNM = 'ORSW050'
                 title_short = 'RSCAT-L2-50km'
              elseif (get_res(tbd) .eq. 25000) then
                 APNM = 'ORSW025'
                 title_short = 'RSCAT-L2-25km'
              else
                 print '(A,I7)','Invalid RapidScat WVC spacing: ', get_res(tbd)
                 stop
              endif
              references = 'RapidScat Wind Product User Manual, ' // &
                   'http://www.osi-saf.org/, http://www.knmi.nl/scatterometer/'
           case default
              print '(A,I7)','Unknown satellite instruments ID: ', get_ins(tbd)
              stop
           end select
        endif

        ! base algorithm version (level 1 software identification)
        if (AVBA .eq. -1) then
           sw_id_l1 = get_sw_id_l1(tbd)
           AVBA     = sw_id_l1
        endif

        ! product algorithm version (level 2 software identification)
        if (AVPA .eq. -1) then
           sw_id_wind = get_sw_id_wind(tbd)
           AVPA       = sw_id_wind

           ! if NetCDF filename contains "_swid" then replace this with the
           ! true software identification
           str_pos = index(filename_nc,"_swid_")
           if (str_pos .gt. 0) then
              write(tmp_str,'(I4.4)') mod(sw_id_wind, 10000)
              filename_nc(str_pos+1:str_pos+4) = tmp_str(1:4)
           endif

           ! determine NetCDF file name without path
           str_pos = -1
           length    = len_trim(filename_nc)
           do i = 1, length
              if (filename_nc(i:i) .eq. '/') str_pos = i
           enddo
           if (str_pos .eq. -1) then
              basename_nc = filename_nc
           else
              basename_nc = filename_nc(str_pos+1:length)
           endif
        endif

        ! pixel size and title
        if (pixel_size .eq. ' ') then
           write(pixel_size,'(F9.1)') real(get_res(tbd)) / 1000.0 ! convert [m] to [km]
           pixel_size = trim(adjustl(pixel_size)) // ' km'
           if (APNM .eq. 'OASWC12' .or. APNM .eq. 'OASC12') then
              title = trim(data_source) // ' Level 2 Coastal' // &
                   ' Ocean Surface Wind Vector Product'
           else
              title = trim(data_source) // ' Level 2 ' // trim(pixel_size) // &
                   ' Ocean Surface Wind Vector Product'
           endif
        endif

        ! start orbit number
        if (LONS .eq. -1) then
           orbit_nr = get_orb(tbd)
           LONS     = orbit_nr
        endif

        ! ascending node longitude and ascending node crossing date/time
        if ( (APNM .eq. 'OASW025' .or. APNM .eq. 'OAS025') .and. mod(ksub,  42) .eq. 21 .or. &
             (APNM .eq. 'OASW012' .or. APNM .eq. 'OAS012') .and. mod(ksub,  82) .eq. 41 .or. &
             (APNM .eq. 'OASWC12' .or. APNM .eq. 'OASC12') .and. mod(ksub,  82) .eq. 41 .or. &
             (APNM .eq. 'OASWC06') .and. mod(ksub, 162) .eq. 81 .or. &
             (APNM .eq. 'OASWC05') .and. mod(ksub, 198) .eq. 99) then ! ASCAT

           if (LSVL .lt. -999.0) then ! not yet determined

              ! find lat and lon of sub-satellite point
              ! this is the mean position of the first nodes of the left and right swath
              call goto_subset(tbd, ksub)
              latl = get_lat(tbd) ! node 21/41
              lonl = get_lon(tbd)
              call goto_subset(tbd, ksub+1)
              latr = get_lat(tbd) ! node 22/42
              lonr = get_lon(tbd)
              call goto_subset(tbd, ksub)

              lat_sub = (latl + latr) / 2.0
              if (lonl - lonr .gt.  180.0) lonr = lonr + 360.0
              if (lonl - lonr .lt. -180.0) lonr = lonr - 360.0
              lon_sub = (lonl + lonr) / 2.0

              ! calculate the lon and time of passing the equator by taking the lat and date/time of
              ! the first point with lat below 70 degrees and not too far away from a lookup table entry
              if (abs(lat_sub) .lt. 70.0 .and. mod(abs(lat_sub)+0.5,10.0) .lt. 1.0) then

                 ! see how many degrees in latitude we have done since the last ascending pass
                 ! over the equator
                 if (get_trk(tbd) .lt.  90.0 .or. &
                      get_trk(tbd) .gt. 270.0) then
                    lut_index = int(mod(lat_sub+360.0,360.0)) / 10 ! ascending direction
                    diff = mod(lat_sub+360.0,360.0) - real(lut_index * 10)
                 else
                    lut_index = int(180.0-lat_sub) / 10            ! descending direction
                    diff = (180.0-lat_sub) - real(lut_index * 10)
                 endif
                 delta_lon = rel_lon_vs_lat_metop(lut_index) + &
                      (diff / 10.0) * (rel_lon_vs_lat_metop(lut_index+1) - rel_lon_vs_lat_metop(lut_index))

                 LSVL = lon_sub - delta_lon
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0          ! LSVL in range -180..180
                 lsvl_360 = LSVL
                 if (lsvl_360 .lt. 0.0) lsvl_360 = lsvl_360 + 360.0 ! eq_cross_lon in range 0..360
                 write(eq_cross_lon,'(F8.3)') lsvl_360

                 delta_time = rel_time_vs_lat_metop(lut_index) + &
                      (diff / 10.0) * (rel_time_vs_lat_metop(lut_index+1) - rel_time_vs_lat_metop(lut_index))
                 date_tmp%year   = year
                 date_tmp%month  = month
                 date_tmp%day    = day
                 time_tmp%hour   = hour
                 time_tmp%minute = minute
                 time_tmp%second = second
                 call increment_datetime(date_tmp, time_tmp, -1.0 * delta_time)
                 write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                      time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 write(eq_cross_date,102) date_tmp%year, date_tmp%month, date_tmp%day
                 write(eq_cross_time,103) time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 rev_orb_period  = '6081.7'
                 orb_inclination = '98.7'
              endif
           endif
        elseif (APNM .eq. 'OSWW100' .and. mod(ksub, 19) .eq. 10 .or. &
                APNM .eq. 'OSWW050' .and. mod(ksub, 38) .eq. 19 .or. &
                APNM .eq. 'OSWW025' .and. mod(ksub, 76) .eq. 38) then  ! SeaWinds

           if (LSVL .lt. -999.0) then ! not yet determined

              ! find lat and lon of sub-satellite point
              ! this is the position of node 10 in the 100-km product
              ! or the mean position of nodes 38 and 39 in the 25-km product
              if (APNM .eq. 'OSWW100') then
                 lat_sub = get_lat(tbd) ! node 10
                 lon_sub = get_lon(tbd)
              else
                 call goto_subset(tbd, ksub)
                 latl = get_lat(tbd) ! node 38
                 lonl = get_lon(tbd)
                 call goto_subset(tbd, ksub+1)
                 latr = get_lat(tbd) ! node 39
                 lonr = get_lon(tbd)
                 call goto_subset(tbd, ksub)

                 lat_sub = (latl + latr) / 2.0
                 if (lonl - lonr .gt.  180.0) lonr = lonr + 360.0
                 if (lonl - lonr .lt. -180.0) lonr = lonr - 360.0
                 lon_sub = (lonl + lonr) / 2.0
              endif

              ! calculate the lon and time of passing the equator by taking the lat and date/time of
              ! the first point with lat below 30 degrees and not too far away from a lookup table entry
              ! skip the first 4 rows, since the 100-km WVCs don't contain information of four 25-km rows
              if (abs(lat_sub) .lt. 30.0 .and. mod(abs(lat_sub)+0.5,10.0) .lt. 1.0 .and. kmsg .gt. 4) then

                 ! see how many degrees in latitude we have done since the last ascending pass
                 ! over the equator
                 if (get_trk(tbd) .lt.  90.0 .or. &
                      get_trk(tbd) .gt. 270.0) then
                    lut_index = int(mod(lat_sub+360.0,360.0)) / 10 ! ascending direction
                    diff = mod(lat_sub+360.0,360.0) - real(lut_index * 10)
                 else
                    lut_index = int(180.0-lat_sub) / 10            ! descending direction
                    diff = (180.0-lat_sub) - real(lut_index * 10)
                 endif
                 delta_lon = rel_lon_vs_lat_quik(lut_index) + &
                      (diff / 10.0) * (rel_lon_vs_lat_quik(lut_index+1) - rel_lon_vs_lat_quik(lut_index))

                 LSVL = lon_sub - delta_lon

                 delta_time = rel_time_vs_lat_quik(lut_index) + &
                      (diff / 10.0) * (rel_time_vs_lat_quik(lut_index+1) - rel_time_vs_lat_quik(lut_index))

                 ! for 100-km data, correct the time by a half 25-km WVC passing time (approx. 3.8 / 2 = 1.9 s)
                 ! for older 100-km data, the WVC time corresponds to the time of the last underlying 25-km row
                 ! the correction for this is another 4 seconds
                 if (APNM .eq. 'OSWW100' .and. AVPA .lt. 20) delta_time = delta_time + 5.9 ! older data
                 if (APNM .eq. 'OSWW100' .and. AVPA .ge. 20) delta_time = delta_time + 1.9 ! newer data

                 ! small time correction for 50-km data
                 if (APNM .eq. 'OSWW050') delta_time = delta_time + 1.9

                 date_tmp%year   = year
                 date_tmp%month  = month
                 date_tmp%day    = day
                 time_tmp%hour   = hour
                 time_tmp%minute = minute
                 time_tmp%second = second
                 call increment_datetime(date_tmp, time_tmp, -1.0 * delta_time)
                 write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                      time_tmp%hour, time_tmp%minute, int(time_tmp%second)

                 ! go back one orbit for LSVL and LSVT when LSVL is after the measurement start
                 if (LSVT .gt. SSBT) then
                    LSVL = LSVL - rel_lon_vs_lat_quik(36)
                    call increment_datetime(date_tmp, time_tmp, -1.0 * rel_time_vs_lat_quik(36))
                    write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                         time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 endif
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0          ! LSVL in range -180..180
                 lsvl_360 = LSVL
                 if (lsvl_360 .lt. 0.0) lsvl_360 = lsvl_360 + 360.0 ! eq_cross_lon in range 0..360
                 write(eq_cross_lon,'(F8.3)') lsvl_360
                 write(eq_cross_date,102) date_tmp%year, date_tmp%month, date_tmp%day
                 write(eq_cross_time,103) time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 rev_orb_period  = '6059.1'
                 orb_inclination = '98.6'
              endif
           endif
        elseif (APNM .eq. 'OOSW050' .and. mod(ksub, 36) .eq. 18 .or. &
                APNM .eq. 'OOSW025' .and. mod(ksub, 76) .eq. 38) then  ! OSCAT

           if (LSVL .lt. -999.0) then ! not yet determined

              ! find lat and lon of sub-satellite point
              ! this is the mean position of nodes 18 and 19 in the 50-km product
              ! or the mean position of nodes 38 and 39 in the 25-km product
              call goto_subset(tbd, ksub)
              latl = get_lat(tbd) ! node 18/38
              lonl = get_lon(tbd)
              call goto_subset(tbd, ksub+1)
              latr = get_lat(tbd) ! node 19/39
              lonr = get_lon(tbd)
              call goto_subset(tbd, ksub)

              lat_sub = (latl + latr) / 2.0
              if (lonl - lonr .gt.  180.0) lonr = lonr + 360.0
              if (lonl - lonr .lt. -180.0) lonr = lonr - 360.0
              lon_sub = (lonl + lonr) / 2.0

              ! calculate the lon and time of passing the equator by taking the lat and date/time of
              ! the first point with lat below 30 degrees and not too far away from a lookup table entry
              if (abs(lat_sub) .lt. 30.0 .and. mod(abs(lat_sub)+0.5,10.0) .lt. 1.0) then

                 ! see how many degrees in latitude we have done since the last ascending pass
                 ! over the equator
                 if (get_trk(tbd) .lt.  90.0 .or. &
                      get_trk(tbd) .gt. 270.0) then
                    lut_index = int(mod(lat_sub+360.0,360.0)) / 10 ! ascending direction
                    diff = mod(lat_sub+360.0,360.0) - real(lut_index * 10)
                 else
                    lut_index = int(180.0-lat_sub) / 10            ! descending direction
                    diff = (180.0-lat_sub) - real(lut_index * 10)
                 endif
                 delta_lon = rel_lon_vs_lat_oce2(lut_index) + &
                      (diff / 10.0) * (rel_lon_vs_lat_oce2(lut_index+1) - rel_lon_vs_lat_oce2(lut_index))

                 LSVL = lon_sub - delta_lon

                 delta_time = rel_time_vs_lat_oce2(lut_index) + &
                      (diff / 10.0) * (rel_time_vs_lat_oce2(lut_index+1) - rel_time_vs_lat_oce2(lut_index))

                 ! the location and timing of 25 km WVCs is slightly different, apply
                 ! a small correction to obtain equal LSVL and LSVT for 25 km and 50 km
                 if (APNM .eq. 'OOSW025') then
                    LSVL = LSVL + 0.265
                    delta_time = delta_time - 7.4
                 endif

                 date_tmp%year   = year
                 date_tmp%month  = month
                 date_tmp%day    = day
                 time_tmp%hour   = hour
                 time_tmp%minute = minute
                 time_tmp%second = second
                 call increment_datetime(date_tmp, time_tmp, -1.0 * delta_time)
                 write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                      time_tmp%hour, time_tmp%minute, int(time_tmp%second)

                 ! go back one orbit for LSVL and LSVT when LSVL is after the measurement start
                 if (LSVT .gt. SSBT) then
                    LSVL = LSVL - rel_lon_vs_lat_oce2(36)
                    call increment_datetime(date_tmp, time_tmp, -1.0 * rel_time_vs_lat_oce2(36))
                    write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                         time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 endif
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0          ! LSVL in range -180..180
                 lsvl_360 = LSVL
                 if (lsvl_360 .lt. 0.0) lsvl_360 = lsvl_360 + 360.0 ! eq_cross_lon in range 0..360
                 write(eq_cross_lon,'(F8.3)') lsvl_360
                 write(eq_cross_date,102) date_tmp%year, date_tmp%month, date_tmp%day
                 write(eq_cross_time,103) time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 rev_orb_period  = '5958.6'
                 orb_inclination = '98.3'
              endif
           endif

        elseif (APNM .eq. 'ORSW050' .and. mod(ksub, 21) .eq. 11 .or. &
                APNM .eq. 'ORSW025' .and. mod(ksub, 42) .eq. 21      ) then  ! RapidScat

           if (LSVL .lt. -999.0) then ! not yet determined

              ! find lat and lon of sub-satellite point
              ! this is the position of node 11 in the 50-km product
              ! or the mean position of nodes 21 and 22 in the 25-km product
              if (APNM .eq. 'ORSW050') then
                 lat_sub = get_lat(tbd) ! node 11
                 lon_sub = get_lon(tbd)
              else
                 call goto_subset(tbd, ksub)
                 latl = get_lat(tbd) ! node 21
                 lonl = get_lon(tbd)
                 call goto_subset(tbd, ksub+1)
                 latr = get_lat(tbd) ! node 22
                 lonr = get_lon(tbd)
                 call goto_subset(tbd, ksub)

                 lat_sub = (latl + latr) / 2.0
                 if (lonl - lonr .gt.  180.0) lonr = lonr + 360.0
                 if (lonl - lonr .lt. -180.0) lonr = lonr - 360.0
                 lon_sub = (lonl + lonr) / 2.0
              endif

              ! calculate the lon and time of passing the equator by taking the
              ! lat and date/time of
              ! the first point with lat below 20 degrees and not too far away
              ! from a lookup table entry
              if (abs(lat_sub) .lt. 20.0 .and. mod(abs(lat_sub)+0.5,10.0) .lt. 1.0) then

                 ! see how many degrees in latitude we have done since the last ascending pass
                 ! over the equator
                 if (get_trk(tbd) .lt.  90.0 .or. &
                     get_trk(tbd) .gt. 270.0) then
                    lut_index = int(mod(lat_sub+360.0,360.0)) / 10 ! ascending direction
                    diff = mod(lat_sub+360.0,360.0) - real(lut_index * 10)
                 else
                    lut_index = int(180.0-lat_sub) / 10            ! descending direction
                    diff = (180.0-lat_sub) - real(lut_index * 10)
                 endif
                 delta_lon = rel_lon_vs_lat_iss(lut_index) + &
                      (diff / 10.0) * (rel_lon_vs_lat_iss(lut_index+1) - rel_lon_vs_lat_iss(lut_index))

                 LSVL = lon_sub - delta_lon

                 delta_time = rel_time_vs_lat_iss(lut_index) + &
                      (diff / 10.0) * (rel_time_vs_lat_iss(lut_index+1) - rel_time_vs_lat_iss(lut_index))

                 ! the timing of 50 km WVCs is slightly different, apply
                 ! a small correction to obtain equal LSVT for 25 km and 50 km
                 if (APNM .eq. 'ORSW050') then
                    delta_time = delta_time + 1.0
                 endif

                 date_tmp%year   = year
                 date_tmp%month  = month
                 date_tmp%day    = day
                 time_tmp%hour   = hour
                 time_tmp%minute = minute
                 time_tmp%second = second
                 call increment_datetime(date_tmp, time_tmp, -1.0 * delta_time)
                 write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                      time_tmp%hour, time_tmp%minute, int(time_tmp%second)

                 ! go back one orbit for LSVL and LSVT when LSVT is after the
                 ! measurement start
                 if (LSVT .gt. SSBT) then
                    LSVL = LSVL - rel_lon_vs_lat_iss(36)
                    call increment_datetime(date_tmp, time_tmp, -1.0 * rel_time_vs_lat_iss(36))
                    write(LSVT,101) date_tmp%year, date_tmp%month, date_tmp%day, &
                         time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 endif
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .gt.  180.0) LSVL = LSVL - 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0
                 if (LSVL .lt. -180.0) LSVL = LSVL + 360.0          ! LSVL in range -180..180
                 lsvl_360 = LSVL
                 if (lsvl_360 .lt. 0.0) lsvl_360 = lsvl_360 + 360.0 ! eq_cross_lon in range 0..360
                 write(eq_cross_lon,'(F8.3)') lsvl_360
                 write(eq_cross_date,102) date_tmp%year, date_tmp%month, date_tmp%day
                 write(eq_cross_time,103) time_tmp%hour, time_tmp%minute, int(time_tmp%second)
                 rev_orb_period  = '5567.1'
                 orb_inclination = '51.6'
              endif
           endif

        endif

        ! fill wind cell information, apply scaling factors if appropriate
        ! ice_prob, ice_age and bs_distance are not always available
        ! in the original seawinds format (only implemented as hack by files
        ! reprocessed by SDP, in which case they are stored in some unused fields)
        model_dir(krc,kcc)   = nc_real_to_short(get_mdir(tbd),sf_wind_dir)
        model_speed(krc,kcc) = nc_real_to_short(get_mspd(tbd),sf_wind_speed)
        ice_prob(krc,kcc)    = nc_real_to_short(get_iceprb(tbd),sf_ice_prob)
        ice_age(krc,kcc)     = nc_real_to_short(get_iceage(tbd),sf_ice_age)
        wvc_quality(krc,kcc) = get_qcflags(tbd)

        ! take selected wind solution and convert direction to oceanographic convention
        wind_dir(krc,kcc)    = nc_real_to_short(get_dir(tbd),sf_wind_dir)
        wind_speed(krc,kcc)  = nc_real_to_short(get_spd(tbd),sf_wind_speed)
        bs_distance(krc,kcc) = nc_real_to_short(get_bsdist(tbd),sf_bs_distance)

     enddo
  enddo msgloop

  ! creation date NetCDF is current system time
  call date_and_time(tmp_str, tmp_str2)
  write(creation_date,105) tmp_str(1:4), tmp_str(5:6), tmp_str(7:8)
  write(creation_time,106) tmp_str2(1:2), tmp_str2(3:4), tmp_str2(5:6)

  ! service type
  if (index(filename_nc,"_eps_") .gt. 0) then
     service_type = "eps"
  elseif (index(filename_nc,"_ear_") .gt. 0) then
     service_type = "ear"
  else
     service_type = "N/A"
  endif

  ! close input file
  call close_BUFR_file(bufr_file_attributes, error_flag)

  ! create NetCDF file
  if (krow .le. 0) then
     print '(A)','No data containing valid winds found in this file, no NetCDF output created'
     stop
  endif
  error_flag = nc_create(trim(filename_nc), unit_nc)
  if (error_flag .ne. 0) stop

  ! define dimensions
  error_flag = nc_def_dim(unit_nc, 'NUMROWS' // achar(0), krow, dimids(0))
  if (error_flag .ne. 0) stop
  error_flag = nc_def_dim(unit_nc, 'NUMCELLS' // achar(0), swath_width, dimids(1))
  if (error_flag .ne. 0) stop
  startids(0) = 0
  startids(1) = 0
  lengths(0) = krow
  lengths(1) = swath_width

  ! define scatterometer fields
  ! time
  call nc_create_var_int(unit_nc, 'time', 2, dimids, ' ', 'time', &
                         'seconds since 1990-01-01 00:00:00', 1.0_r8_, 0.0_r8_, &
                         0.0_r8_, real(-undef_int,r8_), varid_time, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_time, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! latitude
  call nc_create_var_int(unit_nc, 'lat', 2, dimids, ' ', 'latitude', &
                         'degrees_north', sf_lat_lon, 0.0_r8_, -90.0_r8_, 90.0_r8_, &
                         varid_lat, error_flag)
  if (error_flag .ne. 0) stop

  ! longitude
  call nc_create_var_int(unit_nc, 'lon', 2, dimids, ' ', 'longitude', &
                         'degrees_east', sf_lat_lon, 0.0_r8_, 0.0_r8_, 360.0_r8_, &
                         varid_lon, error_flag)
  if (error_flag .ne. 0) stop

  ! cross track cell number
  call nc_create_var_short(unit_nc, 'wvc_index', 2, dimids, ' ', 'cross track wind vector cell number', &
                           '1', 1.0_r8_, 0.0_r8_, 0.0_r8_, 999.0_r8_, &
                           varid_wvc_index, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_wvc_index, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! model wind speed
  call nc_create_var_short(unit_nc, 'model_speed', 2, dimids, ' ', 'model wind speed at 10 m', &
                           'm s-1', sf_wind_speed, 0.0_r8_, 0.0_r8_, 50.0_r8_, &
                           varid_model_speed, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_model_speed, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! model wind direction
  call nc_create_var_short(unit_nc, 'model_dir', 2, dimids, ' ', 'model wind direction at 10 m', &
                           'degree', sf_wind_dir, 0.0_r8_, 0.0_r8_, 360.0_r8_, &
                           varid_model_dir, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_model_dir, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! ice probability
  call nc_create_var_short(unit_nc, 'ice_prob', 2, dimids, ' ', 'ice probability', &
                           '1', sf_ice_prob, 0.0_r8_, 0.0_r8_, 1.0_r8_, &
                           varid_ice_prob, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_ice_prob, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! ice age
  call nc_create_var_short(unit_nc, 'ice_age', 2, dimids, ' ', 'ice age (a-parameter)', &
                           'dB', sf_ice_age, 0.0_r8_, -50.0_r8_, 50.0_r8_, &
                           varid_ice_age, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_ice_age, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! wind vector cell quality
  call nc_create_var_int(unit_nc, 'wvc_quality_flag', 2, dimids, ' ', 'wind vector cell quality', &
                         ' ', 1.0_r8_, 0.0_r8_, 0.0_r8_, 8388607.0_r8_, &
                         varid_wvc_quality, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_wvc_quality, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! add attributes describing the properties of the quality flag
  error_flag = nc_put_att_int_arr(unit_nc, varid_wvc_quality, 'flag_masks' // achar(0), &
                                  17, flag_masks)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'distance_to_gmf_too_large '                 // &
                       'data_are_redundant '                        // &
                       'no_meteorological_background_used '         // &
                       'rain_detected '                             // &
                       'rain_flag_not_usable '                      // &
                       'small_wind_less_than_or_equal_to_3_m_s '    // &
                       'large_wind_greater_than_30_m_s '            // &
                       'wind_inversion_not_successful '             // &
                       'some_portion_of_wvc_is_over_ice '           // &
                       'some_portion_of_wvc_is_over_land '          // &
                       'variational_quality_control_fails '         // &
                       'knmi_quality_control_fails '                // &
                       'product_monitoring_event_flag '             // &
                       'product_monitoring_not_used '               // &
                       'any_beam_noise_content_above_threshold '    // &
                       'poor_azimuth_diversity '                    // &
                       'not_enough_good_sigma0_for_wind_retrieval'

  error_flag = nc_put_att_text(unit_nc, varid_wvc_quality, 'flag_meanings', tmp_str)
  if (error_flag .ne. 0) stop

  ! wind speed
  call nc_create_var_short(unit_nc, 'wind_speed', 2, dimids, ' ', 'wind speed at 10 m', &
                           'm s-1', sf_wind_speed, 0.0_r8_, 0.0_r8_, 50.0_r8_, &
                           varid_wind_speed, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_wind_speed, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! wind direction
  call nc_create_var_short(unit_nc, 'wind_dir', 2, dimids, ' ', 'wind direction at 10 m', &
                           'degree', sf_wind_dir, 0.0_r8_, 0.0_r8_, 360.0_r8_, &
                           varid_wind_dir, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_wind_dir, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! backscatter distance
  call nc_create_var_short(unit_nc, 'bs_distance', 2, dimids, ' ', 'backscatter distance', &
                           '1', sf_bs_distance, 0.0_r8_, -50.0_r8_, 50.0_r8_, &
                           varid_bs_distance, error_flag)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'lat lon'
  error_flag = nc_put_att_text(unit_nc, varid_bs_distance, 'coordinates', tmp_str)
  if (error_flag .ne. 0) stop

  ! define global attributes
  write(tmp_str,'(A)') title
  error_flag = nc_put_att_text(unit_nc, nc_global, 'title', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') title_short
  error_flag = nc_put_att_text(unit_nc, nc_global, 'title_short_name', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'CF-1.4'
  error_flag = nc_put_att_text(unit_nc, nc_global, 'Conventions', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'EUMETSAT/OSI SAF/KNMI'
  error_flag = nc_put_att_text(unit_nc, nc_global, 'institution', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') data_source
  error_flag = nc_put_att_text(unit_nc, nc_global, 'source', tmp_str)
  if (error_flag .ne. 0) stop

  error_flag = nc_put_att_int(unit_nc, nc_global, 'software_identification_level_1' // achar(0), &
                              int(sw_id_l1,i4_))
  if (error_flag .ne. 0) stop

  error_flag = nc_put_att_int(unit_nc, nc_global, 'instrument_calibration_version' // achar(0), &
                              int(ins_cal_vers,i4_))
  if (error_flag .ne. 0) stop

  error_flag = nc_put_att_int(unit_nc, nc_global, 'software_identification_wind' // achar(0), &
                              int(sw_id_wind,i4_))
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') pixel_size
  error_flag = nc_put_att_text(unit_nc, nc_global, 'pixel_size_on_horizontal', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') service_type
  error_flag = nc_put_att_text(unit_nc, nc_global, 'service_type', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'O'   ! Operational
  error_flag = nc_put_att_text(unit_nc, nc_global, 'processing_type', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'ovw' ! Ocean Vector Winds
  error_flag = nc_put_att_text(unit_nc, nc_global, 'contents', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') basename_nc
  error_flag = nc_put_att_text(unit_nc, nc_global, 'granule_name', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'L2'  ! Level 2
  error_flag = nc_put_att_text(unit_nc, nc_global, 'processing_level', tmp_str)
  if (error_flag .ne. 0) stop

  error_flag = nc_put_att_int(unit_nc, nc_global, 'orbit_number' // achar(0), &
                              int(orbit_nr,i4_))
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') start_date
  error_flag = nc_put_att_text(unit_nc, nc_global, 'start_date', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') start_time
  error_flag = nc_put_att_text(unit_nc, nc_global, 'start_time', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') stop_date
  error_flag = nc_put_att_text(unit_nc, nc_global, 'stop_date', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') stop_time
  error_flag = nc_put_att_text(unit_nc, nc_global, 'stop_time', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') eq_cross_lon
  error_flag = nc_put_att_text(unit_nc, nc_global, 'equator_crossing_longitude', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') eq_cross_date
  error_flag = nc_put_att_text(unit_nc, nc_global, 'equator_crossing_date', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') eq_cross_time
  error_flag = nc_put_att_text(unit_nc, nc_global, 'equator_crossing_time', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') rev_orb_period
  error_flag = nc_put_att_text(unit_nc, nc_global, 'rev_orbit_period', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') orb_inclination
  error_flag = nc_put_att_text(unit_nc, nc_global, 'orbit_inclination', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'N/A'
  error_flag = nc_put_att_text(unit_nc, nc_global, 'history', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') references
  error_flag = nc_put_att_text(unit_nc, nc_global, 'references', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') 'Orbit period and inclination are constant values. ' // &
                       'All wind directions in oceanographic convention (0 deg. flowing North)'
  error_flag = nc_put_att_text(unit_nc, nc_global, 'comment', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') creation_date
  error_flag = nc_put_att_text(unit_nc, nc_global, 'creation_date', tmp_str)
  if (error_flag .ne. 0) stop

  write(tmp_str,'(A)') creation_time
  error_flag = nc_put_att_text(unit_nc, nc_global, 'creation_time', tmp_str)
  if (error_flag .ne. 0) stop

  ! leave define mode
  error_flag = nc_enddef(unit_nc)
  if (error_flag .ne. 0) stop

  ! put scatterometer data
  call nc_put_array_int_2d(unit_nc, varid_time, startids, lengths, time, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_int_2d(unit_nc, varid_lat, startids, lengths, lat, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_int_2d(unit_nc, varid_lon, startids, lengths, lon, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_wvc_index, startids, lengths, wvc_index, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_model_speed, startids, lengths, model_speed, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_model_dir, startids, lengths, model_dir, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_ice_prob, startids, lengths, ice_prob, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_ice_age, startids, lengths, ice_age, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_int_2d(unit_nc, varid_wvc_quality, startids, lengths, wvc_quality, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_wind_speed, startids, lengths, wind_speed, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_wind_dir, startids, lengths, wind_dir, error_flag)
  if (error_flag .ne. 0) stop

  call nc_put_array_short_2d(unit_nc, varid_bs_distance, startids, lengths, bs_distance, error_flag)
  if (error_flag .ne. 0) stop

  ! close NetCDF file
  error_flag = nc_close(unit_nc)
  if (error_flag .ne. 0) stop

  ! format statements
  101 format (I4.4,'-',I2.2,'-',I2.2,'T',I2.2,':',I2.2,':',I2.2,'.000')
  102 format (I4.4,'-',I2.2,'-',I2.2)
  103 format (I2.2,':',I2.2,':',I2.2)
  105 format (A4,'-',A2,'-',A2)
  106 format (A2,':',A2,':',A2)

end program Bufr2Nc
