module bufr_data_extraction_module
  !  #[ documentation
  !
  ! this module bundles knowledge about the BUFR file and template
  ! and allows easier extraction of parameters.
  !
  ! Last Modified on: $Date$
  ! 
  !  #]
  !  #[ modules used
  use BufrIO, only: bufr_data_type

  use numerics, only: missing_real_r4=>missing_real, &
                      missing_indicator_real_r4=>missing_indicator_real, &
                      missing_indicator_integer, missing_int, &
                      r4_, r8_

  use DateTimeMod, only: ymd2julian
  !  #]
  !  #[ variables and parameters
  implicit none

  integer, parameter :: bufr_template_undefined      = 0
  integer, parameter :: bufr_template_ERS_OLD        = 1
  integer, parameter :: bufr_template_ERS_OLD_APP    = 2
  integer, parameter :: bufr_template_ERS_NEW        = 3
  integer, parameter :: bufr_template_ERS_NEW_APP    = 4
  integer, parameter :: bufr_template_ASCAT          = 5
  integer, parameter :: bufr_template_Seawinds       = 6
  integer, parameter :: bufr_template_GenSeawinds    = 7
  integer, parameter :: bufr_template_OldGenSeawinds = 8
  integer, parameter :: bufr_template_CFOSAT = 9

  ! for template definitions see:
  ! http://www.knmi.nl/scatterometer/publications/pdf/SeaWinds_Product_Manual.pdf
  ! further info on different template versions derived from the Bufr2Asc.F90 tool
  ! and the Bufr2Bin.F90 tool
  
  ! position of relevant BUFR data descriptors
  !                       for BUFR template number   1   2   3   4   5   6   7   8   9
  ! software identification 1 
  ! This is zero when BUFR file was processed with -research option
  ! in which case data contains 2DVAR analysis wind
  integer, dimension(9), parameter :: pos_si1  = (/ -1, -1,  2,  2,  3,  5,  5,  5,  5 /)
  ! software identification 2
  integer, dimension(9), parameter :: pos_si2  = (/ -1, -1,  2,  2, 83,  5, 94,  5,  5 /)
  ! satellite id
  integer, dimension(9), parameter :: pos_sat  = (/  1,  1,  1,  1,  4,  1,  1,  1,  1 /)
  ! instrument id
  integer, dimension(9), parameter :: pos_ins  = (/ -1, -1, -1, -1,  5,  3,  3,  3,  3 /)
  ! satellite track direction (direction of flight)
  integer, dimension(9), parameter :: pos_trk  = (/  2,  2,  5,  5,  6,  2,  2,  2,  2 /)
  ! pixel size / resolution
  integer, dimension(9), parameter :: pos_res  = (/ -1, -1, -1, -1, 15,  6,  6,  6,  6 /)
  ! date time
  integer, dimension(9), parameter :: pos_dtg  = (/  4,  4, 19, 19,  7,  9,  9,  9,  9 /)
  ! latitude of wvc
  integer, dimension(9), parameter :: pos_lat  = (/ 10, 10, 25, 25, 13, 15, 15, 15, 15 /)
  ! longitude of wvc
  integer, dimension(9), parameter :: pos_lon  = (/ 11, 11, 26, 26, 14, 16, 16, 16, 16 /)
  ! wvc number (along a row)
  integer, dimension(9), parameter :: pos_cln  = (/ -1, -1, -1, -1, 17, 20, 20, 20, 20 /)
  ! nr of ambiguities of the WVC solutions
  integer, dimension(9), parameter :: pos_nram = (/ -1, -1, -1, -1, -1, 24, 24, 24, 24 /)
  ! index of selected wind vector
  integer, dimension(9), parameter :: pos_idx  = (/ -1, -1, -1, -1, 91, 25,102, 98, 25 /)
  ! wind vector cell quality
  integer, dimension(9), parameter :: pos_qcf  = (/ -1, -1, -1, -1, 89, 21,100, 96, 21 /)
  ! wind speed (calculated using index of selected wind vector)
  !integer, dimension(8), parameter :: pos_spd = (/ -1, -1, -1, -1, -1, -1, -1, -1 /)
  ! wind dir (calculated using index of selected wind vector)
  !integer, dimension(8), parameter :: pos_dir = (/ -1, -1, -1, -1, -1, -1, -1, -1 /)
  ! model wind speed
  integer, dimension(9), parameter :: pos_mspd = (/ -1, 41, -1, 56, 85, 23, 96, 22, 23 /)
  ! model wind dir
  integer, dimension(9), parameter :: pos_mdir = (/ -1, 42, -1, 57, 86, 22, 97, 21, 22 /)
  ! orbit number
  integer, dimension(9), parameter :: pos_orb  = (/ -1, -1, -1, -1, 16,  8,  8,  8,  8 /)
  ! ice probability
  ! NOTE: for seawinds SDP abuses an existing field to store this number
  ! (original definition is: "Seawinds Probability of Rain"), same hold 
  ! for CFOSAT
  ! See lines 1661-1664 in SDP/sdp/SdpIcemodel.F90 for this hack.
  integer, dimension(9), parameter :: pos_iceprb = (/ -1, -1, -1, -1, 87,  27, 98, -1, 27 /)
  ! ice age 
  ! NOTE: for seawinds SDP abuses an existing field to store this number
  !  (original definition is: "Attenuation Correction On Sigma-0 (from Tb)"
  ! See lines 1661-1664 in SDP/sdp/SdpIcemodel.F90 for this hack.
  integer, dimension(9), parameter :: pos_iceage = (/ -1, -1, -1, -1, 88,  30, 99, -1, 30 /)
  ! brightness temperature
  integer, dimension(9), parameter :: pos_btem   = (/ -1, -1, -1, -1, -1,  53, -1, -1, 53 /)
  ! backscatter distance (calculated using index of selected wind vector)
  !integer, dimension(8), parameter :: pos_bsdist = (/ -1, -1, -1, -1, -1,  -1, -1, -1 /)
  ! 2DVAR analysis speed when BUFR file was processed with -research option
  ! CFOSAT has not been tested for this option yet
  integer, dimension(9), parameter :: pos_aspd = (/ -1, -1, -1, -1, 23,  61, 61, 61, 61 /)
   ! 2DVAR analysis direction when BUFR file was processed with -research option
  ! CFOSAT has not been tested for this option yet
  integer, dimension(9), parameter :: pos_adir = (/ -1, -1, -1, -1, 37,  76, 76, 76, 76 /)

  ! The following constants define how to calculate the index for a given antenna
  ! property. The index is given by a+(i-1)*b where a is the constant index offset,
  ! b the variable index offset, and i the antenna counter.
  
  ! Variable index offset when looping over the different antennas
  ! [only defined for seawinds/rapidscat for now]
  !                              for BUFR template number  1   2   3   4   5   6   7   8   9
  integer, dimension(9), parameter :: var_ant_offset = (/ -1, -1,  5, -1, 14, 15, 15, -1, 15 /)
  
  
  ! Constant index offset when looping over the different antennas
  ! [only defined for seawinds/rapidscat for now]
  !                               for BUFR template number  1   2   3   4   5   6   7   8   9
  integer, dimension(9), parameter :: pos_azi_angle   = (/ -1, -1, 13, -1, 23, 63, 38, -1, 63 /)
  integer, dimension(9), parameter :: pos_inc_angle   = (/ -1, -1, 12, -1, 22, 64, 39, -1, 64 /)
  integer, dimension(9), parameter :: pos_polaris     = (/ -1, -1, -1, -1, -1, 65, 40, -1, 65 /)
  integer, dimension(9), parameter :: pos_sigma0      = (/ -1, -1, 14, -1, 24, 66, 41, -1, 66 /)
  integer, dimension(9), parameter :: pos_atten_value = (/ -1, -1, 14, -1, -1, 62, 41, -1, 62 /)
  integer, dimension(9), parameter :: pos_beam_lat    = (/ -1, -1, -1, -1, -1, 60, -1, -1, 60 /)
  integer, dimension(9), parameter :: pos_beam_lon    = (/ -1, -1, -1, -1, -1, 61, -1, -1, 61 /)
  integer, dimension(9), parameter :: pos_kpa         = (/ -1, -1, -1, -1, -1, 67, -1, -1, 67 /)
  integer, dimension(9), parameter :: pos_kpb         = (/ -1, -1, -1, -1, -1, 68, -1, -1, 68 /)
  integer, dimension(9), parameter :: pos_kpc         = (/ -1, -1, -1, -1, -1, 69, -1, -1, 69 /)
  integer, dimension(9), parameter :: pos_kp          = (/ -1, -1, -1, -1, 25, -1, -1, -1, -1 /)
  ! saved variable to store the code for the current BUFR template
  integer, save :: bufr_tmpl = 0

  ! print information about BUFR template only once
  logical, save :: tmpl_printed = .false.

  ! saved variables to store the subset number and its offset in the values array
  integer, save :: current_subset = 0
  integer, save :: iaux = 0

  ! possible angle conventions:
  integer, parameter :: angle_convention_meteorological = 1
  integer, parameter :: angle_convention_oceanographic  = 2
  !integer, parameter :: angle_convention_mathematical   = 3 ! not yet used
  ! oceanographic:  0 deg. = pointing/flowing towards the north
  !                 clockwise increment (seen from above)
  !                 (components u=0, v>0)
  ! meteorological: 0 deg. = pointing/flowing towards the south, from the north
  !                 clockwise increment (seen from above)
  !                 (components u=0, v<0)
  ! mathematical:   0 deg. = pointing/flowing towards the east
  !                 counterclockwise increment (seen from above)
  !                 (components u>0, v=0)
  integer, save :: angle_convention_to_report = angle_convention_meteorological
  
  ! in this program, we convert all data from the BUFR module to default
  ! size reals (4 bytes, generally)
  ! to prevent rounding errors when checking for missing values, the
  ! value of missing_indicator_real is also rounded
  real, parameter :: missing_indicator_real = real(missing_indicator_real_r4)
  !  #]
  !  #[ interfaces and overloading
  ! overload this one to accept both 4 and 8 byte reals
  interface BufrReal2Int
     module procedure BufrReal2Int_r4, BufrReal2Int_r8
  end interface BufrReal2Int
  !  #]
contains
  !-------------------------------------------------------------
  subroutine goto_subset(tbd,subset)
    !  #[ start reading data from this subset
    ! a little helper routine that calculates iaux, i.e. the start
    ! index of the current subset of values in the BUFR message
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: subset

    ! pointer to the position just before the start of this subset
    iaux = tbd%nelements*(subset-1)

    current_subset = subset

    ! note that current_subset and iaux are global module variables,
    ! this prevents recalculation of iaux many times.

  end subroutine goto_subset
    !  #]
  subroutine report_angle_conv_oceanographic()
    !  #[ choose oceanographic angle convention
    ! this is stored in a global module variable
    ! Note that scatterometry winds normaly are reported in
    ! meteorological convention (which is the default here as well)
    ! but netcdf output files seem to require oceanographic convention.
    angle_convention_to_report = angle_convention_oceanographic

  end subroutine report_angle_conv_oceanographic
    !  #]
  function missing_real(x) result(m)
    !  #[ custom version that accepts default reals
    real, intent(in) :: x
    logical          :: m
    m=.false.
    if (x .ge. (missing_indicator_real*0.99)) m=.true.
  end function missing_real
    !  #]
  function BufrReal2Int_r8(val) result(res)
    !  #[ convert real*8 to int
    ! in this program, we convert all data from the BUFR module to default
    ! size reals (4 bytes, generally)
    ! to prevent rounding errors when checking for missing values, the
    ! value of missing_indicator_real is also rounded
    real(r8_), intent(in) :: val
    integer               :: res  ! result
    
    if ( val >= missing_indicator_real*0.99 ) then
       res    = missing_indicator_integer
    else
       res    = nint(val)
    endif
  end function BufrReal2Int_r8
  !  #]
  function BufrReal2Int_r4(val) result(res)
    !  #[ convert real*4 to int
    ! in this program, we convert all data from the BUFR module to default
    ! size reals (4 bytes, generally)
    ! to prevent rounding errors when checking for missing values, the
    ! value of missing_indicator_real is also rounded
    real(r4_), intent(in) :: val
    integer               :: res  ! result
    
    if ( val >= missing_indicator_real*0.99 ) then
       res    = missing_indicator_integer
    else
       res    = nint(val)
    endif
  end function BufrReal2Int_r4
    !  #]
  function get_BUFR_template_code(tbd) result(bufr_template_code)
    !  #[ get BUFR template code from BUFR message
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: bufr_template_code ! result

    bufr_template_code = bufr_template_undefined
    if (tbd%unexpanded_descr(1) .eq. 312201) then
       ! Data in old ERS format without appendix (ESA winds reported)
       bufr_template_code = bufr_template_ERS_OLD
       if (tbd%nunexpanded .gt. 1) then
          if (tbd%unexpanded_descr(2) .eq. 1031) then
             ! Data in old ERS format with appendix
             bufr_template_code = bufr_template_ERS_OLD_APP
          endif
       endif
    else if (tbd%unexpanded_descr(1) .eq. 312021) then
       ! Data in new ERS format without appendix (ESA winds reported)
       bufr_template_code = bufr_template_ERS_NEW
       if (tbd%nunexpanded .gt. 1) then
          if (tbd%unexpanded_descr(2) .eq. 1031) then
             ! Data in new ERS format with appendix
             bufr_template_code = bufr_template_ERS_NEW_APP
          endif
       endif
    else if (tbd%unexpanded_descr(1) .eq. 312061) then
       ! Data in ASCAT format
       bufr_template_code = bufr_template_ASCAT
    else if (tbd%unexpanded_descr(1) .eq. 312028) then
       ! Data in SeaWinds format
       bufr_template_code = bufr_template_Seawinds
    else if (tbd%unexpanded_descr(1) .eq. 312221) then
       ! Data in SeaWinds format with OLD generic wind section
       bufr_template_code = bufr_template_OldGenSeawinds
    else if (tbd%unexpanded_descr(1) .eq. 312223) then
       ! Data in SeaWinds format with generic wind section
       bufr_template_code = bufr_template_GenSeawinds
    else if (tbd%unexpanded_descr(1) .eq. 312034 .or. tbd%unexpanded_descr(1) .eq. 312196) then
       ! Data in CFOSAT format
       bufr_template_code = bufr_template_CFOSAT
    else
       print '(2A,I7)','Unknown scatterometer data format, ', &
            'first table D data descriptor =', tbd%unexpanded_descr(1)
       stop
    endif

    if (.not. tmpl_printed) &
         print '(A)',trim(get_BUFR_template_descr(bufr_template_code))
    tmpl_printed = .true.

  end function get_BUFR_template_code
    !  #]
  function get_BUFR_template_descr(bufr_template_code) result(description)
    !  #[ convert template code to text description
    integer, intent(in) :: bufr_template_code
    character(len=256)  :: description ! result

    description(:) = ' '
    select case(bufr_template_code)
    case(bufr_template_ERS_OLD)
       description = 'Data in old ERS format without appendix (ESA winds reported)'
    case(bufr_template_ERS_OLD_APP)
       description = 'Data in old ERS format with appendix'
    case(bufr_template_ERS_NEW)
       description = 'Data in new ERS format without appendix (ESA winds reported)'
    case(bufr_template_ERS_NEW_APP)
       description = 'Data in new ERS format with appendix'
    case(bufr_template_ASCAT)
       description = 'Data in ASCAT format'
    case(bufr_template_Seawinds)
       description = 'Data in SeaWinds format'
    case(bufr_template_OldGenSeawinds)
       description = 'Data in SeaWinds format with OLD generic wind section'
    case(bufr_template_GenSeawinds)
       description = 'Data in SeaWinds format with generic wind section'
    case(bufr_template_CFOSAT)
       description = 'Data in CFOSAT format'
    case default
       description = 'unknown template'
    end select
    
  end function get_BUFR_template_descr
    !  #]
  subroutine set_BUFR_template(tbd)
    !  #[ choose the BUFR template to use
    type(bufr_data_type), intent(in) :: tbd
    ! set the saved global module variable
    bufr_tmpl = get_BUFR_template_code(tbd)
  end subroutine set_BUFR_template
    !  #]
  function get_sw_id_l1(tbd) result(sw_id_l1)
    !  #[ get software identification 1
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: sw_id_l1 ! result
    if (pos_si1(bufr_tmpl) .le. 0) then
      sw_id_l1 = 0
    else
      sw_id_l1 = BufrReal2Int(tbd%values(iaux+pos_si1(bufr_tmpl)))
    endif
  end function get_sw_id_l1
    !  #]
  function get_sw_id_wind(tbd) result(sw_id_wind)
    !  #[ get software identification 2
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: sw_id_wind ! result
    if (pos_si2(bufr_tmpl) .le. 0) then
      sw_id_wind = 0
    else
      sw_id_wind = BufrReal2Int(tbd%values(iaux+pos_si2(bufr_tmpl)))
    endif
  end function get_sw_id_wind
    !  #]
  function get_res(tbd) result(res)
    !  #[ get resolution/wvc size in [m]
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: res ! result
    integer :: sat ! local variable
    if (pos_res(bufr_tmpl) .le. 0) then
      res = 0
    else
      res = BufrReal2Int(tbd%values(iaux+pos_res(bufr_tmpl)))
    endif
    sat = get_sat(tbd)
    if (sat .le. 2) then ! ERS
       res = 25000 ! fill manually since WVC spacing is missing in BUFR
    endif
  end function get_res
    !  #]
  function get_trk(tbd) result(trk)
    !  #[ get satellite track angle
    type(bufr_data_type), intent(in) :: tbd
    real                             :: trk ! result
    trk = real(tbd%values(iaux+pos_trk(bufr_tmpl)))
  end function get_trk
    !  #]
  function get_ascending(tbd) result(ascending)
    !  #[ get ascending/decending boolean
    type(bufr_data_type), intent(in) :: tbd
    logical                          :: ascending ! result
    real :: trk
    trk = get_trk(tbd)
    if ((trk .lt. 90.) .or. (trk .gt. 270)) then   ! ascending flight direction
       ascending = .true.
    else
       ascending = .false.
    endif
  end function get_ascending
    !  #]
  function get_sat(tbd) result(sat)
    !  #[ get satellite id
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: sat ! result
    sat = BufrReal2Int(tbd%values(iaux+pos_sat(bufr_tmpl)))
  end function get_sat
    !  #]
  function get_ins(tbd) result(ins)
    !  #[ get instrument id
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: ins ! result
    integer :: sat ! local variable
    if (pos_ins(bufr_tmpl) .le. 0) then
      ins = 0
    else
      ins = BufrReal2Int(tbd%values(iaux+pos_ins(bufr_tmpl)))
    endif
    sat = get_sat(tbd)
    if (sat .le. 2) then ! ERS
       ins = 142 ! fill manually since instrument id is missing in BUFR
    endif
    if (sat .eq. 421) then ! Oceansat-2
       ins = 288 ! fill manually since instrument id is missing in BUFR
    endif
    if (sat .eq. 422) then ! ScatSat-1
       ins = 288 ! fill manually since instrument id is missing in BUFR
    endif
    if (sat .eq. 555 .or. sat .eq. 802) then ! CFOSAT
       ins = 943 ! fill manually since instrument id is missing in BUFR
    endif
    if (sat .eq. 990 .or. sat .eq. 502 .or. sat .eq. 503 .or. sat .eq. 504) then ! HY-2A/HY-2B/HY-2C
       ins = 686 ! fill manually since instrument id is missing in BUFR
    endif
    if (sat .eq. 991 .or. sat .eq. 801) then ! ISS
       ins = 314 ! fill manually since instrument id is missing in BUFR
    endif
    ! NOTE: WMO defines these id codes for scat satellites/instruments:
    !  ers1/2       : sat_id=1, 2 
    !     ami       : instr_id=142
    !  metop B,A,C  : sat_id=3, 4, 5
    !     ascat     : instr_id=190
    !  quikscat     : sat_id=281
    !     seawinds  : instr_id=8 [short id] 313 [long id]
    !  adeos2       : sat_id=121
    !     seawinds  : instr_id=8 [short id] 313 [long id]
    !  Oceansat-2   : sat_id=421
    !     oscat     : instr_id=288 [long id] no short id
    !  ScatSat-1    : sat_id=422
    !     oscat     : instr_id=288 [long id] no short id
    !  HY-2A        : sat_id=502   (proposed 26 Jan 2015)
    !     HSCAT     : instr_id=686 (proposed 26 Jan 2015)
    !  HY-2B        : sat_id=503
    !     HSCAT     : instr_id=686 (proposed 26 Jan 2015)
    !  HY-2C        : sat_id=504
    !     HSCAT     : instr_id=686 (proposed 26 Jan 2015)
    !  ISS          : sat_id=801   (proposed 26 Jan 2015)
    !     RapidScat : instr_id=314 (proposed 26 Jan 2015)
    !  CFOSAT       : sat_id=802
    !     RFSCAT    : instr_id=943

  end function get_ins
    !  #]
  function get_year(tbd) result(year)
    !  #[ get the year field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: year ! result
    year  = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)))
  end function get_year
    !  #]
  function get_month(tbd) result(month)
    !  #[ get the month field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: month ! result
    month = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+1))
  end function get_month
    !  #]
  function get_day(tbd) result(day)
    !  #[ get the day field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: day ! result
    day = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+2))
  end function get_day
    !  #]
  function get_date_str(tbd, sep) result(date_str)
    !  #[ get date string for a given wvc
    type(bufr_data_type), intent(in) :: tbd
    character(len=1), optional :: sep ! separator to be used
    character(len=20)  :: date_str ! result

    ! local variables
    integer:: year, month, day

    year  = get_year(tbd)
    month = get_month(tbd)
    day   = get_day(tbd)

    if (missing_int(year) .or. missing_int(month) .or. missing_int(day)) then
       date_str = "[missing]"
       return
    end if

    if (present(sep)) then
       ! write time information to variable in format YYYY-MM-DD
       write(date_str,'(I4.4,A1,I2.2,A,I2.2)', err=999) year, sep, month, sep, day
    else
       ! write time information to variable in format YYYYMMDD
       write(date_str,'(I4.4,2I2.2)', err=999) year, month, day
    end if
    return

999 print *,'ERROR: conversion problem while creating date_str'
    stop

  end function get_date_str
    !  #]
  function get_julian_daynumber(tbd) result(julian)
    !  #[ get the julian day number
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: julian ! result
    
    ! local variables
    integer:: year, month, day
    
    year  = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)))
    month = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+1))
    day   = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+2))

    if (missing_int(year) .or. missing_int(month) .or. missing_int(day)) then
       julian = missing_indicator_integer
       return
    end if

    call ymd2julian(year,month,day,julian)

  end function get_julian_daynumber
    !  #]
  function get_hour(tbd) result(hour)
    !  #[ get the hour field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: hour ! result
    hour = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+3))
  end function get_hour
    !  #]
  function get_minute(tbd) result(minute)
    !  #[ get the minute field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: minute ! result
    minute = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+4))
  end function get_minute
    !  #]
  function get_second(tbd) result(second)
    !  #[ get the second field
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: second ! result
    second = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+5))
  end function get_second
    !  #]
  function get_time_str(tbd, sep) result(time_str)
    !  #[ get time string for a given wvc
    type(bufr_data_type), intent(in) :: tbd
    character(len=1), optional       :: sep ! separator to be used
    character(len=20)                :: time_str ! result

    ! local variables
    integer:: hour, minute, second

    hour   = get_hour(tbd)
    minute = get_minute(tbd)
    second = get_second(tbd)

    if ( missing_int(hour) .or. missing_int(minute) .or. &
         missing_int(second)) then
       time_str = "[missing]"
       return
    end if

    if (present(sep)) then
       ! write time information to variable in format HH:MM:SS [if sep=':']
       write(time_str,'(I2.2,A1,I2.2,A1,I2.2)', err=999) hour, sep, minute, sep, second
    else
       ! write time information to variable in format HHMMSS
       write(time_str,'(3I2.2)', err=999) hour, minute, second
    end if
    return

999 print *,'ERROR: conversion problem while creating time_str'
    stop

  end function get_time_str
    !  #]
  function get_time_seconds_in_day(tbd) result(time_i)
    !  #[ get time_i as number of seconds into the day,
    ! 0 = 0:00:00 and 86399 = 23:59:59
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: time_i ! result

    ! local variables
    integer:: hour, minute, second

    hour   = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+3))
    minute = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+4))
    second = BufrReal2Int(tbd%values(iaux+pos_dtg(bufr_tmpl)+5))

    if ( missing_int(hour) .or. missing_int(minute) .or. &
         missing_int(second)) then
       time_i = missing_indicator_integer
       return
    end if

    time_i = hour * 3600 + minute * 60 + second

  end function get_time_seconds_in_day
    !  #]
  function get_lat(tbd) result(lat)
    !  #[ get latitude
    type(bufr_data_type), intent(in) :: tbd
    real                             :: lat ! result

    lat = real(tbd%values(iaux+pos_lat(bufr_tmpl)))
  end function get_lat
    !  #]
  function get_lon(tbd) result(lon)
    !  #[ get longitude
    type(bufr_data_type), intent(in) :: tbd
    real                             :: lon ! result

    lon = real(tbd%values(iaux+pos_lon(bufr_tmpl)))
  end function get_lon
    !  #]
  function get_cln(tbd) result(cln)
    !  #[ get wvc cell number
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: cln ! result

    ! local variable
    integer :: res

    if ( (bufr_tmpl .eq. bufr_template_ERS_OLD    ) .or. &
         (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) .or. &
         (bufr_tmpl .eq. bufr_template_ERS_NEW    ) .or. &
         (bufr_tmpl .eq. bufr_template_ERS_NEW_APP)      ) then
       ! ERS did not have a cell number in its BUFR template, so derive
       ! it from the subset number for this instrument
       cln = mod(current_subset,19)
       if (cln .eq. 0) cln = 19
       return
    end if

    ! NOTE: for the ascat 6.25 km product we have a problem
    ! with the CROSS-TRACK CELL NUMBER field in the template. This one is
    ! only 7 bits wide, so can only hold numbers 0-126 (127 is reserved
    ! as missing value), but we need to number 162 cells here.
    ! Therefore a workaround is implemented for now that assumes the subset
    ! number equals the cross-track cell number for BUFR messages with 
    ! this template, for grid sizes lower than 10000 m (10 km)
    ! in this way also future high resolution experimental products can be handled 
    if (bufr_tmpl .eq. bufr_template_ASCAT) then
       res = get_res(tbd)
       if (res .lt. 10000) then
          cln = current_subset
          return
       end if
    end if

    cln = BufrReal2Int(tbd%values(iaux+pos_cln(bufr_tmpl)))

    ! if the cell number is missing, assign the subset number
    if (cln .eq. missing_indicator_integer) then
      cln = current_subset
    endif

  end function get_cln
    !  #]
  function get_idx(tbd) result(idx)
    !  #[ index of selected wind vector
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: idx ! result

    ! local variables
    real :: idx_real

    if (bufr_tmpl .eq. bufr_template_ERS_OLD .or. &
        bufr_tmpl .eq. bufr_template_ERS_OLD_APP .or. &
        bufr_tmpl .eq. bufr_template_ERS_NEW .or. &
        bufr_tmpl .eq. bufr_template_ERS_NEW_APP) then
       idx = 1
    else
       idx_real = real(tbd%values(iaux+pos_idx(bufr_tmpl)))
       idx = BufrReal2Int(idx_real)
    endif

  end function get_idx
    !  #]
  function get_qcflag(tbd) result(qcflag)
    !  #[ get overall wvc quality boolean
    type(bufr_data_type), intent(in) :: tbd
    logical                          :: qcflag ! result

    ! local variables
    integer :: qcf, bit1_to_inspect, bit2_to_inspect
    real :: qcf_real

    ! assume data is good
    qcflag = .false.
    if (pos_qcf(bufr_tmpl) .le. 0) then
       ! for the several ERS templates this flag is not (yet) defined
       ! TODO: look if this is available in these templates
       return
    endif

    qcf_real = real(tbd%values(iaux+pos_qcf(bufr_tmpl)))
    qcf = BufrReal2Int(qcf_real)

    bit1_to_inspect = -1
    bit2_to_inspect = -1
    if (bufr_tmpl .eq. bufr_template_Seawinds .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       bit1_to_inspect = 10 ! KNMI QC flag
       bit2_to_inspect = 2  ! NOAA rain flag
    else ! valid for GenSeawinds and ascat
       bit1_to_inspect = 17 ! KNMI QC flag
    end if

    if (missing_int(qcf)) then
       qcflag = .true.
    else
       ! do not use data for which QC flag is set
       if (bit1_to_inspect .ne. -1) then
          qcflag = btest(qcf, bit1_to_inspect)
          if (qcflag) return
       end if
       if (bit2_to_inspect .ne. -1) then
          qcflag = btest(qcf, bit2_to_inspect)
          if (qcflag) return
       end if
    endif

  end function get_qcflag
    !  #]
  function get_qcflags(tbd) result(qcflags)
    !  #[ get all wvc quality flags
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: qcflags ! result

    ! local variables
    integer :: qcflags_tmp

    if (pos_qcf(bufr_tmpl) .le. 0) then
      qcflags_tmp = 0
    else
      qcflags_tmp = BufrReal2Int(tbd%values(iaux+pos_qcf(bufr_tmpl)))
    endif
    qcflags = 0

    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! bitwise convert WVC quality from QuikSCAT format to generic format
       if (.not. missing_int(qcflags_tmp)) then
          if (btest(qcflags_tmp,15)) qcflags = ibset(qcflags,22) ! qual_sigma0
          if (btest(qcflags_tmp,14)) qcflags = ibset(qcflags,21) ! azimuth
          if (btest(qcflags_tmp,12)) qcflags = ibset(qcflags,19) ! mon_flag
          if (btest(qcflags_tmp,11)) qcflags = ibset(qcflags,18) ! mon_value
          if (btest(qcflags_tmp,10)) qcflags = ibset(qcflags,17) ! knmi_qc
          if (btest(qcflags_tmp, 9)) qcflags = ibset(qcflags,16) ! var_qc
          if (btest(qcflags_tmp, 8)) qcflags = ibset(qcflags,15) ! land
          if (btest(qcflags_tmp, 7)) qcflags = ibset(qcflags,14) ! ice
          if (btest(qcflags_tmp, 6)) qcflags = ibset(qcflags,13) ! retrieval
          if (btest(qcflags_tmp, 5)) qcflags = ibset(qcflags,12) ! large
          if (btest(qcflags_tmp, 4)) qcflags = ibset(qcflags,11) ! small
          if (btest(qcflags_tmp, 3)) qcflags = ibset(qcflags,10) ! rain_fail
          if (btest(qcflags_tmp, 2)) qcflags = ibset(qcflags, 9) ! rain_detect
       endif
    else ! SeaWinds generic or ASCAT format
       qcflags = qcflags_tmp
    end if

  end function get_qcflags
    !  #]
  function get_spd(tbd,closest) result(spd)
    !  #[ get wind speed of selected solution
    type(bufr_data_type), intent(in) :: tbd
    logical, intent(in), optional    :: closest
    real                             :: spd ! result
    
    ! local variables
    integer :: idx, pos_spd

    idx = get_idx(tbd)

    if (missing_int(idx)) then
       spd = missing_indicator_real
       return
    end if
    
    if (present(closest)) then
       if (closest) then
          ! force usage of closest solutions, in stead of selected one
          idx=1
       end if
    end if

    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind speed - SeaWinds data
       pos_spd = 26 + idx * 5
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD) then
       pos_spd = 27 ! report ESA wind result
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) then
       pos_spd = 32 ! assume first reported one is selected one?
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW) then
       pos_spd = 42 ! report ESA wind result
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW_APP) then
       pos_spd = 47 ! assume first reported one is selected one?
    else ! method for seawinds with (old/new) generic section and for ascat
       ! get wind speed - data with generic wind section
       pos_spd = pos_idx(bufr_tmpl) - 2 + idx * 4
    end if
    spd = real(tbd%values(iaux+pos_spd))
  end function get_spd
    !  #]
  function get_dir(tbd,closest) result(dir)
    !  #[ get wind dir of selected solution
    type(bufr_data_type), intent(in) :: tbd
    logical, intent(in), optional    :: closest
    real                             :: dir ! result

    ! local variables
    integer :: idx, pos_dir
    
    idx = get_idx(tbd)

    if (missing_int(idx)) then
       dir = missing_indicator_real
       return
    end if

    if (present(closest)) then
       if (closest) then
          ! force usage of closest solutions, in stead of selected one
          idx=1
       end if
    end if

    if (bufr_tmpl .eq. bufr_template_Seawinds .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind direction - SeaWinds data
       pos_dir = 28 + idx * 5
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD) then
       pos_dir = 28 ! report ESA wind result
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) then
       pos_dir = 33 ! assume first reported one is selected one?
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW) then
       pos_dir = 43 ! report ESA wind result
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW_APP) then
       pos_dir = 48 ! assume first reported one is selected one?
    else ! method for seawinds with (old/new) generic section and for ascat
       ! get wind direction - data with generic wind section
       pos_dir = pos_idx(bufr_tmpl) - 1 + idx * 4
    end if
    dir = real(tbd%values(iaux+pos_dir))

    if (angle_convention_to_report .eq. angle_convention_oceanographic) then
       ! convert model wind direction to oceanographic convention
       if (.not. missing_real(dir)) then
          dir = dir + 180.0
          if (dir .gt. 360.0) dir = dir - 360.0
       endif
    end if
    
  end function get_dir
    !  #]
  function get_mle(tbd,closest) result(mle)
    !  #[ get MLE of selected solution
    type(bufr_data_type), intent(in) :: tbd
    logical, intent(in), optional    :: closest
    real                             :: mle ! result
    
    ! local variables
    integer :: idx, pos_mle

    idx = get_idx(tbd)

    if (missing_int(idx)) then
       mle = missing_indicator_real
       return
    end if
    
    if (present(closest)) then
       if (closest) then
          ! force usage of closest solutions, in stead of selected one
          idx=1
       end if
    end if

    if (bufr_tmpl .eq. bufr_template_Seawinds .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind dir uncertainty - SeaWinds data
       pos_mle = 29 + idx * 5
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD) then
       mle = missing_indicator_real ! no MLE information
       return
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) then
       pos_mle = 34 ! assume first reported one is selected one?
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW) then
       mle = missing_indicator_real ! no MLE information
       return
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW_APP) then
       pos_mle = 49 ! assume first reported one is selected one?
    else ! method for seawinds with (old/new) generic section and for ascat
       ! get wind speed - data with generic wind section
       pos_mle = pos_idx(bufr_tmpl) + idx * 4
    end if
    mle = real(tbd%values(iaux+pos_mle))
  end function get_mle
    !  #]
  function get_mspd(tbd) result(mspd)
    !  #[ get model wind speed
    type(bufr_data_type), intent(in) :: tbd
    real                             :: mspd ! result
    if (pos_mspd(bufr_tmpl) .le. 0) then
      mspd = missing_indicator_real
    else
      mspd = real(tbd%values(iaux+pos_mspd(bufr_tmpl)))
    endif
  end function get_mspd
    !  #]
  function get_mdir(tbd) result(mdir)
    !  #[ get model wind dir
    type(bufr_data_type), intent(in) :: tbd
    real                             :: mdir ! result
    if (pos_mdir(bufr_tmpl) .le. 0) then
      mdir = missing_indicator_real
    else
      mdir = real(tbd%values(iaux+pos_mdir(bufr_tmpl)))
    endif

    if (angle_convention_to_report .eq. angle_convention_oceanographic) then
       ! convert model wind direction to oceanographic convention
       if (.not. missing_real(mdir)) then
          mdir = mdir + 180.0
          if (mdir .gt. 360.0) mdir = mdir - 360.0
       endif
    end if

  end function get_mdir
    !  #]
  function get_sat_name(tbd) result(satellite)
    !  #[ get name of satellite from bufr message
    type(bufr_data_type), intent(in) :: tbd       ! input
    character(len=25)                :: satellite ! result

    ! local variable
    integer :: sat

    satellite(:) = ' '
    sat = get_sat(tbd)
    select case(sat)
    case(1)
       satellite = 'ERS-1'
    case(2)
       satellite = 'ERS-2'
    case(3)
       satellite = 'Metop-B'
    case(4)
       satellite = 'Metop-A'
    case(5)
       satellite = 'Metop-C'
    case(281)
       satellite = 'QuikSCAT'
    case(421)
       satellite = 'Oceansat-2'
    case(422)
       satellite = 'ScatSat-1'
    case(555,802)
       satellite = 'CFOSAT'
    case(990,502)
       satellite = 'HY-2A'
    case(503)
       satellite = 'HY-2B'
    case(504)
       satellite = 'HY-2C'
    case(991,801)
       satellite = 'ISS'
    case default
       print '(A,I7)','get_sat_name: Unknown satellite identifier: ', sat
       stop
    end select
    
  end function get_sat_name
    !  #]
  function get_sensor(tbd) result(sensor)
    !  #[ get name of scat instrument from bufr message
    type(bufr_data_type), intent(in) :: tbd    ! input
    character(len=25)                :: sensor ! result
    
    ! local variable
    integer :: sat

    sensor = ' '
    sat = get_sat(tbd)
    select case(sat)
    case(1,2)
       sensor    = 'SCAT'
    case(3,4,5)
       sensor    = 'ASCAT'
    case(281)
       sensor    = 'SeaWinds'
    case(421,422)
       sensor    = 'OSCAT'
    case(555,802)
       sensor    = 'CFOSAT'
    case(990,502,503,504)
       sensor    = 'HSCAT'
    case(991,801)
       sensor    = 'RapidScat'
    case default
       print '(A,I7)','Unknown satellite identifier: ', sat
       stop
    end select

  end function get_sensor
    !  #]
  function get_wvcs_per_row(tbd) result(wvcs_per_row)
    !  #[ get wvcs_per_row
    type(bufr_data_type), intent(in) :: tbd    ! input
    integer                          :: wvcs_per_row ! result

    ! local variables
    integer :: sat, res

    sat = get_sat(tbd)
    res = get_res(tbd)
    wvcs_per_row = 0
    if (sat .le. 2) then         ! ERS
       wvcs_per_row = 19
    elseif (sat .le. 5) then     ! ASCAT
       if (res .eq. 25000) then
          wvcs_per_row = 46
          ! 25 km, add 4 wvcs to keep some space between left and right swath
       else
          wvcs_per_row = 86
          ! 12.5 km, add 4 wvcs to keep some space between left and right swath
       endif
    elseif (sat .eq. 281) then   ! QuikSCAT
       wvcs_per_row = 1900000 / res
    elseif (sat .eq. 421) then   ! Oceansat-2
       if (res .eq. 25000) then
          wvcs_per_row = 76
       else
          wvcs_per_row = 36
       endif
    elseif (sat .eq. 555 .or. sat .eq. 802) then   ! CFOSAT
       if (res .eq. 25000) then
          wvcs_per_row = 42
       else
          wvcs_per_row = 21
       endif
    else
       print '(A,I7)','Unable to compute number of WVCs per row '//&
            'for satellite: ', sat
       stop 1
    endif
    
  end function get_wvcs_per_row
    !  #]
  function get_orb(tbd) result(orb)
    !  #[ get orbit number
    type(bufr_data_type), intent(in) :: tbd
    integer                          :: orb ! result
    if (pos_orb(bufr_tmpl) .le. 0) then
      orb = 0
    else
      orb = BufrReal2Int(tbd%values(iaux+pos_orb(bufr_tmpl)))
    endif
  end function get_orb
    !  #]
  function get_iceprb(tbd) result(iceprb)
    !  #[ get ice probability
    type(bufr_data_type), intent(in) :: tbd
    real                             :: iceprb ! result
    if (pos_iceprb(bufr_tmpl) .le. 0) then
      iceprb = missing_indicator_real
    else
      iceprb = real(tbd%values(iaux+pos_iceprb(bufr_tmpl)))
    endif
  end function get_iceprb
    !  #]
  function get_iceage(tbd) result(iceage)
    !  #[ get ice age
    type(bufr_data_type), intent(in) :: tbd
    real                             :: iceage ! result
    if (pos_iceage(bufr_tmpl) .le. 0) then
      iceage = missing_indicator_real
    else
      iceage = real(tbd%values(iaux+pos_iceage(bufr_tmpl)))
    endif
  end function get_iceage
    !  #]
  function get_bsdist(tbd,closest) result(bsdist)
    !  #[ get backscatter distance of selected solution
    type(bufr_data_type), intent(in) :: tbd
    logical, intent(in), optional    :: closest
    real                             :: bsdist ! result
    
    ! local variables
    integer :: idx, pos_bsdist
    
    idx = get_idx(tbd)

    if (missing_int(idx)) then
       bsdist = missing_indicator_real
       return
    end if

    if (present(closest)) then
       if (closest) then
          ! force usage of closest solutions, in stead of selected one
          idx=1
       end if
    end if

    if (bufr_tmpl .eq. bufr_template_Seawinds .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind speed - SeaWinds data
       ! For SeaWinds: wind direction error contains (hopefully) the MLE value
       ! At least this is stated in Anton's original code.
       ! original field definition is:
       !   "Formal Uncertainty In Wind Direction"
       ! See lines 905-910 in SDP/sdp/sdpExtract.F90 for this hack.
       ! Here the comment says:
       !        ! Hack for Seawinds/OSCAT NOAA format
       !        ! Note: %conedistance is only present
       !        ! in %error%dir in reprocessed data
       !        ! (BUFR file ending *_25 or *_12)
       ! The actual hack is implemented here on line 727 of file
       ! SDP/sdp/SdpInversion.F90
       pos_bsdist = 29 + idx * 5
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD) then
       bsdist = missing_indicator_real
       return
    else if (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) then
       pos_bsdist = 34 ! assume first reported one is selected one?
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW) then
       bsdist = missing_indicator_real
       return
    else if (bufr_tmpl .eq. bufr_template_ERS_NEW_APP) then
       pos_bsdist = 49 ! assume first reported one is selected one?
    else
       ! get wind speed - data with generic wind section
       pos_bsdist = pos_idx(bufr_tmpl) + 0 + idx * 4
    end if
    bsdist = real(tbd%values(iaux+pos_bsdist))

! POSSIBLE TODO: inspect the BUFR headers to make sure the current
! BUFR message was indeed a reprocessed one by KNMI, and if not
! fill the bsdist result with missing for the seawinds NOAA template

  end function get_bsdist
    !  #]
  function get_aspd(tbd) result(aspd)
    !  #[ get analysis wind speed
    type(bufr_data_type), intent(in) :: tbd
    real                             :: aspd ! result

    ! local variable
    integer :: sw_id_l1

    sw_id_l1 = get_sw_id_l1(tbd)
    aspd = missing_indicator_real
    if (sw_id_l1 .eq. 0) aspd = real(tbd%values(iaux+pos_aspd(bufr_tmpl)))

  end function get_aspd
    !  #]
  function get_adir(tbd) result(adir)
    !  #[ get analysis wind dir
    type(bufr_data_type), intent(in) :: tbd
    real                             :: adir ! result

    ! local variable
    integer :: sw_id_l1

    sw_id_l1 = get_sw_id_l1(tbd)

    adir = missing_indicator_real
    if (sw_id_l1 .eq. 0) then
       adir = real(tbd%values(iaux+pos_adir(bufr_tmpl)))
       
       if (angle_convention_to_report .eq. angle_convention_oceanographic) then
          ! convert model wind direction to oceanographic convention
          if (.not. missing_real(adir)) then
             adir = adir + 180.0
             if (adir .gt. 360.0) adir = adir - 360.0
          endif
       end if
    end if
    
  end function get_adir
    !  #]
  function get_azi_angle(tbd, beam) result(azi_angle)
    !  #[ get antenna azimuth angle [deg]
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: azi_angle ! result [deg]

    azi_angle = real(tbd%values(iaux + pos_azi_angle(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_azi_angle
    !  #]
  function get_inc_angle(tbd, beam) result(inc_angle)
    !  #[ get antenna incidence angle [deg]
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: inc_angle ! result [deg]

    inc_angle = real(tbd%values(iaux+pos_inc_angle(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_inc_angle
    !  #]
  function get_polarisation(tbd, beam) result(pol)
    !  #[ get normalized radar cross section [dB]
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    integer                          :: pol ! code table 2104

    ! ERS and ASCAT always use VV
    if ( (bufr_tmpl .eq. bufr_template_ASCAT)       .or. &
         (bufr_tmpl .eq. bufr_template_ERS_OLD)     .or. &
         (bufr_tmpl .eq. bufr_template_ERS_OLD_APP) .or. &
         (bufr_tmpl .eq. bufr_template_ERS_NEW)     .or. &
         (bufr_tmpl .eq. bufr_template_ERS_NEW_APP)       ) then
       pol = 1
       return
    end if
       
    pol = nint(real(tbd%values(iaux+pos_polaris(bufr_tmpl) + &
                               (beam-1)*var_ant_offset(bufr_tmpl) )))

    ! the meaning of the codes is:
    !002104 0007 00000000 01 HORIZONTAL POLARISATION
    !            00000001 01 VERTICAL POLARISATION
    !            00000002 01 RIGHT CIRCULAR POLARISATION
    !            00000003 01 LEFT CIRCULAR POLARSZATION
    !            00000004 01 HORIZONTAL AND VERTICAL POLARISATION
    !            00000005 01 RIGHT AND LEFT CIRCULAR POLARISATION
    !            00000015 01 MISSING VALUE

  end function get_polarisation
    !  #]
  function get_sigma0(tbd, beam) result(sigma0)
    !  #[ get normalized radar cross section [dB]
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: sigma0 ! result [dB]

    sigma0 = real(tbd%values(iaux+pos_sigma0(bufr_tmpl) + &
                             (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_sigma0
    !  #]
  function get_item(tbd, item) result(value)
    !  #[ get value of item_nr in BUFR message
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: item
    real                             :: value ! result

    value = real(tbd%values(iaux+item))

  end function get_item
    !  #]
  function get_kpa(tbd, beam) result(kpa)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: kpa ! result

    write (*,*) 'XXX bufr_l2_reader - get_kpa: bufr_tmpl=',bufr_tmpl

    kpa = real(tbd%values(iaux+pos_kpa(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_kpa
    !  #]
  function get_kpb(tbd, beam) result(kpb)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: kpb ! result

    kpb = real(tbd%values(iaux+pos_kpb(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_kpb
    !  #]
  function get_kpc(tbd, beam) result(kpc)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: kpc ! result

    kpc = real(tbd%values(iaux+pos_kpc(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_kpc
    !  #]

  function get_kp(tbd, beam) result(kp)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: kp  ! result

    kp  = real(tbd%values(iaux+pos_kp(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_kp
    !  #]

  function get_atten_value(tbd, beam) result(atten_value)
    !  #[ get_atten_value
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: atten_value ! result [dB]

    atten_value = real(tbd%values(iaux+pos_atten_value(bufr_tmpl) + &
                             (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_atten_value
    !  #]
  function get_btem(tbd) result(btem)
    !  #[ get btem
    type(bufr_data_type), intent(in) :: tbd
    real                             :: btem ! result
    if (pos_btem(bufr_tmpl) .le. 0) then
      btem = missing_indicator_real
    else
      btem = real(tbd%values(iaux+pos_btem(bufr_tmpl)))
    endif
  end function get_btem
    !  #]
!-----------------------------------------------------------------------------
! Add by LZ, for CFOSAT AND seawinds
  function get_ambigus4wsp(tbd, idx) result(am_wsp)
    !  #[ get the four ambiguity wind speed in order
    type(bufr_data_type), intent(in)   :: tbd
    integer, intent(in)                :: idx
    real                               :: am_wsp

    ! local var
    integer  :: pos_spd

    am_wsp = missing_indicator_real
    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind speed 
       pos_spd = 26 + idx * 5
       if (.not. missing_real(real(tbd%values(iaux+pos_spd)))) then 
          am_wsp = real(tbd%values(iaux+pos_spd))
       endif
    else
       print *, 'It is not CFOSAT or Seawinds product.'
    endif

  end function get_ambigus4wsp
    !  #]

  function get_ambigus4wdir(tbd, idx) result(am_wdir)
    !  #[ get the four ambiguity wind speed in order
    type(bufr_data_type), intent(in)   :: tbd
    integer, intent(in)                :: idx
    real                               :: am_wdir

    ! local var
    integer  :: pos_dir

    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
       ! get wind speed 
       pos_dir = 28 + idx * 5
       if (missing_real(real(tbd%values(iaux+pos_dir)))) then 
         am_wdir = missing_indicator_real
         return
       else
         am_wdir = real(tbd%values(iaux+pos_dir))
       endif
    else
       print *, 'It is not CFOSAT or Seawinds product.'
    endif

    if (angle_convention_to_report .eq. angle_convention_oceanographic) then
       ! convert model wind direction to oceanographic convention
       if (.not. missing_real(am_wdir)) then
          am_wdir = am_wdir + 180.0
          if (am_wdir .gt. 360.0) am_wdir = am_wdir - 360.0
       endif
    end if

  end function get_ambigus4wdir
    !  #]
  function get_nr_ambigs(tbd) result(nr_ambigs)
    !  #[ get the number of ambiguities of this WVCs
    type(bufr_data_type), intent(in) :: tbd

    ! local
    integer :: nr_ambigs

    nr_ambigs = BufrReal2Int(tbd%values(iaux+pos_nram(bufr_tmpl)))

  end function
    !  #]

!  cone distances of the ambigs solutions
  function get_ambigus4dist(tbd, idx) result(amdist)
    !  #[ get backscatter distance of ambiguity solutions
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: idx
    real                             :: amdist
    integer   :: pos_dist

    ! hack to store MLE in NOAA BUFR, conedistance stored in error_dir
    amdist = missing_indicator_real
    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then
      
       pos_dist = 29 + idx * 5
       if (.not. missing_real(real(tbd%values(iaux+pos_dist)))) then 
          amdist = real(tbd%values(iaux+pos_dist))
       endif
    else
       print *, 'It is not CFOSAT or Seawinds product.'
    endif    

  end function get_ambigus4dist
    !  #]

!  cone distances of the ambigs solutions
  function get_ambigus4MLE(tbd, idx) result(amMLE)
    !  #[ get backscatter distance of ambiguity solutions
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: idx
    real                             :: amMLE
    integer   :: pos_MLE

    amMLE = missing_indicator_real
    if (bufr_tmpl .eq. bufr_template_Seawinds  .or. &
        bufr_tmpl .eq. bufr_template_CFOSAT) then

       pos_MLE = 30 + idx * 5
       if (.not. missing_real(real(tbd%values(iaux+pos_MLE)))) then
          amMLE = real(tbd%values(iaux+pos_MLE))
       endif
    else
       print *, 'It is not CFOSAT or Seawinds product.'
    endif

  end function get_ambigus4MLE
    !  #]
  function get_beam_lat(tbd, beam) result(beam_lat)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: beam_lat ! result [deg]

    beam_lat = real(tbd%values(iaux+pos_beam_lat(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_beam_lat
    !  #]
  function get_beam_lon(tbd, beam) result(beam_lon)
    !  #[
    type(bufr_data_type), intent(in) :: tbd
    integer, intent(in)              :: beam ! counter, starts at 1
    real                             :: beam_lon ! result [deg]

    beam_lon = real(tbd%values(iaux+pos_beam_lon(bufr_tmpl) + &
                                (beam-1)*var_ant_offset(bufr_tmpl) ))

  end function get_beam_lon
    !  #]
  !-------------------------------------------------------------
end module bufr_data_extraction_module
